/*
class Engine for the operation of engine(step mover)
author:	I.B.Nikolaev@inp.nsk.su
the description of protocol PIV-485 for KSD-485 on page
http://www.inp.nsk.su/~muchnoi/ksm.html
*/

#ifndef IBN_ENGINE_H
#define IBN_ENGINE_H
#include<iostream>



#include<vector>
#include<map>
//#include<algorythm>

#include"Error.h"

#include<signal.h>

#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <termios.h>
//#include <sys/ioctl.h>
#include <sys/times.h>
#include<unistd.h>

#include<errno.h>
using namespace std;
namespace modules
{


	class Port
	{
		const string name;
		int fd;
		struct termios tnew,told;
		static void timeout(int)
		{
			cerr << "timeout of tty open" << endl;
		}
	public:

		struct bad_tty : public std::runtime_error
		{
			bad_tty(const char * str) : runtime_error(str){}
		};

		int Getfd(void) const { return fd; }
		Port(char * device, int timeout = 0) : name(device)
		{
			//Port::TTY_NAME = device;
			//set alarm handler and save ald sigaction in struct oact
			static struct sigaction act,oact;
			act.sa_handler = Port::timeout;
			sigfillset(&(act.sa_mask));
			sigaction(SIGALRM,&act,&oact);

			// 
			alarm(timeout);

			//open tty
			fd = open( device, O_RDWR | O_NOCTTY | O_NONBLOCK);
			//fd = open( device, O_RDWR | O_NOCTTY);

			//exit(0);
			if ( fd == -1 )	throw camac::system_error(errno);

			if(!isatty(fd)) throw bad_tty(("device is not tty: " + name).c_str());

			// reset sigaction
			alarm(0);
			sigaction(SIGALRM, &oact,&act);


			//tune the terminal with saving old options
			tcgetattr(fd, &told);
			cfmakeraw(&tnew );
			cfsetispeed(&tnew, B19200);
			cfsetospeed(&tnew, B19200);
			tcsetattr(fd, TCSANOW, &tnew);

			if(fcntl(fd,F_SETFL,O_NONBLOCK) == -1) throw camac::system_error(errno);
		}

		~Port(void)
		{
			tcsetattr(fd,TCSAFLUSH,&told);
			close(fd);
		}
	};


class Engine    {

	//define types for unsigned char
	typedef vector<u8_t> Pocket_t;

	//define operation bytes
	static const u8_t START = 0xAA;
	static const u8_t STOP = 0xAB;
	static const u8_t SHIFT = 0xAC;

	//define delay after sending pocket
	static const unsigned WAIT = 100000; //mksec

	//define maximal size of all pockets
	static const unsigned MAX_POCKET_SIZE = 32;

	const int Port;

	u8_t Address; //logical address of engine
	double Current;
	double Velocity;

	//Engine commands
	enum COMMAND    {
		CMD_SETADR = 0,
		CMD_IDENT = 1,  			//ask information about device (engine)
		CMD_REPEAT = 2, 			//repeat last unswer
		CMD_STATUS = 3,         	//return status byte (struct Status)
		CMD_MOVE = 4,           	//move with acceleration
		CMD_MOVE_NO_ACCEL = 5,  	//move with no acceleration
		CMD_SET_CONFIG = 6,     	//set confgiguration
		CMD_SET_VELOCITY = 7,   	//set velocity
		CMD_STOP = 8,           	//prinuditelnaya stop
		CMD_OFF = 9,            	//tern off the current
		CMD_SAVE = 10,          	//save parameters in energy independent memory
		CMD_IMPULSE = 11,       	//give impulses during the moving, operate only on next moving command
		CMD_REMAIN = 12,        	//how many steps remain
		CMD_GET_CONFIG = 13,    	//read configuration
		CMD_GET_VELOCITY = 14,  	//read velocity
		CMD_TEST_FREQ = 15,         //measure internal generator frequancy
		CMD_CALIBRATE = 16,         //set timer period for moving with precision velocity
		CMD_MOVE_PREC = 17,         //move with precision velocity
		CMD_OUTPUT_IMPULSE = 18,    //give output impulse for a external device
		CMD_SET_COORDINATE = 19,    //set absolute coordinate
		CMD_GET_COORDINATE = 20,    //get absolute coordinate
		CMD_SYNCHRO = 21            //synchronized few engines
	};


//function use system call write, check system error and if need
//generate exception
inline void send(u8_t);

//read one byte from port with system call read
inline u8_t receive(void);

//send one byte with checking operation byte
inline void send_byte(u8_t);

//receive byte with transform;
inline u8_t receive_byte(void);

//send coommand to engine with special protocol
inline void send_pocket(const Pocket_t &);

//recive answer from engine with special protocol
inline void receive_pocket(Pocket_t &);

public:

static void set32(Pocket_t &p, u32_t data, unsigned n)    {
	if(n > p.size() - 4) throw std::invalid_argument("set32");
    p[n+3] = data & 0xff;
	p[n+2] = (data >> 8) & 0xff;
	p[n+1] = (data >> 16) & 0xff;
	p[n] = (data >> 24) & 0xff;
}

static void set16(Pocket_t &p, u16_t data, unsigned n)    {
	if(n > p.size() - 2) throw std::invalid_argument("set16");
	p[n+1] = data & 0xff;
	p[n] = (data >> 8) & 0xff;
}

static int32_t get32(const Pocket_t &p,unsigned n)    {
    if(n > p.size() - 4) throw std::invalid_argument("get32");
	return	(int32_t(p[n]) << 24 ) + (int32_t(p[n+1]) << 16) + (int32_t(p[n+2]) << 8) + p[n+3];
}

static int16_t get16(const Pocket_t &p,unsigned n)    {
	if(n > p.size() - 2) throw std::invalid_argument("get16");
	return	(int16_t(p[n]) << 8 ) + p[n+1];
}

struct bad_pocket : std::runtime_error {
	bad_pocket(const char * str): runtime_error(str){}
};

struct Status; //status byte - answer of engine in most case
struct Ident;  //information about controller
class Config; //configuration of engine

Engine(int port, int adr = 0): Port(port)	{
	Address = adr;
};

~Engine(void){};

unsigned GetAddress(void)  const  {
	return Address;
}

void SetAddress(unsigned adr)  {
    Address = adr;
}
void Repeat(void){}

Status GetStatus(void);
void Move(int);
void SetSpeed();
double GetSpeed();
void SetCoordinate(int);
int GetCoordinate(void);
Config GetConfig(void);
Ident IdentDevice(int adr = 0);



};


//implementation of class Engine

inline void Engine::send(u8_t b)    {
	int n; //amount of writen byte;
	n = write(Port,&b,1);
	if(n == -1) throw camac::system_error(errno);
	if(n != 1)  throw std::runtime_error("Can not write byte");
};


inline u8_t Engine::receive(void)    {
    u8_t b;
	int n; //amount of read byte;
	n = read(Port,&b,1);
	if(n == -1) throw camac::system_error(errno);
	if(n != 1)  throw std::runtime_error("Can not read byte");
	return b;
}

inline void Engine::send_byte(u8_t b)  {
	if(b == START || b == STOP || b == SHIFT)  {
		send(SHIFT);
		b-=START;
	}
	send(b);
};


inline u8_t Engine::receive_byte(void)    {
	u8_t b = receive();
	while(b == SHIFT) b = receive() + START;
    return b;
}



inline void Engine::send_pocket(const Pocket_t &p)    {
	//calculate control summ
	u8_t control_sum = Address;
	for(unsigned int i = 0; i < p.size(); i++) control_sum^=p[i];
	send(START);
	send_byte(Address);
	for(unsigned i = 0; i < p.size(); i++) send_byte(p[i]);
	send_byte(control_sum);
	send(STOP);
	usleep(WAIT);
};



inline void Engine::receive_pocket(Pocket_t & p)  {
    p.reserve(MAX_POCKET_SIZE);
	u8_t control_sum = receive_byte();
	if(control_sum != Address)  return ;
    u8_t b;
	while( (b = receive_byte()) != STOP)  {
		p.push_back(b);
        control_sum^=b;
        //if(p.size() > 1000) throw bad_pocket();
	}
	if(control_sum!=0) throw bad_pocket("wrong control sum");
}


struct Engine::Status {
	bool end;
	bool prec;
	bool dat;
	bool kplus;
	bool kminus;
	bool go;
	bool ready;
	Status(void){};
	Status(u8_t b)   {
		ready = b & 0x80;
		go = b & 0x40;
		kminus = b & 0x20;
		kplus = b & 0x10;
		dat = b & 0x8;
		prec = b & 0x4;
		end = b & 0x2;
	}
	~Status(void){}
	void print(void)    {
		cout << "status: ";
		cout << "end: " << end;
		cout << ", p vel: " << prec;
		cout << ", dat: " << dat;
		cout << ", K+: " << kplus ;
		cout << ", K-: " << kminus;
		cout << ", go: " << go ;
		cout << ", ready: " << ready <<endl;
	}
};


struct Engine::Ident    {
	unsigned Version;
	unsigned SerialNumber;
	Ident(void) {}
	Ident(Pocket_t p)   {
        Version = Engine::get16(p,2);
		SerialNumber = Engine::get16(p,4);
	}
};

class Engine::Config   {
    static const double TIME = 1/30.0;
	u8_t MoveCurrent;  	// 
    u8_t StopCurrent;  	// 
    u8_t MoveDelay;    	//       
    u8_t Cfg;           //config

public:
    struct bad_current{};
	enum CFG   {
		ACC_LEAVE = 0x1,
		LEAVE_K = 0x2,
		SOFT_K = 0x4,
		DAT = 0x8,
		KPLUS = 0x10,
		KMINUS = 0x20,
        HALF = 0x8
	};

	double GetMoveCurrent(void) const { return find_current(MoveCurrent); }
	double GetStopCurrent(void) const { return find_current(StopCurrent); }
	double GetMoveDelay(void) const { return MoveDelay*TIME; }

	bool AccLeave(void) const { return Cfg & ACC_LEAVE; }	//leave ends with accseleration
	bool LeaveK(void) { return Cfg & LEAVE_K;	}           //leave ends
	bool SoftK(void) { return Cfg & SOFT_K; }             //soft ends
	bool Dat(void) { return Cfg & DAT; }             		// |
	bool Kplus(void) { return Cfg & KPLUS; }             	// |true - datchic normally open; false  -- normally close
	bool Kminus(void) { return Cfg & KMINUS; }            // |
	bool Half(void) { return Cfg & HALF; }              	//mode true: 8-phase, false: 4-phase

	static double find_current(u8_t b)   {
        switch(b)    {
			case 0: return 0;
			case 1: return 0.2;
			case 2: return 0.3;
			case 3: return 0.5;
			case 4: return 0.6;
			case 5: return 1.0;
			case 6: return 2.0;
			case 7: return 3.5;
			default: throw bad_current();
		}
	}

	static u8_t code_current(double I)  {
		if(I < 0.2) return 0;
		if(I < 0.3) return 1;
		if(I < 0.5) return 2;
		if(I < 0.6) return 3;
		if(I < 1.0) return 4;
		if(I < 2.0) return 5;
		if(I < 3.5) return 6;
        return 7;
	}
	Config(void)    {
	}
	Config(Pocket_t &p) {
		if(p.size() != 4) throw Engine::bad_pocket("wrong size of the config pocket");
        if(p[0] >7) throw Engine::bad_pocket("wrong move current in the config pocket");
		MoveCurrent = p[0];
        if(p[1] >7) throw Engine::bad_pocket("wrong stop current in the config pocket");
		StopCurrent = p[1];
		MoveDelay = p[2];
        Cfg = p[3];
	}
	Config(double mc, double sc, double dt, u8_t cfg) {
		MoveCurrent = code_current(mc);
		StopCurrent = code_current(sc);
		MoveDelay = u8_t(dt/TIME);
		Cfg = cfg;
	}

	Pocket_t Pocket(void) {
		Pocket_t p(4);
        memcpy(&p[0],this,4);
        /*
		p[0] = MoveCurrent;
		p[1] = StopCurrent;
		p[2] = MoveDelay;
		p[3] = Cfg;
		*/
        return p;
	}

};


Engine::Status Engine::GetStatus(void) {
	Pocket_t cmd(1);
	cmd[0] = CMD_STATUS;
	send_pocket(cmd);
	Pocket_t answer;
	receive_pocket(answer);
	return Status(answer[0]);
};

void Engine::Move(int step)   {
	Pocket_t cmd(5);
	cmd[0] = CMD_MOVE;
    set32(cmd,step,1);
	/*
	cmd[4] = step & 0xff;
	cmd[3] = (step >> 8) & 0xff;
	cmd[2] = (step >> 16) & 0xff;
	cmd[1] = (step >> 24) & 0xff;
    */
	send_pocket(cmd);
	Pocket_t answr;
	receive_pocket(answr);
	//Status st(answr[0]);
    //st.print();
};

void Engine::SetCoordinate(int coord)   {
	Pocket_t cmd(5);
	cmd[0] = CMD_SET_COORDINATE;
	set32(cmd,coord,1);
	send_pocket(cmd);
	Pocket_t answer;
    receive_pocket(answer);
	//Status st(answer[0]);
    //st.print();
};

int Engine::GetCoordinate(void) {
	Pocket_t cmd(1);
    cmd[0] = CMD_GET_COORDINATE;
    send_pocket(cmd);
	Pocket_t answer;
	receive_pocket(answer);
	int coord = get32(answer,0);
	if(unsigned(coord) == 0x80000000) cerr << "coordinate is not real" <<endl;
    return coord;
};

Engine::Config Engine::GetConfig(void)  {
	Pocket_t cmd(1);
	cmd[0] = CMD_GET_CONFIG;
    send_pocket(cmd);
	Pocket_t answer;
	receive_pocket(answer);
	Config cfg(answer);
    return cfg;
};

Engine::Ident Engine::IdentDevice(int adr = 0)
{
	if(adr==0) adr = Address;
	Pocket_t cmd(1);
	cmd[0] = CMD_IDENT;
    send_pocket(cmd);
	Pocket_t answer;
    receive_pocket(answer);
	Ident id(answer);
    return id;
}

} //namespace modules
#endif //IBN_ENGINE_H
