#include <iostream>
#include <iomanip>
#include <math.h>
#include <map>

#include <chrono>
#include <thread>

#ifndef IBN_GENERATOR_H
#define IBN_GENERATOR_H

#include "Module.h"

#include "Time.h"
#include "ibn/timer.h"
#include "Log.h"

#include "ModuleInventory.h"


#include <unistd.h>

namespace camac 
{

  /*
   * =====================================================================================
   *        Class:  Generator
   *  Description:  Abstract class for all type of Generator
   * =====================================================================================
   */
  class Generator : virtual public ModuleInventory
  {
    public:
      Generator(void){};
      Generator(std::string name, std::string type, std::string subtype) : ModuleInventory(name,type,subtype) {}
      virtual ~Generator(void){};
      virtual void On(void){} ;
      virtual void Off(void){};
      virtual bool IsOn(void){return true;}
      virtual void SetFrequency(double f) {};
      virtual void GenerateFrequency(double frequency /* Hz */, double time=-1 /* sec */){};
      virtual void SetTime(double time /* sec */){};
      virtual double GetTime(void)const {return 0;}
      virtual void Init(void) {};
      virtual void Bind(camac::Address   adr){};
  };

  /*
   * =====================================================================================
   *        Class:  GeneratorArray
   *  Description:  Sinchronous number of generators. 
   * =====================================================================================
   */
  class GeneratorArray: public Generator
  {
    public:
      GeneratorArray(const std::string  & nm);
      virtual ~GeneratorArray(void);
      void Add(Generator * , double shift=0);
      void On(void);
      void Off(void);
      void SetFrequency(double f);
      void GenerateFrequency(double frequency /* Hz */, double time=-1 /* sec */);
      void SetTime(double time /* sec */);
      double GetTime(void) const;
      void Init(void);
      const char * GetSubType(void) const;
      const char * GetName(void) const;

      void SetShift(double sh);
      void SetShift(Generator * g, double sh);
      double GetShift(void) const;
      double GetShift(Generator * g) const;
    private:
      typedef std::map <Generator*, double> GenList_t;
      std::map <Generator*, double> genlist;
      double Time;
      timespec Ts;
      Generator * last_generator; /*  last added generator */
  };




  inline GeneratorArray::GeneratorArray(const std::string & nm) : ModuleInventory(nm.c_str(),"Generator","GeneratorArray")
  {
  }


  inline GeneratorArray::~GeneratorArray(void)
  {
  }


  inline  void GeneratorArray::Add(Generator * g, double shift)
  {
    genlist[g]=shift;
    last_generator=g; /*  we should remember last generator couse map doesnt remember the order of adding */
  }

  inline void GeneratorArray::Init(void)
  {
    //for(GenList_t::iterator i=genlist.begin(); i!=genlist.end(); i++)
		for(auto & p : genlist) p.first->Init();
  }


  inline void GeneratorArray::On(void)
  {
		for(auto & p : genlist) p.first->On();
  }

  inline void GeneratorArray::Off(void)
  {
		for(auto & p : genlist) p.first->Off();    
  }

  inline void GeneratorArray::SetFrequency(double f)
  {
		for(auto & p : genlist) p.first->SetFrequency(fabs(f)+p.second);
  }


  inline void GeneratorArray::SetTime(double t)
  {
    Time = t;
    Ts.tv_sec = time_t(floor(Time));
    Ts.tv_nsec = (long) ( (Time - Ts.tv_sec)*1e9 );
		for(auto & p : genlist) p.first->SetTime(t);
  }

  inline double GeneratorArray::GetTime(void) const
  {
    return  Time;
  }

  inline void GeneratorArray::GenerateFrequency(double f, double t)
  {
    SetTime(t);
    SetFrequency(f);
    nanosleep(&Ts,0);
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  GeneratorArray
   *      Method:  SetShift
   * Description:  Set shift for the specified generator
   *--------------------------------------------------------------------------------------
   */
  inline void GeneratorArray::SetShift(Generator * g, double shift)
  {
    GenList_t::iterator i=genlist.find(g);
    if(i!=genlist.end()) 
      i->second=shift;
    else 
      throw std::runtime_error("Unable to find generator");
  }



  /*
   *--------------------------------------------------------------------------------------
   *       Class:  GeneratorArray
   *      Method:  GetShift
   * Description:  Get shift for spec. generator
   *--------------------------------------------------------------------------------------
   */
  inline double GeneratorArray::GetShift(Generator * g) const 
  {
    GenList_t::const_iterator i=genlist.find(g);
    if(i!=genlist.end())
      return i->second;
    else
    {
      std::cerr << "Unable to find generator: " << g << " in list" << std::endl;
      return 0;
    }
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  GeneratorArray
   *      Method:  SetShift
   * Description:  Set shift for last generator in Hz
   *--------------------------------------------------------------------------------------
   */
  inline void GeneratorArray::SetShift(double shift)
  {
    genlist[last_generator]=shift;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  GeneratorArray
   *      Method:  GetShift
   * Description:  Get shift for last generator in Hz
   *--------------------------------------------------------------------------------------
   */
  inline double GeneratorArray::GetShift(void) const
  {
    GenList_t::const_iterator i=genlist.find(last_generator);
    return i->second;
  }

  /*
   * =====================================================================================
   *        Class:  Generator_G0605
   *  Description:  Frequency sythesizer developed by Shubin Evgeny. 48-50.
   *                This class is written by Bogomyagkhov Anton (44-20)
   * =====================================================================================
   */
  class Generator_G0605	:  public Module, virtual public ModuleInventory, public Generator
  {
    static constexpr double BASE_FREQUENCY=60750733.65;
    public:
    Generator_G0605(void){}
    Generator_G0605(const char* nm): ModuleInventory(nm, "Generator", "G0605") { }
    Generator_G0605(const char* nm, const camac::Address adr): ModuleInventory(nm, "Generator", "G0605"),Module(adr) { }
    virtual ~Generator_G0605() {}
    inline void Init(void);
    inline void Bind(const camac::Address adr);
    inline void On(void);
    inline void Off(void);
    inline bool IsOn(void) { return ison; }
    inline void SetFrequency(double  F);
    inline double GetFrequency(void);
    inline void SetTime(double t);
    inline double GetTime(void)const;
    inline void GenerateFrequency(double F); /*Generate Frequency during time which defined SetTime();*/
    inline void GenerateFrequency(double f, double time);
    protected:
    inline void Wait(void);
    inline void set_frequency(double);
    inline double get_frequency(void);
    private:
    bool ison;
    double frequency; /*  current frequency */
  };

  void Generator_G0605::Init(void)
  { 
    Z();C(); 
  };

  void Generator_G0605::On(void) 
  { 
    ison=true;
    AF(0,26); 
  }

  void Generator_G0605::Off(void)
  { 
    ison=false;
    AF(0,24); 
  }

  void Generator_G0605::SetFrequency(double  F)
  {	
    frequency = F;
    if(ison) set_frequency(F);
  }

  void Generator_G0605::set_frequency(double  F)
  {	
    F=fabs(F);
    u32_t N,S;
    double NS;
    NS=640.0-960.0*F/BASE_FREQUENCY;			//frequency code
    N=(u32_t)NS;					//integer part
    S=(u32_t)((NS-N)*0200000);		//fractial part
    N+=0200000;						//add 1 in a 17 position
    AF(0,16,&S);					//writing fractial part
    AF(1,16,&N);					//writing integer part
  }

  void Generator_G0605::GenerateFrequency(double F)    
  {
    SetFrequency(F);
    Wait();
  }

  void Generator_G0605::GenerateFrequency(double f, double time)
  {
    SetFrequency(f);
    SetTime(time);
    timespec tm;
    tm.tv_sec = time_t(floor(time));
    tm.tv_nsec = (long) ( (time - tm.tv_sec)*1e9 );
    nanosleep(&tm,0);
  }


  void Generator_G0605::SetTime(double t)
  {	//tau-time per point, mc
    unsigned tau = unsigned(t*1000.);
    u32_t M;
    if((tau>=0.1) && (tau<=100)){
      M=(u32_t)(log(tau*10.)/log(2.));		//code for timer
    }
    else{
      if(tau<0.1) M=0;
      if(tau>100) M=10;
    }
    AF(1,17,&M);			//writing
  }

  double Generator_G0605::GetFrequency(void)
  {		
    return frequency;
  }

  double Generator_G0605::get_frequency(void)
  {		//Read frequency
    u32_t N,S;
    double NS;
    AF(0,0,&S);			//reading fractial part
    AF(1,0,&N);			//reading integer part
    NS=N+(double)S/0200000;
    return (640.0-NS)*BASE_FREQUENCY/960.0 ;
  }

  double Generator_G0605::GetTime(void)	const
  {			//Read time, mc
    u32_t M;
    AF(1,1,&M);
    return(0.1*pow(2.,double(M)));
  }


  void Generator_G0605::Wait(void)	
  {
    unsigned i = 0;
    while(!IsLam(0,8) && IsInit() )
    {
      usleep(10);
      if(i > 400000) if(i%100000 == 0)
        std::cerr <<"Has not being  LAM from generator for " << i/10000. << "sec";
    }
  }

  void Generator_G0605::Bind(const camac::Address adr) 
  {
    Module::Bind(adr);
    //Init();
  };

  /*
   * =====================================================================================
   *        Class:  Generator_DDS300
   * 
   *  Description:  Синтезатор синусоидального сигнала. Позволяет делать шаги по частоте 1мкГц
   *                Разработчик Шубин Е.И. тел. 48-50 
   * 
   *       Author:  Ivan Nikolaev
   *      Created:  11.05.2004 15:53:46 NOVST
   *     Revision:  none
   * =====================================================================================
   */
  class  Generator_DDS300 : public Module, virtual public ModuleInventory, public Generator
  {

    static constexpr double INTERNAL_BASE_FREQUENCY=30e6;
    static constexpr double EXTERNAL_BASE_FREQUENCY=5e6;
    static constexpr unsigned MAX_REF_MULT=20;
    static constexpr unsigned MIN_REF_MULT=4;

    public:

    Generator_DDS300(void){}

    Generator_DDS300(std::string nm/*="unknown"*/)://Module(nm, "Generator", "DDS300"),
      ModuleInventory(nm, "Generator", "DDS300") 
    { 
      frequency=0;
    }

    Generator_DDS300(std::string nm, const camac::Address adr):
      ModuleInventory(nm, "Generator", "DDS300"),
      Module(adr)

    {
      Init();
    }

    virtual ~Generator_DDS300(void){}

    inline void Bind(const camac::Address adr);
    inline void Init(void);
    inline void Reset(void);
    inline void SetTime(double time);
    inline double GetTime(void) const;
    inline void PrintStatus(unsigned debug=0);
    inline void SetExternalBaseFrequency(double f=EXTERNAL_BASE_FREQUENCY,unsigned rfmult=20);
    inline void SetInternalBaseFrequency(double f=INTERNAL_BASE_FREQUENCY, unsigned rfmult=MIN_REF_MULT);
    inline void SetRefMult(unsigned k);
    inline void SetBaseFrequency(double f);
    inline void SetBaseFrequencyOut(void); 
    inline void SetComparatorOut(void);
    inline void  GenerateFrequency(double f);
    inline void SetFrequency(double f);
    inline void GenerateFrequency(double f, double time);
    inline void On(void);
    inline void Off(void);
    inline bool IsOn(void) {return ison;}
    private:
    inline unsigned serial_number(void);
    inline void write_stat_reg(unsigned data);
    inline unsigned read_stat_reg(void );
    // Запись  данных  register микросхемы AD9854
    inline void write_reg(u32_t data); /* write data to register of microchip AD9854 */
    inline void write_reg(unsigned a, unsigned data);
    inline void write_ctl_reg(unsigned a, unsigned data);
    inline unsigned read_ctl_reg(unsigned a) const;
    inline void enable_ctl_write(void);
    inline void disable_ctl_write(void);
    inline void set_external(void);
    inline void set_internal(void);
    inline void set_ref_mult(unsigned rfm);
    inline unsigned get_ref_mult(void);
    inline void set_base_frequency(double bf);
    inline double get_base_frequency(void) const;
    inline u64_t frequency2code(double frequency);
    inline double code2frequency(u64_t code);
    inline void update_frequency_step(void);
    inline double get_frequency_step(void) const;
    inline void update(void);
    inline void set_frequency(double f); /*  generator workout frequency */

    double Time;
    timespec ts;
    double base_frequency;
    unsigned ref_mult;
    double system_clock;
    double df;
    double frequency; /*  this is the current freqeuency, Hz */
    bool ison; /*  the state of generator ON/OFF */
    enum SR
    {
      STAT_BASE=0x1,
      STAT_OUT=0x2,
      STAT_WRITE_CTL=0x4
    };
  };

  void Generator_DDS300::Bind(const camac::Address adr) 
  {
    Module::Bind(adr);
    Init();
  };

  void Generator_DDS300::Init(void) 
  {
    ison = false; /*  depolarizer should be turned OFF */
    frequency=0;
    Reset();
    PrintStatus(3);
    //SetExternalBaseFrequency();
    SetInternalBaseFrequency();
    PrintStatus(3);
  }


  void Generator_DDS300::Reset(void)
  {
    write_stat_reg(0);
    //usleep(10000);
    //write_ctl_reg(0x1d,0x0);
    //usleep(10000);
    //write_ctl_reg(0x1e,0x14);
    //usleep(10000);
    //write_ctl_reg(0x1f,0x0);
    //usleep(10000);
    //write_ctl_reg(0x20,0x40);
    //update();
  }


  void Generator_DDS300::SetTime(double time)   
  {
    Time = time;
    ts.tv_sec = time_t(floor(Time));
    ts.tv_nsec = (long) ( (Time - ts.tv_sec)*1e9 );
  };

  double Generator_DDS300::GetTime(void) const { return Time;}

  void Generator_DDS300::PrintStatus(unsigned debug) 
  {
    unsigned w = 10;
    std::cout <<  LOG(debug)<<"Generator: Status Register: " << std::hex << read_stat_reg() <<"\n";
    std::cout <<  LOG(debug)<<"Generator: Control Register: " << "\n";
    std::cout <<  LOG(debug)<<"Generator: "<<std::setw(w) << "addres (hex)"  << std::setw(w) << " data " <<"\n";
    std::cout <<  LOG(debug)<<"Generator: "<<std::setw(w) << std::hex << 0x1d << std::setw(w) << std::hex << read_ctl_reg(0x1d) << "\n";
    std::cout <<  LOG(debug)<<"Generator: "<<std::setw(w) << std::hex << 0x1e << std::setw(w) << std::hex << read_ctl_reg(0x1e) << "\n";
    std::cout <<  LOG(debug)<<"Generator: "<<std::setw(w) << std::hex << 0x1f << std::setw(w) << std::hex << read_ctl_reg(0x1f) << "\n";
    std::cout <<  LOG(debug)<<"Generator: "<<std::setw(w) << std::hex << 0x20 << std::setw(w) << std::hex << read_ctl_reg(0x20) << "\n";
    std::cout <<  std::dec;
    std::cout <<  LOG(debug)<<"\n";
    std::cout <<  LOG(debug)<< "Generator: REF_MULT: " << get_ref_mult() << "\n";
    std::cout <<  LOG(debug)<< "Generator: Base frequency: " << base_frequency << " Hz" << "\n" ;
    std::cout <<  LOG(debug)<< "Generator: System clock: " << system_clock << " Hz" << "\n";
    std::cout <<  LOG(debug)<< "Generator: Frequency step: " << df << " Hz" << "\n";
    std::cout <<  LOG(debug)<< "Generator: state " << ison << std::endl;
  };


  void Generator_DDS300::SetExternalBaseFrequency(double f,unsigned rfmult) 
  {
    if(!IsInit()) return;
    usleep(100000);
    set_external();
    set_base_frequency(f);
    set_ref_mult(rfmult);
    update_frequency_step();
    usleep(100000);
    while(rfmult!=get_ref_mult())
    {
      Reset();
      usleep(100000);
      std::cerr << "WARNING: Genrator: Wrong REF_MULT in DDS-300. Try to write again" << std::endl;
      set_external();
      set_base_frequency(f);
      set_ref_mult(rfmult);
      update_frequency_step();
      usleep(100000);
    }
  };

  void Generator_DDS300::SetInternalBaseFrequency(double f, unsigned rfmult)
  {
    set_internal();
    set_base_frequency(f);
    set_ref_mult(rfmult);
    update_frequency_step();
  }

  void Generator_DDS300::SetRefMult(unsigned k)
  {
    set_ref_mult(k);
    update_frequency_step();
  };

  void Generator_DDS300::SetBaseFrequency(double f)
  {
    set_base_frequency(f);
    update_frequency_step();
  }



  void Generator_DDS300::SetBaseFrequencyOut(void)   
  {
    write_stat_reg(read_stat_reg() & ~STAT_OUT );
  };

  void Generator_DDS300::SetComparatorOut(void)
  {
    write_stat_reg(read_stat_reg() | STAT_OUT );
  };


  void Generator_DDS300::SetFrequency(double f)
  {
    frequency=f;
    if(ison) set_frequency(f);
  }

  void  Generator_DDS300::GenerateFrequency(double f)
  {
    SetFrequency(f);
    nanosleep(&ts,0);
  };

  void Generator_DDS300::GenerateFrequency(double f, double time)
  {
    //timespec tm;
    //tm.tv_sec = time_t(floor(time));
    //tm.tv_nsec = (long) ( (time - tm.tv_sec)*1e9 );
    SetFrequency(f);
    using namespace std::chrono;
    auto d = nanoseconds(long(time*1e9));
    std::this_thread::sleep_for(d);
    //nanosleep(&tm,0);
  }

  void Generator_DDS300::On(void)
  {  
    ison=true;
  }

  void Generator_DDS300::Off(void)
  {
    ison=false;
    set_frequency(0);
  }


  unsigned Generator_DDS300::serial_number(void) 
  {
    u32_t data;
    AF(3,1,&data);
    return data;
  }

  // Управление статусным регистром.

  void Generator_DDS300::write_stat_reg(unsigned data) 
  {
    AF(2,16,&data);
  }

  unsigned Generator_DDS300::read_stat_reg(void ) 
  {
    u32_t data;
    AF(2,0,&data);
    return data;
  }




  // Запись  данных  register микросхемы AD9854
  void Generator_DDS300::write_reg(u32_t data) 
  { 
    AF(0,16,&data);
  }

  void Generator_DDS300::write_reg(unsigned a, unsigned data) 
  {
    write_reg( (a<<8) | (data & 0xff)  )  ;
  };

  void Generator_DDS300::write_ctl_reg(unsigned a, unsigned data)   
  {
    enable_ctl_write();
    write_reg( (a<<8) | (data & 0xff) )  ;
  };

  unsigned Generator_DDS300::read_ctl_reg(unsigned a) const 
  { 
    u32_t data = a<<8;
    AF(1,16,&data);
    AF(0,0,&data);
    return data&0xff;
  }



  void Generator_DDS300::enable_ctl_write(void)   
  {
    write_stat_reg(read_stat_reg() | STAT_WRITE_CTL );
  };

  void Generator_DDS300::disable_ctl_write(void)
  {
    write_stat_reg(read_stat_reg() & ~STAT_WRITE_CTL );
  }


  void Generator_DDS300::set_external(void) 
  {
    write_stat_reg(read_stat_reg()| STAT_BASE);
  }

  void Generator_DDS300::set_internal(void) 
  {
    write_stat_reg(read_stat_reg() & ~STAT_BASE);
  }

  void Generator_DDS300::set_ref_mult(unsigned rfm)
  {
    if(rfm > MAX_REF_MULT) rfm=MAX_REF_MULT;
    if(rfm < MIN_REF_MULT) rfm=MIN_REF_MULT;
    ref_mult=rfm;
    unsigned CR_old = read_ctl_reg(0x1e) & ~0x1f;
    unsigned CR_new = (CR_old | (rfm & 0x1f));
    write_ctl_reg(0x1e, CR_new);
    update();
  };


  unsigned Generator_DDS300::get_ref_mult(void)
  {
    return read_ctl_reg(0x1e)&0x1f;
  }


  void Generator_DDS300::set_base_frequency(double bf) { base_frequency=bf; }


  double Generator_DDS300::get_base_frequency(void) const { return base_frequency; }

  //запись адреса Register (по-видимому Controrl Register)


  u64_t Generator_DDS300::frequency2code(double frequency) 
  { //FTW
    return u64_t (floor(frequency/df));
  }

  double Generator_DDS300::code2frequency(u64_t code)
  {
    return double(code*df);
  }

  //void set_frequency_step(double fgen, unsigned k) 
  void Generator_DDS300::update_frequency_step(void) 
  {
    system_clock=base_frequency*ref_mult;
    u64_t p=0x1;
    df=(system_clock/(p<<48));
  }

  double Generator_DDS300::get_frequency_step(void) const { return df;}



  /*
   *--------------------------------------------------------------------------------------
   *       Class:  Generator_DDS300
   *      Method:  update
   * Description:  Write from buffer to work register
   *--------------------------------------------------------------------------------------
   */
  void Generator_DDS300::update(void) 
  { 
    AF(0,26); 
  };


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  Generator_DDS300
   *      Method:  set_frequency
   * Description:  Low level camac operation. Generator generate the frequency
   *--------------------------------------------------------------------------------------
   */
  void Generator_DDS300::set_frequency(double f)	
  {
    f=fabs(f);
    unsigned reg=0x4;
    u64_t data=frequency2code(f);
    for (unsigned i = 0; i<6; i++,reg++)
    {
      write_reg( reg, data>>(8*(5-i)))  ;
    }
    update();
  }

  /*
   * =====================================================================================
   *        Class:  Generator_DDS500
   * 
   *  Description: Selected depolarizer 
   *               
   * 
   *       Author:  Ivan Nikolaev
   *      Created:  11.05.2004 15:53:46 NOVST
   *     Revision:  none
   * =====================================================================================
   */
  class  Generator_DDS500 : public Module, virtual public ModuleInventory, public Generator
  {
    static constexpr double INTERNAL_BASE_FREQUENCY=30e6;
    static constexpr double EXTERNAL_BASE_FREQUENCY=5e6;
    static constexpr double DELAY_STEP=5;

    public:


    Generator_DDS500(std::string nm="unknown"):
      ModuleInventory(nm, "Generator", "DDS500")
    {
    }

    Generator_DDS500(std::string nm, const camac::Address adr):
      ModuleInventory(nm, "Generator", "DDS500"),
      Module(adr)
    {
      Init();
    }

		virtual ~Generator_DDS500(void)
		{
			std::clog << LOG(DEBUG_LOG) << "~Generator_DDS500" << std::endl;
		}

    inline void Bind(const camac::Address adr);
    inline void Init(void);
    inline void On(void);
    inline void Off(void);
    inline bool IsOn(void) {return ison;}
    inline void SetFrequency(double freq);
    inline double GetFrequency(void) const;
    inline void SetTime(double time /*  in sec */);
    inline double GetTime(void) const;
    inline void GenerateFrequency(double f, double time=-1);
    inline void SetExternalBaseFrequency(void);
    inline void SetInternalBaseFrequency(void);
    inline void Reset(void); /*  reset all configurations */

    protected:
    inline void set_delay_mode(void);
    inline void set_gate(double gate /* ns */);
    inline void  set_phase_auto_synchronization(void);
    inline void set_phase_shift(double phase);
    inline void set_work_channel(unsigned ch=0);
    inline void set_gate  (double gate1 /* ns */, double  gate2 /* ns */);
    /* Write data from camac register to microscheme registers */
    inline void update(void);

    inline void set_frequency(double freq);

    inline void write_buffer_register(unsigned kod);

    inline void write_status_register(unsigned kod);


    unsigned MF; /*  multiplication factor  */
    double FMAX; /*  maximum frequency, Hz  */
    double F0;   /*  base frequency (internal or external) */
    double DF;   /*  quant frequency */
    double K;    /*  just koef for code calculation */

    inline void set_mult_factor(unsigned k);

    private:
    double frequency; /* generating frequency, Hz */
    double delay;  /*  delay in ns */
    double Time; /*  generation time in s. */
    bool  ison; /*  state of depolarizer */
    timespec Ts;
  };

  void Generator_DDS500::Bind(const camac::Address adr) 
  {
    Module::Bind(adr);
    Init();
  };

  void Generator_DDS500::Init(void) 
  {
    SetExternalBaseFrequency();
		//SetInternalBaseFrequency();
    set_mult_factor(1);
    set_delay_mode();
    frequency=0;
    set_gate(500); 
    set_frequency(0);
    set_phase_shift(0);
    Off();
  }

  void Generator_DDS500::On(void)
  {
    ison=true;
    AFD(5,16,0xc0);
    SetFrequency(frequency);
  }

  void Generator_DDS500::Off(void)
  {
    ison=false;
    AFD(5,16,0x0);
    set_frequency(0);
    set_phase_shift(0);
  }


  void Generator_DDS500::SetTime(double time)   
  {
    Time = time;
    Ts.tv_sec = time_t(floor(Time));
    Ts.tv_nsec = (long) ( (Time - Ts.tv_sec)*1e9 );
  };

  double Generator_DDS500::GetTime(void) const
  {
    return Time;
  }

  void Generator_DDS500::GenerateFrequency(double f, double time)
  {
    if(time>0) SetTime(time);
    SetFrequency(f);
    nanosleep(&Ts,0);
  }

  void Generator_DDS500::SetExternalBaseFrequency(void)
  {
    F0 = EXTERNAL_BASE_FREQUENCY;
    write_status_register(0xC0);
  }


  void  Generator_DDS500::SetInternalBaseFrequency(void)
  {
    F0=INTERNAL_BASE_FREQUENCY;
    write_status_register(0);
  }

  void Generator_DDS500::Reset(void) /*  reset all configurations */
  {
    AF(0,9);
  }

  void Generator_DDS500::set_frequency(double freq)
  {
    freq=fabs(freq);
    set_work_channel(3);
    unsigned kod=unsigned(freq*K); /*  calculate code */
    unsigned b10=kod & 0xffff; /* low two byte */
    unsigned b32=kod >> 16;    /* high tow byte */
    AFD(2,16,0x04); //write for frequency register;
    AF(0,16,&b10);
    AF(1,16,&b32);
    update();
    set_phase_shift(-2.*M_PI*freq*delay*1e-9-M_PI);
  }


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  Generagor_DDS500
   *      Method:  set_frequency
   * Description:  Independently writes frequency for two channel.
   *--------------------------------------------------------------------------------------
   */
  //void Generator_DDS500::set_frequency(double f1, double f2)
  //{
  //}

  void Generator_DDS500::SetFrequency(double freq)
  {
    frequency = freq;
    if(ison) set_frequency(frequency);
  }

  double Generator_DDS500::GetFrequency(void) const { return frequency; }


  void Generator_DDS500::set_delay_mode(void) 
  {
    AFD(5,16,0xC0);
  }

  void Generator_DDS500::set_gate(double gate /* ns */)
  {
    delay=gate/2.+20;
    set_gate(gate/2., gate/2.);
    set_phase_auto_synchronization();
    set_phase_shift(-2.*M_PI*frequency*delay*1e-9-M_PI);
  }

  void  Generator_DDS500::set_phase_auto_synchronization(void)
  {
    AFD(2,16,0x02); //write for address 0x02
    AFD(0,16,0x2000); // turn on bit FR2<13>=1
  };

  void Generator_DDS500::set_phase_shift(double phase) 
  {
    unsigned kod = unsigned(phase*(1<<14)/(2.*M_PI));
    /* Control channel 1 */
    AFD(2,16,0x00); //write for address 0x00
    AFD(0,16,0x80); //turn on bit CSR<7>=1 channel 1
    update();
    AFD(2,16,0x05); //write for phase register
    AF(0,16,&kod); //write code for phase
    update();
  }

  void Generator_DDS500::set_work_channel(unsigned ch) 
  {
    AFD(2,16,0x0); //write CSR for addres 0x00
    AFD(0,16,0xf0); //turn on bit csr<6>=1 (channel0) csr<7>=1 (channel1)
    update();
  }

  void Generator_DDS500::set_gate  (double gate1 /* ns */, double  gate2 /* ns */)
  {
    if(gate1<0) gate1=0; 
    if(gate2<0) gate2=0; 
    unsigned int kod1=unsigned(gate1/DELAY_STEP);
    unsigned int kod2=unsigned(gate2/DELAY_STEP);
    unsigned int kod= (kod2 << 8 ) | kod1;
    AF(3,16,&kod);
  }



  /* Write data from camac register to microscheme registers */
  void Generator_DDS500::update(void) 
  {
    AF(0,26); //update dds
  }

  void Generator_DDS500::write_buffer_register(unsigned kod)
  {
    unsigned b10=kod & 0xffff;
    unsigned b32=kod >> 16;
    AF(0,16,&b10);
    AF(1,16,&b32);
  }

  void Generator_DDS500::write_status_register(unsigned kod)
  {
    AF(5,16,&kod);
  }



  void Generator_DDS500::set_mult_factor(unsigned k)
  {
    MF=k;
    FMAX=MF*F0;
    DF = FMAX*pow(2.,-32.);
    K=1.0/DF;
  }

  //typedef Generator_DDS500 GeneratorOLD;
}

#endif
