/*
 * =====================================================================================
 *
 *       Filename:  Calculator.cpp
 *
 *    Description:  Do some usefull calculations. Implementation
 *
 *        Version:  1.0
 *        Created:  07/21/2010 01:32:47 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physics, Novosibirsk, Russia
 *
 * =====================================================================================
 */

#include <algorithm>

#include "Log.h"
#include "Formula.h"
#include "V4Par.h"

#include "data/ArraySerializaton.h"
#include "measurement/EnergyCalibrationConfig.h"

#include "Calculator.h"
#include "vepp4/Stap.h"

using namespace std;

namespace data 
{

  using namespace measurement;

  PhysValue frequency2energy(const PhysValue & fd, const PhysValue &f0, int n)
  {
    using namespace phys;
    PhysValue en;
    double x=fd.value/f0.value;
    en.value =  (n  + x)*me/muea;
    en.error = me/muea*sqrt(sq(fd.error) + sq(x*f0.error))/f0.value;
    return en;
  }

  PhysValue energy2frequency(const PhysValue & en, const PhysValue & f0, int n)
  {
    using namespace phys;
    PhysValue fd;
    fd.value = f0.value * (muea * en.value/me - n );
    fd.error = f0.value*sqrt ( sq(en.error/me*muea) + sq(f0.error/f0.value*(en.value*muea/me-n+1)));
    return fd;
  }




  PhysValue  GetDelta(const PhysValue & N1, const PhysValue & N2 )
  {
    // 1 - N2/N1
    using math::sq;
    if ( N1.value == 0 ) return PhysValue(0,0);
    return PhysValue(1.-N2.value/N1.value, sqrt( sq(N2.error/N1.value)+sq(N1.error*N2.value/sq(N1.value))));
  }



  void energy2frequency(const EnergyCalibrationConfig * e, EnergyCalibrationConfig * f)
  {
    *f=*e;
    f->depolarizer.initial = energy2frequency
      (
       e->depolarizer.initial, 
       e->depolarizer.revolution_frequency.value,
       e->depolarizer.harmonic_number
      ).value;
    f->depolarizer.final = energy2frequency
      (
       e->depolarizer.final, 
       e->depolarizer.revolution_frequency.value,
       e->depolarizer.harmonic_number
      ).value;
    f->depolarizer.step = energy2frequency
      (
       data::PhysValue(0,e->depolarizer.step),
       e->depolarizer.revolution_frequency.value,
       e->depolarizer.harmonic_number
      ).error;
    f->depolarizer.width = energy2frequency
      (
       data::PhysValue(0,e->depolarizer.width),
       e->depolarizer.revolution_frequency.value,
       e->depolarizer.harmonic_number
      ).error;
    f->depolarizer.scan_time = e->depolarizer.step/e->depolarizer.speed;
    f->depolarizer.speed=f->depolarizer.step/f->depolarizer.scan_time;
  }

  void frequency2energy(const EnergyCalibrationConfig *f, EnergyCalibrationConfig *e)
  {
    *e=*f;
    e->depolarizer.initial = frequency2energy
      (
       f->depolarizer.initial, 
       f->depolarizer.revolution_frequency.value,
       f->depolarizer.harmonic_number
      ).value;
    e->depolarizer.final = frequency2energy
      (
       f->depolarizer.final, 
       f->depolarizer.revolution_frequency.value,
       f->depolarizer.harmonic_number 
      ).value;
    e->depolarizer.step = frequency2energy
      (
       data::PhysValue(0,f->depolarizer.step),
       f->depolarizer.revolution_frequency.value,
       f->depolarizer.harmonic_number
      ).error;
    e->depolarizer.width = frequency2energy
      (
       data::PhysValue(0,f->depolarizer.width),
       f->depolarizer.revolution_frequency.value,
       f->depolarizer.harmonic_number
      ).error;
    double scan_time = f->depolarizer.step/f->depolarizer.speed;
    e->depolarizer.speed=e->depolarizer.step/scan_time;
  }

  void energy2frequency(const DepolarizerConfig *e, DepolarizerConfig *f)
  {
    *f=*e;
    f->initial = energy2frequency
      (
       e->initial, 
       e->revolution_frequency,
       e->harmonic_number
      ).value;
    f->final = energy2frequency
      (
       e->final, 
       e->revolution_frequency,
       e->harmonic_number 
      ).value;
    f->step = energy2frequency
      (
       data::PhysValue(0,e->step),
       e->revolution_frequency.value, //VERY STRANGE I MUST CHECK IT LATER
       e->harmonic_number
      ).error;
    f->width = energy2frequency
      (
       data::PhysValue(0,e->width),
       e->revolution_frequency.value, //VERY STRANGE I MUST CHECK IT LATER
       e->harmonic_number
      ).error;
    f->scan_time = e->step/e->speed;
    f->speed=f->step/f->scan_time;
  }

  void frequency2energy(const DepolarizerConfig *f, DepolarizerConfig *e)
  {
    *e=*f;
    e->initial = frequency2energy
      (
       f->initial, 
       f->revolution_frequency,
       f->harmonic_number
      ).value;
    e->final = frequency2energy
      (
       f->final, 
       f->revolution_frequency,
       f->harmonic_number 
      ).value;
    e->step = frequency2energy
      (
       data::PhysValue(0,f->step),
       f->revolution_frequency,
       f->harmonic_number
      ).error;
    e->width = frequency2energy
      (
       data::PhysValue(0,f->width),
       f->revolution_frequency,
       f->harmonic_number
      ).error;
    double scan_time = f->step/f->speed;
    e->speed=e->step/scan_time;
  }

  double frequency2energy(const DepolarizerConfig *c, double f)
  {
    std::cout << LOG(6) << c->revolution_frequency << "  " << c->harmonic_number << " " << f << '\n';
    return frequency2energy(f,c->revolution_frequency,c->harmonic_number).value;
  }

  double energy2frequency(const DepolarizerConfig *c, double e)
  {
    return energy2frequency(e,c->revolution_frequency,c->harmonic_number).value;
  }

  double dfrequency2energy(const DepolarizerConfig *c, double df)
  {
    return frequency2energy ( data::PhysValue(0,df), c->revolution_frequency.value, c->harmonic_number).error;
  }

  double denergy2frequency(const DepolarizerConfig *c, double de)
  {
    return energy2frequency ( data::PhysValue(0,de), c->revolution_frequency.value, c->harmonic_number).error;
  }




  void GetNMR(EnergyCalibrationConfig *c)
  {
    V4nmr nmr;
    c->nmr=nmr.H;
  }

  void GetRF(EnergyCalibrationConfig *c)
  {
    c->depolarizer.revolution_frequency=data::PhysValue(vepp4::Stap().RF/222.,818924*1e-10); 
    //take error from stability of RF
  }

  void GetRF(DepolarizerConfig *c)
  {
		vepp4::Stap s;
    c->RF=s.RF;
    c->revolution_frequency=c->RF/222.;
  }

  Calculator::Calculator(void)
  {
    alpha_t=0;
    N=15;
  }

  void Calculator::CalculateNewEnergy(EnergyCalibrationConfig *o, EnergyCalibrationConfig *c)
  {
    c->energy = data::PhysValue(o->energy.value*c->nmr/o->nmr, o->energy.error);
    if(c->nmr<10 || o->nmr>10000 || o->nmr<10 ) c->energy=o->energy;
		if(c->energy.value==0)
		{
			vepp4::Stap stap;
			c->energy=stap.Eset;
		}
    c->energy.value+=(c->temperature - o->temperature)*c->alpha_t;
    c->depolarizer.direction=-o->depolarizer.direction;
		if(c->depolarizer.direction==0) c->depolarizer.direction=+1;
    c->depolarizer.harmonic_number=o->depolarizer.harmonic_number;
		c->depolarizer.expected=c->energy.value;
  }

  void Calculator::CalculateEnergyRange(EnergyCalibrationConfig *c)
  {
    measurement::DepolarizerConfig &d = c->depolarizer;
    //double shift = std::max(fabs(shift), fabs(d.range)*0.5);
    double shift = fabs(d.range)*0.5;
    if (shift < 0.001) shift=0.025+d.step+std::max(3*d.step, N*c->count_time*d.speed);
    d.initial=c->energy.value - c->depolarizer.direction*shift;
    d.final=c->energy.value + c->depolarizer.direction*shift;
  }

  std::ostream & operator<<(std::ostream &os, const Calculator & cal)
  {
    os << std::setw(20) << cal.GetTemperatureShiftParameter() << std::setw(20) << " # temperature shift\n";
    os << std::setw(20) <<  cal.GetN() << std::setw(20) << " # points before jump\n";
    return os;
  }

  std::istream & operator>>(std::istream & is, Calculator & cal)
  {
    double tmp;
    is >> tmp ; is.ignore(1024,'\n');
    cal.SetTemperatureShiftParameter(tmp);
    is >> tmp ; is.ignore(1024,'\n');
    cal.SetN(int(tmp));
    return is;
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_load
   *  Description:  Create PhysValue load of counter 
   * =====================================================================================
   */
  PhysValue get_load( double N0 /*  number of tryes or bunch flyes */, double N  /*  number of counter hit */)
  {
    return PhysValue(N,sqrt(N*(1.-N/N0)));
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_real_count
   *  Description:  Calculate rate of particles using counter rate.
   * =====================================================================================
   */
  PhysValue get_real_count
    (
     double N0 /*  number of tryes or bunch flyes */, 
     double N  /*  number of counter hit */
    ) 
    {
      return PhysValue( - N0 * log ( 1. - N/N0),  sqrt( N / (1. - N/N0) ));
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_rate_prob
   *  Description:  Return rate probability using probability to hit the counter
   * =====================================================================================
   */
  PhysValue get_rate_prob( data::PhysValue p /*  probability to hit counter */)
  {
    PhysValue mu; /*  parameter of Poisson distribution (real count prob) */
    mu.value = - log(1.-p.value);
    mu.error = fabs(p.error/(1.-p.value));
    return mu; 
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_real_count
   *  Description:  Return real count using revolution frequency and hit counter rate
   * =====================================================================================
   */
  PhysValue get_real_count
    ( 
     double f0 /*  revolution frequency */,
     data::PhysValue f /*  counter load */
    )
    {
      return get_rate_prob(f/f0)*f0;
    }


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_count_prob
   *  Description:  Calculate probability to hit the counter.
   * =====================================================================================
   */
  PhysValue get_count_prob
    (
     const PhysValue & f /* internal scattering rate, Hz */, 
     const PhysValue & f0 /* revolution frequency, Hz */
    )
    {
      double P0=exp(-f.value/f0.value); //probability not to hit counter
      return PhysValue(1.-P0,P0*f.error/f0.value);
    }


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate correlated events using probability to hit counter
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents
    (
     const data::PhysValue & p12, //probability of coincidence
     const data::PhysValue & p1 , //probability to hit one of counter in pair
     const data::PhysValue & p2   //probability to hit second counter in pair
    )    
    {
      /* calculate some useful values */
      double up=p12.value-p1.value*p2.value;  /* numerator */
      double down=1.0+p12.value-p1.value-p2.value; /* denominator */
      double down2=down*down;
      double dp12=(down-up)/down2; /* derivative  of up/down on p12 */
      double dp1=(-p2.value*down - up*(-1.0))/down2; /*  derivative of up/down on p1 */
      double dp2=(-p1.value*down - up*(-1.0))/down2; /*  derivative of up/down on p2 */
      double cor=up/down; /* value itself (correlated events probability) */
      /* calculate error of value cor */
      double ecor=sqrt(sq(dp12*p12.error) + sq(dp1*p1.error) + sq(dp2*p2.error));
      return PhysValue(cor,ecor);
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate number of correlated events using my (I.Nikolaev) theory.
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* load of coincidence, Hz */, 
     const data::PhysValue & N1 /* load of a counter, Hz */, 
     const data::PhysValue & N2 /* load of another counter, Hz */)
    {
      /*  now convert correlated events probability to counting rate in Hz */
      return get_rate_prob(GetCorrelatedEvents(N12/f0,N1/f0,N2/f0))*f0;
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate number of correlated events using my (I.Nikolaev) theory.
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents2
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */)
    {
      /* calculate probability to hit the counters */
      PhysValue p12=get_count_prob(N12,f0); /*  conicidence */
      PhysValue p1=get_count_prob(N1,f0); /*  first counter */
      PhysValue p2=get_count_prob(N2,f0); /*  second counter */
      /*  now convert correlated events probability to counting rate in Hz */
      return get_rate_prob(GetCorrelatedEvents(p12,p1,p2))*f0;
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate Random events
   * =====================================================================================
   */
  PhysValue GetRandomEvents
    ( 
     const data::PhysValue & p12, //probability of coincidence
     const data::PhysValue & p1 , //probability to hit one of counter in pair
     const data::PhysValue & p2   //probability to hit second counter in pair
    )
    {
      return (p1-p12)/(1.-p2);
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate number of random events.
   * =====================================================================================
   */
  PhysValue GetRandomEvents
    (
     double N0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    )
    {
      return get_rate_prob(GetRandomEvents(N12/N0,N1/N0,N2/N0))*N0;
    }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate number of random events.
   * =====================================================================================
   */
  PhysValue GetRandomEvents2
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    )
    {
      /* calculate probability to hit the counters */
      PhysValue p12=get_count_prob(N12,f0); /*  conicidence */
      PhysValue p1=get_count_prob(N1,f0); /*  first counter */
      PhysValue p2=get_count_prob(N2,f0); /*  second counter */
      return get_rate_prob(GetRandomEvents(p12,p1,p2))*f0;
    }
}
