#ifndef IBN_DATA_CALCULATOR_H
#define IBN_DATA_CALCULATOR_H
/*
 * =====================================================================================
 * 
 *        Filename:  Calculator.h
 * 
 *     Description:  Calculate measurement parameters 
 * 
 *         Version:  1.0
 *         Created:  11.10.2004 17:08:47 NOVST
 *        Revision:  none
 *        Compiler:  gcc
 * 
 *          Author:  Ivan Nikolaev ()
 *         Company:  BINP
 *           Email:  I.B.Nikolaev@inp.nsk.su
 * 
 * =====================================================================================
 */
#include <iostream>
#include "Data.h"

#include "measurement/EnergyCalibrationConfig.h"

namespace data 
{

  using namespace measurement;
  inline double sq(double x) { return x*x; }
  PhysValue frequency2energy(const PhysValue & fd, const PhysValue &f0, int n);
  PhysValue energy2frequency(const PhysValue & en, const PhysValue & f0, int n);
  PhysValue  GetDelta(const PhysValue & N1, const PhysValue & N2 );
  void energy2frequency(const EnergyCalibrationConfig * e, EnergyCalibrationConfig * f);
  void frequency2energy(const EnergyCalibrationConfig *f, EnergyCalibrationConfig *e);
  void energy2frequency(const DepolarizerConfig *e, DepolarizerConfig *f);
  void frequency2energy(const DepolarizerConfig *f, DepolarizerConfig *e);
  double frequency2energy(const DepolarizerConfig *c, double f);
  double energy2frequency(const DepolarizerConfig *c, double e);
  double dfrequency2energy(const DepolarizerConfig *c, double df);
  double denergy2frequency(const DepolarizerConfig *c, double de);
  inline double dfrequency2energy(const DepolarizerConfig &c, double df) { return dfrequency2energy(&c, df); }
  inline double denergy2frequency(const DepolarizerConfig &c, double de) { return denergy2frequency(&c, de); }
  void GetNMR(EnergyCalibrationConfig *c);
  void GetRF(EnergyCalibrationConfig *c);
  void GetRF(DepolarizerConfig *c);

  class Calculator 
  {
    public:
      Calculator(void);
      double GetTemperatureShiftParameter(void) const {return alpha_t;}
      void SetTemperatureShiftParameter(double a=0) { alpha_t =a; }
      unsigned GetN(void) const { return N; }
      void SetN(size_t n) {   N = n; }
      void CalculateNewEnergy(EnergyCalibrationConfig *o, EnergyCalibrationConfig *c);
      void CalculateEnergyRange(EnergyCalibrationConfig *c);
      double alpha_t; //temperature drift
      size_t N; //amount of points before the expected depolarization moment.
    private:
  };
  std::ostream & operator<<(std::ostream &os, const Calculator & cal);
  std::istream & operator>>(std::istream & is, Calculator & cal);

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_load
   *  Description:  Create PhysValue load of counter 
   * =====================================================================================
   */
  PhysValue get_load( double N0 /*  number of tryes or bunch flyes */, double N  /*  number of counter hit */); 

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_real_count
   *  Description:  Calculate rate of particles (poisson parameters) using counter rate.
   * =====================================================================================
   */
  PhysValue get_real_count( double N0 /*  number of tryes or bunch flyes */, double N  /*  number of counter hit */); 


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_rate_prob
   *  Description:  Return rate probability using probability to hit the counter
   * =====================================================================================
   */
  PhysValue get_rate_prob( data::PhysValue p /*  probability to hit counter */);



  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_real_count
   *  Description:  Return real count using revolution frequency and hit counter rate
   * =====================================================================================
   */
  PhysValue get_real_count( double f0 /*  revolution frequency */, data::PhysValue f /*  count rate */);



  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  get_count_prob
   *  Description:  Calculate probability to hit the counter.
   * =====================================================================================
   */
  PhysValue get_count_prob ( const PhysValue & f /* internal scattering rate, Hz */, 
      const PhysValue & f0 /* revolution frequency, Hz */);




  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate correlated events using probability to hit counter
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents
    ( 
     const data::PhysValue & p12, //probability of coincidence
     const data::PhysValue & p1 , //probability to hit one of counter in pair
     const data::PhysValue & p2   //probability to hit second counter in pair
    ); 


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate number of correlated events using my (I.Nikolaev) theory.
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    );

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetCorrelatedEvents
   *  Description:  Calculate number of correlated events using my (I.Nikolaev) theory.
   * =====================================================================================
   */
  PhysValue GetCorrelatedEvents2
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    );

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate Random events
   * =====================================================================================
   */
  PhysValue GetRandomEvents
    ( 
     const data::PhysValue & p12, //probability of coincidence
     const data::PhysValue & p1 , //probability to hit one of counter in pair
     const data::PhysValue & p2   //probability to hit second counter in pair
    ); 

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate number of random events.
   * =====================================================================================
   */
  PhysValue GetRandomEvents
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    );

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  GetRandomEvents
   *  Description:  Calculate number of random events.
   * =====================================================================================
   */
  PhysValue GetRandomEvents2
    (
     double f0, /* revolution frequency, Hz */
     const data::PhysValue & N12 /* rate of coincidence, Hz */, 
     const data::PhysValue & N1 /* rate of a counter, Hz */, 
     const data::PhysValue & N2 /* rate of another counter, Hz */
    );
}

#endif
