#include <iostream>
#include <iomanip>
#include <string>
#include <sstream>
#include <sys/time.h>
#include <stdlib.h>
#include <algorithm>
#include <thread>


#include <boost/algorithm/string.hpp>
#include <KDB/kdb.h>

#include "data/Data.h"
#include "EnergyCalibrationStorage.h"
#include "Calculator.h"
#include "Utils.h"
#include "ibn/timer.h"
#include "Time.h"
#include "Log.h"
#include "V4Par.h"
#include "vepp4/Status.h"
#include "ArraySerializaton.h"

using namespace std;
data::PhysValue to_phys_value(const std::string & s)
	{
		std::istringstream is(s);
		data::PhysValue value;
		is >> value;
		return value;
	}

namespace data
{

  EnergyCalibrationStorage::EnergyCalibrationStorage (unsigned nbunch, unsigned ncounter)
  {
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::EnergyCalibrationStorage: Start" << std::endl;
    isdepolon=false; /*  whether depolarizer is switch on */
    success=false; /*  whether energy calibration successful or not */
    isfit=false; /*  wheter data is fited */
    N = 0; /*  Number of points should be zero  */

    nmr.resize(7); /*  this is vector of scalar (doule) */
    NMR.resize(7); /*  this is vector of arrays */

    numbunch = nbunch;
    numcounter = ncounter;
    numpair = ncounter/2;
    InitVacuum();
    InitArrayList();

    std::cout<<LOG(4)<<"EnergyCalibrationStorage::EnergyCalibrationStorage: Finish" << std::endl;
  }


  void EnergyCalibrationStorage::InitArrayList(void)
  {
		CounterArrayList.clear();
    BunchArrayList.clear();
    CounterBunchArrayList.clear();
    PairCounterBunchArrayList.clear();
    FixedArrayList.clear();

		CounterArrayList.push_back(&CounterPosition);

    BunchArrayList.push_back(&CountSum);
    BunchArrayList.push_back(&CoincSum);
    BunchArrayList.push_back(&CorSum);
    BunchArrayList.push_back(&RndSum);
    BunchArrayList.push_back(&RndCoincSum);

    BunchArrayList.push_back(&DeltaSum);
    BunchArrayList.push_back(&DeltaCoincSum);
    BunchArrayList.push_back(&DeltaCorSum);
    BunchArrayList.push_back(&DeltaRndSum);
    BunchArrayList.push_back(&DeltaRndCoincSum);

    BunchArrayList.push_back(&I);
    BunchArrayList.push_back(&SigmaX);
    BunchArrayList.push_back(&SigmaZ);
    BunchArrayList.push_back(&SigmaL);

    /*  Normalized  bunch current */
    BunchArrayList.push_back(&CountSumNormI); /* sum counting rate */
    BunchArrayList.push_back(&CoincSumNormI);  /* sum of coincidence */
    BunchArrayList.push_back(&CorSumNormI);   /* sum of correlated */
    BunchArrayList.push_back(&RndSumNormI); 
    BunchArrayList.push_back(&RndCoincSumNormI); 

    /*  Normalized on square of bunch current */
    BunchArrayList.push_back(&CountSumNormI2); /* sum counting rate */
    BunchArrayList.push_back(&CoincSumNormI2);  /* sum of coincidence */
    BunchArrayList.push_back(&CorSumNormI2);   /* sum of correlated */
    BunchArrayList.push_back(&IRatio);
    
    CounterBunchArrayList.push_back(&CounterLoad);
    CounterBunchArrayList.push_back(&Count);
    CounterBunchArrayList.push_back(&Rnd);
    CounterBunchArrayList.push_back(&DeltaCount);
    CounterBunchArrayList.push_back(&DeltaRnd);

    PairCounterBunchArrayList.push_back(&CoincLoad);
    PairCounterBunchArrayList.push_back(&Coinc);
    PairCounterBunchArrayList.push_back(&Cor);
    PairCounterBunchArrayList.push_back(&DeltaCoinc);
    PairCounterBunchArrayList.push_back(&DeltaCor);
    PairCounterBunchArrayList.push_back(&RndCoinc);
    PairCounterBunchArrayList.push_back(&DeltaRndCoinc);


    FixedArrayList.push_back(&t);
    FixedArrayList.push_back(&RF);
    FixedArrayList.push_back(&DepFr);
    FixedArrayList.push_back(&MDepFr);
    FixedArrayList.push_back(&Isum);

    /* Add NMR arrays to global list */
    for(unsigned i=0; i<NMR.size(); ++i) FixedArrayList.push_back(&NMR[i]);

    FixedArrayList.push_back(&MediumVacuumPresure);

    //TEMPORARY FOR DEBUG THE CRASH
    //for(map <string, Array>::iterator it=Vacuum.begin(); it!=Vacuum.end(); ++it)
    //  FixedArrayList.push_back(&(it->second));

    set_geometry(numbunch, numcounter, numpair);
  }

  EnergyCalibrationStorage::~EnergyCalibrationStorage(void)
  {
  }


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  Clear
   * Description:  Delete all data.
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::Clear(void)   
  {
    clog << LOG(5) << "EnergyCalibrationStorage::Clear " << endl;
    N=0;
		for(auto a : ArrayList)
		{
			a->clear();
		}
    InitVacuum();
  }



  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  reserve
   * Description:  Reserve memory for fast following push_back operation.
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::reserve(unsigned num)
  {
    std::clog << LOG(5) << "EnergyCalibrationStorage::reserve: num=" << num << std::endl;
    for(list<Array*>::iterator p=ArrayList.begin();p!=ArrayList.end(); ++p)
    {
      (*p)->reserve(num);
    }
    for(std::map<string,data::Array>::iterator p=Vacuum.begin();p!=Vacuum.end();++p)
    {
      p->second.reserve(num);
    }
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  resize
   * Description:  Reserve memory for fast following push_back operation.
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::resize(unsigned num)
  {
    N=num;
    for(list<Array*>::iterator p=ArrayList.begin();p!=ArrayList.end(); p++)
      (*p)->resize(num);
    for(std::map<string,data::Array>::iterator p=Vacuum.begin();p!=Vacuum.end();p++)
    {
      p->second.resize(num);
    }
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  push_back
   * Description:  Add new zero point
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::push_back(void)
  {
		for(auto p : ArrayList) p->push_back(0, 0);
		for(auto p : Vacuum)    p.second.push_back(0, 0);
    N++;
  }

  
  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  set_geometry
   * Description:  Set new geometry number of bunches, number of counters, number of pair
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::set_geometry(unsigned nbunch, unsigned ncounter, unsigned npair)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::set_geometry: nbunch=" << nbunch << ", ncounter=" << ncounter << ", npair=" << npair << endl;
    ArrayList.clear(); /* empty list of all arrays */ 

    numbunch = nbunch;
    numcounter = ncounter;
    numpair = npair;

    cfg.bunch_norm.resize(numbunch);

		for(auto  a : CounterArrayList)
		{
			a->resize(numcounter);
			for(unsigned counter=0;counter<numcounter; counter++)
			{
				ArrayList.push_back(&(*a)[counter]);
			}
		}

    /*  for bunch only */
    for(list< vector<Array> * >::iterator i=BunchArrayList.begin(); i!=BunchArrayList.end(); ++i)
    {
      vector<Array> & ba = *(*i);
      ba.resize(numbunch);
      /*  now add this new phys arrays to list of all arrays */
      for(unsigned bunch=0; bunch<numbunch; bunch++)
        ArrayList.push_back(&ba[bunch]);
    }

    /*  and for counters */
    for(list< vector <vector<Array> > * >::iterator i=CounterBunchArrayList.begin(); i!=CounterBunchArrayList.end(); ++i) 
    {
      vector< vector<Array> > & cba = *(*i);
      cba.resize(numbunch);
      for(unsigned bunch=0; bunch<numbunch; bunch++)
      {
        cba[bunch].resize(numcounter);
        /*  now add this new phys arrays to list of all arrays */
        for(unsigned pair=0; pair<numcounter; pair++)
          ArrayList.push_back(&cba[bunch][pair]);
      }
    }

    /* and for pairs */
    for(list< vector <vector<Array> > * >::iterator i=PairCounterBunchArrayList.begin(); i!=PairCounterBunchArrayList.end(); ++i) 
    {
      vector< vector<Array> > & cba = *(*i);
      cba.resize(numbunch);
      for(unsigned bunch=0; bunch<nbunch; bunch++)
      {
        cba[bunch].resize(numpair);
        /*  now add this new phys arrays to list of all arrays */
        for(unsigned pair=0; pair<numpair; pair++)
          ArrayList.push_back(&cba[bunch][pair]);
      }
    }

    /*  add to array list geometry independent array list */
    for(list<Array*>::iterator p=FixedArrayList.begin(); p!=FixedArrayList.end(); ++p)
      ArrayList.push_back(*p);
  } //end of set_geometry



  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  Save2kadrs
   * Description:  Save result of energy calibration into kadrs file for Anton B.
   *--------------------------------------------------------------------------------------
   */

  void EnergyCalibrationStorage::Save2kadrs(void)
  {
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2kadrs" << endl;
    string filename = "/home/vepp4/kadrs/v4-energy-calibration.pol";
    ofstream of(filename.c_str());
    of.precision(15);
    of << "# Energy measurement by resonance depolariation method." << endl;
    of << "# Data generated by polarimeter" << endl;
    of << "VERSION: " << Storage::VERSION << endl;
    of << "Measurement date: " << begin_time.tv_sec << endl;
    time_t deptime = fit_param.size()>0 ? time_t(fit_param[0].value) : 0;
    of << "Unixtime: " << begin_time.tv_sec+deptime << endl;
    of << "Run: " << cfg.run << endl;
    of << "Mode: " << "???" << endl;
    of << "Set energy, MeV: " << "???" << endl;
    of << "Calibration status: " << success << endl;
    of << "Direction of calibration: " << "???" << endl;
    of << "Measured energy, MeV: " << E << endl;
    of << "Measured frequency, Hz: " << Fr << endl;
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2kadrs DONE" << endl;
  }




  void EnergyCalibrationStorage::SaveCurrent(std::string  savefilename) //add current data at file	
  {
    boost::filesystem::path path(savefilename);
		ibn::timer t;
		//if(saving_thread) saving_thread->join();
		//saving_thread.reset(new std::thread([=]{SaveBinary(path);}));
    SaveBinary(path);
  }


  //=============================================================================================================================
  bool EnergyCalibrationStorage::Save2DB(const char * table_name) //сохранить в базу данных
  {

    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 	START" << std::endl;

    KDBconn * kdb_conn = kdb_open();
    if (!kdb_conn)
    {
      kdb_close( kdb_conn );
      throw std::runtime_error("can't connect to calibration data base");
    }
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 1" << std::endl;
    char buf_char[128];
    strncpy(buf_char, table_name, 127);
    unsigned TableId=kdb_get_id(kdb_conn, buf_char); //получили ID таблицы которому соответствует название таблицы
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 2" << std::endl;



    union
    {
      // this union is to convert mode into digit and back 
      //   int32_t declare in calibration data base library 
      //   by E.Baldin 
      //
      char s[8] ;
      int32_t i;
    };
    int32_t buf[50];
    sprintf(s,"%s", mode.c_str());
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 2.1" << std::endl;
    buf[0] = 0;
    buf[1] = i;
    buf[2] = int(100*setE);
    //                      buf[3] = EC.date;
    buf[3] = cfg.date;
    buf[4] = cfg.run;
    buf[5] = success;
    buf[6] = cfg.depolarizer.direction;
    double DELTA_SCALE=1e6;
    double SLOPE_SCALE=DELTA_SCALE*1e3;
    double ENERGY_SCALE =1e6;
    double FREQUENCY_SCALE = 1e3;
    double TIME_SCALE = 1e3;
    buf[7] = int(ENERGY_SCALE * E.value); //в элекрон вольтах
    buf[8] = int(ENERGY_SCALE * E.error);
    buf[9] = int(FREQUENCY_SCALE * Fr.value); //в милигерцах
    buf[10] = int(FREQUENCY_SCALE * Fr.error);
    buf[11] = 0;


    buf[12] = int(TIME_SCALE  * fit_param[0].value + 0.5); //время деполяризации в мсек
    buf[13] = int(TIME_SCALE  * fit_param[0].error + 0.5);
    buf[14] = int(DELTA_SCALE * fit_param[1].value + 0.5);//скачек в десятых процента
    buf[15] = int(DELTA_SCALE * fit_param[1].error + 0.5);
    buf[16] = int(DELTA_SCALE * fit_param[2].value + 0.5);//начальный уровень  в десятых процента
    buf[17] = int(DELTA_SCALE * fit_param[2].error + 0.5);
    buf[18] = int(SLOPE_SCALE * fit_param[3].value + 0.5); //наклон 1 в десятых процента на 1000 секунд
    buf[19] = int(SLOPE_SCALE * fit_param[3].error + 0.5);
    buf[20] = int(SLOPE_SCALE * fit_param[4].value + 0.5); //наклон 2 в десятых процента на 1000 секунд
    buf[21] = int(SLOPE_SCALE * fit_param[4].error + 0.5);
    buf[22] = int(1000 * chi2); // хи квадрат, нормированный на степени свободы
    buf[23] = int(cfg.count_time + 0.5);
    buf[24] = int(cfg.depolarizer.scan_time*1000 + 0.5);
    buf[25] = int(cfg.depolarizer.revolution_frequency.value + 0.5);
    buf[26] = int(ENERGY_SCALE * cfg.depolarizer.initial + 0.5);
    buf[27] = int(ENERGY_SCALE * cfg.depolarizer.final + 0.5);
    buf[28] = int(ENERGY_SCALE * cfg.depolarizer.step + 0.5);
    buf[29] = int(ENERGY_SCALE * cfg.depolarizer.speed + 0.5);
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 2.3" << std::endl;
    buf[30] = 0;
    buf[31] = int(cfg.depolarizer.attenuation + 0.5);
    buf[32] = int(cfg.depolarizer.RF + 0.5);

		vepp4::Status vs;

    buf[33] = vs.cycle_time;
    buf[34] = vs.energy_time;
    buf[35] = cfg.reason;  /* 0 -- none,  1 - regular, 2 - MODE changed , 3 -- CYCLE , 4 -- OTHER*/
    int len = 36;
    //Время begintime в базе данных должно соответствовать
    //началу измерения, которое всегдя одно, таким образом мы можем
    //отделять одну калибровку от другой    
    //		time_t tm = time_t(fit_param[0].value);//time_t(begin_time);
    time_t tm = time_t(begin_time.tv_sec);
    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 3" << std::endl;
    if (!kdb_write(kdb_conn,TableId, tm, buf,len))
    {
      kdb_close(kdb_conn);
      std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: 4" << std::endl;
      std::string str = "can't insert data into calibration data base table " + std::string(table_name);
      //throw std::runtime_error(str.c_str());
      std::cerr << LOG(0) << str << '\n';
      return false;
    }
    std::cout << LOG(1) << "Inserting done" << std::endl;


    issave = true;

    kdb_close(kdb_conn); //закрыть соединение с сервером, где хрантся база данных

    std::cout<<LOG(4)<<"EnergyCalibrationStorage::Save2DB: FINISH" << std::endl;
    return true;

  }

  void EnergyCalibrationStorage::PostLoad(void)
  {
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::PostLoad: BEGIN" << endl;
    /*  Post load code to calculate some more data */
    FillCalculableData(t.size());
    InitArrayList();
    RemoveDeadVacuumChannels();
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::PostLoad: END" << endl;
  }


  void EnergyCalibrationStorage::RemoveDeadVacuumChannels(void)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::RemoveDeadVacuumChannels" << endl;
    //check the vacuum parameters.
    list < Array * > vlst;
    auto vacuum_size=t.size();
    for(std::map<std::string, Array>::iterator it=Vacuum.begin();it!=Vacuum.end();++it)
    {
      bool iszero=false;
      vacuum_size = std::min(it->second.size(), vacuum_size);
      for(unsigned i=0; i!=it->second.size();++i)
        if(it->second[i].value==0) 
        {
          iszero = true;
          break;
        }
      if(!iszero)
        vlst.push_back(&it->second);
    }
    MediumVacuumPresure.clear();
    for(unsigned i=0;i!=vacuum_size;++i)
    {
      ibn::averager <double> averager;
      for(list<Array*>::iterator it = vlst.begin(); it!=vlst.end(); ++it) averager.add((*(*it))[i].value);
      MediumVacuumPresure.push_back(averager.average(), 0);
    }
  }


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  InitVacuum
   * Description:  Init Vacuum variables with no data. When drawing vacuum data it is
   * usefull to see blanc canvas instead of zero point data to be sure no data present.
   * So correct showing of the empty array is the responsibility of the gui library
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::InitVacuum(void)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::InitVacuum: BEGIN" << endl;
    MediumVacuumPresure.clear();
    Vacuum.clear();
    V4Vacuum v4vacuum;
    for(std::map<std::string, double>::iterator vcm=v4vacuum.P.begin(); vcm!=v4vacuum.P.end();++vcm)
    {
      Vacuum[vcm->first];
    }
    clog << LOG(5) << "EnergyCalibrationStorage::InitVacuum: END" << endl;
  }



  void EnergyCalibrationStorage::FillCalculableData(unsigned point)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::FillCalculableData: BEGIN" << endl;
    if(point>=t.size()) point=t.size()-1;// try to fill point which doesnt exist

    /*  calculate total current */
    Isum.reserve(t.size()); //this is done for fast adding new points
    for(unsigned i=Isum.size(); i<=point; i++)
    {
      data::PhysValue sum(0);
      for(unsigned bunch=0; bunch < numbunch; bunch++)
      {
        if(I[bunch].size() <= i) continue; //skip no bunch data
        sum+=I[bunch][i];
      }
      Isum.push_back(sum);
    }

    FillCorrelatedData(point);
    FillNormalizedData(point);
    clog << LOG(5) << "EnergyCalibrationStorage::FillCalculableData: END" << endl;
  }



  
  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  FillCorrelatedData
   * Description:  Calculate and fill data  for correlated events for single data point.
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::FillCorrelatedData(unsigned point)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::FillCorrelatedData: BEGIN" << endl;
    if(t.size()<=point) point=t.size()-1;
    unsigned max_size = t.size();
    for(unsigned bunch=0; bunch<numbunch; bunch++)
    {
      clog <<LOG(6) <<  "bunch=" << bunch << endl;
      for(unsigned pair=0; pair<numpair; pair++)
      {
        Cor[bunch][pair].reserve(max_size);
        DeltaCor[bunch][pair].reserve(max_size);
        RndCoinc[bunch][pair].reserve(max_size);
        DeltaRndCoinc[bunch][pair].reserve(max_size);
      }
      CorSum[bunch].reserve(max_size);
      DeltaCorSum[bunch].reserve(max_size);
      RndCoincSum[bunch].reserve(max_size);
      DeltaRndCoincSum[bunch].reserve(max_size);
      CorSumNormI2[bunch].reserve(max_size);
      //Prepare for Random events
      for(unsigned counter=0; counter<numcounter; counter++)
      {
        Rnd[bunch][counter].reserve(max_size);
        DeltaRnd[bunch][counter].reserve(max_size);
      }
      RndSum[bunch].reserve(max_size);
      DeltaRndSum[bunch].reserve(max_size);
    }
    double f0=818924; /* revolution frequency  WARING: it should be fixed */
    /*  if point > current correleted array size 
     *  then fill missed data */
		std::clog <<LOG(6) <<  "cor00.size=" << Cor[0][0].size()<<  "  point=" <<  point << std::endl;
    for(unsigned i = Cor[0][0].size(); i<=point; ++i)
    {
      /* fill correlated data for every pair and bunch 
       * we must do it before delta calculation
       * because delta needs data filled for all bunches */
      vector < PhysValue > ncor_sum(numbunch); /* create accumulator for every bunch */
      vector < PhysValue > rnd_sum(numbunch); /* create accumulator for random events  for every bunch */
      vector < PhysValue > rnd_coinc_sum(numbunch); // sum of random coincidence
      for(unsigned bunch=0; bunch < numbunch; bunch++)
      {
        ncor_sum[bunch]=0; //init cor sum
        rnd_sum[bunch]=0;  //init rnd sum
        rnd_coinc_sum[bunch]=0; 
        for(unsigned pair=0; pair < numpair; pair++)
        {
          /*  I need some code to protect when no coincidence data present */
          if(Coinc[bunch][pair].size()<= i || 
              Count[bunch][2*pair].size()<=i || 
              Count[bunch][2*pair+1].size() <= i)
            continue;
					//std::cout << "bunch=" << bunch << " pair=" << pair << std::endl;
          //Fill correlated events
          Cor[bunch][pair].push_back
            (
             GetCorrelatedEvents(f0, CoincLoad[bunch][pair][i], CounterLoad[bunch][2*pair][i], CounterLoad[bunch][2*pair+1][i])
            );
          //Fill random for internal counters
          Rnd[bunch][2*pair].push_back
            (
             GetRandomEvents(f0, CoincLoad[bunch][pair][i], CounterLoad[bunch][2*pair][i], CounterLoad[bunch][2*pair+1][i])
            );
          //Fill random for external counters
          Rnd[bunch][2*pair+1].push_back
            (
             GetRandomEvents(f0, CoincLoad[bunch][pair][i], CounterLoad[bunch][2*pair+1][i], CounterLoad[bunch][2*pair][i])
            );
          RndCoinc[bunch][pair].push_back(Rnd[bunch][2*pair].back()*Rnd[bunch][2*pair+1].back()/cfg.depolarizer.revolution_frequency);
					//std::cout << "bunch=" << bunch << " pair=" << pair << std::endl;
          /*  sum events for all counter */
          ncor_sum[bunch]+=Cor[bunch][pair][i];
          rnd_sum[bunch]+=Rnd[bunch][2*pair].back() + Rnd[bunch][2*pair+1].back();
          rnd_coinc_sum[bunch]+=RndCoinc[bunch][pair].back();
        }
      }
			std::clog << LOG(DEBUG_LOG) << "After  bunch loop " << std::endl;
			std::clog << LOG(DEBUG_LOG) << "norm_bunch.size = " << cfg.bunch_norm.size() << std::endl;

      /*  Now calculate delta */
      for(unsigned bunch=0; bunch < numbunch; bunch++)
      {
			std::clog << LOG(DEBUG_LOG) << " Calculating delta bunch = " << bunch << std::endl;
        if(I[bunch].size() <= i) continue;
				for(int bn=0;bn<cfg.bunch_norm.size(); bn++)
				{
					//std::cout << cfg.bunch_norm[bn] << std::endl;
				}
        unsigned norm_bunch=cfg.bunch_norm[bunch]-1;
        for(unsigned pair=0; pair < numpair; pair++)
        {
					std::clog << LOG(DEBUG_LOG) << "DElta COr=  pair" << pair << std::endl;
					std::clog << LOG(DEBUG_LOG) << "  bunch = " << bunch << " pair = " << pair << "  norm_bunch=" << norm_bunch << std::endl;
					std::clog << LOG(DEBUG_LOG) << "Cor[bunch][pair][i]  = " <<  Cor[bunch][pair][i] << std::endl;
					std::clog << LOG(DEBUG_LOG) << "Cor[norm_bunch][pair][i]  = " <<  Cor[norm_bunch][pair][i] << std::endl;
          DeltaCor[bunch][pair].push_back(GetDelta(Cor[bunch][pair][i],Cor[norm_bunch][pair][i]));
					std::clog << LOG(DEBUG_LOG) << "DEltaRndCoinc=  pair" << pair << std::endl;
          DeltaRndCoinc[bunch][pair].push_back(GetDelta(RndCoinc[bunch][pair][i], RndCoinc[norm_bunch][pair][i]));
        }
        for(unsigned counter=0; counter < numcounter; counter++)
        {
					std::clog << LOG(DEBUG_LOG) << "DEltaRnd=  counter =" << counter << std::endl;
          DeltaRnd[bunch][counter].push_back(GetDelta(Rnd[bunch][counter][i],Rnd[norm_bunch][counter][i]));
        }
				std::clog << LOG(DEBUG_LOG) << "CorSum ="<< std::endl;
        CorSum[bunch].push_back(ncor_sum[bunch]);
				std::clog << LOG(DEBUG_LOG) << "RndSum ="<< std::endl;
        RndSum[bunch].push_back(rnd_sum[bunch]);
				std::clog << LOG(DEBUG_LOG) << "RndCoincSum"<< std::endl;
        RndCoincSum[bunch].push_back(rnd_coinc_sum[bunch]);
				std::clog << LOG(DEBUG_LOG) << "DeltaCorSum"<< std::endl;
        DeltaCorSum[bunch].push_back(GetDelta(ncor_sum[bunch],ncor_sum[norm_bunch]));
				std::clog << LOG(DEBUG_LOG) << "DeltaRndSum"<< std::endl;
        DeltaRndSum[bunch].push_back(GetDelta(rnd_sum[bunch],rnd_sum[norm_bunch]));
				std::clog << LOG(DEBUG_LOG) << "DeltaRndCoincSum"<< std::endl;
        DeltaRndCoincSum[bunch].push_back(GetDelta(rnd_coinc_sum[bunch], rnd_coinc_sum[norm_bunch]));
				std::clog << LOG(DEBUG_LOG) << "CorSumNormI2"<< std::endl;
        if(I[bunch][i].value!=0)
          CorSumNormI2[bunch].push_back(CorSum[bunch][i]/((I[bunch][i]*I[bunch][i])*1e-6));
        else
          CorSumNormI2[bunch].push_back(0);
      }
    }
    clog << LOG(5) << "EnergyCalibrationStorage::FillCorrelatedData: END" << endl;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  FillNormalizedData
   * Description:  Fill normalized data
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::FillNormalizedData(unsigned point)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::FillNormalizedData: BEGIN" << endl;
    if(t.size()<=point) point=t.size()-1; 
    /*  if point > current correleted array size 
     *  then fill missed data */
    for(unsigned bunch=0; bunch < numbunch; bunch++)
    {
      CountSumNormI[bunch].reserve(t.size());
      CoincSumNormI[bunch].reserve(t.size());
      CorSumNormI[bunch].reserve(t.size());
      RndSumNormI[bunch].reserve(t.size());
      RndCoincSumNormI[bunch].reserve(t.size());

      IRatio[bunch].reserve(t.size());
      CountSumNormI2[bunch].reserve(t.size());
      CoincSumNormI2[bunch].reserve(t.size());
      CorSumNormI2[bunch].reserve(t.size());
    }
    for(unsigned i = CountSumNormI[0].size(); i<=point; i++)
    {
      for(unsigned bunch=0; bunch < numbunch; bunch++)
      {
        if(I[bunch].size() <= i) continue;
        if(I[bunch][i].value != 0 ) 
        {
          CountSumNormI[bunch].push_back(CountSum[bunch][i]/I[bunch][i]*1e3);
          CoincSumNormI[bunch].push_back (CoincSum [bunch][i]/I[bunch][i]*1e3);
          CorSumNormI[bunch].push_back  (CorSum [bunch][i]/I[bunch][i]*1e3);
          RndSumNormI[bunch].push_back  (RndSum [bunch][i]/I[bunch][i]*1e3);
          RndCoincSumNormI[bunch].push_back  (RndCoincSum [bunch][i]/I[bunch][i]*1e3);
        }
        else 
        {
          CountSumNormI[bunch].push_back(0);
          CoincSumNormI[bunch].push_back(0);
          CorSumNormI[bunch].push_back(0);
          RndSumNormI[bunch].push_back(0);
          RndCoincSumNormI[bunch].push_back(0);
        }
      }
    }
    for(unsigned i = CountSumNormI2[0].size(); i<=point; i++)
    {
      for(unsigned bunch=0; bunch < numbunch; bunch++)
      {
        if(I[bunch].size() <= i) continue;
        data::PhysValue Isum=0;
        for(unsigned b=0; b<numbunch; b++)
        {
          if(b!=bunch && I[b].size()>i)
            Isum=Isum+I[b][i];
        }
        if(I[bunch][i].value != 0 ) 
        {
          IRatio[bunch].push_back(Isum/I[bunch][i]);
          CountSumNormI2[bunch].push_back(CountSum[bunch][i]/(I[bunch][i]*I[bunch][i]*1e-6));
          CoincSumNormI2[bunch].push_back(CoincSum[bunch][i]/(I[bunch][i]*I[bunch][i]*1e-6));
          CorSumNormI2[bunch].push_back(CorSum[bunch][i]/(I[bunch][i]*I[bunch][i]*1e-6));
        }
        else 
        {
          IRatio[bunch].push_back(0);
          CountSumNormI2[bunch].push_back(0);
          CoincSumNormI2[bunch].push_back(0);
          CorSumNormI2[bunch].push_back(0);
        }
      }
    }
    clog << LOG(5) << "EnergyCalibrationStorage::FillNormalizedData: END" << endl;
  }


  void EnergyCalibrationStorage::ReconsructLoad(void)
  {
    double T = cfg.count_time;
    double N0 = T*818924;
    for(unsigned bunch=0;bunch<numbunch;++bunch)
    {
      for(unsigned counter=0; counter < numcounter; ++counter)
      {
        CounterLoad[bunch][counter].resize(Count[bunch][counter].size());
        for(unsigned point=0; point<CounterLoad[bunch][counter].size(); ++point)
        {
          double N = get_count_prob(Count[bunch][counter][point]*T,N0).value*N0;
          CounterLoad[bunch][counter][point] = get_load(N0,N)/T;
        }
      }
      for(unsigned pair=0; pair < numpair; ++pair)
      {
        CoincLoad[bunch][pair].resize(Coinc[bunch][pair].size());
        for(unsigned point=0; point<CoincLoad[bunch][pair].size(); ++point)
        {
          double N = get_count_prob(Coinc[bunch][pair][point]*T,N0).value*N0;
          CoincLoad[bunch][pair][point] = get_load(N0,N)/T;
        }
      }
    }
  }

  bool EnergyCalibrationStorage::LoadFormat0(std::string file_name)
  {
    clog << LOG(4) << "EnergyCalibrationStorage::LoadFormat0" << std::endl;
    ifstream file(file_name.c_str());
    if(!file)
    {
      cerr << LOG(1) << "Unable to open file " << file_name << "" << std::endl;
    }
    return LoadFormat0(file);
  }

  bool EnergyCalibrationStorage::LoadFormat0(std::istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    char buf[1024];
    file.getline(buf,1024,'\n');
		int version =0;
		cfg.run = 0;
    if(string(buf)=="# Measurement with two counters")
    {
      clog << LOG(4) << "Format0 detected" << std::endl;
      string tmpstring;
      struct tm tm;
      file.getline(buf,1024,'\n'); /*  read the time of energy calibration */
			clog << LOG(4) << buf << endl;
      strptime(buf,"# %c",&tm);
      clog << LOG(4) << 1900+tm.tm_year << " " << tm.tm_mon << " " << tm.tm_mday << " " << tm.tm_hour << " " << tm.tm_min << " " << tm.tm_sec << endl;
			struct tm tm2;
			tm2.tm_year=101;
			tm2.tm_mon = 6;
			tm2.tm_mday= 5;
			tm2.tm_hour = 4;
			tm2.tm_min = 10;
			tm2.tm_sec = 33;
			time_t t2 = mktime(&tm2);
			time_t t1 = mktime(&tm);
			cfg.date = t1;
			cfg.name = "version0";
			version = t1 >= t2 ? 1 : 0;
			cout << "version = " << version << endl;
      N=0;
      reserve(1000); /*  reserve place*/
      char  ch;
      cfg.count_time=1; /*  default count time */
      //bool is_scan_time_calculated=false;
      int depolarizer_on = 0;
      while(file>>ch)
      {
        //file >> ch; 
        if(ch=='#') 
        {
          file.ignore(65535,'\n');  /*  skip comment string */
          continue;
        }
        //if(file.eof()) break;
        else 
        {
          file.putback(ch);
          push_back(); /*  initialize  */
          file >> t.back().value;
          file >> DepFr.back().value;
					cout << N-1 << "  t=" << t.back().value << "  f= " << DepFr.back().value << endl;
          if(fabs(DepFr.back().value)>100 && depolarizer_on<3) 
          {
            depolarizer_on++;
            if(depolarizer_on==2) /*  calculate depolarizer parameters */
            {
              cfg.depolarizer.step = DepFr.back().value - DepFr[t.size()-2].value;
							cout << "cfg.depolariser.step = " << cfg.depolarizer.step << endl << "fr0 = " <<  DepFr[t.size()-2].value << endl;
              cfg.depolarizer.scan_time = t.back().value - t[t.size()-2].value;
              cfg.depolarizer.speed = cfg.depolarizer.step / cfg.depolarizer.scan_time;
            }
          }
					if(version==1) 
					{
						file >> DeltaSum[0].back();
					}
          file >> I[0].back().value;
          file >> I[1].back().value;
					cout << I[0].back().value << " I1 = " << I[1].back().value << endl;
          if(t.size()==2) /*  calculate count time */
          {
            cfg.count_time = t[1].value - t[0].value;
            CoincSum[0][0].error = sqrt(CoincSum[0][0].value/cfg.count_time); 
            CoincSum[1][0].error = sqrt(CoincSum[0][0].value/cfg.count_time); 
            CorSum[0][0].error = sqrt(CoincSum[0][0].value/cfg.count_time); 
            CorSum[1][0].error = sqrt(CoincSum[0][0].value/cfg.count_time); 
          }
          file >> CoincSum[0].back().value; /*  Tsh1 */
          CoincSum[0].back().error = sqrt(CoincSum[0].back().value/cfg.count_time); 
          file >> CoincSum[1].back().value; /*  Tsh2 */
          CoincSum[1].back().error = sqrt(CoincSum[1].back().value/cfg.count_time); 

					if(version==0) file >> DeltaSum[0].back();


          file >> CorSum[0].back().value; /*  Rndm1 */
          CorSum[0].back().error = sqrt(CorSum[0].back().value/cfg.count_time); 
          file >> CorSum[1].back().value; /*  Rndm2 */
          CorSum[1].back().error = sqrt(CorSum[1].back().value/cfg.count_time); 

					cout << "Tsh0 = " << CoincSum[0].back().value << "  Tsh1 = " << CoincSum[1].back().value << endl;

          file.ignore(65535,'\n');
        }
      }
			cfg.depolarizer.revolution_frequency=818924;
			cfg.depolarizer.RF=181801160;
			cfg.depolarizer.harmonic_number=4;
			//modify the Depolarizer frequency
			for(unsigned i=0;i<t.size();i++)
			{
          if(DepFr[i].value < 600000 && DepFr[i].value > 300000)
          {
            DepFr[i].value = - DepFr[i].value; /*  1500 MeV range */
            cfg.depolarizer.harmonic_number=4; 
          }
          if(fabs(DepFr[i].value)<10) DepFr[i].value=0;
			}

			cout << cfg << endl;
      clog << LOG(2) << "Successfully read format 0" << std::endl;
      InitVacuum();
			ShiftTime();
      return true;
    }
    else 
    {
      return false;
    }
  }

  bool EnergyCalibrationStorage::LoadFormat1(std::string file_name)
  {
    clog << LOG(4) << "EnergyCalibrationStorage::LoadFormat2" << std::endl;
    ifstream file(file_name.c_str());
    if(!file)
    {
      cerr << LOG(1) << "Unable to open file " << file_name << "" << std::endl;
    }
    return LoadFormat1(file);
  }


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  parse_line
   *  Description:  Get string and extract key and value data divided by delimiter
   * =====================================================================================
   */
  template <class T>
  void parse_line(string line,string & key, T & value, char delim=':' )
  {
    size_t delim_pos = line.find(delim);
    if(delim_pos == string::npos) return;
    key = line.substr(0,delim_pos);
    string value_str = line.substr(delim_pos+1);
    //value_str.erase(std::remove(value_str.begin(),value_str.end(),' '),value_str.end());
    istringstream is(value_str);
    is >> value;
  }

  bool EnergyCalibrationStorage::LoadFormat01(istream & file)
  {
    file.seekg(ios_base::beg); //start at the begining of file
    string head_line;
    string date_line;
    string mode_line;
    string key;
    string line;
    double width; /*  for frequency width */
    double modulation_frequency; // modulation frequency , Hz now doesn t use
    getline(file,head_line);
    if(head_line=="# Polarization Measurement")
      clog << LOG(3) << "Gues format 01 detected. Continue analizing" << endl;
    getline(file,date_line);
    getline(file,mode_line);
    if(mode_line.find("Mode")==mode_line.npos) 
      return false;
    tm tt;
    strptime(date_line.c_str(),"%c",&tt);
    begin_time.tv_sec = mktime(&tt);
    begin_time.tv_usec=0;
    parse_line(mode_line,key,mode);
    getline(file,line); parse_line(line, key, setE);
    getline(file,line); parse_line(line, key, cfg.count_time);
    getline(file,line); parse_line(line, key, cfg.depolarizer.scan_time);
    getline(file,line); parse_line(line, key, cfg.depolarizer.initial);
    getline(file,line); parse_line(line, key, cfg.depolarizer.final);
    getline(file,line); parse_line(line, key, cfg.depolarizer.step);
    getline(file,line); parse_line(line, key, width);
    getline(file,line); parse_line(line, key, modulation_frequency);
    getline(file,line); //just skip nmr header
    cfg.depolarizer.revolution_frequency=818924;
    cfg.depolarizer.RF=181801160;
    cfg.depolarizer.harmonic_number=4;
    cfg.depolarizer.speed = cfg.depolarizer.step/cfg.depolarizer.scan_time;
    measurement::EnergyCalibrationConfig mcfg;
    data::frequency2energy(&cfg,&mcfg);
    cfg = mcfg;
    char tmpchar;
    file >> tmpchar; // eat the #
    nmr.resize(7);
    for(unsigned n=0;n<nmr.size();++n)
      file >> nmr[n];
    getline(file,line); //skip data header
    char ch;
    while(file>>ch)
    {
      if(ch=='#') 
      {
        file.ignore(65535,'\n');  /*  skip comment string */
        continue;
      }
      else 
      {
        file.putback(ch);
        push_back(); /*  initialize  */
        file >> t.back().value;
        file >> DepFr.back().value;
        file >> DeltaSum[0].back();
        file >> I[0].back().value;
        file >> I[1].back().value;
        file >> CoincSum[0].back().value; /*  Tsh1 */
        CoincSum[0].back().error = sqrt(CoincSum[0].back().value/cfg.count_time); 
        file >> CoincSum[1].back().value; /*  Tsh2 */
        CoincSum[1].back().error = sqrt(CoincSum[1].back().value/cfg.count_time); 
        file >> CorSum[0].back().value; /*  Rndm1 */
        CorSum[0].back().error = sqrt(CorSum[0].back().value/cfg.count_time); 
        file >> CorSum[1].back().value; /*  Rndm2 */
        CorSum[1].back().error = sqrt(CorSum[1].back().value/cfg.count_time);
        for(unsigned counter=0; counter <4; ++counter)
        {
          ibn::valer<data::Array::Value_t> data = Count[0][counter].back();
          file >> data.value;
          data.error = sqrt(data.value/cfg.count_time);
        }
        if(DepFr.back().value < 600000 && DepFr.back().value > 300000)
        {
          DepFr.back().value = -DepFr.back().value; /*  1500 MeV range */
        }
        if(fabs(DepFr.back().value)<10)
          DepFr.back().value=0;
        if(fabs(DeltaSum[0].back().value) == fabs(DeltaSum[0].back().error))
        {
          DeltaSum[0].back()= GetDelta(CoincSum[0].back(), CoincSum[1].back());
        }
        file.ignore(65535,'\n');
      }
    }
    /* say to ::Load that we dont know the run number and loader 
     * will try to calculate it from filename
     */
    //cfg.run=static_cast<unsigned>(-1);     
    cfg.run=0;     
		ShiftTime();
    return true;
  }

  bool EnergyCalibrationStorage::LoadFormat1(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    string head_line, buf_line;
    getline(file,head_line);
    getline(file,buf_line);
    if(head_line=="# Polarization Measurement" && buf_line.find("Run")!=buf_line.npos)
    {
      clog << LOG(1) << "Format 1 detected" << std::endl;
      string tmpstring;
      double data1, data2;
      //file.ignore(655536,'\n');
      file.ignore(37);
      //cout << tm.tm_year << " " << tm.tm_mon << " " << tm.tm_mday << " " << tm.tm_hour << " " << tm.tm_min << " " << tm.tm_sec << endl;
      file >> cfg.run;
      file >> tmpstring;
      file >> setE;
      file >> tmpstring;
      success = tmpstring == "success" ? 1 : 0;
      file >> tmpstring;  /*  read direction of scanning */
      file >> data1 >> data2; /*  read measured energy */
      E = PhysValue(data1,data2); /*  set up measured energy */
      file >> data1 >> data2;  /*  read frequency */
      Fr = PhysValue(data1,data2); /*  set up measured frequency */
      file >> data1; /*  skip calib option */
      /*  read fit param */
      fit_param.resize(0);
      file >> data1 >> data2; /*  read dep time, sec*/
      fit_param.push_back(PhysValue(data1,data2));
      file >> data1 >> data2; /*  delta*/
      fit_param.push_back(PhysValue(data1,data2));
      file >> data1 >> data2; /*  const*/
      fit_param.push_back(PhysValue(data1,data2));
      file >> data1 >> data2; /*  k1*/
      fit_param.push_back(PhysValue(data1,data2));
      file >> data1 >> data2; /*  k2*/
      fit_param.push_back(PhysValue(data1,data2));
      file >> chi2; /*  p chi2 */
      /*  read measurement parameters */
      file >> cfg.count_time; /*  read time of one measurement */
      file >> cfg.depolarizer.scan_time; /*  read time of one measurement */
      file >> cfg.depolarizer.RF; /*  rf frequency */
      file >> cfg.depolarizer.initial; /*  initial frequency */
      file >> cfg.depolarizer.final; /*  initial frequency */
      file >> cfg.depolarizer.step; /* frequency step */
      cfg.depolarizer.speed = cfg.depolarizer.step /cfg.depolarizer.scan_time; /*  calculate frequency scan speed. */
      file >> data1; /*  read frequency width of depolarizer */
      file >> data1; /*  read modulation frequency */
      file >> cfg.depolarizer.attenuation; /*  read attenuation */
      file >> data1; /* read unixtime */
      begin_time.tv_sec = (time_t)data1;
      for(unsigned i = 0 ;i < 5; i++) file >> nmr[i];
      clog << LOG(2) << "Run " << cfg.run << "" << std::endl;
      clog << LOG(2) << "time " << begin_time.tv_sec << " (" << ctime(&begin_time.tv_sec)<< ")" << std::endl;
      clog << LOG(2) << "Energy: " << E << "" << std::endl;
      clog << LOG(2) << "delta: " << fit_param[1] << "" << std::endl;
      resize(0); /*  clear arrays */
      file.ignore(65535,'\n'); /*  skip empty string */
      N=0;
      char ch;
      while(file>>ch)
      {
        //file >> ch; 
        if(ch=='#') 
        {
          file.ignore(65535,'\n');  /*  skip comment string */
          continue;
        }
        //if(file.eof()) break;
        else 
        {
          file.putback(ch);
          file >> data1; /*  read time */
          t.push_back(data1,0);
          file >> data1; /*  frequency,  */
          DepFr.push_back(-data1,0);
          file >> data1; /*  RF  */
          RF.push_back(data1,0);
          file >> data1 >> data2; /*  delta */
          DeltaSum[0].push_back(data1,data2);
          file >> data1 >> data2; /*  current 1 */
          I[0].push_back(data1,data2);
          file >> data1 >> data2; /*  current 2 */
          I[1].push_back(data1,data2);
          file >> data1 >> data2; /*  coincidence (Touschek) */
          CoincSum[0].push_back(data1,sqrt(data1/cfg.count_time)); 
          CoincSum[1].push_back(data2,sqrt(data2/cfg.count_time)); 
          file >> data1 >> data2; /*  random coincidece */
          CorSum[0].push_back(data1,sqrt(data1/cfg.count_time)); 
          CorSum[1].push_back(data2,sqrt(data2/cfg.count_time)); 
          file >> data1 >> data2; /*  Out12 */
          Count[0][1].push_back(data1,sqrt(data1/cfg.count_time));
          Count[1][1].push_back(data2,sqrt(data2/cfg.count_time));
          file >> data1 >> data2; /*  In12 */
          Count[0][0].push_back(data1,sqrt(data1/cfg.count_time));
          Count[1][0].push_back(data2,sqrt(data2/cfg.count_time));
          CountSum[0].push_back(Count[0][0].back()+Count[0][1].back());
          CountSum[1].push_back(Count[1][0].back()+Count[1][1].back());
          //NMR.resize(5);
          for(unsigned i =0 ;i<5;i++)
          {
            file >> data1;
            NMR[i].push_back(data1);
          }
          cfg.depolarizer.revolution_frequency=818924;
          cfg.depolarizer.harmonic_number=5;
          if(E.value > 1700 && E.value <1900)
            cfg.depolarizer.harmonic_number=5;
          if(E.value < 1700 )
            cfg.depolarizer.harmonic_number=4;
          N++;
        }
      }
      cfg.depolarizer.speed = cfg.depolarizer.step/cfg.depolarizer.scan_time;
      measurement::EnergyCalibrationConfig mcfg;
      data::frequency2energy(&cfg,&mcfg);
      cfg = mcfg;
      InitVacuum();
			ShiftTime();
      clog << LOG(2) << "Successfully read first data format" << std::endl;
      return true;
    }
    else 
    {
      return false;
    }
  }


  bool EnergyCalibrationStorage::ParseHeader(std::map<std::string,std::string> & vm)
  {
    map<string,string>::iterator it;
    try 
    {
      //double VERSION=0;
      //if((it=vm.find("VERSION"))!=vm.end()) VERSION = std::stod(it->second);
      //else clog << LOG(2) << "Unable to find VERSION" << endl;
      clog << LOG(5) << "Looking for measurement date" << endl;
      if((it=vm.find("Measurement date"))!=vm.end())
      {
        cout << "\""<<it->second<<"\""<< endl;
        tm t;
        strptime(it->second.c_str(),"%FT%T",&t);
        begin_time.tv_sec = mktime(&t);
        begin_time.tv_usec=0;
      }
      else clog << LOG(2) << "Measurement date" << endl;
      clog << LOG(5) << "Looking for Unixtime" << endl;
      time_t unixtime;
      if((it=vm.find("Unixtime"))!=vm.end()) unixtime = stol(it->second);
      else clog << LOG(2) << "Unable to find Unixtime" << endl;
      cout <<  begin_time.tv_sec << "  unixtie= " << unixtime << endl;
      if((it=vm.find("Run"))!=vm.end()) cfg.run = stoi(it->second);
      else clog << LOG(2) << "Unable to find Run" << endl;
      if((it=vm.find("Mode"))!=vm.end()) mode = it->second;
      else clog << LOG(2)<< "Unable to find Mode" << endl;
      if((it=vm.find("Set energy, Mev"))!=vm.end()) setE = std::stod(it->second);
      else clog <<LOG(2)<< "Unable to find Set energy" << endl;
      if((it=vm.find("Calibration status"))!=vm.end()) success = (it->second) == "success" ? true : false ;
      else clog << LOG(2) << "Unable to find Calibration status" << endl;
      if((it=vm.find("Measured energy, Mev"))!=vm.end()) cfg.energy = to_phys_value(it->second);
      else clog << LOG(2) << "Unable to find Measured energy" << endl;
      cout << "fit param " << endl;
      if((it=vm.find("Number of fit parameters"))!=vm.end()) fit_param.resize(std::stoi(it->second));
      else clog <<LOG(2)<< "Unable to find fit parameters number" << endl;
      for(unsigned i=0; i<fit_param.size(); i++ )
      {
        string parname =" p"+std::to_string(i); 
        if((it=vm.find(parname))!=vm.end()) fit_param[i] = to_phys_value(it->second);
        else clog << LOG(2) << "Unable to find fit parameter " << parname << endl;
      }
      if((it=vm.find("Chi2 "))!=vm.end()) chi2 = std::stod(it->second);
      else clog << LOG(2) <<  "Unable to find  Chi2."  << endl;
      if((it=vm.find("Count time (s/point)"))!=vm.end()) cfg.count_time = std::stod(it->second);
      else clog << LOG(2) << "Unable to find  Count time."<< endl;
      if((it=vm.find("Scan time (s/step)"))!=vm.end()) cfg.depolarizer.scan_time = std::stod(it->second);
      else clog << LOG(2) << "Unable to find  Scan time "  << endl;
      cfg.depolarizer.revolution_frequency = 818924;
      if((it=vm.find("VEPP revolution frequency, Hz"))!=vm.end()) cfg.depolarizer.revolution_frequency.value = std::stod(it->second);
      else clog << LOG(2)<< "Unable to find  revolution frequency." << endl;
      if((it=vm.find("Frequency step,   Hz"))!=vm.end()) cfg.depolarizer.step = std::stod(it->second);
      else clog << LOG(2)<< "Unable to find  Frequency step" << endl;
      if((it=vm.find("Attenuation, dB"))!=vm.end()) cfg.depolarizer.attenuation = std::stod(it->second);
      else clog << LOG(2)<< "Unable to find  Attenuation" << endl;
      if((it=vm.find("Frequency range,  Hz"))!=vm.end()) 
      {
        istringstream is(it->second);
        is >> cfg.depolarizer.initial;
        char c;
        is >> c; //skip minus
        is >> cfg.depolarizer.final;
      }
      else 
      {
        clog << LOG(2)<< "Unable to find  Frequency step" << endl;
        if((it=vm.find("Initial frequency,  Hz"))!=vm.end())
          cfg.depolarizer.initial = std::stod(it->second);
        if((it=vm.find("Final frequency, Hz"))!=vm.end())
          cfg.depolarizer.final = std::stod(it->second);
      }
      cfg.depolarizer.harmonic_number = 4;
      if((it=vm.find("harmonic_number"))!=vm.end()) 
        cfg.depolarizer.harmonic_number = std::stoi(it->second);
      else clog<< LOG(2) << "Unable to find  number of points" << endl;
      cfg.depolarizer.speed = cfg.depolarizer.step/cfg.depolarizer.scan_time;

      if((it=vm.find("Number of NMR channels"))!=vm.end())
      {
        nmr.resize(std::stoi(it->second));
        std::istringstream iss(vm["nmr"]);
        for(int i=0;i<nmr.size();i++)
        {
          iss >> nmr[i];
        }
        NMR.resize(nmr.size());
      }
      else clog<< LOG(2) << "Unable to find  nmr channels number" << endl;

      if((it=vm.find("Number of points"))!=vm.end()) N = std::stoi(it->second);
      else clog<< LOG(2) << "Unable to find  number of points" << endl;

      clog << LOG(1) << cfg << endl;
    }
		catch(...)
		{
			std::cerr << LOG(0) << "ERROR:  Unknown exception in EnergyCalibrationStorage::ParseHeader(std::map<std::string,std::string> & vm)" << std::endl;
		}
    return true;
  }


  void fill_variable_map(istream & is, std::map<std::string, std::string> & vm, char delim=':', std::string break_line="Number of points")
  {
    string line;
    //std::regex nmr_re(R"()")
    while (getline(is,line))
    {
      size_t delim_pos = line.find(delim);
      if(delim_pos == string::npos) continue; //not found the key and value
      string key = line.substr(0,delim_pos);
      string value = line.substr(delim_pos+1); //from delim_pos to end;
      value.erase(std::remove(value.begin(),value.end(),' '),value.end());
      vm[key]=value;
      clog << LOG(5) << key << "********"<< value << endl;
      if(line.find("Number of NMR channels")!=string::npos)
      {
        getline(is,line); //skip commented code
        getline(is,line); //here is nmr data
        vm["nmr"] = line;
      }
      if(line.find(break_line)!=string::npos) break; //read file until found break line
    }
  }


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  extract_data
   *  Description:  Extract the data from string line "<some description> : <data>"
   *  or from line "<data>"
   * =====================================================================================
   */
  template <class T> void extract_data(string line, T & value, char delim=':')
  {
      size_t delim_pos = line.find(delim);
      string data_string;
      data_string = line.substr(0,delim_pos);
      if(delim_pos != string::npos) string value = line.substr(delim_pos+1); 
      istringstream is(data_string);
      is >> value;
      return;
  }

  bool EnergyCalibrationStorage::LoadFormat32(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    file.ignore(65535,'\n'); // ignore two header
    file.ignore(65535,'\n');
    string line;
    getline(file,line);
    if(line.find("VERSION")==string::npos || line.find("3.2") == string::npos)
      return false;
    double version=0;
    istringstream tmpis(line.substr(line.find(":")+1));
    tmpis >> version;
    clog << LOG(1) << "Version 3.2 format detected" << endl;
    map <string, string> vm; /*  map with headers */
    fill_variable_map(file,vm);
    if(!ParseHeader(vm)) return false;
    measurement::EnergyCalibrationConfig cfg_with_MeV;
    cfg.depolarizer.harmonic_number=5;
    data::frequency2energy(&cfg, &cfg_with_MeV);
    cfg = cfg_with_MeV;
    char ch;
    unsigned pnt=0;
    while(file>>ch)
    {
      if(ch=='#') 
      {
        file.ignore(65535,'\n');  /*  skip comment string */
        continue;
      }
      else 
      {
        file.putback(ch);
        //init the data
        t.push_back(0);
        DepFr.push_back(0);//  frequency 
        RF.push_back(0);
        for(unsigned bunch=0; bunch <numbunch; ++bunch)
        {
          I[bunch].push_back(0);
          SigmaX[bunch].push_back(0);
          SigmaZ[bunch].push_back(0);
          SigmaL[bunch].push_back(0);
          for(unsigned counter=0; counter < numcounter; counter++)
          {
            Count[bunch][counter].push_back(0);
            CounterLoad[bunch][counter].push_back(0);
            DeltaCount[bunch][counter].push_back(0);
          }
          CountSum[bunch].push_back(0);
          for(unsigned pair=0; pair < numpair; pair++)
          {
            Coinc[bunch][pair].push_back(0);
            CoincLoad[bunch][pair].push_back(0);
            DeltaCoinc[bunch][pair].push_back(0);
          }
          CoincSum[bunch].push_back(0);
          DeltaCoincSum[bunch].push_back(0);
          DeltaSum[bunch].push_back(0);
        }
        data::PhysValue data;
        file >> t[pnt];  
        file >> DepFr[pnt]; 
        DepFr[pnt].value = -DepFr[pnt].value;
        file >> RF[pnt];
        data::PhysValue Tsh1, Tsh2, Rnd1, Rnd2, Out1, Out2, In1, In2, sigmax, sigmaz, sigmal;
        file >> DeltaCoincSum[0][pnt] >> I[0][pnt] >> I[1][pnt] >> Tsh1 >> Tsh2 >> Rnd1 >> Rnd2 >> Out1 >> Out2 >> In1 >> In2 >> sigmax >> sigmaz >> sigmal;
        Count[0][0][pnt]=In1;
        Count[1][0][pnt]=In2;
        Count[0][1][pnt]=Out1;
        Count[0][1][pnt]=Out2;
        CountSum[0][pnt]=In1+Out1;
        CountSum[1][pnt]=In2+Out2;
        DeltaSum[0][pnt]=data::GetDelta(CountSum[0][pnt], CountSum[1][pnt]);
        DeltaSum[1][pnt]=data::GetDelta(CountSum[1][pnt], CountSum[0][pnt]);
        pnt++; //increment the point.
      }
    }
		InitVacuum();
		ShiftTime();
    return true;
  }

  bool EnergyCalibrationStorage::LoadFormat34(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    file.ignore(65535,'\n'); // ignore two header
    file.ignore(65535,'\n');
    string line;
    getline(file,line);
    if(line.find("VERSION")==string::npos || line.find("3.4") == string::npos)
      return false;
    double version=0;
    istringstream tmpis(line.substr(line.find(":")+1));
    tmpis >> version;
    clog << LOG(1) << "Version 3.4 format detected" << endl;
    map <string, string> vm; /*  map with headers */
    fill_variable_map(file,vm);
    if(!ParseHeader(vm)) return false;
    measurement::EnergyCalibrationConfig cfg_with_MeV;
    data::frequency2energy(&cfg, &cfg_with_MeV);
    cfg = cfg_with_MeV;
    char ch;
    while(file>>ch)
    {
      if(ch=='#') 
      {
        file.ignore(65535,'\n');  /*  skip comment string */
        continue;
      }
      else 
      {
        file.putback(ch);
        data::PhysValue data;
        file >> data; t.push_back(data);//read time with errors 
        file >> data; DepFr.push_back(data);//  frequency 
        file >> data;  RF.push_back(data);
        for(int bunch=0; bunch<3; bunch++)
        {
          file >> data; I[bunch].push_back(data); /*  current */
          file >> data; DeltaSum[bunch].push_back(data);//delta
          file >> data; Count[bunch][1].push_back(data);//external counter
          CounterLoad[bunch][1].push_back(data);
          file >> data; Count[bunch][0].push_back(data);//internal counter
          CounterLoad[bunch][0].push_back(data);
          for(unsigned counter=2; counter < numcounter; counter++)
          {
            Count[bunch][counter].push_back(0);
            CounterLoad[bunch][counter].push_back(0);
          }
          CountSum[bunch].push_back(Count[bunch][0].back() + Count[bunch][1].back());
        }
        for(unsigned bunch=3; bunch < numbunch; ++bunch)
        {
          I[bunch].push_back(0); /*  current */
          DeltaSum[bunch].push_back(0);//delta
          for(unsigned counter=0; counter < numcounter; counter++)
          {
            Count[bunch][counter].push_back(0);
            CounterLoad[bunch][counter].push_back(0);
          }
          CountSum[bunch].push_back(0);
        }
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaX[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaZ[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaL[bunch].push_back(data);

        for(unsigned bunch =0; bunch < numbunch; bunch++)
        {
          // fill deltacount
          for(unsigned counter=0;counter<numcounter;++counter)
          {
            DeltaCount[bunch][counter].push_back(0);
          }
          // fill coincidence
          for(unsigned pair=0; pair<numpair; ++pair)
          {
            Coinc[bunch][pair].push_back(0);
            CoincLoad[bunch][pair].push_back(0);
          }

          CoincSum[bunch].push_back(0);
          DeltaCoincSum[bunch].push_back(0);
        }
      }
    }
		InitVacuum();
		ShiftTime();
    return true;
  }

  bool EnergyCalibrationStorage::LoadFormat4(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    file.ignore(65535,'\n'); // ignore two header
    file.ignore(65535,'\n');
    string line;
    getline(file, line);
    if(line.find("VERSION")==line.npos || line.find("4.0")==line.npos) 
      return false;
    clog << LOG(1) << "Detect format 4.0 " << endl;
    map <string, string> vm; /*  map with headers */
    fill_variable_map(file,vm);
    if(!ParseHeader(vm)) return false;
    char ch;
    while(file>>ch)
    {
      if(ch=='#') 
      {
        file.ignore(65535,'\n');  /*  skip comment string */
        continue;
      }
      else 
      {
        file.putback(ch);
        data::PhysValue data;
        file >> data; t.push_back(data);//read time with errors 
        clog << LOG(3) << data << endl;
        file >> data; DepFr.push_back(data);//  frequency 
        file >> data;  RF.push_back(data);
        for(unsigned bunch=0; bunch<numbunch; bunch++)
        {
          file >> data; I[bunch].push_back(data); /*  current */
          file >> data; DeltaSum[bunch].push_back(data);//delta
          file >> data; Count[bunch][1].push_back(data);//external counter
          CounterLoad[bunch][1].push_back(data);
          file >> data; Count[bunch][0].push_back(data);//internal counter
          CounterLoad[bunch][0].push_back(data);
          file >> data; Coinc[bunch][0].push_back(data);//coincidence
          CoincLoad[bunch][0].push_back(data);
          for(unsigned counter=2; counter < numcounter; ++counter)
          {
            Count[bunch][counter].push_back(0);
            CounterLoad[bunch][counter].push_back(0);
          }
          for(unsigned pair=1; pair < numpair; ++pair)
          {
            Coinc[bunch][pair].push_back(0);
            CoincLoad[bunch][pair].push_back(0);
          }
          data::PhysValue sum=0;
          for(unsigned counter=0;counter<numcounter;++counter) sum += Count[bunch][counter].back();
          CountSum[bunch].push_back(sum);
          sum=0;
          for(unsigned pair=0;pair<numpair;++pair) sum += Coinc[bunch][pair].back();
          CoincSum[bunch].push_back(sum);
        }
        //now calculate the delta for the coincidence
        for(unsigned bunch=0;bunch<numbunch; bunch++)
        {
          unsigned nb=cfg.bunch_norm[bunch]-1; //norm bunch
          DeltaCoincSum[bunch].push_back(data::GetDelta(CoincSum[bunch].back(), CoincSum[nb].back()));
        }
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaX[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaZ[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaL[bunch].push_back(data);
      }
    }
    InitVacuum();
		ShiftTime();
		MDepFr=DepFr;
    isfit = true;
    return true;
  }

  bool EnergyCalibrationStorage::LoadFormat41(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    file.ignore(65535,'\n'); // ignore two header
    file.ignore(65535,'\n');
    string line;
    getline(file, line);
    if(line.find("VERSION")==line.npos || line.find("4.1")==line.npos) 
      return false;
    clog << LOG(1) << "Detect format 4.1 " << endl;
    map <string, string> vm; /*  map with headers */
    fill_variable_map(file,vm);
    if(!ParseHeader(vm)) return false;
    char ch;
    while(file>>ch)
    {
      if(ch=='#') 
      {
        file.ignore(65535,'\n');  /*  skip comment string */
        continue;
      }
      else 
      {
        file.putback(ch);
        data::PhysValue data;
        file >> data; t.push_back(data);//read time with errors 
        file >> data; DepFr.push_back(data);//  frequency 
        file >> data;  RF.push_back(data);
        for(unsigned bunch=0; bunch<numbunch; bunch++)
        {
          file >> data; I[bunch].push_back(data); /*  current */
          file >> data; DeltaSum[bunch].push_back(data);//delta
          file >> data; Count[bunch][1].push_back(data);//external counter
          CounterLoad[bunch][1].push_back(data);
          file >> data; Count[bunch][0].push_back(data);//internal counter
          CounterLoad[bunch][0].push_back(data);
          file >> data; Coinc[bunch][0].push_back(data);//coincidence
          CoincLoad[bunch][0].push_back(data);
          //set zero to other counters
          for(unsigned counter=2; counter < numcounter; ++counter)
          {
            Count[bunch][counter].push_back(0);
            CounterLoad[bunch][counter].push_back(0);
          }
          //set zero to other pair
          for(unsigned pair=1; pair < numpair; ++pair)
          {
            Coinc[bunch][pair].push_back(0);
            CoincLoad[bunch][pair].push_back(0);
          }
          data::PhysValue sum=0;
          //sum counters load (2 counters)
          for(unsigned counter=0;counter<numcounter;++counter) 
            sum += Count[bunch][counter].back();
          CountSum[bunch].push_back(sum);
          sum=0; //sum all paris load (only one pair)
          for(unsigned pair=0;pair<numpair;++pair) sum += Coinc[bunch][pair].back();
          CoincSum[bunch].push_back(sum);
        }
        //now calculate the delta for the coincidence
        for(unsigned bunch=0;bunch<numbunch; bunch++)
        {
          unsigned nb=cfg.bunch_norm[bunch]-1; //norm bunch
          DeltaCoincSum[bunch].push_back(data::GetDelta(CoincSum[bunch].back(), CoincSum[nb].back()));
        }
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaX[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaZ[bunch].push_back(data);
        file >> data; 
        for(unsigned bunch=0;bunch<numbunch;++bunch) SigmaL[bunch].push_back(data);
        for(unsigned n=0;n<NMR.size();++n)
        {
          file >> data;
          NMR[n].push_back(data);
        }
      }
    }
    InitVacuum();
		ShiftTime();
    return true;
  }


	void EnergyCalibrationStorage::ShiftTime(void)
	{
		clog << LOG(3) << "Shifting time and frequencies for old data" << endl;
		auto dcfg  = cfg.depolarizer;
		int direction = dcfg.final-dcfg.initial > 0 ? 1 : -1;
		for(unsigned i=0;i<t.size();i++)
		{
			t[i] = ibn::valer<double>(t[i].value - cfg.count_time/2.0,  cfg.count_time/2.0);
			//t[i] = ibn::valer<double>(t[i].value - cfg.count_time/2.0,  0);
			DepFr[i] = DepFr[i].value == 0 ? 0 : ibn::valer<double>(DepFr[i].value - direction*fabs(denergy2frequency(dcfg, dcfg.speed))*cfg.count_time/2.0,  0);
		}
	}


	void EnergyCalibrationStorage::Save(std::string filename)
	{
		Save2XML(filename);
	}
}//end of namespace data



