/*
 * =====================================================================================
 *
 *       Filename:  EnergyCalibrationStorageSerialization.cpp
 *
 *    Description: Serialization of Energy calibration storage. This is to suppress boost warnings 
 *
 *        Version:  1.0
 *        Created:  07/05/2010 10:45:19 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#include <iostream>
#include <sstream>
#include <fstream>

#include <boost/format.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>

#include <boost/serialization/list.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/map.hpp>


#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/bzip2.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include <boost/iostreams/filter/zlib.hpp>


#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>


#include "Log.h"

#include "EnergyCalibrationStorage.h"

#include "ArraySerializaton.h"


using namespace std;

namespace data
{

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  SaveBinary
   * Description:  Backup data into special file with binary format
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::SaveBinary(boost::filesystem::path fp) 
  { 
    clog << LOG(DEBUG_LOG) <<  "EnergyCalibrationStorage::SaveBinary into " << fp.string() << "\n";
    namespace fs=boost::filesystem;
    fs::path tmpfile(fp.string()+".tmp");
    fs::path dir = fp.branch_path();
    if(!fs::exists(dir) || !fs::is_directory(dir))
    {
      cerr << LOG(NORMAL_LOG) << "WARNING: Unable to save temporary data into unexisting dir: " << dir.string() << endl;
      return;
    }
		std::ofstream ofs(tmpfile.string());
    if(!ofs)
    {
      cerr << LOG(NORMAL_LOG) <<  "WARNING: Unable to open file " << tmpfile.string() << " for saving" << std::endl;
      return;
    }
    {
			try 
			{
				boost::archive::binary_oarchive oa(ofs);
				oa << *this;
			}
			catch (std::exception e)
			{
				std::cerr << LOG(NORMAL_LOG)<< "Saving binary data to file: " << tmpfile.string()<< ": " << e.what() <<std::endl;
			}
    }
    ofs.close();
    if(fs::exists(tmpfile)) fs::rename(tmpfile, fp);
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::SaveBinary DONE\n";
  }

  void EnergyCalibrationStorage::Save2FSSerial(std::string filename) 
  {
    std::ofstream ofs(filename);
    if(!ofs)
    {
      std::cerr << "WARNING: Unable to open file " << filename << " for saving" << std::endl;
      return;
    }
    boost::archive::text_oarchive oa(ofs);
    oa << *this;
    ofs.close();
  }

  void EnergyCalibrationStorage::Save2FSSerial(boost::filesystem::path fp) 
  {
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2FSSerial" << endl;
    namespace fs=boost::filesystem;
    fs::path dir = fp.branch_path();
    if(!fs::exists(dir) || !fs::is_directory(dir))
    {
      cerr << LOG(NORMAL_LOG) << "WARNING: Unable to save temporary data into unexisting dir: " << dir.string() << endl;
      return;
    }
    std::ofstream ofs(fp.string());
    if(!ofs)
    {
      std::cerr << "WARNING: Unable to open file " << fp.string() << " for saving" << std::endl;
      return;
    }
    boost::archive::text_oarchive oa(ofs);
    oa << *this;
    ofs.close();
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2FSSerial FINISH" << endl;
  }

  bool EnergyCalibrationStorage::LoadFormat7(std::istream & ifs)
  {
    ifs.clear();
    ifs.seekg(ios_base::beg); //start from the begining of the file
    ifs.clear();
    try 
    {
      clog << LOG(NORMAL_LOG) << "Trying to load format more or equal 7..."<<std::endl;
      boost::archive::text_iarchive ia(ifs);
      ia >> *this;
    } 
    catch(boost::archive::archive_exception & e) 
    {
      LOG << "failed " << e.what() << endl;
      return false;
    }
		clog << LOG(NORMAL_LOG) << "Successfully load OK" << endl;
    return true;
  }

  bool EnergyCalibrationStorage::LoadBinary(std::istream & ifs)
  {
    ifs.clear();
    ifs.seekg(ios_base::beg); //start from the begining of the file
    ifs.clear();
    try 
    {
      clog << LOG(1) << "Trying to load binary format..." << flush;
      boost::archive::binary_iarchive ia(ifs);
      ia >> *this;
    } 
    catch(boost::archive::archive_exception & e) 
    {
      LOG << "failed " << e.what() << endl;
      return false;
    }
		LOG << "OK" << endl;
    return true;
  }

  bool EnergyCalibrationStorage::LoadFormat6(std::istream & ifs)
  {
    ifs.clear();
    ifs.seekg(ios_base::beg); //start from the begining of the file
    ifs.clear();
    try 
    {
      clog << LOG(1) << "Trying to load format 5 and 6...";
      boost::archive::text_iarchive ia(ifs, ARCHIVE_MAGIC_NUMBER);
			//static const int archive_version;
      //boost::archive::text_iarchive ia(ifs, archive_version);
      ia >> *this;
    } 
    catch(boost::archive::archive_exception & e) 
    {
      LOG << "failed " << e.what() << endl;
      return false;
    }
    return true;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  Save4Utug
   * Description:  Сохраним данные для УТЮГА
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::Save4Utug(void) 
  {
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save4Utug" << endl;
    ofstream utug_file("/home/vepp4/kadrs/vepp4energy.dat");
    time_t deptime = fit_param.size()>0 ? time_t(fit_param[0].value) : 0;
    time_t depolarization_time = begin_time.tv_sec+deptime;
    utug_file.precision(15);
    utug_file << depolarization_time << "  # " << ctime(&depolarization_time);
    utug_file << cfg.run << " # energy calibrations run" << endl;
    utug_file << E       << " # measured energy and error, MeV" << endl;
    utug_file << Fr      << " # measured frequency and error, Hz " << endl;
    utug_file.close();
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save4Utug DONE" << endl;
  }

  void EnergyCalibrationStorage::Restore(const std::string & loadfilename)
  {
    using namespace std;
    EnergyCalibrationStorage * d = this;
    ifstream ifs(loadfilename);
    if(!ifs)
    {
      cerr << "Unable to open file: " << loadfilename << "  for restore EnergyCalibrationStorage " << endl;
    }
    boost::archive::binary_iarchive ia(ifs);
    try
    {
      ia >> *d;
      issave=false;
    } 
    catch(boost::archive::archive_exception & e)
    {
      cerr << "Unable to load data from file " << loadfilename << endl;
      cerr << "ERROR: " << e.what() << endl;
    }
    PostLoad();
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  EnergyCalibrationStorage
   *      Method:  Load
   * Description:  Load data from file
   *--------------------------------------------------------------------------------------
   */
  void EnergyCalibrationStorage::Load(std::string loadfilename) 
  {
    clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Load\n";
    namespace fs = boost::filesystem;
    fs::path path(loadfilename);
    if(!fs::exists(path)) 
    {
      cerr << LOG(1) << "File " << loadfilename << " does not exists" << endl;
      return;
    }
    using namespace boost::iostreams;
    std::unique_ptr<istream> ifs;
    string ext = fs::extension(path); //read file extension
    string basename = fs::basename(path); //read basename
    if(ext==".bz2" || ext==".gz" || ext==".zip") // check is it gz archive
    {
      clog << LOG(1) << "File " << loadfilename << " is zipped try to unzip..." << endl;
      ifstream zf(path.string(),ios_base::in | ios_base::binary); //open zipped file
      if(!zf)
      {
        clog << LOG(1) << " Unable to open file " << path.string() << endl;
      }
      filtering_streambuf<input> in;
      if(ext==".bz2") in.push(bzip2_decompressor());
      if(ext==".gz")  in.push(gzip_decompressor());
      if(ext==".zip") in.push(zlib_decompressor());
      in.push(zf);
      stringstream * sb = new stringstream;
      copy(in,*sb);
      ifs.reset(sb);
      ext = fs::extension(basename); //check the new extension
    }
    else 
    {
      ifs.reset(new ifstream(path.string()));
      if(!(*ifs))	
      {
        cerr << LOG(NORMAL_LOG) << "Unable to open file: " << loadfilename << "  for load EnergyCalibrationStorage " << endl;
      }
    }
    //xml could be identifyed by extension
    if(ext==".xml") if(LoadFormatXML(*ifs)) goto POSTLOAD;
    if(LoadFormat0(*ifs))  goto POSTLOAD;
    if(LoadFormat01(*ifs)) goto POSTLOAD; 
    if(LoadFormat1(*ifs))  goto POSTLOAD;
    if(LoadFormat32(*ifs)) goto POSTLOAD;
    if(LoadFormat34(*ifs)) goto POSTLOAD;
    if(LoadFormat4(*ifs))  goto POSTLOAD;
    if(LoadFormat41(*ifs)) goto POSTLOAD;
    if(LoadFormat6(*ifs))  goto POSTLOAD;
    if(LoadFormat7(*ifs))  goto POSTLOAD;
    if(LoadBinary(*ifs))   goto POSTLOAD;
    clog<< LOG(NORMAL_LOG) << "ERROR: Unable to load any format" << endl;
    return;
POSTLOAD:
      PostLoad();
    /* try to calculate run number for some formats
     * wich does not conntect run number information
     * try to calculate it frome filename */
    //if(cfg.run==static_cast<unsigned>(-1))
    if(cfg.run==0)
    {
      fs::path p(fs::basename(path));
			cfg.run = stoi(p.string());
    }
		//Save2XML((PCFG.work_dir/"tmp.xml").string());
    NMR.resize(5);
    for(auto & nmr : NMR) nmr.resize(t.size());
		std::cout << LOG(NORMAL_LOG) << "run=" << cfg.run << "  E=" << (boost::format("%11.6f +- %8.6f MeV") % E.value % E.error) << "   H = " << NMR[0][0].value << " Gs" <<  std::endl;
  }

  template <class Archive> 
    void EnergyCalibrationStorage::serialize(Archive & ar, const unsigned int version)
    {
			std::clog << LOG(INFO_LOG) << "EnergyCalibrationStorage::serialize: version =  " << version << std::endl;
      ar & numbunch;
			std::clog << LOG(INFO_LOG+1) << "Number of bunches = " << numbunch << std::endl;
      ar & numcounter;
			std::clog << LOG(INFO_LOG+1) << "Number of counter = " << numcounter << std::endl;
      ar & cfg;
			std::clog << LOG(INFO_LOG+1) << "cfg is loaded = " << std::endl;
      ar & begin_time.tv_sec;
      ar & begin_time.tv_usec;
			std::clog << LOG(INFO_LOG+1) << "begin_time =  is loaded = " << begin_time.tv_sec <<  "  " << ctime(&begin_time.tv_sec);
      ar & mode;
			std::clog << LOG(INFO_LOG+1) << "mode =  " << mode <<  std::endl;
      ar & setE;
			std::clog << LOG(INFO_LOG+1) << "setE =  " << setE <<  std::endl;
      ar & calib;
			std::clog << LOG(INFO_LOG+1) << "calib =  " << calib <<  std::endl;
      ar & calib_fr;
			std::clog << LOG(INFO_LOG+1) << "calib_fr =  " << calib_fr <<  std::endl;
      ar & success;
			std::clog << LOG(INFO_LOG+1) << "succes is loaded success=" << success <<  std::endl;
      ar & isdepolon;
      ar & dep_fr_num;
      ar & E;
      ar & Fr;
      ar & veppFr;
      ar & chi2;
      ar & isfit;
      ar & fit_param;
      ar & t;
			std::clog << LOG(INFO_LOG+1) << "time is loaded t.size()=" << t.size() <<  std::endl;
      ar & DepFr;
      ar & MDepFr;
      ar & RF;
      ar & Count;
      ar & CountSum;
      ar & DeltaCount;
      ar & DeltaSum;
      ar & Coinc;
      ar & CoincSum;
      ar & DeltaCoinc;
      ar & DeltaCoincSum;
      ar & I;
      ar & SigmaX & SigmaZ & SigmaL;
      ar & NMR;
			if(version < 6 ) goto POSTSER;
			ar & MediumVacuumPresure;
			ar & Vacuum;
			if(version < 8) goto POSTSER;
			ar & CounterLoad;
			ar & CoincLoad;
			if(version < 9) goto POSTSER;
			ar & cfg.shiftman;
			if(version < 10) goto POSTSER;
			if(version < 11) goto POSTSER;
			ar & CounterPosition; //counter position
POSTSER:
			if(version < 6) InitVacuum();
			if(version < 8) ReconsructLoad();
			if(version < 10) ShiftTime();
			if(version < 11) 
			{
				MDepFr=DepFr;
				CounterPosition.resize(numcounter);
				for(int c=0;c<numcounter;c++)  CounterPosition[c].resize(t.size());
			}
			std::clog << LOG(DEBUG_LOG) << "data::EnergyCalibrationStorageSerialization::serialize END version = " << version << std::endl;
    }


    void EnergyCalibrationStorage::SaveConfig(string filename)
    {
			std::ofstream ofs(filename);
      if(!ofs)
      {
        cerr << LOG(NORMAL_LOG) << "Unable to open file: " << filename << endl;
        return;
      }
      boost::archive::xml_oarchive oarch(ofs);
      oarch << boost::serialization::make_nvp("EnergyCalibrationConfig",cfg.depolarizer);
    }

} //end of namespace data

namespace measurement
{
	template<class Archive>
		void EnergyCalibrationConfig::serialize(Archive & ar,  const unsigned int version)
		{
			std::clog << LOG(INFO_LOG) << "EnergyCalibrationConfig::serialize BEGIN " << version << std::endl;
			ar & BOOST_SERIALIZATION_NVP(run);
			std::clog  << LOG(INFO_LOG+1) << " run = " << run << std::endl;
			ar & BOOST_SERIALIZATION_NVP(name);
			std::clog  << LOG(INFO_LOG+1) << " name = " << name << std::endl;
			ar & BOOST_SERIALIZATION_NVP(date);
			std::clog  << LOG(INFO_LOG+1) << " date = " << date << std::endl;
			if(version < 11) ar & BOOST_SERIALIZATION_NVP(depolarizer.direction);//move direction to DepolarizerConfig serialization
			if(version < 11) std::clog  << LOG(INFO_LOG+1) << " direction = " << depolarizer.direction << std::endl;
			//ar & BOOST_SERIALIZATION_NVP(depolarizer.direction); //move direction to DepolarizerConfig serialization
			ar & BOOST_SERIALIZATION_NVP(count_time);
			std::clog  << LOG(INFO_LOG+1) << " count_time = " << count_time << std::endl;
			//ar & BOOST_SERIALIZATION_NVP(depolarizer.direction); //move direction to DepolarizerConfig serialization
			if(version < 11) ar & BOOST_SERIALIZATION_NVP(depolarizer.direction); //move direction to DepolarizerConfig serialization
			if(version < 11) std::clog  << LOG(INFO_LOG+1) << " direction2 = " << depolarizer.direction << std::endl;
			ar & BOOST_SERIALIZATION_NVP(depolarizer);
			ar & BOOST_SERIALIZATION_NVP(bunch_norm);
			std::clog  << LOG(INFO_LOG+1) << " bunch_norm.size  = " << bunch_norm.size() << std::endl;
			for(int j=0;j<bunch_norm.size();j++) std::clog << LOG(INFO_LOG+1) << "  bunch " << j+1 << " / " << bunch_norm[j] << std::endl;
			if(version < 11) 
			{ 
				bunch_norm = {2, 1, 4, 3};
			}
			//std::clog  << "bunch_norm: size = " << bunch_norm.size() << ": ";
			//for(int bn=0;bn<bunch_norm.size(); bn++) std::clog << bunch_norm[bn] << " ";
			//std::clog << std::endl;
			ar & BOOST_SERIALIZATION_NVP(nmr);
			std::clog  << LOG(INFO_LOG+1) << " nmr  = " << nmr << std::endl;
			ar & BOOST_SERIALIZATION_NVP(temperature);
			std::clog  << LOG(INFO_LOG+1) << " temperature  = " << temperature << std::endl;
			ar & BOOST_SERIALIZATION_NVP(alpha_t);
			std::clog  << LOG(INFO_LOG+1) << " alpha_t  = " << alpha_t << std::endl;
			ar & BOOST_SERIALIZATION_NVP(N);
			std::clog  << LOG(INFO_LOG+1) << " N  = " << N << std::endl;
			ar & BOOST_SERIALIZATION_NVP(delta);
			std::clog  << LOG(INFO_LOG+1) << " delta  = " << delta << std::endl;
			if(version < 11)  return;
			ar & BOOST_SERIALIZATION_NVP(type);
			std::clog  << LOG(INFO_LOG+1) << " type  = " << type << std::endl;
			ar & BOOST_SERIALIZATION_NVP(experiment_name);
			std::clog  << LOG(INFO_LOG+1) << " experiment_name  = " << experiment_name << std::endl;
			ar & BOOST_SERIALIZATION_NVP(pass);
			std::clog  << LOG(INFO_LOG+1) << " pass  = " << pass << std::endl;
			ar & BOOST_SERIALIZATION_NVP(energy);
			std::clog  << LOG(INFO_LOG+1) << " energy  = " << energy << std::endl;
			ar & BOOST_SERIALIZATION_NVP(fit_method);
			std::clog  << LOG(INFO_LOG+1) << " fit_method  = " << fit_method << std::endl;
			ar & BOOST_SERIALIZATION_NVP(fit_graph);
			std::clog  << LOG(INFO_LOG+1) << " fit_graph  = " << fit_graph << std::endl;
			ar & BOOST_SERIALIZATION_NVP(shiftman);
			std::clog  << LOG(INFO_LOG+1) << " shiftman  = " << shiftman << std::endl;
			std::clog << LOG(INFO_LOG) << "EnergyCalibrationConfig::serialize END version = " << version << std::endl;
      if(version < 12) return;
			ar & BOOST_SERIALIZATION_NVP(reason);
			std::clog  << LOG(INFO_LOG+1) << " reason  = " << reason << std::endl;
		}
}
