#include <boost/property_tree/xml_parser.hpp>

#include "Utils.h"

#include "EnergyCalibrationStorage.h"
#include "data/ArraySerializaton.h"
#include <ibn/timer.h>

namespace data
{

	void EnergyCalibrationStorage::Save2XML(std::string  savefilename) //сохранить измеренные данные в файл
	{
		std::clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2XML BEGIN (save to " << savefilename << ")"  << std::endl;
		ibn::timer timer; //start measurement of saving time
		N=t.size();
		boost::property_tree::ptree main_tree;
		main_tree.put("savingdata.EnergyCalibration.HEAD.numbunch", numbunch);
		auto & tree = main_tree.get_child("savingdata.EnergyCalibration");
		auto & head_tree = tree.get_child("HEAD");
		head_tree.put("version", to_string(Storage::VERSION));
		head_tree.put("run", cfg.run);
		head_tree.put("energy", cfg.energy); 
		head_tree.put("mode", mode);
		head_tree.put("numcounter",  numcounter);
		head_tree.put("attenuation", cfg.depolarizer.attenuation);
		head_tree.put("scan_speed", cfg.depolarizer.speed);
		head_tree.put("scan_step", cfg.depolarizer.step);
		head_tree.put("scan_initial_freq", cfg.depolarizer.initial);
		head_tree.put("scan_final_freq", cfg.depolarizer.final);
		head_tree.put("RF", cfg.depolarizer.RF);
		head_tree.put("nmr", cfg.nmr);
		head_tree.put("direction", cfg.depolarizer.direction);
		head_tree.put("alpha_t", cfg.alpha_t);
		head_tree.put("temperature", cfg.temperature);
		head_tree.put("name", cfg.name);
		head_tree.put("N", N);
		head_tree.put("begin_time_sec", begin_time.tv_sec);
		head_tree.put("begin_time_usec", begin_time.tv_usec);
		head_tree.put("date", cfg.date);
		head_tree.put("setE", setE);
		head_tree.put("calib", calib);
		head_tree.put("calib_fr", calib_fr);
		if(success == true) head_tree.put("success", "1");
		else head_tree.put("success", "0");
		head_tree.put("count_time", cfg.count_time);
		head_tree.put("harmonic_number", cfg.depolarizer.harmonic_number);
		head_tree.put("dep_fr_num", dep_fr_num);
		head_tree.put("isdepolon", isdepolon);
		head_tree.put("E", E);
		head_tree.put("Fr", Fr);
		head_tree.put("veppFr", veppFr);
		head_tree.put("chi2", chi2);
		head_tree.put("revolution_frequency", cfg.depolarizer.revolution_frequency);


		tree.put("FIT.parameters_number", fit_param.size());
		{
			ostringstream os;
			for(auto p : fit_param) os << p << "\t";
			tree.put("FIT.isfit", isfit ? "yes" : "no");
			tree.put("FIT.parameters", os.str());
		}
		boost::property_tree::ptree data_tree;
		boost::property_tree::ptree delta_tree;
		boost::property_tree::ptree count_tree;


		unsigned w=50; //ширина при форматировании
		unsigned wpoint=5; //размер поля для номера точки
		unsigned precision=10; //точность после запятой в научном формате. 
		using std::ios_base;
		ios_base::fmtflags format=ios_base::left | ios_base::scientific | ios_base::showpos;

		auto put_bunch_counter = [&](boost::property_tree::ptree & tree, const std::string & total_name,  const BunchCounterArray_t & BCA,  const BunchArray_t & BA)
		{
			for (unsigned counter=0 ; counter<numcounter ; ++counter) 
			{
				std::ostringstream ostr;
				ostr.precision(precision);
				ostr.setf(format);
				ostr<<std::endl;
				for(unsigned i=0 ; i<N; ++i)
				{
					ostr << std::setw(wpoint) << i << " ";
					for (unsigned bunch=0 ; bunch<numbunch ; ++bunch)
					{
						ostr << std::setw(w) << BCA[bunch][counter][i] << " ";
					}
					ostr << std::endl;
				}
				tree.put(std::string("counter") + "-" + std::to_string(counter+1),  ostr.str());
			}
			std::ostringstream ostr;
			ostr.setf(format);
			ostr.precision(precision);
			ostr<<std::endl;
			for(unsigned i=0; i<N; i++)
			{
				ostr << std::setw(wpoint) << i << " ";
				for(unsigned b=0 ; b<numbunch ; b++)
				{
					ostr << std::setw(w) << BA[b][i] << " ";
				}
				ostr << std::endl;
			}
			//ostr<<std::endl;
			tree.put(total_name,  ostr.str());
		};

		put_bunch_counter(count_tree, "total_count", Count, CountSum);
		put_bunch_counter(delta_tree, "total_delta", DeltaCount, DeltaSum);

		data_tree.add_child("COUNT", count_tree);
		data_tree.add_child("DELTA", delta_tree);

		{//write depolarizer data 
			std::ostringstream ostr;
			ostr.setf(format);
			ostr.precision(precision);
			ostr<<std::endl;
			for(unsigned i=0 ; i<N ; i++)
			{
				ostr << std::setw(wpoint) << i << " ";
				ostr << std::setw(w) << DepFr[i]  << " ";
				PhysValue f = DepFr[i];
				ostr << std::setw(w) << data::frequency2energy(f,cfg.depolarizer.revolution_frequency,cfg.depolarizer.harmonic_number) << " "; //энергия соотвествующая данной частоте деполяриз.
				double x = DepFr[i].value/cfg.depolarizer.revolution_frequency.value;
				ostr << std::setw(w) << x-floor(x) << " ";
				ostr << std::setw(w) << 0 << " ";
				ostr << std::endl;
			}
			data_tree.put("depolarizer", ostr.str());
		}

		{ //print time
			std::ostringstream ostr;
			ostr.setf(format);
			ostr.precision(precision);
			ostr<<std::endl;
			for(unsigned i=0 ; i<N ; i++)
			{
				ostr << std::setw(wpoint) << i << " ";
				ostr << std::setw(w) << t[i] << " "; //time
				//human readable time
				char time_string[1024];
				time_t t_tmp = time_t(t[i].value + begin_time.tv_sec);
				strftime(time_string,1024,"%F-%T",localtime(&t_tmp));
				ostr << std::setw(w) <<  time_string << " "; 
				ostr << std::endl;
			}
			data_tree.put("time", ostr.str());
		}

		{ 
			std::ostringstream ostr;
			ostr.precision(precision);
			ostr.setf(format);
			ostr<<std::endl;
			for(unsigned i=0 ; i<N ; i++ )
			{
				ostr << std::setw(wpoint) << i << " ";
				for (unsigned b=0 ; b<numbunch ; ++b)
				{    
					ostr << std::setw(w) << I[b][i] << " ";
				}
				ostr<< std::endl;
			}
			data_tree.put("VEPP4.BEAM.current", ostr.str());
		}
		/* =============== SAVE BEAM SIZE ============================== */
		{ 
			std::ostringstream ostr;
			ostr.precision(precision);
			ostr.setf(format);
			ostr<<std::endl;
			for(unsigned i=0 ; i<N ; i++ )
			{
				ostr << std::setw(wpoint) << i << " ";
				for (unsigned b=0 ; b<numbunch ; ++b)
				{
					ostr << std::setw(w) << SigmaX[b][i] << " ";
					ostr << std::setw(w) << SigmaZ[b][i] << " ";
					ostr << std::setw(w) << SigmaL[b][i] << " ";
				}
				ostr<<std::endl;
			}
			data_tree.put("VEPP4.BEAM.beam_size", ostr.str());
		} 
		/* =============== NMR data ==================*/
		{
			std::ostringstream ostr;
			ostr.setf(format);
			ostr.precision(precision);
			ostr << std::endl;
			for(unsigned i=0 ; i<N ; i++)
			{
				ostr << std::setw(wpoint) << i << " ";
				for (unsigned j=0 ; j<NMR.size() ; ++j)
				{    
					ostr << std::setw(w)<< NMR[j][i] << " "; //вобъем данные в массив
				}
				ostr << std::endl;
			} 
			data_tree.put("VEPP4.nmr_number", NMR.size());
			data_tree.put("VEPP4.nmr", ostr.str());
		}

		tree.add_child("DATA", data_tree);

		//boost::property_tree::xml_writer_settings<char> settings('\t', 1);
		boost::property_tree::write_xml(savefilename, main_tree);
		std::clog << LOG(DEBUG_LOG) << "EnergyCalibrationStorage::Save2XML END" << std::endl;
	}

	template <typename T>
		std::vector<T> as_vector(boost::property_tree::ptree const& pt, boost::property_tree::ptree::key_type const& key)
		{
			std::vector<T> r;
			std::string s = pt.get<std::string>(key);
			std::stringstream ss(s);
			T data;
			while(ss>>data) r.push_back(data);
			return r;
		}

  void EnergyCalibrationStorage::XMLLoad(std::string loadfilename) 
	{
		std::ifstream file(loadfilename);
		XMLLoad(file);
	}

  bool EnergyCalibrationStorage::LoadFormatXML(const std::string & loadfilename)
  {
    clog << LOG(5) << "EnergyCalibrationStorage::LoadFormatXML: BEGIN" << endl;
    try 
    {
      XMLLoad(loadfilename);
    }
    catch(std::exception e)
    {
      cerr << LOG(1) << "Unable to read xml file: " << loadfilename << endl;
      return false;
    }
    return true;
  }

  void EnergyCalibrationStorage::XMLLoad(std::istream & file) 
	{
		std::clog<<LOG(VERBOSE2_LOG)<<"EnergyCalibrationStorage::Load: START" << std::endl;
		boost::property_tree::ptree main_tree;
		//std::clog << LOG(VERBOSE_LOG) << "Try to load file " << loadfilename << std::endl;
		try
		{
			boost::property_tree::xml_parser::read_xml(file, main_tree);

			int version = 0;
			auto & tree = main_tree.get_child("savingdata.EnergyCalibration");
			auto & head_tree = tree.get_child("HEAD"); 
			numbunch = head_tree.get<unsigned>("numbunch");
			numcounter = head_tree.get<unsigned>("numcounter");
			std::clog << LOG(VERBOSE_LOG) << "numbunch=" << numbunch << " numcounter = " << numcounter << std::endl;
			cfg.depolarizer.attenuation = head_tree.get<double>("attenuation");
			cfg.depolarizer.speed       = head_tree.get<double>("scan_speed");
			cfg.depolarizer.step        = head_tree.get<double>("scan_step");
			cfg.depolarizer.initial     = head_tree.get<double>("scan_initial_freq");
			cfg.depolarizer.final       = head_tree.get<double>("scan_final_freq");
			cfg.depolarizer.RF          = head_tree.get<double>("RF");
			cfg.nmr                     = head_tree.get<double>("nmr");
			cfg.energy                  = head_tree.get<ibn::valer<double>>("energy");


			cfg.depolarizer.direction          = head_tree.get<int>("direction");
			cfg.alpha_t            = head_tree.get<double>("alpha_t");
			cfg.temperature        = head_tree.get<double>("temperature");
			cfg.name               = head_tree.get<std::string>("name");
			N	                     = head_tree.get<unsigned>("N");
			std::clog << LOG(VERBOSE_LOG) << "Number of points N = " << N << std::endl;
			cfg.run                = head_tree.get<unsigned>("run");
			auto version_node = head_tree.get_child_optional("version");
			if( !version_node )
			{
				version=9;
			}
			else 
			{
				version=head_tree.get<int>("version");
			}
			//std::clog << LOG(VERBOSE_LOG) << "format version in XML file: " << version << endl;
			begin_time.tv_sec      = head_tree.get<unsigned>("begin_time_sec");
			begin_time.tv_usec     = head_tree.get<unsigned>("begin_time_usec");
			cfg.date	 						 = head_tree.get<time_t>("date");
			mode                   = head_tree.get<std::string>("mode");
			setE                   = head_tree.get<double>("setE");
			//isfit                  = head_tree.get<bool>("isfit");
			calib                  = head_tree.get<double>("calib");
			calib_fr               = head_tree.get<double>("calib_fr");
			switch(head_tree.get<int>("success"))
			{
				case 0:
					success = false;
					break;
				case 1:
					success = true;
					break;
				default:
					success = false;
					break;
			}
			cfg.count_time         						 = head_tree.get<double>("count_time");
			cfg.depolarizer.harmonic_number    = head_tree.get<int>("harmonic_number");
			dep_fr_num         								 = head_tree.get<int>("dep_fr_num");
			isdepolon          								 = head_tree.get<bool>("isdepolon");
			E                                  = head_tree.get<ibn::valer<double>>("E");
			Fr                                 = head_tree.get<ibn::valer<double>>("Fr");
			veppFr             								 = head_tree.get<double>("veppFr");
			chi2               								 = head_tree.get<double>("chi2");
			cfg.depolarizer.revolution_frequency = head_tree.get<ibn::valer<double>>("revolution_frequency");

			auto & fit_tree =  tree.get_child("FIT"); 
			auto isfit_str = fit_tree.get<std::string>("isfit");
			if(isfit_str=="yes" || isfit_str == "1" || isfit_str == "YES")
			{
				isfit = true;
				fit_param.resize(fit_tree.get<unsigned>("parameters_number"));
				fit_param = as_vector<ibn::valer<double>>(fit_tree, "parameters");
			}
			else
			{
				isfit = false;
				fit_param.resize(0);
			}

			/*--------------------------------------------------------------------------------
				set number of bunches,  number of counter and number of counter pairs
				-------------------------------------------------------------------------------*/

			set_geometry(numbunch,  numcounter,  numcounter/2);
			resize(N);

			/*-----------------------------------------------------------------------------
			 *  Now read the data per point
			 *-----------------------------------------------------------------------------*/

			auto read_bunch_array = [&](const std::string & path, const std::string & chan_name,  const std::string & total_name, BunchCounterArray_t & BC, BunchArray_t & B )
			{
				auto & pt = tree.get_child(path); 
				for(unsigned c=0;c<numcounter; c++)
				{
					std::istringstream iss(pt.get<std::string>(chan_name+std::to_string(c+1)));
					for(unsigned i=0; i<N;i++)
					{
						unsigned n;
						iss >> n;
						for(unsigned b=0;b<numbunch;b++)
						{
							iss>>BC[b][c][i];
						}
					}
				}
				{
					std::istringstream iss(pt.get<std::string>(total_name));
					for(unsigned i=0; i<N;i++)
					{
						unsigned n;
						iss >> n;
						for(unsigned b=0;b<numbunch;b++)
						{
							iss >>B[b][i];
						}
					}
				}
			};

			read_bunch_array("DATA.COUNT", "counter-", "total_count",  Count,  CountSum);
			read_bunch_array("DATA.DELTA", "counter-", "total_delta",  DeltaCount,  DeltaSum);

			/*-----------------------------------------------------------------------------
			 *  Read time and parameters of the depolarizer 
			 *-----------------------------------------------------------------------------*/
			auto   read_single = [&](const std::string & path, Array & A)
			{
				std::istringstream iss(tree.get<std::string>(path));
				for(unsigned i=0; i<N;i++)
				{
					unsigned n;
					iss >> n;
					iss >> A[i];
					iss.ignore(65535,'\n');
				}
			};

			read_single("DATA.time",  t);
			read_single("DATA.depolarizer",  DepFr);


			/*-----------------------------------------------------------------------------
			 *  Read accelerator parameters
			 *-----------------------------------------------------------------------------*/
			//NMR
			{
				unsigned nmr_size = tree.get<unsigned>("DATA.VEPP4M.nmr_number");
				NMR.resize(nmr_size);
				for(auto & nmr : NMR) nmr.resize(N);
				std::istringstream iss(tree.get<std::string>("DATA.VEPP4M.nmr"));
				for(unsigned i=0;i<N;i++)
				{
					unsigned n;
					iss >> n;
					for(unsigned ch=0;ch<NMR.size();ch++) iss >> NMR[ch][i];
				}
			}

			auto & beam_tree = tree.get_child("DATA.VEPP4M.BEAM");
			//current
			{
				std::istringstream iss (beam_tree.get<std::string>("current"));
				for(unsigned i=0;i<N;i++)
				{
					unsigned n;
					iss >> n;
					for(unsigned b=0;b<numbunch;b++) iss>>I[b][i];
				}
			}
			//size;
			{
				std::istringstream iss (beam_tree.get<std::string>("beam_size"));
				for(unsigned i=0;i<N;i++)
				{
					unsigned n;
					iss >> n;
					for(unsigned b=0;b<numbunch;b++) iss>>SigmaX[b][i] >> SigmaZ[b][i] >> SigmaL[b][i];
				}
			}

			if(version < 10) ShiftTime();
			std::cout << LOG(1) << "Measured energy = " << cfg.energy << " MeV in run " << cfg.run << " at " << ctime(&cfg.date);
			std::cout << LOG(1) << "Scan mode \"" << cfg.name << "\"  scan_speed = " << cfg.depolarizer.speed << "  MeV/s,  attenuation = " << cfg.depolarizer.attenuation << " dB" << std::endl;
			std::clog<<LOG(5)<<"EnergyCalibrationStorage::Load: END" << std::endl;
		}
		catch(boost::property_tree::xml_parser_error e)
		{
			std::cout << "ERROR: xml_parser_error: " << e.what() << std::endl;
		}
		catch(boost::property_tree::ptree_bad_data e)
		{
			std::cout << "ERROR reading   ptree_bad_data: " << e.what() << std::endl;
			throw e;
		}
		catch(boost::property_tree::ptree_bad_path e)
		{
			std::cout << "ERROR reading  ptree_bad_path: " << e.what() << std::endl;
		}
		catch(boost::property_tree::ptree_error e)
		{
			std::cout << "ERROR reading  ptree_error: " << e.what() << std::endl;
		}
		catch(std::runtime_error e)
		{
			std::cout << e.what() << std::endl;
		}
	}

  bool EnergyCalibrationStorage::LoadFormatXML(istream & file)
  {
    file.seekg(ios_base::beg); //start from the begining of the file
    //ofstream os("/tmp/ecrun.xml");
    //os << file.rdbuf();
		//os.close();
    //return LoadFormatXML("/tmp/ecrun.xml");
		XMLLoad(file);
		return true;
  }
}
