/*
 * =====================================================================================
 *
 *       Filename:  FCN.h
 *
 *    Description:  FCN collection
 *
 *        Version:  1.0
 *        Created:  13.11.2012 15:05:43
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */


#ifndef IBN_POLARIMETER_FIT_COUNT_CURRENT_FITTER_H
#define IBN_POLARIMETER_FIT_COUNT_CURRENT_FITTER_H

#include "CountFunctions.h"
#include "data/Array.h"
#include "FitParConfiguration.h"
#include "Log.h"
#include "gui/EnergyCalibrationView.h"

#include "Fitter.h"
#include <TF1.h>
#include <TGraph.h>
#include <TGraphErrors.h>
#include <TAxis.h>

namespace fit
{
  template <class JumpFunction>
    class CountCurrentFitter : public Fitter
    {
      unique_ptr<TGraph> graph_function;
      public:
      data::Array * TIME;
      data::Array * CURRENT;
      data::Array * COUNT;
      data::Array fit_result;
      std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;
      mutable Count<JumpFunction> CountFunction;
      CountCurrentFitter(void)
      {
        type = MINUIT2;
        name = "COUNT_CURRENT";
        title = "Count and current fit";
        defpar.Add("T",        0, 100); //depolarization moment
        defpar.Add("DELTA",    0, 0.1, -1, +1); //depolarization effect
        defpar.Add("TAU",      1,   1); //depolarization time
        defpar.Add("POWER",    1, 0.1); //depolarizer power
        defpar.Add("I0",       1e-3, 1e-4); //initial current in A
        defpar.Add("TAUTSH", 1e4, 1e3, 0, 1e10); //touschek beam life time
        defpar.Add("TAUBG",  1e4, 1e3, 0, 1e10); //background beam life time
        defpar.Add("PTSH",     1, 0.1,0,1); //touschek registration probability
        defpar.Add("PBG",      1, 0.1,0,1); //background registration probability
        inipar=defpar;
        TIME=0;
        CURRENT=0;
        COUNT=0;
      }
      CountCurrentFitter(const std::vector<FitParConfiguration> & pcfg){};
      CountCurrentFitter(const data::Array * t, const data::Array  * I, const data::Array * count, const std::vector<FitParConfiguration> & pcfg){};
      ~CountCurrentFitter(void){};
      void Fit(void);
      void SetParameters(double * par);
      void SetParameters(const vector<double> & par) const;
      double operator() (const std::vector<double> & par) const;
      double Up() const;
      void SetMinPar(void);
      void DrawFunction(void);
      ROOT::Minuit2::MnUserParameters FindJump(void);
    };

  typedef CountCurrentFitter<HeavisideJump> HeavisideCountCurrentFitter;
  typedef CountCurrentFitter<AtanJump>      AtanCountCurrentFitter;
  typedef CountCurrentFitter<ErfJump>       ErfCountCurrentFitter;

  template <class JumpFunction>
  void CountCurrentFitter<JumpFunction>::Fit(void)
  {
    data::EnergyCalibrationStorage * EC = dynamic_cast<data::EnergyCalibrationStorage*>(fStorage);
    TIME = &EC->t;
    auto theData = fView->GetData(fDataName);
    if(theData == nullptr) 
    {
      clog << LOG(NORMAL_LOG) << "Unable to find data " << fDataName << std::endl;
      return;
    }
    COUNT  = theData->data;
    fBunch = theData->bunch;
    CURRENT = &EC->I[fBunch];
    using namespace ROOT::Minuit2;
    MnUserParameters newpar=inipar;
    if(!inipar.Parameter(newpar.Index("T")).IsFixed()) newpar = FindJump();
    if(tmin==tmax)
    {
      tmin = (*TIME)[0].value;
      tmax = (*TIME)[TIME->size()-1].value;
    }
    MnMigrad migrad(*this, newpar);
    FunctionMinimum minimum = migrad();
    minpar=minimum.UserParameters();
    std::cout << LOG(1) << " chi2 = " << (*this)(minpar.Params()) << endl;
    std::cout << LOG(1) << " free pars = " << migrad.VariableParameters() << endl;
    std::cout << LOG(1) << " data size = " << TIME->size();
    std::cout << LOG(1) << " NDF = " << TIME->size() - migrad.VariableParameters() << endl;
    SetMinPar();
  }

  template <class JumpFunction>
  void CountCurrentFitter<JumpFunction>::SetParameters(double * par)
  {
    CountFunction.SetDepolarizationMoment(par[0]);
    CountFunction.SetDelta(par[1]);
    CountFunction.SetDepolarizationTime(par[2]);
    CountFunction.SetDepolarizerPower(par[3]);
    CountFunction.SetInitialCurrent(par[4]);
    CountFunction.SetTouschekLifeTime(par[5]);
    CountFunction.SetBackgroundLifeTime(par[6]);
    CountFunction.SetTouschekRegistrationProbability(par[7]);
    CountFunction.SetBackgroundRegistrationProbability(par[8]);
  }

  template <class JumpFunction>
  void CountCurrentFitter<JumpFunction>::SetParameters(const vector<double> & par) const
  {
    CountFunction.SetDepolarizationMoment(par[0]);
    CountFunction.SetDelta(par[1]);
    CountFunction.SetDepolarizationTime(par[2]);
    CountFunction.SetDepolarizerPower(par[3]);
    CountFunction.SetInitialCurrent(par[4]);
    CountFunction.SetTouschekLifeTime(par[5]);
    CountFunction.SetBackgroundLifeTime(par[6]);
    CountFunction.SetTouschekRegistrationProbability(par[7]);
    CountFunction.SetBackgroundRegistrationProbability(par[8]);
  }


  template <class JumpFunction>
  double CountCurrentFitter<JumpFunction>::operator() (const std::vector<double> & par) const
  {
    SetParameters(par);
    double chi2=0;
    for(unsigned i=0;i<TIME->size();i++)
    {
      //cout << "new cycle" << endl;
      double t = (*TIME)[i].value;
      double I = 1e6*CountFunction.GetCurrent(t); //convert Ampers to mkA
      //cout << " I = " << I << endl;
      double dI = (*CURRENT)[i].error;
      //cout << " dI = " << dI << endl;
      chi2+=pow((I-(*CURRENT)[i].value)/dI,2.0);
      double C = CountFunction.GetAverageCount(t,20);
      //double C = CountFunction(t);
      double dC = (*COUNT)[i].error;
      chi2+=pow((C-(*COUNT)[i].value)/dC,2.0);
      //cout << t << " count = " << CountFunction(t) << "     average = " << C << endl;
      //cout << "chi2 " << chi2 << endl;
      if(std::isnan(CountFunction(t)))
      {
        for(auto x : par)
        {
          cout << x << " ";
        }
        cout << endl;
      }
      //cout << i << " t=" << t << " I=" << I << " dI=" << dI << " count=" << C << "  dC" << dC <<  " " << endl;
      //cout << t << " " << (*CURRENT)[i].value << " " << (*COUNT)[i].value << endl;
    }
    return chi2;
  } 

  template <class JumpFunction>
  double CountCurrentFitter<JumpFunction>::Up() const { return 1.; }

  template <class JumpFunction>
  void CountCurrentFitter<JumpFunction>::SetMinPar(void)
  {
    CountFunction.SetParameters(&minpar.Params()[0]);
  }

  template <class JumpFunction>
  void CountCurrentFitter<JumpFunction>::DrawFunction(void)
  {
    //root_function.reset(new TF1("CounterCurrent", &CountFunction,0,1e5,9));
    //cout << "Draw function" << endl;
    //root_function->SetParameters(&(CountFunction.GetParameters()[0]));
    //for(int i=0;i<7;i++)
    //{
    //  cout << "p"<<i << "=" << root_function->GetParameter(i) << endl;
    //}
    //root_function->Draw("same");
    //for(unsigned i=0;i<TIME->size();i++)
    //{
    //  double t = (*TIME)[i].value;
    //  cout << t << " C = " << CountFunction(t)  << "    average = " << CountFunction.GetAverageCount(t,20) << endl;
    //}
    int N=1000;
    graph_function.reset(new TGraph(N));
    for(int i=0;i <N;i++)
    {
      double t = tmin + (tmax-tmin)/N*i; 
      double C = CountFunction.GetAverageCount(t,20);
      double c = CountFunction(t);
      graph_function->SetPoint(i,t, C);
      cout << t << " C = " << C  << "  averaging affect" <<  (C-c)/c*100 << endl;
    }
    graph_function->SetLineWidth(2);
    graph_function->SetLineColor(kRed);
    graph_function->Draw("l");

    /*  теперь уберём наклон */
    TGraphErrors * gdata =new TGraphErrors(COUNT->size());
    TGraph * gfit =new TGraphErrors(graph_function->GetN());
    gfit->SetLineColor(kRed);
    gfit->SetLineWidth(2);
    Count<JumpFunction> NotDepCount;
    NotDepCount.SetParameters(&CountFunction.GetParameters()[0]);
    NotDepCount.SetDepolarizationMoment(tmax);
    for(unsigned i=0;i<COUNT->size();i++)
    {
      double t =(*TIME)[i].value; 
      double norm = NotDepCount.GetAverageCount(t,20);
      gdata->SetPoint(i,t , (*COUNT)[i].value/norm-1);
      gdata->SetPointError(i, (*TIME)[i].error, (*COUNT)[i].error/norm);
    }
    for(int i=0;i<graph_function->GetN();i++)
    {
      double t =graph_function->GetX()[i]; 
      double norm = NotDepCount.GetAverageCount(t,20);
      gfit->SetPoint(i, t, graph_function->GetY()[i]/norm - 1);
    }
    TCanvas * c = new TCanvas;
    c->SetTitle("antislope");
    gdata->SetMarkerStyle(20);
    gdata->Draw("ap");
    gdata->SetTitle("antislope");
    gdata->GetXaxis()->SetTitle("t, sec");
    gfit->Draw("l");
  }

  /*  Ищет скачок, предполагая, что время обозначено параметром T */
  template <class JumpFunction>
  ROOT::Minuit2::MnUserParameters CountCurrentFitter<JumpFunction>::FindJump(void)
  {
    using namespace ROOT::Minuit2;
    //Заполним список минимумов
    std::list<FunctionMinimum> fmlist;
    for(unsigned i=0; i<TIME->size();i++)
    {
      double t =  (*TIME)[i].value; //возьмем время в определённой точке
      MnMigrad migrad(*this, inipar);
      migrad.SetValue("T",t);
      FunctionMinimum fm = migrad();
      fmlist.push_back(fm);
    }
    //в этом списке найдем самый минимальный
    return min_element 
      (
       begin(fmlist), end(fmlist), 
       [](const FunctionMinimum & f1, const FunctionMinimum & f2) 
       { 
       return f1.Fval() < f2.Fval(); 
       }
      )->UserParameters();
  }


}

#endif

