/*
 * =====================================================================================
 *
 *       Filename:  CountFunctions.h
 *
 *    Description:  Collection of function to fit count rate
 *
 *        Version:  1.0
 *        Created:  13.11.2012 11:37:51
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */

#ifndef IBN_POLARIMETER_FIT_COUNT_FUNCTIONS_H
#define IBN_POLARIMETER_FIT_COUNT_FUNCTIONS_H
#include "fit/PolarizationFunctions.h"
#include "Formula.h"
#include "ibn/integral.h"
#include <limits>
namespace fit
{
  template <class JumpFunction>
    class Current
    {
      public:
        Polarization<JumpFunction> P; //polarization vs time
        double I0;   //начальный ток
        double Ttsh; //время жизни по тушеку
        double Tbg;  //время жизни по фону
        Current(double td=0, double delta=0, double tau=1, double power=1, double i0=1, double tau_tsh=10000, double tau_bg=10000) :
          P(Polarization<JumpFunction>(td,delta,tau,power,1./tau_bg)),
          I0(i0),
          Ttsh(tau_tsh),
          Tbg(tau_bg)
        {

        }
        Current(Polarization<JumpFunction> pf,  double i0, double tau_tsh, double tau_bg) :
          P(pf), I0(i0), Ttsh(tau_tsh),  Tbg(tau_bg) {}
        double operator()(double t)
        {
          //if(!std::isnormal(Tbg) || !std::isnormal(Ttsh)) return 0;
          double e = exp(-t/Tbg);
          double result =  I0*e/(1.+(1.-e)*Tbg/Ttsh - P.integral(0,t)/Ttsh);
          //double result =  I0*e/(1.+(1.-e)*Tbg/Ttsh);
          //if(std::isnan(result)) return 0;
          //if(std::isinf(result)) return std::numeric_limits<double>::max()/1e100;
          //if(result < 0 || !std::isnormal(result) ) result =0;
          return result;
        }

        double operator()(double *x, double *p)
        {
          SetParameters(p);
          return (*this)(x[0]);
        }

        double polarization(double t) 
        {
          return P.polarization(t);
        }

        void SetInitialCurrent(double i0) {I0=i0;}
        void SetTouschekLifeTime(double tau) {Ttsh=fabs(tau);}
        void SetBackgroundLifeTime(double tau)
        {
          //if(tau==0) Tbg = std::numeric_limits<double>::min();
          Tbg=fabs(tau); 
          P.SetBackground(1./Tbg);
        }

        void SetDepolarizationMoment(double t) {P.SetDepolarizationMoment(t);}
        void SetDelta(double d) {P.SetDelta(d);}
        void SetDepolarizationTime(double tau) {P.SetDepolarizationTime(tau);}
        void SetDepolarizerPower(double p) {P.SetDepolarizerPower(p);}
        void SetJumpFunction(JumpFunction jf) {P.SetJumpFunction(jf);}
        void SetPolarizationFunction(Polarization<JumpFunction> pf) {P=pf;} 

        double GetInitialCurrent(void) {return I0;}
        double GetTouschekLifeTime(void) {return Ttsh;}
        double GetBackgroundLifeTime(void) {return Tbg;}

        double GetDepolarizationMoment(void) {return P.GetDepolarizationMoment();}
        double GetDelta(void) {return P.GetDelta();}
        double GetDepolarizationTime(void) {return P.GetDepolarizationTime();}
        double GetDepolarizerPower(void) {return P.GetDepolarizerPower();}
        const JumpFunction & GetJumpFunction(void) {return P.GetJumpFunction();}
        const Polarization<JumpFunction> & GetPolarizationFunction(void) {return P;} 

        void SetParameters(double *par)
        {
          SetDepolarizationMoment(par[0]);
          SetDelta(par[1]);
          SetDepolarizationTime(par[2]);
          SetDepolarizerPower(par[3]);
          SetInitialCurrent(par[4]);
          SetTouschekLifeTime(par[5]);
          SetBackgroundLifeTime(par[6]);
        }

        vector<double> GetParameters(void)
        {
          vector<double> p(9);
          p[0] = GetDepolarizationMoment();
          p[1] = GetDelta();
          p[2] = GetDepolarizationTime();
          p[3] = GetDepolarizerPower();
          p[4] = GetInitialCurrent();
          p[5] = GetTouschekLifeTime();
          p[6] = GetBackgroundLifeTime();
          return p;
        }
    };

  typedef Current<HeavisideJump> HeavisideCurrent;
  typedef Current<AtanJump> AtanCurrent;
  typedef Current<ErfJump> ErfCurrent;

  template <class A> 
    class B
    {
      public:
        B(void){}
        double operator()(double x) { return x;}
    };

  template < class JumpFunction >
    class Count
    {
      Current<JumpFunction>  C;
      double PROB_TSH; //вероятност зарегистрировать тушек
      double PROB_BG; //вероятность зарегистрировать фон
      public:
        Count(double td=0, double delta=0, double tau=1, double power=1, double I0=1, double Ttsh=10000, double Tbg=10000, double prob_tsh=1, double prob_bg=1) : 
          C(Current<JumpFunction>(td,delta,tau,power,I0,Ttsh, Tbg)),
          PROB_TSH(prob_tsh), PROB_BG(prob_bg) {}
        Count(Current<JumpFunction> cf, double prob_tsh, double prob_bg) : C(cf), PROB_TSH(prob_tsh), PROB_BG(prob_bg) {}
        double operator()(double t)
        {
          double I = fabs(C(t));
          double I0 = C.I0;
          double Ie = phys::e*phys::f0;
        //  cout << "op t = " << t << " I = " << I << " Ttsh=" << C.Ttsh << " Tbg=" << C.Tbg << " Ptsh=" << PROB_TSH << " Pbg=" << PROB_BG << " pol = " << C.polarization(t);
          double result  = PROB_TSH*I*I/(I0*Ie*C.Ttsh)*(1.0 - C.polarization(t))  + PROB_BG*I/(Ie*C.Tbg);
         // cout << " result = " << result << endl;
          return result;
        }

        double operator()(double *x, double *p)
        {
          SetParameters(p);
          return (*this)(x[0]);
        }

        void SetDepolarizationMoment(double t) {C.SetDepolarizationMoment(t);}
        void SetDelta(double d) {C.SetDelta(d);}
        void SetDepolarizationTime(double tau) {C.SetDepolarizationTime(tau);}
        void SetDepolarizerPower(double p) {C.SetDepolarizerPower(p);}
        void SetInitialCurrent(double i0) {C.SetInitialCurrent(i0);}
        void SetTouschekLifeTime(double tau) {C.SetTouschekLifeTime(fabs(tau)); }
        void SetBackgroundLifeTime(double tau)
        {
          //if(tau==0) tau==std::numeric_limits<double>::min();
          C.SetBackgroundLifeTime(fabs(tau));
        }
        void SetTouschekRegistrationProbability(double p) { PROB_TSH=fabs(p);}
        void SetBackgroundRegistrationProbability(double p) { PROB_BG=fabs(p);}

        void SetCurrentFunction(Current<JumpFunction> cf) { C = cf;}
        void SetPolarizationFunction(Polarization<JumpFunction> pf) { C.SetPolarizationFunction(pf); }
        void SetJumpFunction(JumpFunction jf) {C.SetJumpFunction(jf);}

        double GetDepolarizationMoment(void) {return C.GetDepolarizationMoment();}
        double GetDelta(void) {return C.GetDelta();}
        double GetDepolarizationTime(void) {return C.GetDepolarizationTime();}
        double GetDepolarizerPower(void) {return C.GetDepolarizerPower();}
        double GetInitialCurrent(void) {return C.GetInitialCurrent();}
        double GetTouschekLifeTime(void) {return C.GetTouschekLifeTime(); }
        double GetBackgroundLifeTime(void) {return C.GetBackgroundLifeTime();}
        double GetTouschekRegistrationProbability(void) { return PROB_TSH;}
        double GetBackgroundRegistrationProbability(void) { return PROB_BG;}

        const Current<JumpFunction> &  GetCurrentFunction(void) {return C;}
        const Polarization<JumpFunction> & GetPolarizationFunction(void) {return C.GetPolarizationFunction(); }
        const JumpFunction & GetJumpFunction(void) {return C.GetJumpFunction();}


        void SetParameters(double *par)
        {
          SetDepolarizationMoment(par[0]);
          SetDelta(par[1]);
          SetDepolarizationTime(par[2]);
          SetDepolarizerPower(par[3]);
          SetInitialCurrent(par[4]);
          SetTouschekLifeTime(par[5]);
          SetBackgroundLifeTime(par[6]);
          SetTouschekRegistrationProbability(par[7]);
          SetBackgroundRegistrationProbability(par[8]);
        }

        vector<double> GetParameters(void)
        {
          vector<double> p(9);
          p[0] = GetDepolarizationMoment();
          p[1] = GetDelta();
          p[2] = GetDepolarizationTime();
          p[3] = GetDepolarizerPower();
          p[4] = GetInitialCurrent();
          p[5] = GetTouschekLifeTime();
          p[6] = GetBackgroundLifeTime();
          p[7] = GetTouschekRegistrationProbability();
          p[8] = GetBackgroundRegistrationProbability();
          return p;
        }


        double GetCurrent(double t) { return C(t);}

        double integral(double t1, double t2)
        {
          return ibn::dgaus(*this, t1, t2,1e-10);
        }

        double GetAverageCount(double t, double DeltaT)
        {
          double c = (*this)(t);
          if(std::isnan(c) || std::isinf(c) ) return 0;
          //cout << " t = " << t << " Tsh=" << C.Ttsh << " Tbg=" << C.Tbg;
          double average= integral(t-DeltaT, t)/DeltaT;
          //cout << " average=" << average  << endl;
          return average;
        }
    };

  typedef Count<HeavisideJump> HeavisideCount;
  typedef Count<AtanJump> AtanCount;
  typedef Count<ErfJump> ErfCount;

}//end of fit
#endif
