// =====================================================================================
//
//       Filename:  CurrentFitter.h
//
//    Description:  
//
//        Version:  1.0
//        Created:  20.04.2016 19:37:40
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#pragma once
#include <memory>

#include <TF1.h>

#include "data/Array.h"
#include "Fitter.h"
#include "CountFunctions.h"
#include "gui/EnergyCalibrationView.h"
namespace fit
{
  template <class JumpFunction>
    class CurrentFitter : public Fitter
  {
    unique_ptr<TF1> root_function;
    public:
    data::Array * TIME;
    data::Array * CURRENT;
    data::Array fit_result;
    //    TFitterMinuit * minuit=nullptr;
    std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;
    mutable Current<JumpFunction> CurrentFunction;
    CurrentFitter(void)
    {
      type = MINUIT2;
      name = "CURRENT";
      title = "Current fit";
      defpar.Add("T",        0, 100); //depolarization moment
      defpar.Add("DELTA",    0, 0.1); //depolarization effect
      defpar.Add("TAU",      1,   1); //depolarization time
      defpar.Add("POWER",    1, 0.1); //depolarizer power
      defpar.Add("I0",       1, 0.1); //initial current
      defpar.Add("TAUTSH", 1e4, 1e3); //touschek beam life time
      defpar.Add("TAUBG",  1e4, 1e3); //background beam life time
      inipar=defpar;
      TIME=0;
      CURRENT=0;
    }
    ~CurrentFitter(void){};
    void Fit(void);
    void SetParameters(double * par);
    void SetParameters(const vector<double> & par) const;
    double operator() (const std::vector<double> & par) const;
    double Up() const;
    void SetMinPar(void);
    void DrawFunction(void);
    /*  Ищет скачок, предполагая, что время обозначено параметром T */
    ROOT::Minuit2::MnUserParameters FindJump(void);
  };
  typedef CurrentFitter<HeavisideJump> HeavisideCurrentFitter;
  typedef CurrentFitter<AtanJump>      AtanCurrentFitter;
  typedef CurrentFitter<ErfJump>       ErfCurrentFitter;

  template <class JumpFunction>
    void CurrentFitter<JumpFunction>::Fit(void)
    {
      data::EnergyCalibrationStorage * EC = dynamic_cast<data::EnergyCalibrationStorage*>(fStorage);
      TIME = &EC->t;
      auto theData = fView->GetData(fDataName);
      if(theData == nullptr) 
      {
        clog << LOG(NORMAL_LOG) << "Unable to find data " << fDataName << std::endl;
        return;
      }
      CURRENT  = theData->data;
      fBunch = theData->bunch;
      using namespace ROOT::Minuit2;
      MnMigrad migrad(*this, inipar);
      FunctionMinimum minimum = migrad();
      minpar=minimum.UserParameters();
      CurrentFunction.SetParameters(&minpar.Params()[0]);
    }

  template <class JumpFunction>
    void CurrentFitter<JumpFunction>::SetParameters(double * par)
    {
      CurrentFunction.SetDepolarizationMoment(par[0]);
      CurrentFunction.SetDelta(par[1]);
      CurrentFunction.SetDepolarizationTime(par[2]);
      CurrentFunction.SetDepolarizerPower(par[3]);
      CurrentFunction.SetInitialCurrent(par[4]);
      CurrentFunction.SetTouschekLifeTime(par[5]);
      CurrentFunction.SetBackgroundLifeTime(par[6]);
    }

  template <class JumpFunction>
    void CurrentFitter<JumpFunction>::SetParameters(const vector<double> & par) const
    {
      CurrentFunction.SetDepolarizationMoment(par[0]);
      CurrentFunction.SetDelta(par[1]);
      CurrentFunction.SetDepolarizationTime(par[2]);
      CurrentFunction.SetDepolarizerPower(par[3]);
      CurrentFunction.SetInitialCurrent(par[4]);
      CurrentFunction.SetTouschekLifeTime(par[5]);
      CurrentFunction.SetBackgroundLifeTime(par[6]);
    }


  template <class JumpFunction>
    double CurrentFitter<JumpFunction>::operator() (const std::vector<double> & par) const
    {
      SetParameters(par);
      double chi2=0;
      for(unsigned i=0;i<TIME->size();i++)
      {
        double t = (*TIME)[i].value;
        double I = CurrentFunction(t);
        double dI = (*CURRENT)[i].error;
        chi2+=pow((I-(*CURRENT)[i].value)/dI,2.0);
      }
      return chi2;
    } 

  template <class JumpFunction>
    double CurrentFitter<JumpFunction>::Up() const { return 1.; }

  template <class JumpFunction>
    void CurrentFitter<JumpFunction>::SetMinPar(void)
    {
      CurrentFunction.SetParameters(&minpar.Params()[0]);
    }

  template <class JumpFunction>
    void CurrentFitter<JumpFunction>::DrawFunction(void)
    {
      root_function.reset(new TF1("CounterCurrent", &CurrentFunction,0,1e5,7));
      cout << "Draw function" << endl;
      root_function->SetParameters(&(CurrentFunction.GetParameters()[0]));
      for(int i=0;i<7;i++)
      {
        cout << "p"<<i << "=" << root_function->GetParameter(i) << endl;
      }
      root_function->Draw("same");
    }

    /*  Ищет скачок, предполагая, что время обозначено параметром T */
  template <class JumpFunction>
    ROOT::Minuit2::MnUserParameters CurrentFitter<JumpFunction>::FindJump(void)
    {
      using namespace ROOT::Minuit2;
      //Заполним список минимумов
      std::list<FunctionMinimum> fmlist;
      cout << "In find jump " << endl;
      for(unsigned i=0; i<TIME->size();i++)
      {
        double t =  (*TIME)[i].value; //возьмем время в определённой точке
        cout << i << " " << t << " " << endl;
        MnMigrad migrad(*this, inipar);
        migrad.SetValue("T",t);
        fmlist.push_back(migrad());
        cout << i << " " << t << " " << endl;
      }
      //в этом списке найдем самый минимальный
      return min_element 
        (
         begin(fmlist), end(fmlist), 
         [](const FunctionMinimum & f1, const FunctionMinimum & f2) 
         { 
         return f1.Fval() < f2.Fval(); 
         }
        )->UserParameters();
    }


};

