/*
 * =====================================================================================
 *
 *       Filename:  FitConfig.cpp
 *
 *    Description:  Fit Method Configuration
 *
 *        Version:  1.0
 *        Created:  07/11/2010 12:49:27 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */

#include <map>
#include <list>

#include "Functions.h"
#include "FitConfig.h"

namespace fit
{

  std::map < std::string, FitMethodConfiguration >  FitMethodList;
  std::list <unsigned> BadPointList;

	void InitFitMethodList(void)	
	{
    FitMethodConfiguration  * cfg = &FitMethodList["INSTANT_DEPOLARIZATION"];
    cfg->name = "INSTANT_DEPOLARIZATION";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::FIND_JUMP;
    cfg->func = &instant_depolarization;
    cfg->par.Add("T",0);
    cfg->par.Add("DELTA",0);
    cfg->par.Add("DN",0);
    cfg->par.Add("LIFE_TIME",5000);

    cfg = &FitMethodList["LINEAR"];
    cfg->name = "LINEAR";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::FIND_JUMP;
    cfg->func = &linear_func;
    cfg->par.Add("T",0,100);
    cfg->par.Add("DELTA",0,0.01);
    cfg->par.Add("CONST",0,0.1);
    cfg->par.Add("SLOPE1",0,1e-5);
    cfg->par.Add("SLOPE2",0,1e-5);

    cfg = &FitMethodList["LINEAR_JUMP"];
    cfg->name = "LINEAR_JUMP";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::FIND_JUMP;
    cfg->func = &lin_super;
    cfg->par.Add("T",0,100);
    cfg->par.Add("DELTA",0,0.01);
    cfg->par.Add("CONST",0,0.1);
    cfg->par.Add("SLOPE1",0,1e-5);
    cfg->par.Add("SLOPE2",0,1e-5);

    cfg = &FitMethodList["EXP_JUMP"];
    cfg->name = "EXP_JUMP";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::FIND_JUMP;
    cfg->func = &exp_func;
    cfg->par.Add("T",0,100);
    cfg->par.Add("DELTA",0,0.05);
    cfg->par.Add("CONST",0,0.05);
    cfg->par.Add("SLOPE1",0,1e-5);
    cfg->par.Add("SLOPE2",0,1e-5);

    //std::vector<FitParConfiguration> ljp;
    ROOT::Minuit2::MnUserParameters ljp;
    ljp.Add("T",0);
    ljp.Add("DELTA",0);
    ljp.Add("CONST",0);
    ljp.Add("SLOPE1",0);
    ljp.Add("SLOPE2",0);
    ljp.Add("TAU",100);

    cfg = &FitMethodList["ATAN_OLD"];
    cfg->name = "ATAN_OLD";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &atan_old;
    cfg->par = ljp;

    cfg = &FitMethodList["ATAN"];
    cfg->name = "ATAN";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &atan_super;
    cfg->par = ljp;

    cfg = &FitMethodList["ERF"];
    cfg->name = "ERF";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &erf_super;
    cfg->par = ljp;


    //std::vector<FitParConfiguration> eerfp;
    ROOT::Minuit2::MnUserParameters eerfp;
    eerfp = ljp;
    eerfp.Add("POWER",1,0.01);

    cfg = &FitMethodList["EXPERF"];
    cfg->name = "EXPERF";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &experf;
    cfg->par = eerfp;

    cfg = &FitMethodList["EXPERF_SUPER"];
    cfg->name = "EXPERF_SUPER";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &experf_super;
    cfg->par = eerfp;

    //std::vector<FitParConfiguration> fullp;
    ROOT::Minuit2::MnUserParameters fullp;
    fullp.Add("T",0);
    fullp.Add("DELTA",0);
    fullp.Add("DN",0);
    fullp.Add("TAU",100);
    fullp.Add("LIFE_TIME",5000);

    cfg = &FitMethodList["ATAN_FULL"];
    cfg->name = "ATAN_FULL";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &atan_full;
    cfg->par = fullp;

    cfg = &FitMethodList["ATAN_FIXED_SLOPE"];
    cfg->name = "ATAN_FIXED_SLOPE";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = &atan_fixed_slope;
    cfg->par = fullp;

    cfg = &FitMethodList["BACKGROUND"];
    cfg->name = "BACKGROUND";
    cfg->type = FitMethodConfiguration::FCN;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = 0;
    cfg->par.Add("kT",1e3,100,0,1e6);
    cfg->par.Add("kG1",1e3,100,0,1e6);
    cfg->par.Add("kG2",1e3,100,0,1e6);
    cfg->par.Add("kG3",0,1e3,0,1e6);
    cfg->par.Add("kT2",0,1e-2);
    cfg->par.Add("V0",0,1e-3,-0.03,+0.03);
    cfg->par.Add("V1",0,1e-3,-0.03,+0.03);
    cfg->par.Add("Vp",1,0.01,-2,+2); 
    //cfg->par.back().fixed=true;
    cfg->par.Fix("Vp");

    cfg = &FitMethodList["BACKGROUND2"];
    cfg->name = "BACKGROUND2";
    cfg->type = FitMethodConfiguration::FCN;
    cfg->flag = FitMethodConfiguration::NONE;
    cfg->func = 0;
    cfg->par.Add("kT",1e3,100,0,1e6); //touschek?
    cfg->par.Add("kG",0,100,0,1e6); //gas scattering
    cfg->par.Add("kI",0,0.01, 0, 10); //sum current dpnds
    cfg->par.Add("kP",100, 100, 0, 1e6); //vacuum dependancek
    cfg->par.Add("V1",0,1e-3); //volume meas bunch1
    cfg->par.Add("V2",0,1e-3); //volume mease bunch2
    cfg->par.Add("Vp",1,0.01,-2,+2);  //power of volume
    cfg->par.Add("kP2",0,10,0,1e6);  //power of volume


    cfg = &FitMethodList["LIFETIME"];
    cfg->name = "LIFETIME";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &lifetime;
    cfg->par.Add("p0",0,1);
    cfg->par.Add("A",1e3,1);
    cfg->par.Add("T0",0,1);
    cfg->par.Add("TAU",1e4,1);

    cfg = &FitMethodList["PPOLINOM"];
    cfg->name = "PPOLINOM";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &ppolinom;
    cfg->par.Add("p0",0);
    cfg->par.Add("p1",0);
    cfg->par.Add("p2",0);
    cfg->par.Add("p3",0);

    cfg = &FitMethodList["BARRIER"];
    cfg->name="BARRIER";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &barrier;
    cfg->par.Add("T1",10);
    cfg->par.Add("T2",20);
    cfg->par.Add("LVL0",0);
    cfg->par.Add("LVL1-LVL0",0);
    cfg->par.Add("LVL2-LVL0",0);

    cfg = &FitMethodList["COUNT"];
    cfg->name="COUNT";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &count;
    cfg->par.Add("N0",1e9);
    cfg->par.Add("TB",10000);

    cfg = &FitMethodList["COUNT2"];
    cfg->name="COUNT2";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &count2;
    cfg->par.Add("I0",1);
    cfg->par.Add("TB",10000);

    cfg = &FitMethodList["TSH_CURRENT"];
    cfg->name="TSH_CURRENT";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &tsh_current;
    cfg->par.Add("I0",1);
    cfg->par.Add("TB",10000);

    cfg = &FitMethodList["TSH_BG_CURRENT"];
    cfg->name="TSH_BG_CURRENT";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &tsh_bg_current;
    cfg->par.Add("I0",1);
    cfg->par.Add("TB",10000);
    cfg->par.Add("BG",0);

    cfg = &FitMethodList["EXP_CURRENT"];
    cfg->name="EXP_CURRENT";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &exp_current;
    cfg->par.Add("I0",1);
    cfg->par.Add("TB",10000);

    cfg = &FitMethodList["TSH_POL_BG_CURRENT"];
    cfg->name="TSH_POL_BG_CURRENT";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = cfg->NONE;
    cfg->func = &tsh_pol_bg_current;
    cfg->par.Add("T",0);
    cfg->par.Add("DELTA",0.001);
    cfg->par.Add("I",1);
    cfg->par.Add("TAU_BG",10000);
    cfg->par.Add("TAU_TSH",10000);

    cfg = &FitMethodList["COUNT4"];
    cfg->name="COUNT4";
    cfg->type = FitMethodConfiguration::FUNC;
    cfg->flag = FitMethodConfiguration::FIND_JUMP;
    cfg->func = &count4;
    cfg->par.Add("T",0);
    cfg->par.Add("DELTA",0.001);
    cfg->par.Add("I",1);
    cfg->par.Add("TAU_BG",10000);
    cfg->par.Add("TAU_TSH",10000);
    cfg->par.Add("p",0.1);

    //cfg = &FitMethodList["COUNT_CURRENT"];
    //cfg->name = "COUNT_CURRENT";
    //cfg->type = FitMethodConfiguration::FCN;
    //cfg->flag = FitMethodConfiguration::NONE;
    //cfg->func = 0;
    //cfg->par.Add("T",0, 100); //depolarization moment
    //cfg->par.Add("DELTA",0, 0.1); //depolarization effect
    //cfg->par.Add("TAU",1,1); //depolarization time
    //cfg->par.Add("POWER",1, 0.1); //depolarizer power
    //cfg->par.Add("I0",1,0.1); //initial current
    //cfg->par.Add("TAUTSH",1e4,1e3); //touschek beam life time
    //cfg->par.Add("TAUBG",1e4,1e3);//background beam life time
    //cfg->par.Add("PTSH",1,0.1);//touschek registration probability
    //cfg->par.Add("PBG",1,0.1); //background registration probability
    //cfg->fitter = new HeavisideCountCurrentFitter();

  }
}//namespace fit
