/*
 * =====================================================================================
 *
 *       Filename:  FitManager.cpp
 *
 *    Description:  Manange the fit procedures.
 *
 *        Version:  1.0
 *        Created:  02.11.2010 17:33:02
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#include <regex>

#include "FitManager.h"
#include "data/EnergyCalibrationStorage.h"
#include "gui/EnergyCalibrationView.h"

#include <TSystem.h>
#include <TStyle.h>
#include <TVirtualFitter.h>
#include <TF1.h>

#include <algorithm>
#include <list>

#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>

#include "Functions.h"
#include "CountCurrentFitter.h"
#include "CurrentFitter.h"
#include "RootFunctionFitter.h"

#include "RelaxationFitter.h"
#include "RelaxationFitter2.h"
#include "LinearJumpFitter.h"
#include "JumpFitter.h"
#include "SimpleAtanFitter.h"

#include "Time.h"

namespace fit
{
  double COUNT_TIME=20;
  double SCAN_TIME=1;
  const static unsigned MINFITPOINT=4;
  const static double SUCCESS_LEVEL=3.0;
  FitManager * FIT_MANAGER = 0;
  void * COUNTER_CURRENT_FITTER = 0;



  FitManager::FitManager(data::EnergyCalibrationStorage * ec )	 
  {
    std::clog<<LOG(3)<<"FitManager::FitManager START\n";
		//ROOT::Minuit2::MnPrint::SetLevel(0); //this code doest work
    InitFitMethodList();
    EC = ec;
    cfg = FitMethodList["LINEAR_JUMP"];
    clog << LOG(4) << "Find fit method: " << cfg.name <<  " with " << cfg.par.Params().size() << " parameters" << endl;
    fit_function=0;
    du_name="none";
    canvas = 0;
    graph  = 0;
    InitFitFunction();
    gStyle->SetOptFit();
    gStyle->SetStatX(0.27+gStyle->GetStatW());
    gStyle->SetStatY(0.9);

    /*  New engine to fit the data using Minuit2 */
    typedef std::unique_ptr<Fitter> uF_t;
    ROOT::Minuit2::MnUserParameters upar;
    upar.Add("T",0,100);
    upar.Add("DELTA",0,0.05);
    upar.Add("CONST",0,0.05);
    upar.Add("SLOPE1",0,1e-4);
    upar.Add("SLOPE2",0,1e-4);
    clog << LOG(5) << "FitManager::FitManager: Setup LINEAR_JUMP"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("LINEAR_JUMP",&linear_jump2,0,1e4,5), upar)));
    clog << LOG(5) << "FitManager::FitManager: Setup OLD_LINEAR_JUMP"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("OLD_LINEAR_JUMP",&linear_func,0,1e4,5), upar)));


    ROOT::Minuit2::MnUserParameters atan_par;
    atan_par.Add("T",0,100);
    atan_par.Add("DELTA",0,0.05);
    atan_par.Add("CONST",0,0.05);
    atan_par.Add("SLOPE1",0,1e-4);
    atan_par.Add("SLOPE2",0,1e-4);
    atan_par.Add("TAU",50,50);
    clog << LOG(5) << "FitManager::FitManager: Setup ATAN"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("ATAN",&atan_old,0,1e5,6), atan_par)));

		//relaxation of the polarization
    ROOT::Minuit2::MnUserParameters relaxation_par;
    relaxation_par.Add("T",0,0.01);
    relaxation_par.Add("DELTA",0,0.05);
    relaxation_par.Add("DN",0,0.05);
    relaxation_par.Add("TAU_POL",100,0.05);
    relaxation_par.Add("TAU_TSH",5000,0.05);
    relaxation_par.Add("SHIFT",0,0.01);
    clog << LOG(5) << "FitManager::FitManager: Setup RELAXATION"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("RELAXATION",&relaxation,0,1e5,6), relaxation_par,  false)));

    clog << LOG(5) << "FitManager::FitManager: Setup SMART_RELAXATION"<< std::endl;
    FitList.push_back(uF_t(new RelaxationFitter));

    clog << LOG(5) << "FitManager::FitManager: Setup SMART_RELAXATION2"<< std::endl;
    FitList.push_back(uF_t(new RelaxationFitter2));

    clog << LOG(5) << "FitManager::FitManager: Setup LINEAR_JUMP2"<< std::endl;
    FitList.push_back(uF_t(new LinearJumpFitter));

    clog << LOG(5) << "FitManager::FitManager: Setup EXP_JUMP"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("EXP_JUMP",&exp_func,0,1e4,5), upar)));
    clog << LOG(5) << "FitManager::FitManager: Setup HeavisideCountCurrentFitter"<< std::endl;
    FitList.push_back(uF_t(new HeavisideCountCurrentFitter));
    clog << LOG(5) << "FitManager::FitManager: Setup HeavisideCurrentFitter"<< std::endl;
    FitList.push_back(uF_t(new HeavisideCurrentFitter));
    clog << LOG(5) << "FitManager::FitManager: Setup SimpleAtan"<< std::endl;
    FitList.push_back(uF_t(new SimpleAtanFitter));
    clog << LOG(5) << "FitManager::FitManager: Setup SIMPLE_ATAN2"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("SIMPLE_ATAN2",&SimpleAtan::delta_average,0,1e5,6), atan_par)));
    ROOT::Minuit2::MnUserParameters double_linear_jump_par;
    double_linear_jump_par.Add("T",0,100);
    double_linear_jump_par.Add("DT", 0,100);
    double_linear_jump_par.Add("DELTA1",0,0.05);
    double_linear_jump_par.Add("DELTA2",0,0.05);
    double_linear_jump_par.Add("CONST", 0,0.05);
    double_linear_jump_par.Add("SLOPE1",0,1e-4);
    double_linear_jump_par.Add("SLOPE2",0,1e-4);
    double_linear_jump_par.Add("SLOPE3",0,1e-4);
    clog << LOG(5) << "FitManager::FitManager: Setup DOULBLE_LINEAR_JUMP"<< std::endl;
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("DOUBLE_LINEAR_JUMP",&double_linear_jump,0,1e5,8), double_linear_jump_par)));

    ROOT::Minuit2::MnUserParameters experf_par;
    experf_par.Add("T",0,100);
    experf_par.Add("DELTA",0,0.05);
    experf_par.Add("CONST",0,0.05);
    experf_par.Add("SLOPE1",0,1e-4);
    experf_par.Add("SLOPE2",0,1e-4);
    experf_par.Add("TAU",100.,10,  0.0, 10000.);
    experf_par.Add("POWER",1.0, 0.5);
    FitList.push_back(uF_t(new RootFunctionFitter(new TF1("SIMPLE_EXPERF",&experf,0,1e4,7), experf_par)));

    ROOT::Minuit2::MnUserParameters experf2_par;
    experf2_par.Add("T",0,100);
    experf2_par.Add("DELTA",0,0.05);
    experf2_par.Add("DN",0,0.05);
    experf2_par.Add("TAU_TSH",6000,5000,  0, 50000);
    experf2_par.Add("POWER",1.0, 0.5, 0, 100);
    experf2_par.Add("TAU",100,50, 0, 5000);
		//experf2_par.SetLimits("TAU", 0, 5000);
		//experf2_par.SetLimits("TAU_TSH", 100, 100000);
    FitList.push_back(uF_t(new RootFunctionFitter("EXPERF2", experf2, experf2_par, false)));

    clog << LOG(5) << "FitManager::FitManager: Setup SMART_RELAXATION"<< std::endl;
    FitList.push_back(uF_t(new JumpFitter));

    clog << LOG(5) << "FitManager::FitManager: FindFitter"<< std::endl;
    fitter = FindFitter("LINEAR_JUMP");
    std::clog<<LOG(3)<<"FitManager::FitManager FINISH\n";
  }

  FitManager::~FitManager(void)
  {
  }

  void FitManager::SetName(const std::string s)
  {
    du_name=s;
  }

  void FitManager::SetGraph(TGraphErrors * gr)
  {
    graph = gr;
  }

  void FitManager::SetCanvas(TCanvas * c)
  {
    canvas = c;
  }

  TGraph * FitManager::GetGraph(void)
  {
    return graph;
  }

  TCanvas * FitManager::GetCanvas(void)
  {
    return canvas;
  }

  void FitManager::SetFitFunction(std::string f)
  {
    fitter  = FindFitter(f);
    if(!fitter)
    {
      auto p = FitMethodList.find(f);
      if(p!=FitMethodList.end())
      {
        cfg=p->second;
      }
    }
  }

  std::string FitManager::GetFunctionName(void) const 
  {
		if(fitter) return fitter->name;
    return cfg.name;
  }

  void FitManager::Fit ( const char * formula , double tmin , double tmax )	
  {
    if ( ! graph ) 
    {
      std::cerr <<LOG(1)<< "FitManager: No graphic to fit\n";
      return;
    }
    gStyle->SetOptFit();
    gStyle->SetStatX(0.27+gStyle->GetStatW());
    gStyle->SetStatY(0.9);
    TF1 *func = new TF1("formula", formula,tmin, tmax);
    const unsigned NP = cfg.par.Params().size();
    for(unsigned i = 0; i<NP; i++)
    {
      func->SetParameter(i, cfg.par.Value(i));
    }
    std::cout << LOG(1) << "FitManaer: Use formula " << formula << " for fit\n";
    graph->Fit("formula", "RE");
    if ( canvas )   
    {
      canvas->Modified();
      canvas->Update();
    }
    canvas=0;
    graph=0;
    fit_result.resize(NP);
    for(unsigned i = 0; i<NP; i++)
    {
      fit_result[i].value = func->GetParameter(i);
      fit_result[i].error = func->GetParError(i);
    }
  }


  double FitManager::Fit(TGraphErrors * graph, double * par, const char * opt, const char * gopt)   
  {
    fit_function->SetParameters(par);
    fit_function->SetRange(tmin,tmax);
    graph->Fit(cfg.name.c_str(),opt,gopt);
    const unsigned NP = cfg.par.Params().size();
    fit_result.resize(NP);
    for(unsigned i = 0; i<fit_result.size(); i++)
    {
      fit_result[i].value = fit_function->GetParameter(i);
      fit_result[i].error = fit_function->GetParError(i);
    }
    return fit_function->GetChisquare()/fit_function->GetNDF();
  }

  void FitManager::Fit(double tmin, double tmax)	
  {
    std::cout<<LOG(4)<<"FitManager::Fit START\n";
    if(gROOT->FindObject(du_name.c_str()))
    { 
      Fit(graph,tmin,tmax);
    }
    else
    {
      std::cerr << LOG(1) << "ERROR: FitManager: Display Unit " << du_name << " for fiting does not exist\n";
      graph=0;
      canvas=0;
      du_name="none";
    }
    std::cout<<LOG(4)<<"FitManager::Fit FINISH\n";
  }


  int FitManager::InitFitFunction(void)	
  {
    if ( fit_function != 0 ) delete fit_function;
    const unsigned NP = cfg.par.Params().size();
    fit_function = new TF1(cfg.name.c_str(), cfg.func, tmin, tmax, NP);
    fit_function->SetNpx(1000);
    clog << LOG(INFO_LOG) << "Init fit function " << cfg.name << " with " << NP << " parameters" << endl;
    cout << LOG(INFO_LOG) 
      << setw(3)  << '#'
      << setw(10) << "Name"
      << setw(10) << "Initial"
      << setw(10) << "lower limit"
      << setw(10) << "upper limit"
      << setw(10) << "fixed"
      << setw(10) << "has limit" << endl;
    for(unsigned i =0 ;i < NP; ++i) 
    {
      const ROOT::Minuit2::MinuitParameter & p = cfg.par.Parameter(i);
      fit_function->SetParName(i, p.Name());
      fit_function->SetParameter(i, p.Value());
      fit_function->SetParLimits(i, p.LowerLimit(), p.UpperLimit());
      if(p.IsFixed()) fit_function->FixParameter(i, p.Value());
      cout << LOG(INFO_LOG) 
        << setw(3)  << i
        << setw(10) << p.Name()
        << setw(10) << p.Value() 
        << setw(10) << p.LowerLimit()
        << setw(10) << p.UpperLimit()
        << setw(10) << p.IsFixed() 
        << setw(10) << p.HasLimits() << endl;
    }
    return 0;
  }

  bool FitManager::FindJump(TGraphErrors * graph, double * par_find)
  {
    clog << LOG(3) << "FitManager::FindJump\n";
    bool find=false;
    double chi_tmp=1e200;
    double chi=chi_tmp-1;
    const char * fit_options ="RQ";
    for(double t = tmin; t< tmax; t+=EC->cfg.count_time )
    {
      clog << LOG(4) << "checking jump at time " << t << "\n";
      vector <double> par_tmp(cfg.par.Params());
      par_tmp[0] = t;
      chi_tmp = Fit(graph,&par_tmp[0],fit_options);
      if ( chi_tmp < chi)	
      {
        find=true;
        chi=chi_tmp;
        fit_function->GetParameters(par_find);
      }
      if(canvas)  
      {
        canvas->cd();
        canvas->Modified();
        canvas->Update();
      }
      gSystem->ProcessEvents();
    }
    return find;
  }

  void FitManager::SetCountTime(void)
  {
    COUNT_TIME = EC->cfg.count_time;
  }


  void FitManager::Fit(TGraphErrors * graph, double tmn, double tmx)
  {
		return;
    float stat_shift_x=0.27,stat_shift_y=0.9;
    gStyle->SetOptFit();
    gStyle->SetStatX(stat_shift_x+gStyle->GetStatW());
    gStyle->SetStatY(stat_shift_y);
    COUNT_TIME = EC->cfg.count_time;
    SCAN_TIME=EC->cfg.depolarizer.scan_time;
    tmin=tmn;
    tmax=tmx;
    InitFitFunction();
    std::vector <double> par_find(cfg.par.Params());
    std::vector <double> par_err(cfg.par.Errors());
    if(cfg.flag == FitMethodConfiguration::FIND_JUMP)
    {
      //Если время зафиксировано, то не будем перебирать
      //все времена
      if (cfg.par.Parameter(cfg.par.Index("T")).IsFixed())
      {
        EC->chi2 = Fit(graph,&par_find[0],"RE");
      } 
      else 
      {
        if(FindJump(graph,&par_find[0]))  
          EC->chi2 = Fit(graph,&par_find[0],"RE");
      }
    } 
    else  
    {
      EC->chi2 = Fit(graph,&par_find[0],"RE");
      fit_function->GetParameters(&par_find[0]);
    }
    fit_result.resize(par_find.size());
    for(unsigned i=0;i<fit_result.size();i++)
    {
      fit_result[i] =  data::PhysValue(fit_function->GetParameter(i), fit_function->GetParError(i));
      par_err[i]=fit_result[i].error;
      par_find[i]=fit_result[i].value;
    }
    FillEnergyCalibrations(par_find,par_err);
  }


  //void FitManager::SetPar (const std::vector < FitParConfiguration> &p)
  void FitManager::SetParameters(const ROOT::Minuit2::MnUserParameters &p)
  {
    cfg.par = p;
  }

  ROOT::Minuit2::MnUserParameters FitManager::GetParameters(void) const
  {
    //return cfg.par;
    if(fitter) return fitter->inipar;
    else
    {
      return cfg.par;
    }
  }

  void FitManager::FillEnergyCalibrations(const std::vector <double>  & par, const std::vector <double> &par_err )
  {
    EC->isfit=true;
		if(std::isnan(par[0]))
		{
			EC->isfit=false;
			EC->success=false;
			return;
		}
    EC->success = (par[1]/par_err[1] >= SUCCESS_LEVEL) ? true : false; //Скачек считаем успешным, если его величина больше 3-х сигма.
    data::PhysValue td(par[0], par_err[0]);
    EC->cfg.date=time_t(EC->begin_time.tv_sec + td.value);
    data::PhysValue fd = getLinearAproximatedFrequency(td,EC->t,EC->DepFr);

    auto td_double = std::chrono::duration<double> (td.value);
    auto td_duration = std::chrono::duration_cast < std::chrono::system_clock::duration > (td_double);
    auto td_clock = EC->Begin_time + td_duration;

    std::cout << LOG(1) << boost::format("Linear approximation: %20.3f Hz") % fd.value << std::endl;
    if( EC->HRDepFreq )
    {
      double hr_fd = fuzzy_find(*EC->HRDepFreq, td_clock);
      double hr_fd2 = linear_find(*EC->HRDepFreq, td_clock);
      std::cout << LOG(1) << boost::format("High resolution frequency: %20.3f Hz") % hr_fd << std::endl;
      std::cout << LOG(1) << boost::format("High resolution frequency (linear approximation): %20.3f Hz") % hr_fd2 << std::endl;
    }



    /* Ошибка деполяризатора складывается
     * из ошибки определения средней частоты деполяризации, это 
     * то, что возвращает функция getLinearAproximatedFrequency
     * плюс ошибка от того, что мы не знаем где именно внутри полосы
     * деполяризатора произошла деполяризация
     */
    data::PhysValue rf = getLinearAproximatedFrequency(td,EC->t,EC->RF);
    std::cout <<  LOG(1) << "Radio frequency = " << boost::format("%10.1f Hz") % rf.value  << std::endl;
    if ( fabs(rf.value/222. - EC->cfg.depolarizer.revolution_frequency.value) > 10  )	
    {
      std::cout << LOG(1) << "WARNING: Something wrong with RF measurement. Using default value of revolution frequency: " << EC->cfg.depolarizer.revolution_frequency.value << " You MUST recalculate the ENERGY from measured frequency." << '\n';
    } 
    else 
    {
      EC->cfg.depolarizer.revolution_frequency.value = rf.value/222.;
      EC->cfg.depolarizer.revolution_frequency.error = 0;
    }

    data::PhysValue en = EC->frequency2energy(fd);

    EC->fit_param.resize(par.size());
    for(unsigned i = 0; i < EC->fit_param.size() ; i++)	
    {
      data::PhysValue tmp(par[i], par_err[i]);
      EC->fit_param[i] = tmp;
    }
    EC->Fr = fd;
    EC->E = en;
    EC->cfg.depolarizer.direction = EC->cfg.depolarizer.initial < EC->cfg.depolarizer.final ? +1 : -1;
    std::cout.precision(8);
    std::cout << LOG(1) << fit::TimeString(td)      << std::endl;	
		if(fd.value!=0)
		{
			std::cout << LOG(1) << fit::FrequencyString(fd) << std::endl;
			std::cout << LOG(1) << fit::EnergyString(en) << std::endl;
		}
  }


  data::PhysValue getLinearAproximatedFrequency(data::PhysValue td, const data::Array &Tm, const data::Array &MFr)
  {
    //This function calculate depolarization frequency and it's error using the vector of MFr
    using namespace math;
    unsigned i;
    unsigned Npoint = MFr.size();
    for( i = 0; i<Npoint-1 && td.value > Tm[i].value; i++);
    if(i>0) i--;
    double f1  = MFr[i].value;
    double f2  = MFr[i+1].value;
    double t1  = Tm[i].value;
    double t2  = Tm[i+1].value;
    double df = f2-f1;
    double dt = t2-t1;
    data::PhysValue fr(f1 + (td.value-t1)*df/dt, sqrt(sq(MFr[i].error) + sq(td.error * df/dt) )); 
    return fr;
  }

  void FitManager::FitFCN(void) 
  {
    TMinuit * minuit = new TMinuit(3);
    vector <string> find_vector;
    boost::split(find_vector,du_name,boost::is_any_of("-"));
    string data_name = find_vector[0];
    for(unsigned i=1;i<find_vector.size()-1;i++) data_name=data_name+"-"+find_vector[i];
    cout << LOG(1) << "Data name to fit " << data_name << "\n";
    map < string, vector < data::Array >* > data_list;
    data_list["normalized2-counter-sum-vs-Ir"] = &EC->CountSumNormI2;
    data_list["normalized2-coincidence-sum-vs-Ir"] = &EC->CoincSumNormI2;
    data_list["normalized2-correlated-sum-vs-Ir"] = &EC->CorSumNormI2;
    map < string, vector < data::Array >* >::iterator it = data_list.find(data_name);
    if(it!=data_list.end())
    {
      bunchData = it->second;
      fcn_fit_graph_name=data_name;
    }
    else
    {
      bunchData  = &EC->CountSumNormI2;
      fcn_fit_graph_name="normalized2-counter-sum-vs-Ir";
      clog << LOG(1) << "Unable to find fit data. Return to default: " << fcn_fit_graph_name << "\n";
    }
    minuit->SetFCN(FitManager::normalized_rate_fcn);
    int ierflg=0;
    // Assign the names for the parameters
    //vector <string> par_name = RootFunctionParName["BACKGROUND"];
    //std::vector<fit::FitParConfiguration> & pcfg = FitMethodList["BACKGROUND"].par;
    //std::vector<fit::FitParConfiguration> & pcfg = cfg.par;
    ROOT::Minuit2::MnUserParameters & pcfg = cfg.par;
    const unsigned  NP  = pcfg.Params().size();
    for(unsigned i = 0; i<NP;++i)
    {
      ROOT::Minuit2::MinuitParameter p = pcfg.Parameter(i);
      minuit->mnparm(i,p.Name(),p.Value(), p.Error(), p.LowerLimit(),p.UpperLimit(),ierflg);
      if ( p.IsFixed()) minuit->FixParameter(i);
    }
    cout << LOG(1) << "Bad point list: \n";
    for(list<unsigned>::iterator it=BadPointList.begin();it!=BadPointList.end();it++) cout << *it << "  " ;
    cout << LOG(1) << "\n";
    FIT_MANAGER = this;
    minuit->Migrad();
    minuit->Migrad();
    fit_result.resize(NP);
    vector <double> par(NP);
    for(unsigned i=0;i<NP;++i)
    {
      minuit->GetParameter(i,fit_result[i].value, fit_result[i].error);
      par[i] = fit_result[i].value;
    }

    vector <double> scale(fit_result.size());
    vector <string> unit (fit_result.size());
    data::PhysValue kT = fit_result[0]; scale[0]=1e-3; unit[0]="kHz/mA^2";
    data::PhysValue kG = fit_result[1]; scale[1]=1e-3; unit[1]="kHz/mA";
    data::PhysValue kG2 = fit_result[2];scale[2]=1e-3; unit[2]="kHz/mA^2";
    data::PhysValue kG3 = fit_result[3];scale[3]=1e-3; unit[3]="kHz/mA^3";
    scale[4]=1e-3; unit[4]="1/mA";
    scale[5]=1; unit[5]="mm^2";
    scale[6]=1; unit[6]="mm^2";
    scale[7]=1;    unit[7]="";
    double total = kT.value + kG.value + kG2.value + kG3.value;
    double chi2;
    double chi2ndf;
    minuit->Eval(3,0,chi2,&par[0],ierflg);
    chi2ndf = chi2/(2*EC->t.size()-minuit->GetNumFreePars()) ;
    cout << LOG(1) << "chi2/NDF=" <<chi2ndf << "\n";

    for(unsigned i=0; i<4; ++i)
    {
      double err = fit_result[i].error*sqrt(chi2ndf);
      cout << LOG(1) 
        << setw(4)  << pcfg.Name(i) 
        << setw(12) << fit_result[i].value*scale[i]
        << setw(4)  << "+- "
        << setw(12) << err*scale[i]
        << setw(10) << unit[i]
        << setw(12) << fit_result[i].value/total*100. 
        << setw(4)  << "+- "
        << setw(12) << err/total*100. 
        << " %"  <<endl;
    }
    for(unsigned i=4; i< fit_result.size();++i)
    {
      double err = fit_result[i].error*sqrt(chi2ndf);
      cout << LOG(1) 
        << setw(4) << pcfg.Name(i) 
        << setw(12) << fit_result[i].value*scale[i]
        << setw(4) << "+- "
        << setw(12) << err*scale[i]
        << setw(10) << unit[i] << endl;
    }

    delete minuit;
  }

  void FitManager::FitFCN2(void) 
  {
    ROOT::Minuit2::MnUserParameters & pcfg = cfg.par;
    const unsigned  NP  = pcfg.Params().size();
    TMinuit * minuit = new TMinuit(NP);
    vector <string> find_vector;
    boost::split(find_vector,du_name,boost::is_any_of("-"));
    string data_name = find_vector[0];
    for(unsigned i=1;i<find_vector.size()-1;i++) data_name=data_name+"-"+find_vector[i];
    cout << LOG(1) << "Data name to fit " << data_name << "\n";
    map < string, vector < data::Array >* > data_list;
    data_list["normalized-counter-sum-vs-I"] = &EC->CountSumNormI;
    data_list["normalized-coincidence-sum-vs-I"] = &EC->CoincSumNormI;
    data_list["normalized-correlated-sum-vs-I"] = &EC->CorSumNormI;
    data_list["normalized-random-sum-vs-I"] = &EC->RndSumNormI;
    data_list["normalized-random-coinc-sum-vs-I"] = &EC->RndCoincSumNormI;
    map < string, vector < data::Array >* >::iterator it = data_list.find(data_name);
    if(it!=data_list.end())
    {
      bunchData = it->second;
      fcn_fit_graph_name=data_name;
      cout << "Set fcn_fit_graph_name = " << it->first <<  " while "<< data_name << " requested" << endl;
    }
    else
    {
      bunchData  = &EC->CountSumNormI;
      fcn_fit_graph_name="normalized-counter-sum-vs-I";
      clog << LOG(1) << "Unable to find fit data. Return to default: " << fcn_fit_graph_name << "\n";
    }
    minuit->SetFCN(FitManager::norm_rate_fcn);
    int ierflg=0;
    for(unsigned i = 0; i<NP;++i)
    {
      ROOT::Minuit2::MinuitParameter p = pcfg.Parameter(i);
      minuit->mnparm(i,p.Name(),p.Value(), p.Error(), p.LowerLimit(),p.UpperLimit(),ierflg);
      if ( p.IsFixed()) minuit->FixParameter(i);
    }
    cout << LOG(1) << "Bad point list: \n";
    for(list<unsigned>::iterator it=BadPointList.begin();it!=BadPointList.end();it++)
    {
      cout << *it << "  " ;
    }
    cout << LOG(1) << "\n";
    FIT_MANAGER = this;

    minuit->Migrad();
    fit_result.resize(NP);
    vector <double> par(NP);
    for(unsigned i=0;i<par.size();++i)
    {
      minuit->GetParameter(i,fit_result[i].value, fit_result[i].error);
      par[i] = fit_result[i].value;
    }
    vector <double> scale(fit_result.size());
    vector <string> unit (fit_result.size());
    data::PhysValue & kT = fit_result[0]; scale[0]=1e-3; unit[0]="kHz/mA^2";
    data::PhysValue & kG = fit_result[1]; scale[1]=1e-3; unit[1]="kHz/mA/nTorr";
    data::PhysValue & kI = fit_result[2]; scale[2]=1;  unit[2]="1/mA";
    data::PhysValue & kP = fit_result[3]; scale[3]=1e-3; unit[3]="kHz/mA/nTorr";
    scale[4]=1; unit[4]="";
    scale[5]=1; unit[5]="";
    scale[6]=1; unit[6]="";
    //data::PhysValue & kP2 = fit_result[7]; scale[3]=1e-3; unit[3]="kHz/mA^2/nTorr^2";
    double chi2;
    double chi2ndf;
    minuit->Eval(3,0,chi2,&par[0],ierflg);
    //NO fit the vacuum dependance
    TGraph * vacuum_graph = new TGraph(EC->MediumVacuumPresure.size(), EC->Isum.value(), EC->MediumVacuumPresure.value());
    TCanvas * vacuum_canvas = new TCanvas;
    vacuum_canvas->SetBorderMode(0);
    vacuum_graph->Draw("al");
    vacuum_graph->Fit("pol1");
    clog << "Getvacuum fun" << endl;
    TF1 * vacuum_fun = vacuum_graph->GetFunction("pol1");
    data::PhysValue PatI0(vacuum_fun->GetParameter(0),vacuum_fun->GetParError(0));
    data::PhysValue PpI(vacuum_fun->GetParameter(1), vacuum_fun->GetParError(1));
    cout << "Vacuum at zero current: " << PatI0*1e-3<<  " nTorr" << endl;
    cout << "Vacuum changing koef: " << PpI << " nTorr/mA" << endl;
    cout << "  PpI/PatI0 1/mA" << endl;
    double I1mA =1000; //mkA
    double P1 = vacuum_fun->Eval(I1mA)*1e-3 ;
    cout <<  "Vacuum at 1mA: " << P1 << " nTorr" << endl;
    chi2ndf = chi2/(2*EC->t.size()-minuit->GetNumFreePars()) ;
    double P0 = EC->MediumVacuumPresure[0].value*1e-3; //begining vacuum in nTorr
    double PL = EC->MediumVacuumPresure.back().value*1e-3; //last vacuum in nTorr
    data::PhysValue total = kT+kP*P1+kG*P1*(kI+1.0); //all is normalized on 1mA.
    cout << "Begining vacuum: " << P0 << "nTorr, last vacuum: " << PL << " nTorr. Changing: " << PL-P0 << " nTorr " << (PL-P0)/P0*100 << "\%" << endl;
    cout <<  "Total load for 1mA at the begining: " << total << " kHz" << endl;
    double I0 = EC->Isum[0].value*1e-3;//begin current, mA
    double IL = EC->Isum.back().value*1e-3;//last current mA
    cout <<  "Total begin current: " << I0 << " mA, -> " << IL << " dI: " 
      <<  IL-I0 << " mA  " <<  (IL-I0)/I0*100. << " %" << endl;
    cout << LOG(1) << "chi2/NDF=" <<chi2ndf << "\n";
    if(chi2ndf < 1) chi2ndf=1;
    double scale_factor = chi2ndf < 1? 1: sqrt(chi2ndf);
    for(unsigned i=0; i<fit_result.size(); ++i)
    {
      fit_result[i].error=fit_result[i].error*scale_factor;
      total.error = total.error*scale_factor;
      cout << setw(4)  << cfg.par.Name(i);
      switch (i)
      {
        case 0:
        case 7:
          cout 
            << setw(27) << fit_result[i]*scale[i]
            << setw(15) << unit[i]
            << setw(27) << fit_result[i]/total*100.
            << " %"; 
          break;
        case 1:
          cout << setw(27) << kG*scale[i] << setw(15) << unit[i];
          cout << setw(27) << kG*P1/total*100. << " %";
          break;
        case  2:
          cout << setw(27) << kI*scale[i]  << setw(15) << unit[i];
          cout << setw(27) << kI*kG*P1/total*100. << " %";
          cout << setw(27) << kI*kG*scale[1]  << " kHz/mA^2/nTorr";
          break;
        case 3:
          cout << setw(27) << kP*scale[i]  << setw(15) << unit[i];
          cout << setw(27) << kP*P1/total*100. << " %";
          break;
        default:
          cout << setw(27) << fit_result[i].value*scale[i]  << setw(15) << unit[i];
          break;
      }
      cout << endl;
    }
    delete minuit;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  FitManager
   *      Method:  rate_sum_bg_fcn
   * Description:  Find background and signal in coincidence rate.
   *--------------------------------------------------------------------------------------
   */
  void FitManager::normalized_rate_fcn(Int_t &npar, Double_t *gin, Double_t &sum, Double_t *par, Int_t iflag) 
  {
    FitManager * fm = FIT_MANAGER; /*  external link to get data */
    data::EnergyCalibrationStorage * ec = fm->EC;
    vector <data::PhysValue> K(par, par + fm->cfg.par.Params().size());
    sum=0;
    for(unsigned bunch=0; bunch < 2; bunch++) 
    {
      for(unsigned i=0; i<ec->IRatio[bunch].size();i++)
      {
        /* check for bad point wich is should not fited */
        if(find(BadPointList.begin(), BadPointList.end(), i) != BadPointList.end()) continue;
        if(ec->t[i].value < fm->tmin || ec->t[i].value > fm->tmax) continue;
        data::PhysValue dx = fm->NormalizedRate(i,bunch,K)  - (*fm->bunchData)[bunch][i];
        sum += (dx.value*dx.value) / (dx.error*dx.error);
      }
    }
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  FitManager
   *      Method:  rate_sum_bg_fcn
   * Description:  Find background and signal in coincidence rate.
   *--------------------------------------------------------------------------------------
   */
  void FitManager::norm_rate_fcn(Int_t &npar, Double_t *gin, Double_t &sum, Double_t *par, Int_t iflag) 
  {
    FitManager * fm = FIT_MANAGER; /*  external link to get data */
    data::EnergyCalibrationStorage * ec = fm->EC;
    vector <data::PhysValue> K(par, par + fm->cfg.par.Params().size());
    sum=0;
    for(unsigned bunch=0; bunch < 2; bunch++) 
    {
      for(unsigned i=0; i<ec->IRatio[bunch].size();i++)
      {
        /* check for bad point wich is should not fited */
        if(find(BadPointList.begin(), BadPointList.end(), i) != BadPointList.end()) continue;
        if(ec->t[i].value < fm->tmin || ec->t[i].value > fm->tmax) continue;
        data::PhysValue dx = fm->NormRate(i,bunch,K)  - (*fm->bunchData)[bunch][i];
        sum += (dx.value*dx.value) / (dx.error*dx.error);
      }
    }
  }

  void FitManager::SetTimeRange(double time_min, double time_max)
  {
    tmin = time_min;
    tmax = time_max;
    fitter->SetTimeRange(tmin, tmax);
  }


  data::PhysValue FitManager::NormalizedRate(unsigned i, unsigned bunch, const vector < data::PhysValue> & par)
  {
    using namespace ibn;
    data::PhysValue kT=par[0];
    data::PhysValue kG=par[1];
    data::PhysValue kG2=par[2];
    data::PhysValue kG3=par[3];
    data::PhysValue kT2 = par[4];
    data::PhysValue V0[2];
    V0[0] = par[5];
    V0[1] = par[6];
    data::PhysValue Vpower = par[7]; /* power of  volume*/
    /*  Touschek contribution */
    data::PhysValue Vini = EC->SigmaX[bunch][0]*EC->SigmaZ[bunch][0];
    data::PhysValue Vcur = EC->SigmaX[bunch][i]*EC->SigmaZ[bunch][i];
    data::PhysValue Vr = (V0[bunch]+1.0)/(Vcur/Vini +V0[bunch]);
    data::PhysValue xT = kT*pow(Vr,Vpower)*(kT2*EC->I[bunch][i]/1e3+1.0);
    /*  Beam gas scattering background */
    data::PhysValue xG = kG/EC->I[bunch][i]*1e3;
    /*  Beam gas background for vacuum */
    data::PhysValue xG2 = kG2*(EC->IRatio[bunch][i]+1.0);
    /*  square wacuum */
    data::PhysValue i2sum(0);
    for(unsigned b=0;b!=EC->numbunch;b++) i2sum= i2sum+EC->I[b][i]*EC->I[b][i]/1e6;
    data::PhysValue xG3 = kG3*i2sum/EC->I[bunch][i]*1e3;
    /*  total */
    data::PhysValue x = xT+xG+xG2+xG3;
    return x;
  }

  data::PhysValue FitManager::NormRate(unsigned i, unsigned bunch, const vector < data::PhysValue> & par)
  {
    data::PhysValue V0[2];
    data::PhysValue kT=par[0];
    data::PhysValue kG=par[1];
    data::PhysValue kI=par[2]; //vacuum coef
    data::PhysValue kP = par[3];
    V0[0] = par[4];
    V0[1] = par[5];
    data::PhysValue Vpower = par[6];
    data::PhysValue kP2 = par[7];
    data::PhysValue Vini = EC->SigmaX[bunch][0]*EC->SigmaZ[bunch][0]; //begining size
    data::PhysValue Vcur = EC->SigmaX[bunch][i]*EC->SigmaZ[bunch][i]; //current beam size
    data::PhysValue Vr = (V0[bunch]+1.0)/(Vcur/Vini +V0[bunch]);
    data::PhysValue xT = kT*pow(Vr,Vpower)*EC->I[bunch][i]*1e-3;
    /*  Beam gas scattering background */
    data::PhysValue xG = kG*EC->MediumVacuumPresure[0]/1e3*(kI*EC->Isum[i]*1e-3+1.0);
    data::PhysValue xP = kP*EC->MediumVacuumPresure[i]/1e3; //normalize koef on nTorr
    data::PhysValue xP2 = kP2*EC->I[bunch][i]*1e-3*EC->MediumVacuumPresure[i]/1e3*EC->MediumVacuumPresure[i]/1e3; //normalize koef on nTorr
    data::PhysValue x = xT+xG+xP+xP2;
    return x;
  }
  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  EnergyString
   *  Description:  Print energy measurement result into string
   * =====================================================================================
   */
  std::string EnergyString(data::PhysValue a)
  {
		return (boost::format("E = %11.6f +- %10.6f MeV") %a.value % a.error).str();
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  FrequencyString
   *  Description:  Print frequency measurement result into string
   * =====================================================================================
   */
  std::string FrequencyString(data::PhysValue a)
  {
		if(a.value==0) return std::string("");
		boost::format fmt("Fd = %+10.3f +- %10.3f Hz");
		return (fmt % a.value % a.error).str() ;//  fmt.str();
	}

  std::string TimeString(data::PhysValue a)
  {
    std::ostringstream s;
    s  << boost::format("Depolarization moment: Td = %5.3f +- %5.3f sec") % a.value % a.error;
    return s.str();
  }

  Fitter * FitManager::FindFitter(const string & method_name)
  {
    auto itfit = find_if
      (
       begin(FitList),
       end(FitList),
       [&method_name](const unique_ptr<Fitter> & f)
       { 
         return f->name==method_name; 
       }
      );
    if(itfit==FitList.end()) return nullptr;
    fitter = itfit->get();
    return itfit->get();
  }

  void FitManager::Fit(gui::EnergyCalibrationView * ECView,  std::string  fit_method, std::string  display_unit_name)
  {
		std::clog << LOG(DEBUG_LOG) << "Fitter::Fit BEGIN" << std::endl;
		if(fit_method=="") fit_method="LINEAR_JUMP";
		auto fitter = FindFitter(fit_method);
		if(fitter==nullptr) 
		{
			std::clog << LOG(NORMAL_LOG) << "ERROR: Unable to find fitter " << fit_method << std::endl;
			return ;
		}
    SetCountTime();
		auto canvas = ECView->GetCanvas(display_unit_name);
		if(canvas==nullptr) 
		{
			std::clog << LOG(NORMAL_LOG) << "ERROR: Unable to find data to fit " << display_unit_name << std::endl;
			return;
		}
    fitter->SetView(ECView);
    fitter->SetData(ECView->GetEnergyCalibrationStorage(),display_unit_name);
    std::clog << "Befire fitter->Fit()"<<std::endl;
    fitter->Fit();
    std::clog << "After fitter->Fit()"<<std::endl;
    fitter->Print();
    canvas->cd();
    fitter->DrawFunction();
    FillEnergyCalibrations(fitter->minpar.Params(),fitter->minpar.Errors());
    canvas->SetMeasured();
    canvas->Modified();
    canvas->Update();
  }

	void FitManager::Fit(gui::EnergyCalibrationView *ECView)
	{
    Fit(ECView,ECView->GetEnergyCalibrationStorage()->cfg.fit_method, PCFG.gui.default_graph);
	}

} //namespace fit
