//#=====================================================================================
//#
//#       Filename:  Fit.h
//#
//#    Description:  Class which controls fit procedure.
//#
//#        Version:  1.0
//#        Created:  03.03.2003
//#       Revision:  none
//#       Compiler:  GNU C/C++
//#
//#         Author:  Ivan B. Nikolaev
//#        Company:  BINP,Russia
//#          Email:  I.B.Nikolaev@inp.nsk.su
//#
//#=====================================================================================

#ifndef IBN_FIT_H
#define IBN_FIT_H

#include <vector>
#include <string>
#include <memory>



#include <TF1.h>
#include <TGraphErrors.h>
#include <TCanvas.h>
#include <TROOT.h>
#include <TMinuit.h>

#include "Log.h"
#include "data/Data.h"
#include "data/Array.h"
#include "FitConfig.h"

#include "Fitter.h"

namespace data {class EnergyCalibrationStorage;}
namespace gui {class EnergyCalibrationView;}

namespace fit	
{
  class FitManager;
  extern FitManager * FIT_MANAGER;

  class FitManager	
  {			
    data::EnergyCalibrationStorage* EC;  /*  Data with energy calibrations */
    std::list<std::unique_ptr<Fitter>>  FitList; /* List of fittters */
    Fitter * fitter;
    std::string function_name; //the name of fit function
    std::string du_name; //the name of what we should fit

    FitMethodConfiguration cfg; //Fit method configuration
    double tmin, tmax; //range for fit

    TF1 * fit_function;
    TGraphErrors * graph;
    TCanvas * canvas;

		//void Fit(data::EnergyCalibrationStorage * ec);

    double Fit(TGraphErrors * graph, double * par, const char * opt, const char * gopt="");
    void Fit(TGraphErrors * graph, double tmn=0, double tmx=0);
    bool FindJump(TGraphErrors * graph, double * par_find);

    public:
		void Fit(gui::EnergyCalibrationView *);
  	void Fit(gui::EnergyCalibrationView * ECView,  std::string  method_name, std::string  display_unit_name);
    void FillEnergyCalibrations(const std::vector <double>  & par, const std::vector <double> &par_err );
    std::string fcn_fit_graph_name;  /*  data for fitting via fcn */

    std::vector <data::PhysValue> fit_result; /*  result of the fit */

    FitManager(data::EnergyCalibrationStorage * ec );

    ~FitManager(void);

    data::EnergyCalibrationStorage * GetEnergyCalibration(void) { return EC; }

    void SetName(const std::string s);

    void SetGraph(TGraphErrors * gr);

    void SetCanvas(TCanvas * c=0);

    TGraph * GetGraph(void);

    TCanvas * GetCanvas(void);

    void SetFitFunction(std::string f);

    std::string GetFunctionName(void) const;

    void SimpleFit(double tmin=0, double tmax=0);

    void Fit ( const char * formula , double tmin =0, double tmax = 0);	


    void Fit(double tmin = 0, double tmax = 0);

    void FitFCN(void) ;
    void FitFCN2(void) ;

    void SetParameters (const ROOT::Minuit2::MnUserParameters &p);
    ROOT::Minuit2::MnUserParameters GetParameters(void)  const;

    int InitFitFunction(void);

    std::vector < data::Array > * bunchData; /* Bunch data for fitting */

    data::PhysValue NormalizedRate(unsigned i, unsigned bunch,  const std::vector < data::PhysValue> & param);
    data::PhysValue NormRate(unsigned i, unsigned bunch,  const std::vector < data::PhysValue> & param);

    static void normalized_rate_fcn(Int_t &npar, Double_t *gin, Double_t &sum, Double_t *par, Int_t iflag);
    static void norm_rate_fcn(Int_t &npar, Double_t *gin, Double_t &sum, Double_t *par, Int_t iflag);

    void SetTimeRange(double time_min=0, double time_max=0);

    Fitter * FindFitter(const std::string & name);
    Fitter * GetFitter(void) { return fitter;}
    void SetCountTime(void);
    const std::list<std::unique_ptr<Fitter>> & GetFitList() { return FitList;};
  };

  data::PhysValue getLinearAproximatedFrequency(data::PhysValue td, const data::Array &Tm, const data::Array &MFr);


  
  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  EnergyString
   *  Description:  Print energy measurement result into string
   * =====================================================================================
   */
  std::string EnergyString(data::PhysValue a);

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  FrequencyString
   *  Description:  Print frequency measurement result into string
   * =====================================================================================
   */
  std::string FrequencyString(data::PhysValue a);

  std::string TimeString(data::PhysValue a);

}


#endif //IBN_FIT_H
