#ifndef IBN_FIT_FUNCTIONS_H
#define IBN_FIT_FUNCTIONS_H

#include <cmath>

#include <map>
#include <list>
#include <string>
#include <vector>

#include <limits>

#include <algorithm>
#include <functional>


#include "ibn/integral.h"
#include "Formula.h"

#include "heaviside.h"
#include "CountFunctions.h"
#include "Log.h"

namespace fit
{
	extern double COUNT_TIME;


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  USFL_LOG
   *  Description:  This function is integral of 1/(t+tau)
   * =====================================================================================
   */
  inline double USFL_LOG(double t, double TAU, double T)
  {
    double x=T/(t+TAU);
    if(x>=1.0) return std::numeric_limits<double>::max();
    return  - log(1.0 - x);
  }

  inline double lnc(double x)
  {
    if(x<=-1) return std::numeric_limits<double>::max();
    if(x==0) return 1.0;
    return log(1.0+x)/x;
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  delta_nodep
   *  Description:  Nodepolarization part of average Delta
   * =====================================================================================
   */
  inline double delta_nodep
    (
     double t   /* time  */,
     double D   /* polarization contribution */,  
     double DN  /* diff of particles */,
     double TAU /* beam life time */,
     double T   /* count time  */
    )
  {
    return D + 2.0*(DN-D)*TAU/(t+TAU)*lnc(-T/(t+TAU));
  }


  inline double instant_depolarization(
      double t  /* current time */, 
      double td /* moment of depolarization */, 
      double D  /* polarization contribution */,
      double DN /* initial relative particle number difference */,
      double TAU /* life time of the beam */,
      double T   /* count time */)
  {
    if(TAU<=0) return std::numeric_limits<double>::max();
    double result=delta_nodep(t,D,DN,TAU,T);
    if(t<=td) return result;
    double tmp=(td+TAU)/(t+TAU);
    if(t>=td+T) return result+D*( 1.0 - 2.*(1.0 - tmp)*lnc(-T/(t+TAU)));
    return result+D*(t-td)/T *( 1.  - 2.*(1.0 - tmp )*lnc(-(t-td)/(t+TAU)));
  }

  inline double instant_depolarization(double *X, double *P)
  {
    return instant_depolarization(
        X[0] /* t current time */,
        P[0] /* Td depolarization moment */,
        P[1] /* DELTA polarization contribution */,
        P[2] /* DN particle difference*/,
        P[3] /* TAU beam life time */,
        COUNT_TIME);
  }

	
	//static unsigned flag;
		
	inline double linearjump(double t,double D, double a, double k1, double k2)
  {
		double T=COUNT_TIME;
		if(t <=0)	return a+k1*(t-T/2.);
		if(t >= T)	return a+D +k2*(t-T/2.);
		return a+D*t/T + (k2*t*t - k1*(T-t)*(T-t))/2./T;
	}


	inline double double_linear_jump(double t, double t1, double t2, double D1, double D2, double a, double k1, double k2, double k3)
  {
		double T=COUNT_TIME;
		if(t <= t1 ) return a+k1*(t-t1 -T/2.);
		if(t > t1 && t < t1 + T) return a+D1*(t-t1)/T + (k2*(t-t1)*(t-t1) - k1*(T-(t-t1))*(T-(t-t1)))/2./T;
		a=a+D1+k2*(t2-t1);
		if(t <= t2 ) return a+k2*(t-t2 -T/2.);
		if(t > t2 && t < t2 + T) return a+D2*(t-t2)/T + (k3*(t-t2)*(t-t2) - k2*(T-(t-t2))*(T-(t-t2)))/2./T;
		if(t >= t2 + T) return a+D2+k3*(t-t2-T/2.);
		return 0;
	}

	inline double double_linear_jump(double *X, double *P)
  {
		double t = X[0];
		double t1 = P[0];
		double t2 = P[1];
		double D1 = P[2];
		double D2 = P[3];
		double a  = P[4];
		double k1 = P[5];
		double k2 = P[6];
		double k3 = P[7];
		return double_linear_jump(t,t1,t2,D1,D2,a,k1,k2,k3);
	}

	inline double expmod(double x)
  {
		if(x==0) return 1;
		return (1.0-exp(-x))/x;
	}

	inline double expjump(double t, double D, double a, double k1, double k2)
  {
		double T=COUNT_TIME;
		if(t <=0 ) return a*exp(k1*t)*expmod(k1*T);
		if(t >=T ) return (a+D)*exp(k2*t)*expmod(k2*T);
		return a*(1.0-t/T)*expmod((T-t)*k1) + (a+D)*t/T*expmod(-t*k2);
	}

	inline double jump(double x,  double D, double a, double k1, double k2)
  {
		double L = COUNT_TIME;
		if(x <= 0) return a + k1*x;
		if(x >= L ) return a + D + k2*x;
		else return a + D/L*x;
	}

	inline double expo(double x, double a, double b, double k)
  {
		return a + b*( 1. - exp ( k * x) );
	}


	inline double linear_func(double *X, double *P)
  {
		double x = *X - P[0];//time
		double &D  = P[1]; //jump value
		double a  = P[2]; //const
		double k1= P[3];	//slope befor depolarization
		double k2= P[4];	//slope after depolarization
		return linearjump(x,D,a,k1,k2);
	}

	inline double linear_jump2(double *X, double *P)
  {
		double T=COUNT_TIME;
		double t = *X - P[0];//time
		double D  = P[1]; //jump value
		double c1  = P[2]; //const
		double c2  = c1+D; //const
		double k1= P[3];	//slope befor depolarization
		double k2= P[4];	//slope after depolarization
		if(t <= -T/2) return c1 + k1*t;
		if(t >=  T/2) return c2 + k2*t;
		return (c1+c2)/2 + (c2-c1)*t/T + k2/2/T*(t+T/2)*(t+T/2) - k1/2/T*(t-T/2)*(t-T/2);
	}

	inline double exp_func(double *X, double *P)
  {
		double x = *X - P[0];//time
		double &D  = P[1]; //jump value
		double a  = P[2]; //const
		double k1= P[3];	//slope befor depolarization
		double k2= P[4];	//slope after depolarization
		return expjump(x,D,a,k1,k2);
	}

	inline double func2jump(double *X, double *P)
  {
		double t = X[0];
		double t1 = P[1];
		double t2 = P[0];
		double D2 = P[2];
		double D1 = D2/10;
		double a = P[3];
		double k1 = P[4];
		double k2 = P[4];
		double k3 = 0;
		return double_linear_jump(t,t1,t2,D1,D2,a,k1,k2,k3);
	}

	inline double func_quick(double *x, double *p)
  {
		if(*x <= p[0]) return p[2] + p[3]*(*x-p[0]);
		if(*x >= p[0] + COUNT_TIME ) return p[2] + p[1] + p[3]*(*x - p[0] - COUNT_TIME);
		else return p[2] + p[1]/COUNT_TIME*(*x - p[0]);
	}

	inline double expo_func (double *x ,double *p)
  {
		return expo ( *x, p[0],p[1],p[2]);
			
	}

	inline double atan_slope(double *x, double *p)	
	{
	    double t=x[0]-p[0];
	    double Delta=p[1];
	    double a=p[2];
	    double k1=p[3];
	    double k2=p[4];
	    double tau=p[5]*sqrt(2./M_PI);
	    double hev=(M_2_PI*atan(t/tau) + 1.)/2.;
	    if(std::isnan(hev)) return 1e300;
	    return a+ Delta*hev + t*(k1+(k2-k1)*hev);
	}


	inline double integral_atan(double x)
	{
	    return x*atan(x) - 0.5*log(x*x+1);
	}

	//Step function made from atan
	inline double hev_atan(double x) 
	{
    return 0;
	}

	inline double int_xhev_atan(double x)
	{
	    return 0.5*(x*x+1.0)*hev_atan(x) -  x/2./M_PI;
	}

	inline double int2_heaviside_atan(double x)
	{
    return 0;
	}

	inline double ihev(double x) 
	{
	    return 1./M_PI*(x*atan(x) - 0.5*log(x*x+1.)) + 0.5*x;
	}

	inline double ixhev(double x)
	{
	    return 0.5/M_PI*((x*x+1.)*atan(x)-x)+0.25*x*x;
	}



  inline double atan_old(double *x, double *p)
  {
    double T=COUNT_TIME;
    //p[0] -- depolarization moment
    double t=x[0]-p[0];
    double D=p[1]; //jump value
    double a=p[2]; //initial level
    double k1=p[3]; //slope before depolarization
    double k2=p[4]; //slope after depolarization
    double tau=p[5]*sqrt(2./M_PI); //depolarization time with some normalization for consistency with exact model.
    return 
      a+
      D*tau/T*(ihev(t/tau)-ihev((t-T)/tau)) +
      0.5*k1/T*(t*t-(t-T)*(t-T)) +
      (k2-k1)*tau/T*( ixhev(t/tau) - ixhev((t-T)/tau));
  }


	/* 
	 * ===  FUNCTION  ======================================================================
	 *         Name:  atan_super
	 *  Description:  atan as long jump plus averaging over the counting time
	 * =====================================================================================
	 */
	inline double atan_super(double *x, double *p)
	{
	    double D=p[1];
	    double C=p[2];
	    double k1=p[3];
	    double k2=p[4];
	    double tau=p[5]*sqrt(2./M_PI);
	    double T=COUNT_TIME/tau;
	    double t=(x[0]-p[0])/tau;
	    return  C + D/T*( int_heaviside_atan(t) - int_heaviside_atan(t-T) )
		+ 0.5*k1*tau/T*( t*t - (t-T)*(t-T)) +
		+ (k2-k1)*tau/T* ( int2_heaviside_atan(t) - int2_heaviside_atan(t-T) );
	}

	inline double atan_fixed_slope(double *x, double *p)
	{
	    double D=p[1]; // 
	    double dN=p[2]; //    
	    double tau=p[3]*sqrt(2./M_PI);
	    double Tb=p[4]; //  .
	    double t=(x[0]-p[0])/tau;

	    double T=COUNT_TIME/tau;
	    double k1=2./Tb*(D-dN);
	    double k2=-2./Tb*dN;
	    double C=2*dN-D;
	    return  C + D/T*( int_heaviside_atan(t) - int_heaviside_atan(t-T) )
		+ 0.5*k1*tau/T*( t*t - (t-T)*(t-T)) +
		+ (k2-k1)*tau/T* ( int2_heaviside_atan(t) - int2_heaviside_atan(t-T) );
	}




  class DeltaNoDep
  {
    public:
     const double D;   /* polarization contribution */
     const double DN;  /* diff of particles */
     const double TAU; /* beam life time */
     const double T;   /* count time  */
     DeltaNoDep(double d, double dn, double tau, double t) 
       : D(d), DN(dn),TAU(tau), T(t)
       {
       }
     inline double operator()(double t)
     {
       if(TAU<=0) return D;
       if(T<=0) return D + 2.0*(DN-D)*TAU/(t+TAU); //no averaging.
       if(t-T+TAU <=0 ) return  std::numeric_limits<double>::max(); //some conditions. shoud be minus infty.
       return D + 2.0*(DN-D)*TAU/T*log((t+TAU)/(t-T+TAU));
     }
  };

  class One_x_plus_a : public std::unary_function<double,double>
  {
    public:
      const double a;
      One_x_plus_a(double a_): a(a_){}
      inline double operator() (double x) const 
      {
        return  1.0/(x+a);
      }
  };

  class Int_hev_atan_mod : public std::unary_function<double,double>
  {
    const double x0;
    public:
      Int_hev_atan_mod(double x): x0(x) {}
      inline double operator() (double x)
      {
        return int_heaviside_atan(x)/(x+x0);
      }
  };





  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  atan_full
   *  Description:  Full implementation of counting rate theory. Only dN<<1 and delta << 1 is needed
   * =====================================================================================
   */
  inline double test_function(double x) { return x;}

  inline double atan_full(double *x, double *p)
  {
	    double D=p[1];  //polarization contribution
	    double DN=p[2]; //particle number difference in polarized and unpolarized bunch
	    double TAUD=p[3]*sqrt(2./M_PI); //depolarization time
	    double TAUB=p[4]/TAUD; //beam time of life.
      double T=COUNT_TIME/TAUD;
	    double t=(x[0]-p[0])/TAUD;
      Int_hev_atan_mod iham(TAUB);
      return delta_nodep(t, D, DN,  TAUB, T) + 
        D*TAUD/T*(int_heaviside_atan(t)-int_heaviside_atan(t-T)) +
        D*TAUD/T*ibn::dgaus(iham,t-T,t,1e-5);
      return test_function(t);
  }

	/* 
	 * ===  FUNCTION  ======================================================================
	 *         Name:  erfexp
	 *  Description:  function with erfexp
	 * =====================================================================================
	 */
	inline double experf(double *x, double *p)
	{
	    double t = x[0]-p[0];
	    double delta = p[1];
	    double C = p[2];
	    double k1 = p[3];
	    double k2 = p[4];
	    double tau = p[5]*M_SQRT2;
	    double power = p[6];//depolarization power
	    //define step function
	    double hev = ( 1.0 - exp( - power * (1. + erf(t/tau)) )  ) / ( 1. - exp(-2.*power) );
	    return C + delta * hev + t*(k1 + (k2-k1)*hev);
	}





	/* 
	 * ===  FUNCTION  ======================================================================
	 *         Name:  erf_super
	 *  Description:  erf based  jump plus averaging over the counting time
	 * =====================================================================================
	 */
	inline double erf_super(double *x, double *p)
	{
	    double D=p[1];
	    double C=p[2];
	    double k1=p[3];
	    double k2=p[4];
	    double tau=p[5]*M_SQRT2;
	    double T=COUNT_TIME/tau;
	    double t=(x[0]-p[0])/tau;
	    return  C + D/T*( int_heaviside_erf(t) - int_heaviside_erf(t-T) )
		+ 0.5*k1*tau/T*( t*t - (t-T)*(t-T)) +
		+ (k2-k1)*tau/T* ( int2_heaviside_erf(t) - int2_heaviside_erf(t-T) );
	}



	inline double int_hev_lin(double x)
	{
	    return x*heaviside(x);
	}

	inline double int2_hev_lin(double x)
	{
	    return 0.5*x*x*heaviside(x);
	}

	inline double lin_super(double *x, double *p)
	{
	    double D=p[1];
	    double C=p[2];
	    double k1=p[3];
	    double k2=p[4];
	    double T=COUNT_TIME;
	    double t=(x[0]-p[0]);
	    return  C + D/T*( int_hev_lin(t) - int_hev_lin(t-T) )
		+ 0.5*k1/T*( t*t - (t-T)*(t-T)) +
		+ (k2-k1)/T* ( int2_hev_lin(t) - int2_hev_lin(t-T) );
	}


	class Hev_experf 
	{
	    const double p;
	    const double NORM;
	    public:
	    Hev_experf(double p_) : p(p_), NORM(1./( 1. - exp(-2.*p) )) {}
	    double operator()(double x) 
	    {
		return ( 1.0 - exp( - 2.0*p * heaviside_erf(x) )  )*NORM;
	    }
	};


	inline double experf_super(double *x, double *p)
	{
	    double tau = p[5]*M_SQRT2;
	    double t = (x[0]-p[0])/tau;
	    double D = p[1];
	    double C = p[2];
	    double k1 = p[3];
	    double k2 = p[4];
	    double P = p[6];//depolarization power
	    Hev_experf hev(P);
	    return C + D*hev(t) + k1*tau*t + (k2-k1)*tau*ibn::dgaus(hev,0,t,1e-5);
	}

  inline double dummy_function(double *x, double *p)
  {
    return 0;
  }

  inline double lifetime(double *x ,double *p)
  {
    return p[0] + p[1]*exp((x[0]-p[2])/p[3]);
  }

  inline double ppolinom(double *x, double *p)
  {
    return p[0]+p[1]*x[0] + p[2]*x[0]*x[0] + p[3]*pow(x[0],3.0);
  }


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  barrier
   *  Description:  Fit border 
   * =====================================================================================
   */
  inline double barrier(double *x, double *p)
  {
    double t = x[0];
    double t1 = p[0];
    double t2 = p[1];
    double c1 = p[2];
    double c2 = p[3];
    double c3 = p[4];
    if( t < t1) return c1;
    if (t > t2) return c1+c3;
    else return c1+c2;
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  count
   *  Description:  Counting rate vs time
   * =====================================================================================
   */
  inline double count(double *x, double *p)
  {
    double N0 = p[0]; //number of particles in the beam
    double tb = p[1]; //beam life time
    double t = x[0]; //time
    return N0/tb/math::sq(1.0+t/tb);
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  count2
   *  Description:  Counting rate vs time fit parameters in mA
   * =====================================================================================
   */
  inline double count2(double *x, double *p)
  {
    double I0 = p[0]*1e-3; //number of particles in the beam in A
    double tb = p[1]; //beam life time
    double t = x[0]; //time
    double ef0 = phys::e*phys::f0;
    return I0/(ef0*tb*math::sq(1.0+t/tb));
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  count3
   *  Description:  Counting rate vs time fit parameters in mA
   * =====================================================================================
   */
  inline double count3(double *x, double *p)
  {
    double I0 = p[0]*1e-3; //number of particles in the beam in A
    double tb = p[1]; //beam life time
    double t = x[0]; //time
    double ef0 = phys::e*phys::f0;
    return I0/(ef0*tb*math::sq(1.0+t/tb));
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  current
   *  Description:  Current vs time fit
   * =====================================================================================
   */
  inline double tsh_current(double *x, double *p)
  {
    double I0 = p[0]*1e3; //number of particles in the beam in A, p0 in mA, I0 in mkA
    double tb = p[1]; //beam life time
    double t = x[0]; //time
    return I0/(1.0+t/tb);
  }


  inline double my_golden_limit(double x)
  {
    if(x==0) return 1.0;
    return (1.0 - exp(-x))/x;
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  current2
   *  Description:  Current fit with background taken into account
   * =====================================================================================
   */
  inline double tsh_bg_current(double *x, double *p)
  {
    double I0 = p[0]*1e3; //number of particles in the beam in A, p0 in mA, I0 in mkA
    double tb = p[1]; //beam life time
    double bg = p[2];  //background contribution bg = Bbg/AN0 = B*tb
    double t = x[0]; //time
    double y = bg*t/tb;
    return I0*exp(-y)/(1.0 + t/tb*my_golden_limit(y));
  }
  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  current3
   *  Description:  Only background
   * =====================================================================================
   */
  inline double exp_current(double *x, double *p)
  {
    double I0 = p[0]*1e3; //number of particles in the beam in A, p0 in mA, I0 in mkA
    double tb = p[1]; //beam life time
    double t = x[0]; //time
    return I0*exp(-t/tb);
  }


  class polarization_hev 
  {
    double T; //depolarization time
    double D; //delta
    double B;  //background contribution
    public:
      polarization_hev(double td, double delta, double bg) : T(td) , D(delta), B(bg) {}
      double operator()(double t)
      {
        return D*exp(-B*t)*(1.0 - heaviside(t-T));
      }
      double integral(double a, double b, double prec = 1e-10)
      {
        return ibn::dgaus(*this, a, b, prec);
      }
  };


  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  tsh_pol_bg_current
   *  Description:  Current fit with background taken into account
   * =====================================================================================
   */
  inline double tsh_pol_bg_current(double *x, double *p)
  {
    double td = p[0]; //moment of the depolarization
    double D = p[1]; //depolarization jump
    double I0 = p[2]*1e3; //Initial current
    double tau_bg = p[3]; //beam life time (background part)
    double tau_tsh = p[4]; //beam life time touschek
    double t = x[0]; //time
    double e = exp(-t/tau_bg);
    polarization_hev ph(td, D, 1./tau_bg);
    double result = I0*e/(1.0 + (1.0-e)*tau_bg/tau_tsh - ph.integral(0,t));
    return result;
  }

  /* 
   * ===  FUNCTION  ======================================================================
   *         Name:  count4
   *  Description:  count with toushek background and polarization
   * =====================================================================================
   */
  inline double count4(double *x, double *p)
  {
    double td = p[0]; //moment of the depolarization
    double D = p[1]; //depolarization jump
    double I0 = p[2]*1e-3; //Initial current
    double tau_bg = p[3]; //beam life time (background part)
    double tau_tsh = p[4]; //beam life time touschek
    double prob = p[5]; //beam life time touschek
    double t = x[0]; //time
    double e = exp(-t/tau_bg);
    polarization_hev ph(td, D, 1./tau_bg);
    double i0 = phys::e*phys::f0; //  
    double I = I0*e/(1.0 + (1.0-e)*tau_bg/tau_tsh - ph.integral(0,t)/tau_tsh); //current part
    double result = I*I*(1.0-D*(1.0 - heaviside(t-td)))/I0/tau_tsh/i0 + I/tau_bg/i0;
    return result*prob;
  }

  class SimpleAtan
  {
    public:
      static double heaviside(double x) 
      {
        return atan(x)*M_1_PI + 0.5;
      };

      static double int1_heaviside(double x)
      {
        return x*heaviside(x) - 0.5*M_1_PI*log(x*x+1.);
      }

      static double int2_heaviside(double x)
      {
        return 0.5*(x*x-1.0)*heaviside(x) + 0.25 + 0.5*M_1_PI*x*(1.0 - log(x*x+1.0));
      }

      static double delta(double t, double TD, double D, double TAU, double C, double K1, double K2)
      {
        double x = (t-TD)/TAU;
        double x0 = -TD/TAU;
        double k1 = K1*TAU;
        double k2 = K2*TAU;
        return C + D*heaviside(x) + k1*x + (k2-k1)*(int1_heaviside(x) - int1_heaviside(x0));
      }


      //double delta_average(double t, double T) const 
      //{
      //  double x = (t-TD)/TAU;
      //  double x0 = -TD/TAU;
      //  double k1 = K1*TAU;
      //  double k2 = K2*TAU;
      //  double dx = T/TAU;
      //  return C - (k2-k1)*int1_heaviside(x0)  + 
      //    D*(int1_heaviside(x)-int1_heaviside(x-dx))/dx +
      //    0.5*k1*(x*x - (x-dx)*(x-dx))/dx +
      //    (k2-k1)*(int2_heaviside(x) - int2_heaviside(x-dx))/dx;
      //}

      static double delta_average(double *X, double *P)
      {
        double t = X[0];
        double TD=P[0];
        double D=P[1];
        double C = P[2];
        double K1 = P[3];
        double K2 = P[4];
        double TAU = P[5];
        double T = COUNT_TIME;
        double x = (t-TD)/TAU;
        double x0 = -TD/TAU;
        double k1 = K1*TAU;
        double k2 = K2*TAU;
        double dx = T/TAU;
        return C - (k2-k1)*int1_heaviside(x0)  + 
          D*(int1_heaviside(x)-int1_heaviside(x-dx))/dx +
          0.5*k1*(x*x - (x-dx)*(x-dx))/dx +
          (k2-k1)*(int2_heaviside(x) - int2_heaviside(x-dx))/dx;
      }


  };

	inline double relaxation(double *X,  double *P)
	{
		double t = X[0];
		//double t0 = P[0];
		double D = P[1];
		double dN = P[2];
		double tau = P[3];
		double tau0 = P[4];
		return 2*dN*tau0/(t+ tau0) + D*tau/(t + tau0)*(1.0 - exp(-2*t/tau)) - D*exp(-2*t/tau);
	}



	inline double experf2(double *X,  double *P)
	{
		double t=X[0];
		double D=P[1];
		double dN = P[2];
		double tau0 = std::abs(P[3]); //touschek tau
		double p = std::abs(P[4]);// depolarizer power
		double tau = std::abs(P[5]); //depolarizer tau

		auto delta = [=](double t){ return D*exp(-p*(1.0 + erf((t-P[0])/tau)));};
		double d = delta(t);
		if(std::isinf(d) || std::isnan(d)) return 1e300;
		double itg = ibn::dgaus(delta,0, t, 1e-5);
		if(std::isinf(itg) || std::isnan(itg)) return 1e300;
		double result =  2*dN*tau0/(t+tau0) - d + 2.0/(t+tau0)*itg;
		return result;
	};

}
#endif
