// =====================================================================================
//
//       Filename:  RelaxationFitter.h
//
//    Description:  Smart Relaxation Fitter
//
//        Version:  1.0
//        Created:  23.02.2015 15:06:41
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#pragma once
#include "Fitter.h"
#include "Minuit2/MnPrint.h"

#include <TF1.h>
#include <TPaveText.h>

#include "data/EnergyCalibrationStorage.h"
#include "gui/EnergyCalibrationView.h"
#include <boost/format.hpp>
#include <ibn/integral.h>

namespace fit
{
	class JumpFitter : public Fitter
	{
		unique_ptr<TF1> CountFunc;
		unique_ptr<TF1> NoSlopeCountFunc;
		unique_ptr<TF1> CurrentFunc;
		unique_ptr<TPaveText> pave;
		unique_ptr<TCanvas> noslope_canvas;
		unique_ptr<TGraphErrors> noslope_count_graph;
		unique_ptr<TGraphErrors> noslope_current_graph;
		gui::EnergyCalibrationView * ECView;
		public:
		data::Array * TIME;
		data::Array * COUNT;
		data::Array * CURRENT;
		data::Array fit_result;
		std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;

		mutable double fTDEP; //depolarization moment
		mutable double fDELTA;  //koef 
		mutable double fTAU_POL; //depolarization time
		mutable double fTAU_TSH; //touschek time,  sec
		mutable double fTAU_BG;  //background time, sec
		mutable double fPTSH; //probability to register touschek
		mutable double fPBG;  //probability to register background
		mutable double fP0;    //initial polarization
		mutable double fPOWER;    //power of depolarizer
		mutable double fI0;

		mutable struct 
		{
			int TDEP;
			int DELTA;
			int TAU_POL;
			int TAU_TSH;
			int TAU_BG;
			int PTSH;
			int PBG;
			int P0;
			int POWER;
			int I0;
		} IDX;


		mutable double CHI2;
		mutable double CHI2_CURRENT;
		mutable double NDF; //number degree of freedom
		mutable double NPOINT; //number of data which play game

		const double PMAX = 8.0/15.0*sqrt(3.0); //maximum polarization degreee (Sokolov Ternov)



		JumpFitter(void)
		{
			type = MINUIT2;
			name = "JUMP_FITTER";
			title = "Fit single channel jump";
			int idx=0;
			defpar.Add("TDEP",0, 100);      								IDX.TDEP = idx++;
			defpar.Add("DELTA",0.01, 0.01); 								IDX.DELTA = idx++;
			defpar.Add("TAU_POL",  0.1, 10 , 0, 100); 				IDX.TAU_POL =idx++;
			defpar.Fix("TAU_POL");
			defpar.Add("TAU_TSH",  3000,  1000, 0, 1e6);    IDX.TAU_TSH = idx++;
			defpar.Add("TAU_BG",  10000,   1000, 0, 1e10);  IDX.TAU_BG = idx++;
			defpar.Add("PTSH",  0.2, 0.5, 0, 1);            IDX.PTSH = idx++;
			defpar.Add("PBG",  0.05, 0.5, 0, 1);            IDX.PBG = idx++;
			defpar.Add("P0",  0.92, 0.5, 0, 0.92);          IDX.P0 = idx++;
			defpar.Add("POWER",  10, 0.5, 0, 1000);  				IDX.POWER = idx++;
			defpar.Fix("POWER");
			defpar.Add("I0",  2000, 500);          				  IDX.I0 = idx++;
			inipar=defpar;
		}

		void SetParameters(const vector<double> & par) const
		{
			fTDEP = par[IDX.TDEP];
			fDELTA = par[IDX.DELTA];
			fTAU_POL =     fabs(par[IDX.TAU_POL]);
			fTAU_TSH =     fabs(par[IDX.TAU_TSH]);
			fTAU_BG =      fabs(par[IDX.TAU_BG]);
			fPTSH =        fabs(par[IDX.PTSH]);
			fPBG =         fabs(par[IDX.PBG]);
			fP0 =          fabs(par[IDX.P0]);
			fPOWER =       fabs(par[IDX.POWER]);
			fI0 = 		     fabs(par[IDX.I0]);
		}

		~JumpFitter(void)
		{
		}

		void SetData(data::EnergyCalibrationStorage * d)
		{
			TIME = &(d->t);
			COUNT = &(d->CountSum[0]);
			CURRENT = &(d->I[0]);
		}

		void SetView(gui::EnergyCalibrationView * view)
		{
			ECView = view;
		}

		/*  Ищет скачок, предполагая, что время обозначено параметром T */
		ROOT::Minuit2::MnUserParameters FindJump(void)
		{
			using namespace ROOT::Minuit2;
			std::list<FunctionMinimum> fmlist; //list of all minimums
			for(unsigned i=0; i<TIME->size();i++)
			{
				double t =  (*TIME)[i].value; //take time
				MnMigrad migrad(*this, inipar);
				migrad.SetValue("TDEP",t);
				FunctionMinimum fm = migrad();
				fmlist.push_back(fm); //fill list of minimum
			}
			//return the best minimum
			return min_element 
				(
				 begin(fmlist), end(fmlist), 
				 [](const FunctionMinimum & f1, const FunctionMinimum & f2) 
				 { 
				 return f1.Fval() < f2.Fval(); 
				 }
				)->UserParameters();
		}

		void Fit(void)
		{
			//gErrorIgnoreLevel = 1001;
			using namespace ROOT::Minuit2;
			if(tmax==0) tmax=(*TIME)[TIME->size()-1].value;
			MnUserParameters newpar=inipar;
			if(!inipar.Parameter(newpar.Index("TDEP")).IsFixed()) newpar = FindJump();
			MnMigrad migrad(*this, newpar);
			FunctionMinimum minimum = migrad();
			minpar=minimum.UserParameters();
			SetMinPar();
			NDF = NPOINT - minpar.VariableParameters();
		}

		double step(double t) const
		{
			t = (t-fTDEP)/fTAU_POL;
			if(t<-1) return 0;
			if(t<+1) return (t+1)*0.5;
			return 1;
		}

		double polarization(double t) const 
		{
			//cout << "t=" << t << "  step=" << step(t) << " pol=" << fP0*exp(-fPOWER*step(t)) << endl;
			return fP0*exp(-fPOWER*step(t));
		}

		double pol_integral(double t) const
		{
			double t1 = fTDEP-fTAU_POL;
			if(t<t1) return fP0*fP0*fTAU_BG*(1.0 - exp(-t/fTAU_BG));
			double k = 1./fTAU_BG + fPOWER/fTAU_POL;
			double t2 = fTDEP+fTAU_POL;
			if(t<t2) return fP0*fP0*fTAU_BG*(1.0 - exp(-t1/fTAU_BG)) + fP0*fP0/k*exp(-t1/fTAU_BG)*(1.0 - exp(-(t-t1)*k));
			return fP0*fP0*fTAU_BG*(1.0 - exp(-t1/fTAU_BG)) + fP0*fP0/k*exp(-t1/fTAU_BG)*(1.0 - exp(-(t2-t1)*k)) + fP0*fP0*exp(-2*fPOWER)*fTAU_BG*(1.0 - exp(-(t-t2)/fTAU_BG));
		}


		double Current(double t) const 
		{
			double pI = pol_integral(t);
			double result = fI0*exp(-t/fTAU_BG)/(1.0+fTAU_BG/fTAU_TSH*(1.0 - exp(-t/fTAU_BG)) - fDELTA*pI/fTAU_TSH);
			if(result<0 || std::isnan(result) || std::isinf(result)) result=0;
			return result;
		}

		double Count(double t,  double jump=1) const 
		{
			const double Ie = 1.602e-19*818924*1e6; //current of single electron in uA
			double I = Current(t);
			double p = polarization(t);
			double N = I/Ie;
			double N0 = fI0/Ie;
			double result  =  fPTSH/fTAU_TSH * N*N/N0 * (1.0 - jump*fDELTA*p*p)  + fPBG/fTAU_BG*N;
			//cout << "ptsh=" << fPTSH << " tau_tsh=" << fTAU_TSH << " N=" << N << " N0=" << N0  << " count = " << result << endl;
			if(result<0 || std::isnan(result) || std::isinf(result)) result=0;
			return result;
		}

		double AverageCount(double t,  double dt,  double jump=1) const
		{
			double result =  ibn::dgaus([this, dt, jump](double t){ return Count(t, jump); },  t-dt/2.0,  t+dt/2.0, 1e-5)/dt;
			if(result<0 || std::isnan(result) || std::isinf(result)) result=0;
			return result;
		}

		double operator() (const std::vector<double> & par) const
		{
			SetParameters(par);
			double chi2=0;
			double chi2_current=0;
			NPOINT=0; //number of data point wich play game
			for(unsigned i=0;i<TIME->size();i++)
			{
				double t = (*TIME)[i].value; //current time
				double dt = (*TIME)[i].error; //current time
				if(t < tmin || tmax < t) continue;
				double N = (*COUNT)[i].value; //count rate
				double dN = (*COUNT)[i].error; //count rate
				double I = (*CURRENT)[i].value;
				double dI = (*CURRENT)[i].error;
				double Iexpected = Current(t);
				chi2+= math::sq((N - AverageCount(t, 2*dt))/dN);
				if(dI!=0)
				{
					chi2_current += math::sq( (I- Iexpected)/dI);
				}
				//cout << I << " " << dI << " " << Iexpected << "  " << chi2_current << endl;
				NPOINT++;
			}
			CHI2 = chi2;
			CHI2_CURRENT = chi2_current;
			//cout << "CHI2_CURRENT = " << CHI2_CURRENT << endl;
			return chi2+chi2_current;
		} 

		double Up() const { return 1.; }

		void SetMinPar(void)
		{
			SetParameters(minpar.Params());
		}

		void Print(void)
		{
			cout << LOG(1) << "Fit result:" << endl;
			cout << LOG(1) << minpar << endl;
			cout << LOG(1) << "CHI2/NDF = " << CHI2/NDF << endl;
			cout << LOG(1) << "CHI2_CURRENT/NDF = " << CHI2_CURRENT/NDF << endl;
			cout << LOG(1) << "NPOINT = " << NPOINT << endl;
		}

		void Draw(int opt)
		{
			if(!pave)
			{
				pave.reset(new TPaveText(0.7, 0.7, 0.9, 0.9, "NDC"));
			}
			pave->Clear();
			//pave->SetTextFont(102);
			pave->SetTextAlign(12);
			pave->SetBorderSize(1);
			pave->SetFillColor(0);
			boost::format fmt("%10s = %10.3f #pm %6.3f");
			for(unsigned i=0;i<minpar.Params().size();i++)
			{
				string s = (fmt %  minpar.GetName(i) % minpar.Params()[i] % minpar.Errors()[i] ).str();
				//cout << i << " " << s << endl;
				pave->AddText(s.c_str());
			}

			double TCOUNT=2*(*TIME)[0].error;
			ECView->Add("counter_sum-1")->Draw();
			pave->Draw();
			CountFunc.reset(new TF1("CountFunc", [this, TCOUNT](double * x,  double *p) { return AverageCount(x[0], TCOUNT); } , 0, tmax, 0));
			CountFunc->Draw("same");

			NoSlopeCountFunc.reset(new TF1("NoSlopeCountFunc", [this, TCOUNT](double * x,  double *p) { return AverageCount(x[0], TCOUNT, 1)-AverageCount(x[0], TCOUNT, 0);  } , 0, tmax, 0));
			NoSlopeCountFunc->SetNpx(1000);

			ECView->Add("current-1")->Draw();
			CurrentFunc.reset(new TF1("CurrentFunc", [this](double * x,  double *p) { return Current(x[0]); } , 0, tmax, 0));
			CurrentFunc->Draw("same");
			ECView->Update();

			
			//noslope_canvas.reset(new TCanvas("noslope_count", "Noslope count"));
			auto noslope_canvas = new TCanvas("noslope_count", "Noslope count");
			noslope_canvas->Divide(1, 2);
			noslope_count_graph.reset(new TGraphErrors);
			noslope_current_graph.reset(new TGraphErrors);
			for(unsigned i=0; i<(*TIME).size();i++)
			{
				double t = (*TIME)[i].value;
				double dt = (*TIME)[i].error;
				double N = (*COUNT)[i].value;
				double dN = (*COUNT)[i].error;
				double Nexp = AverageCount(t, 2*dt, 0); //expected count with now jump
				noslope_count_graph->SetPoint(i, t,  N - Nexp);
				noslope_count_graph->SetPointError(i, dt, dN );
				noslope_current_graph->SetPoint(i, t,  (*CURRENT)[i].value-CurrentFunc->Eval(t));
				noslope_current_graph->SetPointError(i, dt, (*CURRENT)[i].error );
			}
			noslope_count_graph->SetMarkerStyle(21);
			noslope_canvas->cd(1);
			noslope_count_graph->Draw("ap");
			NoSlopeCountFunc->Draw("same");
			noslope_canvas->cd(2);
			noslope_current_graph->SetMarkerStyle(21);
			noslope_current_graph->Draw("ap");
		}
	};
}



