// =====================================================================================
//
//       Filename:  LinearJumpFitter.h
//
//    Description:  Simple linear jump fitter
//
//        Version:  1.0
//        Created:  09.03.2015 05:05:41
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#ifndef IBN_POL_FIT_LINEAR_JUMP_FITTER_H
#define IBN_POL_FIT_LINEAR_JUMP_FITTER_H
#include "Fitter.h"
#include "Minuit2/MnPrint.h"

#include <TF1.h>
#include <TPaveText.h>

#include "data/EnergyCalibrationStorage.h"
#include "gui/EnergyCalibrationView.h"
#include <boost/format.hpp>

namespace fit
{
	class LinearJumpFitter : public Fitter
	{
		unique_ptr<TF1> Func;
		unique_ptr<TPaveText> pave;
		gui::EnergyCalibrationView * ECView;
		public:
		data::Array * fTIME;
		data::Array * fDELTA;
		std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;

		mutable double T;   //depolarization time
		mutable double DELTA;  //jump
		mutable double C1;  //initial level
		mutable double C2; //level after jump C1+DELTA
		mutable double K1;  //slope before jump
		mutable double K2; //slope after jump
		mutable double COUNT_TIME; //count time
		mutable struct 
		{
			int T;
			int DELTA;
			int C1;
			int C2;
			int K1;
			int K2;
		} IDX;


		mutable double CHI2;
		mutable double NDF; //number degree of freedom
		mutable double NPOINT; //number of data which play game

		LinearJumpFitter(void)
		{
			type = MINUIT2;
			name = "SMART_LINEAR_JUMP";
			title = "Linear jump fitter of jump";
			int idx=0;
			defpar.Add("T",0, 0.01); //depolarization effect
			IDX.T = idx++;
			defpar.Add("DELTA",0.01, 0.01); //depolarization effect
			IDX.DELTA = idx++;
			defpar.Add("CONST",   0, 0.01);    //count difference due to different number of particle
			IDX.C1 = idx++;
			defpar.Add("SLOPE1",   0, 0.01);    //count difference due to different number of particle
			IDX.K1 = idx++;
			defpar.Add("SLOPE2",   0, 0.01);    //count difference due to different number of particle
			IDX.K2 = idx++;
			inipar=defpar;
		}

		void SetParameters(const vector<double> & par) const
		{
			T = par[IDX.T];
			DELTA = par[IDX.DELTA];
			C1  = par[IDX.C1];
			C2 = C1 +DELTA;
			K1 = par[IDX.K1];
			K2 = par[IDX.K2];
			COUNT_TIME = ECView->GetStorage()->cfg.count_time;
		}

		~LinearJumpFitter(void)
		{
		}
		

		double Delta(double t) const
		{
			t = t-T;
			if(t <= -COUNT_TIME/2) return C1 + K1*t;
			if(t >=  COUNT_TIME/2) return C2 + K2*t;
			return (C2+C1)/2 + (C2-C1)*t/COUNT_TIME + 0.5*K2*pow(t+COUNT_TIME/2, 2)/COUNT_TIME  - 0.5*K1*pow(t-COUNT_TIME/2, 2)/COUNT_TIME;
		};



		void SetData(data::EnergyCalibrationStorage * d)
		{
			fTIME = &(d->t);
			fDELTA = &(d->DeltaSum[0]);
		}

		void SetView(gui::EnergyCalibrationView * view)
		{
			ECView = view;
		}

		void Fit(void)
		{
			using namespace ROOT::Minuit2;
			if(tmax==0) tmax=(*fTIME)[fTIME->size()-1].value;
			MnMigrad migrad(*this, inipar);
			FunctionMinimum minimum = migrad();
			minpar=minimum.UserParameters();
			SetMinPar();
			NDF = NPOINT - minpar.VariableParameters();
		}

		double operator() (const std::vector<double> & par) const
		{
			SetParameters(par);
			double chi2=0;
			NPOINT=0; //number of data point wich play game
			for(unsigned i=0;i<fTIME->size();i++)
			{
				double t = (*fTIME)[i].value;
				//if(i+1<fTIME->size()) COUNT_TIME = (*fTIME)[i+1].value - t;
				if(t < tmin || tmax < t) continue;
				//delta contribution
				double d = (*fDELTA)[i].value;
				double d_error = (*fDELTA)[i].error;
				chi2 +=pow((d - Delta(t))/d_error,2.0);
				NPOINT++;
			}
			CHI2 = chi2;
			return chi2;
		} 

		double Up() const { return 1.; }

		void SetMinPar(void)
		{
			SetParameters(minpar.Params());
		}

		void Print(void)
		{
			cout << LOG(1) << "Fit result:" << endl;
			cout << LOG(1) << minpar << endl;
			cout << LOG(1) << "CHI2/NDF = " << CHI2/NDF << endl;
		}

		void Draw(int opt)
		{
			if(!pave)
			{
				pave.reset(new TPaveText(0.7, 0.7, 0.9, 0.9, "NDC"));
			}
			pave->Clear();
			pave->SetTextAlign(12);
			pave->SetBorderSize(1);
			pave->SetFillColor(0);
			boost::format fmt("%10s = %10.3f #pm %6.3f");
			for(unsigned i=0;i<minpar.Params().size();i++)
			{
				string s = (fmt %  minpar.GetName(i) % minpar.Params()[i] % minpar.Errors()[i] ).str();
				pave->AddText(s.c_str());
			}

			ECView->Add("delta-1")->Draw();
			pave->Draw();
			Func.reset(new TF1("LinearJumpFunc", [=](double * x,  double *p) { return Delta(x[0]); } , 0, tmax, 0));
			Func->Draw("same");
			ECView->Update();
		}
	};
}
#endif
