/*
 * =====================================================================================
 *
 *       Filename:  PolarizationFunctions.h
 *
 *    Description:  Function diskribes polarization
 *
 *        Version:  1.0
 *        Created:  13.11.2012 11:40:58
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#ifndef IBN_POLARIMETER_POLARIZATION_FUNCTIONS_H
#define IBN_POLARIMETER_POLARIZATION_FUNCTIONS_H

//#include "heaviside.h"
#include "ibn/integral.h"

#include "functional"

namespace fit
{
  /*
   * =====================================================================================
   *        Class:  Polarization
   *  Description:  
   * =====================================================================================
   */
  template < class JumpFunction >
  class Polarization
  {
    public:
      JumpFunction J;
      double T; //depolarization moment in seconds
      double D; //delta (undementional)
      double B; //background contribution 1./sec

      Polarization(double td=0, double delta=0, double tau=1, double power=1, double bg=0) : J(JumpFunction(tau, power)), T(td), D(delta), B(bg) {}
      Polarization(JumpFunction jf, double td, double delta, double bg) : J(jf), T(td), D(delta), B(bg) {}

      double jump(double t)
      {
        return D*J(t-T);
      }

      //функция поляризации в начальный момент времени есть поляризация D
      //после сканирования поляризация равна нулю.
      double polarization(double t)
      {
        return D*(1.0-J(t-T));
      }

      double operator()(double t)
      {
        return exp(-B*t)*polarization(t);
      }

      double integral(double a, double b, double prec = 1e-10)
      {
        return ibn::dgaus(*this, a, b, prec);
      }

      void SetDepolarizationMoment(double t) {T=t;}
      void SetDelta(double d) {D=d;}
      void SetBackground(double bg) {B=fabs(bg);}
      void SetDepolarizationTime(double tau) {J.tau = fabs(tau);}
      void SetDepolarizerPower(double p) {J.power =p;}
      void SetJumpFunction(JumpFunction jf) {jump = jf; }

      double GetDepolarizationMoment(void) {return T;}
      double GetDelta(void) {return D;}
      double GetBackground(void) {return B;}
      double GetDepolarizationTime(void) {return J.tau;}
      double GetDepolarizerPower(void) {return J.power;}
      const JumpFunction & GetJumpFunction(void) {return jump;}
  };


  /*
   * =====================================================================================
   *        Class:  HeavisideJump
   *  Description:  Мгновенный скачок
   * =====================================================================================
   */
  struct HeavisideJump : public std::unary_function<double, double>
  {
    double tau; //время пересечения резонаса
    double power; //сила деполяризатора
    HeavisideJump(double t=0, double p=1): tau(t), power(p) {}
    double operator()(double x)
    {
      if(x<0) return 0;
      if(x==0) return 0.5;
      return 1;
    }
  };

  template <> 
    inline double Polarization<HeavisideJump>::integral(double a, double b, double prec)
    {
      if(!std::isnormal(B))
      {
        if(a<T && b<T) return D*(b-a);
        if(a<T && b>=T) return D*(T-a);
        return 0;
      }
      if(a<T && b<T) return D/B*(exp(-B*a) - exp(-B*b));
      if(a<T && b>=T) return D/B*(exp(-B*a) - exp(-B*T));
      return 0;
    }

  /*
   * =====================================================================================
   *        Class:  AtanJump
   *  Description:  Затянутый скачок описываемый арктангенсом
   * =====================================================================================
   */
  struct AtanJump : public std::unary_function<double, double>
  {
    double tau; //время пересечения резонаса
    double power; //сила деполяризатора
    AtanJump(double t=1, double p=1) : tau(t), power(p) {}
    double operator()(double t)
    {
      return 0.5*(M_2_PI*atan(t/tau) + 1.);
    }
  };
  

  /*
   * =====================================================================================
   *        Class:  ErfJump
   *  Description:  Затянутый скачок описываемый функцией ошибок
   * =====================================================================================
   */
  struct ErfJump : public std::unary_function<double, double>
  {
    double tau; //время пересечения резонаса
    double power; //сила деполяризатора
    ErfJump(double t=1,double p=1) :  tau(t), power(p) {}
    double operator()(double t)
    {
      return 0.5*(M_2_PI*atan(t/tau) + 1.);
    }
  };



  typedef Polarization<HeavisideJump> HeavisidePolarization;
  typedef Polarization<AtanJump> AtanPolarization;
  typedef Polarization<ErfJump> ErfPolarization;
}
#endif

