// =====================================================================================
//
//       Filename:  RelaxationFitter.h
//
//    Description:  Smart Relaxation Fitter
//
//        Version:  1.0
//        Created:  23.02.2015 15:06:41
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#ifndef IBN_POL_FIT_RELAXATION_FITTER_H
#define IBN_POL_FIT_RELAXATION_FITTER_H
#include "Fitter.h"
#include "Minuit2/MnPrint.h"

#include <TF1.h>
#include <TPaveText.h>

#include "data/EnergyCalibrationStorage.h"
#include "gui/EnergyCalibrationView.h"
#include <boost/format.hpp>

namespace fit
{
	class RelaxationFitter : public Fitter
	{
		unique_ptr<TF1> DeltaFunc;
		unique_ptr<TF1> CurrentFunc[2];
		unique_ptr<TF1> CountFunc[2];
		unique_ptr<TF1> VolumeFunc;
		bool VIFIXED;
		bool SKIPVOLUME;

		unique_ptr<TPaveText> pave;
		gui::EnergyCalibrationView * ECView;
		public:
		data::Array * TIME;
		data::Array * DELTA;
		data::Array * CURRENT[2];
		data::Array * COUNT[2];
		data::Array * SIGMAX;
		data::Array * SIGMAZ;
		data::Array fit_result;
		std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;

		mutable double D;  //jump
		mutable double DN; //relative difference in number of particle
		mutable double TAU_POL; //depolarization time
		mutable double TAU_TSH; //touschek time,  sec
		mutable double TAU_BG;  //background time, sec
		mutable double I0[2];  //initial current,  uA
		mutable double PTSH; //probability to register touschek
		mutable double PBG;  //probability to register background
		mutable double BVOL;  //initial beam volume um^2
		mutable double VI;    //volume dependace on current, 1/uA^2
		mutable double P0;    //initial polarization of first bunch
		mutable double P1;    //equilibrium level of  polarization = 0.92/G
		mutable double TAU_SOKOLOV; //polarization time due to Sokolov Termov radiation polarization
		mutable double CURRENT_SHIFT; //unideality of measurement current
		mutable double DELTA_SHIFT;   //undetiality of beam 1 beam 2
		mutable double CURRENT_ERROR_SCALE; //scale of the error of current measurement
		mutable double COUNT_ERROR_SCALE;   //scale of the error of count measurement

		mutable struct 
		{
			int DELTA;
			int DN;
			int TAU_POL;
			int TAU_TSH;
			int TAU_BG;
			int I0;
			int I1;
			int PTSH;
			int PBG;
			int BVOL;
			int VI;
			int P0;
			int TAU_SOKOLOV;
			int CURRENT_SHIFT;
			int DELTA_SHIFT;
			int CURRENT_ERROR_SCALE;
			int COUNT_ERROR_SCALE;
		} IDX;


		mutable double COUNT_TIME;
		mutable double CHI2;
		mutable double CHI2_DELTA;
		mutable double CHI2_DN;
		mutable double CHI2_CURRENT[2];
		mutable double CHI2_COUNT[2];
		mutable double CHI2_VOLUME;
		mutable double NDF; //number degree of freedom
		mutable double NPOINT; //number of data which play game

		const double PMAX = 8.0/15.0*sqrt(3.0); //maximum polarization degreee (Sokolov Ternov)



		RelaxationFitter(void)
		{
			type = MINUIT2;
			name = "SMART_RELAXATION";
			title = "Fit data to find depolarization time for its relaxation";
			int idx=0;
			defpar.Add("DELTA",0.01, 0.01); //depolarization effect
			IDX.DELTA = idx++;
			defpar.Add("DN",   0.01, 0.01);    //count difference due to different number of particle
			IDX.DN = idx++;
			defpar.Add("TAU_POL",  500,   100, 0, 1e6); //depolarization time,  sec
			IDX.TAU_POL =idx++;
			defpar.Add("TAU_TSH",  5000,  1000, 0, 1e6); //touschek lift time,  sec
			IDX.TAU_TSH = idx++;
			defpar.Add("TAU_BG",  10000,   1000, 0, 1e10); //background life time,  sec
			IDX.TAU_BG = idx++;
			defpar.Add("I0",  2000,   1000); //initial current, for the first bunch uA 
			IDX.I0 = idx++;
			defpar.Add("I1",  2000,   1000); //initial current,  uA
			IDX.I1 = idx++;
			defpar.Add("PTSH",  0.2, 0.5, 0, 1);   //probability to register touschek
			IDX.PTSH = idx++;
			defpar.Add("PBG",  0.05, 0.5, 0, 1);   //probability to register background
			IDX.PBG = idx++;
			defpar.Add("BVOL", 200e3);             //beam volume um^2
			IDX.BVOL = idx++;
			defpar.Add("VI", 0);
			IDX.VI = idx++;
			defpar.Add("P0", 0.8, 0.1, -PMAX, PMAX); defpar.Fix("P0"); //initial polarization degree of first bunch
			IDX.P0 = idx++;
			defpar.Add("TAU_SOKOLOV", 87.0); defpar.Fix("TAU_SOKOLOV"); //in hours
			IDX.TAU_SOKOLOV = idx++;
			defpar.Add("CURRENT_SHIFT", 0); defpar.Fix("CURRENT_SHIFT");
			IDX.CURRENT_SHIFT = idx++;
			defpar.Add("DELTA_SHIFT", 0, 0.01); defpar.Fix("DELTA_SHIFT");
			IDX.DELTA_SHIFT = idx++;
			defpar.Add("CUR_ERR_SCALE", 1); defpar.Fix("CUR_ERR_SCALE");
			IDX.CURRENT_ERROR_SCALE = idx++;
			defpar.Add("CNT_ERR_SCALE", 1); defpar.Fix("CNT_ERR_SCALE");
			IDX.COUNT_ERROR_SCALE = idx++;
			inipar=defpar;
		}

		void SetParameters(const vector<double> & par) const
		{
			D = par[IDX.DELTA];
			DN = par[IDX.DN];
			TAU_POL =     fabs(par[IDX.TAU_POL]);
			TAU_TSH =     fabs(par[IDX.TAU_TSH]);
			TAU_BG =      fabs(par[IDX.TAU_BG]);
			I0[0] =       fabs(par[IDX.I0]);
			I0[1] =       fabs(par[IDX.I1]);
			//I0[1] =       fabs(I0[0]*(1.0-DN));
			PTSH =        fabs(par[IDX.PTSH]);
			PBG =         fabs(par[IDX.PBG]);
			BVOL =        fabs(par[IDX.BVOL]);
			VI =              par[IDX.VI];
			P0 =          fabs(par[IDX.P0]);
			TAU_SOKOLOV = fabs(par[IDX.TAU_SOKOLOV])*3600.0;
			P1 = PMAX*TAU_POL/TAU_SOKOLOV;
			CURRENT_SHIFT =   par[IDX.CURRENT_SHIFT]*1e-2;
			DELTA_SHIFT =     par[IDX.DELTA_SHIFT];
			CURRENT_ERROR_SCALE = par[IDX.CURRENT_ERROR_SCALE];
			COUNT_ERROR_SCALE = par[IDX.COUNT_ERROR_SCALE];
		}

		~RelaxationFitter(void)
		{
		}


		void SetData(data::EnergyCalibrationStorage * d)
		{
		}

		void SetView(gui::EnergyCalibrationView * view)
		{
			ECView = view;
		}

		void Fit(void)
		{
      auto EC = ECView->GetEnergyCalibrationStorage();
			TIME = &(EC->t);
			DELTA = &(EC->DeltaSum[0]);
			CURRENT[0] = &(EC->I[0]);
			CURRENT[1] = &(EC->I[1]);
			COUNT[0] = &(EC->CountSum[0]);
			COUNT[1] = &(EC->CountSum[1]);
			SIGMAX  =  & (EC->SigmaX[0]);
			SIGMAZ  =  & (EC->SigmaZ[0]);
      
			using namespace ROOT::Minuit2;
			if(tmax==0) tmax=(*TIME)[TIME->size()-1].value;
			MnMigrad migrad(*this, inipar);
			VIFIXED = migrad.Parameter(inipar.Index("VI")).IsFixed();
			if(inipar.Value("VI") == 0 && VIFIXED) SKIPVOLUME =true;
			else  SKIPVOLUME = false;
			FunctionMinimum minimum = migrad();
			minpar=minimum.UserParameters();
			SetMinPar();
			NDF = 5*NPOINT - minpar.VariableParameters();
		}



		//this is effective counting rate per second per particle in the beam
		//it is constant if pA=pB
		double beta(double t,  double pA,  double pB) const
		{
			return pB/TAU_BG + (pB  + (pA-pB)*exp(-t/TAU_BG))/TAU_TSH;
		}


		double VolumeCorrection(double t) const 
		{
			double b0 = beta(t, 0, 1);
			double Ia = (I0[0]+I0[1])*0.5;//average current
			return VI*Ia*(1.-exp(-t/TAU_BG))*(1./TAU_BG + 1./TAU_TSH)/b0;
		}

		double IntegratedVolumeCorrection(double t) const 
		{
			double b0 = beta(t, 0, 1);
			double Ia = (I0[0]+I0[1])*0.5;//average current
			return VI*Ia*(TAU_BG+TAU_TSH)*(b0 - (1. + log(b0*TAU_BG))/TAU_BG );
		}

		double polarization(double t,  double p0, double p1) const 
		{
			double e = exp(-t/TAU_POL);
			double p = p0*e - p1*(1.-e);
			return p*p;
		}

		//\int_0^t e^{-Bt} \delta(t) dt/TAU_POL undefined
		double indefinite_polarization_integral(double t,  double p0,  double p1) const
		{
			double k = TAU_POL/TAU_BG;
			double x = t/TAU_POL;
			double e1 = exp(-k*x)/k;
			double e2 = exp(-(k+1)*x)/(k+1);
			double e3 = exp(-(k+2)*x)/(k+2);
			return  - p1*p1*e1 +  2.0*p1*(p1+p0)*e2 - (p1+p0)*(p1+p0)*e3;
		}


		double polarization_integral(double t,  double P0,  double P1) const
		{
			return indefinite_polarization_integral(t, P0, P1) - indefinite_polarization_integral(0, P0, P1);
		}


		//delta with two polarization
		double Delta(double t) const 
		{
			double b0 = beta(t, 0, 1.0);
			double b1 = beta(t, PTSH, PBG);
			double b2 = beta(t,2.0*PTSH, PBG);
			double p1 = polarization(t,  P0,   P1); //polarization of first polarized bunch
			double p2 = polarization(t,  0.0,  P1); //polarization of second unpolarized bunch
			double ip1 = polarization_integral(t, P0, P1);
			double ip2 = polarization_integral(t, 0, P1);
			return DELTA_SHIFT -  PTSH/(TAU_TSH*b1)*D*exp(-t/TAU_BG)*(p1-p2) + b2/(TAU_BG*b1*b0)*(DN+D*TAU_POL/TAU_TSH*(ip1-ip2));
      //double n1 = Count(t,I0[0],P0,P1);
      //double n2 = Count(t,I0[1],0,P1);
      //return DELTA_SHIFT+(n1-n2)/n2;
		}


    ibn::valer<double> PolarizationPart(int i) const
    {
      double t = (*TIME)[i].value;
			//double b0 = beta(t, 0, 1.0);
			double b1 = beta(t, PTSH, PBG);
			//double b2 = beta(t,2.0*PTSH, PBG);
			double p1 = polarization(t,  P0,   P1); //polarization of first polarized bunch
			double p2 = polarization(t,  0.0,  P1); //polarization of second unpolarized bunch
			//double ip1 = polarization_integral(t, P0, P1);
			//double ip2 = polarization_integral(t, 0, P1);
			//return DELTA_SHIFT -  PTSH/(TAU_TSH*b1)*D*exp(-t/TAU_BG)*(p1-p2) + b2/(TAU_BG*b1*b0)*(DN+D*TAU_POL/TAU_TSH*(ip1-ip2));
      return ((*DELTA)[i] - ibn::valer<double>(Delta(t),0))/PTSH*(TAU_TSH*b1)/D*exp(t/TAU_BG) + (p1-p2);
    }

    double PolarizationPart(double t) const
    {
			//double b0 = beta(t, 0, 1.0);
			//double b1 = beta(t, PTSH, PBG);
			//double b2 = beta(t,2.0*PTSH, PBG);
			double p1 = polarization(t,  P0,   P1); //polarization of first polarized bunch
			double p2 = polarization(t,  0.0,  P1); //polarization of second unpolarized bunch
      return p1-p2;
    }


		double Current(double t,  double I0,  double P0,  double P1) const 
		{
			double dV = IntegratedVolumeCorrection(t);
			double pI = polarization_integral(t, P0, P1);
			return I0*exp(-t/TAU_BG)/(TAU_BG*beta(t, 0, 1) - D*TAU_POL/TAU_TSH*pI + dV);
		}



		double Volume(double t,  double I0,  double P0,  double P1) const 
		{
			return (1.0 + VI*Current(t, I0,  P0,  P1))/(1.0  + VI*I0);
		}

		double Count(double t, double I0,  double P0,  double P1) const 
		{
			const double Ie = 1.602e-19*818924*1e6; //current of single electron in uA
			double N0 = I0/Ie;
			double I = Current(t, 1.0, P0, P1); //normalized current
			double p = polarization(t, P0, P1);
			return N0*I*( PTSH/TAU_TSH*I*(1.0 - D*p)/Volume(t, I0,  P0, P1)  + PBG/TAU_BG);
		}

		double operator() (const std::vector<double> & par) const
		{
			SetParameters(par);
			double chi2=0;
			double chi2_delta=0;
			double chi2_dn =0;
			double chi2_I[2]={0, 0};
			double chi2_count[2]={0, 0};
			double chi2_volume = 0;
			NPOINT=0; //number of data point wich play game

			double P[2] = {P0, 0}; //polarization of bunches
			double di[2] = {0, CURRENT_SHIFT};

			for(unsigned i=0;i<TIME->size();i++)
			{
				double t = (*TIME)[i].value;
				if(i+1<TIME->size()) COUNT_TIME = (*TIME)[i+1].value - t;
				t-=COUNT_TIME/2.0; //SHIFT time position in order to compesate my wrong graph drawing i'm going to change it in future
				if(t < tmin || tmax < t) continue;
				//delta contribution
				double d = (*DELTA)[i].value;
				double d_error = (*DELTA)[i].error;
				chi2_delta+=pow((d - Delta(t))/d_error,2.0);
				for(int bunch=0;bunch<2;bunch++)
				{
					//current contribution
					double I = (*CURRENT[bunch])[i].value;
					double dI = CURRENT_ERROR_SCALE*(*CURRENT[bunch])[i].error;
					chi2_I[bunch]+=pow((I -Current(t, I0[bunch]*(1.0+di[bunch]),  P[bunch],  P1))/dI, 2.0);
					//count contribution
					double nu = (*COUNT[bunch])[i].value;
					double dnu = COUNT_ERROR_SCALE*(*COUNT[bunch])[i].error;
					if(bunch==0)
					{
						chi2_count[bunch] += pow((nu -Count(t, I0[bunch]*(1.0+di[bunch]), P[bunch], P1))/dnu, 2.0);
					}
				}
				//volume contribution
				double sz = (*SIGMAZ)[i].value;
				double sx = (*SIGMAX)[i].value;
				double dsz = (*SIGMAZ)[i].error;
				double dsx = (*SIGMAX)[i].error;
				double dv = sqrt( sz*sz*dsx*dsx  + sx*sx*dsz*dsz);
				//to calculate volume use average initial current and average initial polarization
        double v = sz*sx;
        double v1 = BVOL*Volume(t,  I0[0], P0, P1);
        double v2 = BVOL*Volume(t,  I0[1], 0,  P1);
				chi2_volume+=pow( (v - 0.5*(v1+v2))/dv  , 2.0);
				NPOINT++;
			}
			chi2 = chi2_delta + chi2_I[0] + chi2_I[1] + chi2_count[0] + chi2_count[1] + chi2_dn;
			if(!SKIPVOLUME) 
			{
				CHI2_VOLUME = chi2_volume;
				chi2+=chi2_volume;
			}
			CHI2 = chi2;
			CHI2_DELTA = chi2_delta;
			CHI2_DN = chi2_dn;
			for(int bunch=0;bunch<2;bunch++)
			{
				CHI2_CURRENT[bunch] = chi2_I[bunch];
				CHI2_COUNT[bunch] = chi2_count[bunch];
			}
			return chi2;
		} 

		double Up() const { return 1.; }

		void SetMinPar(void)
		{
			SetParameters(minpar.Params());
		}

		void Print(void)
		{
			cout << LOG(1) << "Fit result:" << endl;
			cout << LOG(1) << minpar << endl;
			cout << LOG(1) << "CHI2/NDF = " << CHI2/NDF << endl;
			cout << LOG(1) << "CHI2_DELTA/NPOINT = " << CHI2_DELTA/NPOINT << endl;
			cout << LOG(1) << "CHI2_CURRENT1/NPOINT = " << CHI2_CURRENT[0]/NPOINT << endl;
			cout << LOG(1) << "CHI2_CURRENT2/NPOINT = " << CHI2_CURRENT[1]/NPOINT << endl;
			cout << LOG(1) << "CHI2_COUNT1/NPOINT = " << CHI2_COUNT[0]/NPOINT << endl;
			cout << LOG(1) << "CHI2_COUNT2/NPOINT = " << CHI2_COUNT[1]/NPOINT << endl;
			cout << LOG(1) << "CHI2_VOLUME/NPOINT = " << CHI2_VOLUME/NPOINT << endl;
			cout << LOG(1) << "NPOINT = " << NPOINT << endl;
			cout << LOG(1) << "G = " << TAU_POL/TAU_SOKOLOV << endl;
		}

		void Draw(int opt)
		{
			if(!pave)
			{
				pave.reset(new TPaveText(0.7, 0.7, 0.9, 0.9, "NDC"));
			}
			pave->Clear();
			//pave->SetTextFont(102);
			pave->SetTextAlign(12);
			pave->SetBorderSize(1);
			pave->SetFillColor(0);
			boost::format fmt("%10s = %10.3f #pm %6.3f");
			for(unsigned i=0;i<minpar.Params().size();i++)
			{
				string s = (fmt %  minpar.GetName(i) % minpar.Params()[i] % minpar.Errors()[i] ).str();
				//cout << i << " " << s << endl;
				pave->AddText(s.c_str());
			}

			ECView->Add("delta-1")->Draw();
			pave->Draw();
			DeltaFunc.reset(new TF1("RelaxationDeltaFunc", [=](double * x,  double *p) { return Delta(x[0]); } , 0, tmax, 0));
			DeltaFunc->Draw("same");
			if(opt!=0)
			{
				double P[2]={P0, 0}; //polarization
				double di[2] = {0, CURRENT_SHIFT};
				for(unsigned bunch=0;bunch<2; bunch++)
				{
					string bunch_suffix = to_string(bunch);
					string human_bunch_suffix = to_string(bunch+1);
					string delta_fun_name = "RelaxationDeltaFunc"+ bunch_suffix;
					string count_fun_name = "RelaxationCountFunc"+ bunch_suffix;
					string current_graph_name = "current-"+human_bunch_suffix;
					string count_graph_name = "counter_sum-"+human_bunch_suffix;
					CurrentFunc[bunch].reset(new TF1(delta_fun_name.c_str(), [&](double * x,  double *p) {return Current(x[0], I0[bunch]*(1.0 + di[bunch]),  P[bunch], P1);  } , tmin, tmax, 0));
					CountFunc[bunch].reset(new TF1(count_fun_name.c_str(), [&](double * x,  double *p)   {return   Count(x[0], I0[bunch]*(1.0 + di[bunch]),  P[bunch], P1);  } , tmin, tmax, 0));
					ECView->Add(current_graph_name.c_str())->Draw();
					CurrentFunc[bunch]->Draw("same");
					pave->Draw();
					ECView->Add(count_graph_name.c_str())->Draw();
					CountFunc[bunch]->Draw("same");
					pave->Draw();
				}
				ECView->Add("beam_volume-1")->Draw();
				pave->Draw();
				VolumeFunc.reset(new TF1("RelaxationVolumeFunc", [=](double * x,  double *p) 
							{ 
							return BVOL*Volume(x[0], 0.5*(I0[0]+I0[1]), P0/2.0, P1) ; 
							} , tmin, tmax, 0));
				VolumeFunc->Draw("same");
			}

      new TCanvas;
      TGraphErrors * g = new TGraphErrors;
      for(int i=0;i<TIME->size();i++)
      {
        auto d = PolarizationPart(i);
        g->SetPoint(i, (*TIME)[i].value, d.value);
        g->SetPointError(i, (*TIME)[i].error,d.error);
      }
      g->SetMarkerStyle(21);
      g->Draw("ap");
      TF1 * f = new TF1("polpart",[&](double *x, double *p){ return PolarizationPart(x[0]); },0, tmax,0);
      f->Draw("same");
			ECView->Update();
		}
	};
}
#endif
