// =====================================================================================
//
//       Filename:  RelaxationFitter.h
//
//    Description:  Smart Relaxation Fitter
//
//        Version:  1.0
//        Created:  23.02.2015 15:06:41
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#pragma once
#include "Fitter.h"
#include "Minuit2/MnPrint.h"

#include <TF1.h>
#include <TPaveText.h>

#include "data/EnergyCalibrationStorage.h"
#include "gui/EnergyCalibrationView.h"
#include <boost/format.hpp>

namespace fit
{
	class RelaxationFitter2 : public Fitter
	{
    const double Ie = 1.602e-19*818924*1e6; //current of single electron in uA
		unique_ptr<TF1> DeltaFunc;
		unique_ptr<TF1> CurrentFunc[2];
		unique_ptr<TF1> CountFunc[2];
		unique_ptr<TF1> VolumeFunc;
		bool VIFIXED;
		bool SKIPVOLUME;

		unique_ptr<TPaveText> pave;
		gui::EnergyCalibrationView * ECView;
		public:
		data::Array * TIME;
		data::Array * DELTA;
		data::Array * CURRENT[2];
		data::Array * COUNT[2];
		data::Array * SIGMAX;
		data::Array * SIGMAZ;
		data::Array fit_result;
		std::unique_ptr<ROOT::Minuit2::MnApplication> minimizer;

		mutable double D;  //jump
		//mutable double DN; //relative difference in number of particle
		mutable double TAU_POL; //depolarization time
		mutable double TAU_TSH[2]; //touschek time,  sec
		mutable double TAU_BG[2];  //background time, sec
		mutable double I0[2];  //initial current,  uA
		mutable double PTSH[2]; //probability to register touschek
		mutable double PBG[2];  //probability to register background
		mutable double P0[2];    //initial polarization of first bunch
		mutable double BVOL[2];  //initial beam volume um^2
		mutable double VI;    //volume dependace on current, 1/uA^2
		mutable double P1;    //equilibrium level of  polarization = 0.92/G
		mutable double TAU_SOKOLOV; //polarization time due to Sokolov Termov radiation polarization
		mutable double CURRENT_SHIFT; //unideality of measurement current
		mutable double CURRENT_ERROR_SCALE; //scale of the error of current measurement
		mutable double COUNT_ERROR_SCALE;   //scale of the error of count measurement
		mutable double VOLUME_ERROR_SCALE;   //scale of the error of count measurement

		mutable struct 
		{
			int DELTA;
			int DN;
			int TAU_POL;
			int TAU_TSH[2];
			int TAU_BG[2];
			int I0[2];
			int PTSH[2];
			int PBG[2];
			int P0[2];
			int BVOL[2];
			int VI;
			int TAU_SOKOLOV;
			int CURRENT_SHIFT;
			int CURRENT_ERROR_SCALE;
			int COUNT_ERROR_SCALE;
      int VOLUME_ERROR_SCALE;
		} IDX;


		mutable double COUNT_TIME;
		mutable double CHI2;
		mutable double CHI2_DELTA;
		mutable double CHI2_DN;
		mutable double CHI2_CURRENT[2];
		mutable double CHI2_COUNT[2];
		mutable double CHI2_VOLUME;
		mutable double NDF; //number degree of freedom
		mutable double NPOINT; //number of data which play game

		const double PMAX = 8.0/15.0*sqrt(3.0); //maximum polarization degreee (Sokolov Ternov)



		RelaxationFitter2(void)
		{
			type = MINUIT2;
			name = "SMART_RELAXATION2";
			title = "Fit data to find depolarization time for its relaxation";
			int idx=0;
			defpar.Add("DELTA",0.01, 0.01); //depolarization effect
			IDX.DELTA = idx++;
			//defpar.Add("DN",   0.01, 0.01);    //count difference due to different number of particle
			//IDX.DN = idx++;
			defpar.Add("TAU_POL",  1500,   100, 0, 1e6); //depolarization time,  sec
			IDX.TAU_POL =idx++;
			defpar.Add("TAU_TSH1",  5000,  1000, 0, 1e6); //touschek lift time,  sec
			IDX.TAU_TSH[0] = idx++;
			defpar.Add("TAU_TSH2",  5000,  1000, 0, 1e6);  defpar.Fix("TAU_TSH2");
			IDX.TAU_TSH[1] = idx++;
			defpar.Add("TAU_BG1",  10000,   1000, 0, 1e10); //background life time,  sec
			IDX.TAU_BG[0] = idx++;
			defpar.Add("TAU_BG2",  10000,   1000, 0, 1e10); defpar.Fix("TAU_BG2");
			IDX.TAU_BG[1] = idx++;
			defpar.Add("I1",  2000,   1000); //initial current, for the first bunch uA 
			IDX.I0[0] = idx++;
			defpar.Add("I2",  2000,   1000); //initial current,  uA
			IDX.I0[1] = idx++;
			defpar.Add("PTSH1",  0.2, 0.5, 0, 1);   //probability to register touschek
			IDX.PTSH[0] = idx++;
			defpar.Add("PTSH2",  0.2, 0.5, 0, 1);   defpar.Fix("PTSH2");
			IDX.PTSH[1] = idx++;
			defpar.Add("PBG1",  0.05, 0.5, 0, 1);   //probability to register background
			IDX.PBG[0] = idx++;
			defpar.Add("PBG2",  0.05, 0.5, 0, 1);  defpar.Fix("PBG2");
			IDX.PBG[1] = idx++;
			defpar.Add("P1", 0.6, 0.1, -PMAX, PMAX); defpar.Fix("P1"); //initial polarization degree of first bunch
			IDX.P0[0] = idx++;
			defpar.Add("P2",0, 0.1, -PMAX, PMAX); defpar.Fix("P2"); //initial polarization degree of first bunch
			IDX.P0[1] = idx++;
			defpar.Add("BVOL1", 200e3);             //beam volume um^2
			IDX.BVOL[0] = idx++;
			defpar.Add("BVOL2", 200e3);             //beam volume um^2
			IDX.BVOL[1] = idx++;
			defpar.Add("VI", 0);
			IDX.VI = idx++;
			defpar.Add("TAU_SOKOLOV", 87.0); defpar.Fix("TAU_SOKOLOV"); //in hours
			IDX.TAU_SOKOLOV = idx++;
			defpar.Add("CUR_ERR_SCALE", 8); defpar.Fix("CUR_ERR_SCALE");
			IDX.CURRENT_ERROR_SCALE = idx++;
			defpar.Add("CNT_ERR_SCALE", 17); defpar.Fix("CNT_ERR_SCALE");
			IDX.COUNT_ERROR_SCALE = idx++;
			defpar.Add("VOL_ERR_SCALE", 2); defpar.Fix("VOL_ERR_SCALE");
			IDX.VOLUME_ERROR_SCALE = idx++;
			inipar=defpar;
		}

		void SetParameters(const vector<double> & par) const
		{
			D = par[IDX.DELTA];
			TAU_POL =     fabs(par[IDX.TAU_POL]);
      for(int b=0;b<2;b++)
      {
        I0[b] =       fabs(par[IDX.I0[b]]);
        TAU_TSH[b] = fabs(par[IDX.TAU_TSH[b]]);
        TAU_BG[b] =      fabs(par[IDX.TAU_BG[b]]);
        PTSH[b] =        fabs(par[IDX.PTSH[b]]);
        PBG[b] =         fabs(par[IDX.PBG[b]]);
        P0[b] =          fabs(par[IDX.P0[b]]);
        BVOL[b] =        fabs(par[IDX.BVOL[b]]);
      }
      if(inipar.Parameter(IDX.TAU_TSH[1]).IsFixed()) //calculate TAU_TSH for second bunch from first bunch
      {
        double A =BVOL[0]/(TAU_TSH[0]*I0[0]);
        TAU_TSH[1] = BVOL[1]/(A*I0[1]);
      }
      if(inipar.Parameter(IDX.TAU_BG[1]).IsFixed()) 
      {
        TAU_BG[1] = TAU_BG[0];
      }
      if(inipar.Parameter(IDX.PTSH[1]).IsFixed()) 
      {
        PTSH[1]=PTSH[0];
      }
      if(inipar.Parameter(IDX.PBG[1]).IsFixed()) 
      {
        PBG[1]=PBG[0];
      }
      if(inipar.Parameter(IDX.BVOL[1]).IsFixed()) 
      {
        BVOL[1]=BVOL[0];
      }
			VI =              par[IDX.VI];
			TAU_SOKOLOV = fabs(par[IDX.TAU_SOKOLOV])*3600.0;
			P1 = PMAX*TAU_POL/TAU_SOKOLOV;
			CURRENT_ERROR_SCALE = par[IDX.CURRENT_ERROR_SCALE];
			COUNT_ERROR_SCALE = par[IDX.COUNT_ERROR_SCALE];
      VOLUME_ERROR_SCALE = par[IDX.VOLUME_ERROR_SCALE];
		}

		~RelaxationFitter2(void)
		{
		}


		void SetData(data::EnergyCalibrationStorage * d)
		{
		}

		void SetView(gui::EnergyCalibrationView * view)
		{
			ECView = view;
		}

		void Fit(void)
		{
      auto EC = ECView->GetEnergyCalibrationStorage();
			TIME = &(EC->t);
			CURRENT[0] = &(EC->I[0]);
			CURRENT[1] = &(EC->I[1]);
      if(true)
      {
        DELTA = &(EC->DeltaSum[0]);
        COUNT[0] = &(EC->CountSum[0]);
        COUNT[1] = &(EC->CountSum[1]);
      }
      else
      {
        DELTA = &(EC->DeltaCoincSum[0]);
        COUNT[0] = &(EC->CoincSum[0]);
        COUNT[1] = &(EC->CoincSum[1]);
      }
			SIGMAX  =  & (EC->SigmaX[0]);
			SIGMAZ  =  & (EC->SigmaZ[0]);
      
			using namespace ROOT::Minuit2;
			if(tmax==0) tmax=(*TIME)[TIME->size()-1].value;
			MnMigrad migrad(*this, inipar);
			VIFIXED = migrad.Parameter(inipar.Index("VI")).IsFixed();
			if(inipar.Value("VI") == 0 && VIFIXED) SKIPVOLUME =true;
			else  SKIPVOLUME = false;
			FunctionMinimum minimum = migrad();
			minpar=minimum.UserParameters();
			SetMinPar();
			NDF -= minpar.VariableParameters();
		}



		//this is effective counting rate per second per particle in the beam
		//it is constant if pA=pB
		double beta(double t,  double pA,  double pB, int b=0) const
		{
			return pB/TAU_BG[b] + (pB  + (pA-pB)*exp(-t/TAU_BG[b]))/TAU_TSH[b];
		}


		double VolumeCorrection(double t, int b=0) const 
		{
			double b0 = beta(t, 0, 1,b);
			double Ia = (I0[0]+I0[1])*0.5;//average current
			return VI*Ia*(1.-exp(-t/TAU_BG[b]))*(1./TAU_BG[b] + 1./TAU_TSH[b])/b0;
		}

		double IntegratedVolumeCorrection(double t, int b) const 
		{
			double b0 = beta(t, 0, 1,b);
			double Ia = (I0[0]+I0[1])*0.5;//average current
			return VI*Ia*(TAU_BG[b]+TAU_TSH[b])*(b0 - (1. + log(b0*TAU_BG[b]))/TAU_BG[b] );
		}

		double polarization(double t,  double p0, double p1) const 
		{
			double e = exp(-t/TAU_POL);
			double p = p0*e - p1*(1.-e);
			return p*p;
		}

		//\int_0^t e^{-Bt} \delta(t) dt/TAU_POL undefined
		double indefinite_polarization_integral(double t,  double p0,  double p1, int b=0) const
		{
			double k = TAU_POL/TAU_BG[b];
			double x = t/TAU_POL;
			double e1 = exp(-k*x)/k;
			double e2 = exp(-(k+1)*x)/(k+1);
			double e3 = exp(-(k+2)*x)/(k+2);
			return  - p1*p1*e1 +  2.0*p1*(p1+p0)*e2 - (p1+p0)*(p1+p0)*e3;
		}


		double polarization_integral(double t,  double P0,  double P1, int b=0) const
		{
			return indefinite_polarization_integral(t, P0, P1,b) - indefinite_polarization_integral(0, P0, P1,b);
		}


		//delta with two polarization
		double Delta(double t) const 
		{
			//double b0 = beta(t, 0, 1.0);
			//double b1 = beta(t, PTSH, PBG);
			//double b2 = beta(t,2.0*PTSH, PBG);
			//double p1 = polarization(t,  P0,   P1); //polarization of first polarized bunch
			//double p2 = polarization(t,  0.0,  P1); //polarization of second unpolarized bunch
			//double ip1 = polarization_integral(t, P0, P1);
			//double ip2 = polarization_integral(t, 0, P1);
			//return DELTA_SHIFT -  PTSH/(TAU_TSH*b1)*D*exp(-t/TAU_BG)*(p1-p2) + b2/(TAU_BG*b1*b0)*(DN+D*TAU_POL/TAU_TSH*(ip1-ip2));
      double n1 = Count(t,0);
      double n2 = Count(t,1);
      return (n1-n2)/n2;
		}


    ibn::valer<double> PolarizationPart(int i, int b=0) const
    {
      double t = (*TIME)[i].value;
			//double b0 = beta(t, 0, 1.0);
			double b1 = beta(t, PTSH[b], PBG[b],b);
			//double b2 = beta(t,2.0*PTSH, PBG);
			double p1 = polarization(t,  P0[0],   P1); //polarization of first polarized bunch
			double p2 = polarization(t,  P0[1],  P1); //polarization of second unpolarized bunch
			//double ip1 = polarization_integral(t, P0, P1);
			//double ip2 = polarization_integral(t, 0, P1);
			//return DELTA_SHIFT -  PTSH/(TAU_TSH*b1)*D*exp(-t/TAU_BG)*(p1-p2) + b2/(TAU_BG*b1*b0)*(DN+D*TAU_POL/TAU_TSH*(ip1-ip2));
      return ((*DELTA)[i] - ibn::valer<double>(Delta(t),0))/PTSH[b]*(TAU_TSH[b]*b1)/D*exp(t/TAU_BG[b]) + (p1-p2);
    }

    double PolarizationPart(double t) const
    {
			//double b0 = beta(t, 0, 1.0);
			//double b1 = beta(t, PTSH, PBG);
			//double b2 = beta(t,2.0*PTSH, PBG);
			double p1 = polarization(t,  P0[0],   P1); //polarization of first polarized bunch
			double p2 = polarization(t,  P0[1],   P1); //polarization of second unpolarized bunch
      return p1-p2;
    }


		//double Current(double t,  double I0,  double P0,  double P1) const 
		//{
		//	double dV = IntegratedVolumeCorrection(t);
		//	double pI = polarization_integral(t, P0, P1);
		//	return I0*exp(-t/TAU_BG)/(TAU_BG*beta(t, 0, 1) - D*TAU_POL/TAU_TSH*pI + dV);
		//}

		double Current(double t, int b=0) const 
		{
			double dV = IntegratedVolumeCorrection(t,b);
			double pI = polarization_integral(t, P0[b], P1);
			return I0[b]*exp(-t/TAU_BG[b])/(TAU_BG[b]*beta(t, 0, 1) - D*TAU_POL/TAU_TSH[b]*pI + dV);
		}

		double Volume(double t,  int b=0) const 
		{
			return (1.0 + VI*Current(t,b))/(1.0  + VI*I0[b]);
		}

		double Count(double t, int b) const 
		{
			const double Ie = 1.602e-19*818924*1e6; //current of single electron in uA
			double N0 = I0[b]/Ie;
			double I = Current(t, b)/Current(0,b);
			double p = polarization(t, P0[b], P1);
			return N0*I*( PTSH[b]/TAU_TSH[b]*I*(1.0 - D*p)/Volume(t, b)  + PBG[b]/TAU_BG[b]);
		}

		double operator() (const std::vector<double> & par) const
		{
			SetParameters(par);
			double chi2=0;
			double chi2_delta=0;
			double chi2_dn =0;
			double chi2_I[2]={0, 0};
			double chi2_count[2]={0, 0};
			double chi2_volume = 0;
			NPOINT=0; //number of data point wich play game
      NDF=0;


			for(unsigned i=0;i<TIME->size();i++)
			{
				double t = (*TIME)[i].value;
				//if(i+1<TIME->size()) COUNT_TIME = (*TIME)[i+1].value - t;
				//t-=COUNT_TIME/2.0; //SHIFT time position in order to compesate my wrong graph drawing i'm going to change it in future
				if(t < tmin || tmax < t) continue;
				//delta contribution
				double d = (*DELTA)[i].value;
				double d_error = (*DELTA)[i].error;
				chi2_delta+=pow((d - Delta(t))/d_error,2.0);
        NDF++;
				for(int bunch=0;bunch<2;bunch++)
				{
					//current contribution
					double I = (*CURRENT[bunch])[i].value;
					double dI = CURRENT_ERROR_SCALE*(*CURRENT[bunch])[i].error;
					chi2_I[bunch]+=pow((I -Current(t,bunch))/dI, 2.0);
          NDF++;
					//count contribution
					double nu = (*COUNT[bunch])[i].value;
					double dnu = COUNT_ERROR_SCALE*(*COUNT[bunch])[i].error;
					if(bunch==0)
					{
						chi2_count[bunch] += pow((nu -Count(t,bunch))/dnu, 2.0);
            NDF++;
					}
				}
				//volume contribution
				double sz = (*SIGMAZ)[i].value;
				double sx = (*SIGMAX)[i].value;
				double dsz = (*SIGMAZ)[i].error;
				double dsx = (*SIGMAX)[i].error;
				double dv = sqrt( sz*sz*dsx*dsx  + sx*sx*dsz*dsz)*VOLUME_ERROR_SCALE;
				//to calculate volume use average initial current and average initial polarization
        double v = sz*sx;
        double v1 = BVOL[0]*Volume(t,0);
        double v2 = BVOL[1]*Volume(t,1);
				chi2_volume+=pow( (v - 0.5*(v1+v2))/dv  , 2.0);
        NDF++;
				NPOINT++;
			}
			chi2 = chi2_delta + chi2_I[0] + chi2_I[1] + chi2_count[0] + chi2_count[1] + chi2_dn;
			if(!SKIPVOLUME) 
			{
				CHI2_VOLUME = chi2_volume;
				chi2+=chi2_volume;
			}
			CHI2 = chi2;
			CHI2_DELTA = chi2_delta;
			CHI2_DN = chi2_dn;
			for(int bunch=0;bunch<2;bunch++)
			{
				CHI2_CURRENT[bunch] = chi2_I[bunch];
				CHI2_COUNT[bunch] = chi2_count[bunch];
			}
			return chi2;
		} 

		double Up() const { return 1.; }

		void SetMinPar(void)
		{
			SetParameters(minpar.Params());
		}

		void Print(void)
		{
			cout << LOG(1) << "Fit result:" << endl;
			cout << LOG(1) << minpar << endl;
			cout << LOG(1) << "TOTAL_CHI2/NDF = " << CHI2/NDF << endl;
			cout << LOG(1) << "CHI2_DELTA/NPOINT = " << CHI2_DELTA/NPOINT << endl;
			cout << LOG(1) << "CHI2_CURRENT1/NPOINT = " << CHI2_CURRENT[0]/NPOINT << endl;
			cout << LOG(1) << "CHI2_CURRENT2/NPOINT = " << CHI2_CURRENT[1]/NPOINT << endl;
			cout << LOG(1) << "CHI2_COUNT1/NPOINT = " << CHI2_COUNT[0]/NPOINT << endl;
			cout << LOG(1) << "CHI2_COUNT2/NPOINT = " << CHI2_COUNT[1]/NPOINT << endl;
			cout << LOG(1) << "CHI2_VOLUME/NPOINT = " << CHI2_VOLUME/NPOINT << endl;
			cout << LOG(1) << "NPOINT = " << NPOINT << endl;
			cout << LOG(1) << "G = " << TAU_POL/TAU_SOKOLOV << endl;
		}

		void Draw(int opt)
		{
			if(!pave)
			{
				pave.reset(new TPaveText(0.7, 0.7, 0.9, 0.9, "NDC"));
			}
			pave->Clear();
			//pave->SetTextFont(102);
			pave->SetTextAlign(12);
			pave->SetBorderSize(1);
			pave->SetFillColor(0);
			boost::format fmt("%10s = %10.3f #pm %6.3f");
			for(unsigned i=0;i<minpar.Params().size();i++)
			{
				string s = (fmt %  minpar.GetName(i) % minpar.Params()[i] % minpar.Errors()[i] ).str();
				//cout << i << " " << s << endl;
				pave->AddText(s.c_str());
			}

			//ECView->Add("delta-1")->Draw();
			//ECView->Add("delta_coincidence_sum-1")->Draw();
      //TCanvas * c_group = 
      new TCanvas("c_group","delta-1");
      TGraphErrors * gd = new TGraphErrors;

      int ngroup=8;
      for(int i=0;i<DELTA->size()/ngroup;i++)
      {
        //ibn::averager<double> av;
        ibn::phys_averager av;
        ibn::averager<double> tt;
        //std::vector<double> d(ngroup);
        //std::vector<double> e(ngroup);
        double tmin = (*DELTA)[ngroup*i].value;
        double tmax = tmin;
        for(int j =0;j<ngroup;j++) 
        {
          //d[j] = (*DELTA)[ngroup*i+j].value;
          //e[j] = (*DELTA)[ngroup*i+j].error;
          int idx = ngroup*i+j;
          av.add((*DELTA)[idx].value, (*DELTA)[idx].error);
          tt.add((*TIME)[idx].value);
          tmin = std::min(tmin, (*TIME)[idx].value - (*TIME)[idx].error);
          tmax = std::min(tmin, (*TIME)[idx].value + (*TIME)[idx].error);
        }
        gd->SetPoint(i,tt.average(), av.average());
        gd->SetPointError(i,(tmax-tmin)/2, av.sigma_average());
      }
      gd->SetMarkerStyle(21);
      gd->Draw("ap");
      gd->GetXaxis()->SetTitle("time (sec)");
      gd->GetYaxis()->SetTitle("");
      gd->SetTitle("f_{1}/f_{2} - 1");
      gd->GetYaxis()->SetDecimals();

			//pave->Draw();
			DeltaFunc.reset(new TF1("RelaxationDeltaFunc", [=](double * x,  double *p) { return Delta(x[0]); } , 0, tmax, 0));
			DeltaFunc->Draw("same");
			if(opt!=0)
			{
				for(unsigned bunch=0;bunch<2; bunch++)
				{
					string bunch_suffix = to_string(bunch);
					string human_bunch_suffix = to_string(bunch+1);
					string delta_fun_name = "RelaxationDeltaFunc"+ bunch_suffix;
					string count_fun_name = "RelaxationCountFunc"+ bunch_suffix;
					string current_graph_name = "current-"+human_bunch_suffix;
					string count_graph_name = "counter_sum-"+human_bunch_suffix;
					//string count_graph_name = "coincidence_sum-"+human_bunch_suffix;
					CurrentFunc[bunch].reset(new TF1(delta_fun_name.c_str(), [=](double * x,  double *p) {return Current(x[0], bunch);  } , tmin, tmax, 0));
					CountFunc[bunch].reset(new TF1(count_fun_name.c_str(), [=](double * x,  double *p)   {return   Count(x[0], bunch);  } , tmin, tmax, 0));
					ECView->Add(current_graph_name.c_str())->Draw();
					CurrentFunc[bunch]->Draw("same");
					pave->Draw();
					ECView->Add(count_graph_name.c_str())->Draw();
					CountFunc[bunch]->Draw("same");
					pave->Draw();
				}
				ECView->Add("beam_volume-1")->Draw();
				pave->Draw();
				VolumeFunc.reset(new TF1("RelaxationVolumeFunc", [=](double * x,  double *p) 
							{ 
							return 0.5*(BVOL[0]*Volume(x[0], 0) +  BVOL[1]*Volume(x[0], 1)); 
							} , tmin, tmax, 0));
				VolumeFunc->Draw("same");
			}

      //TCanvas *c_pol  = 
      new TCanvas("c_pol","Polarization");
      TGraphErrors * g = new TGraphErrors;
      for(int i=0;i<TIME->size()/ngroup;i++)
      {
        ibn::phys_averager av;
        ibn::averager<double> tt;
        //std::vector<double> d(ngroup);
        //std::vector<double> e(ngroup);
        double tmin = (*DELTA)[ngroup*i].value;
        double tmax = tmin;
        for(int j =0;j<ngroup;j++) 
        {
          //d[j] = (*DELTA)[ngroup*i+j].value;
          //e[j] = (*DELTA)[ngroup*i+j].error;
          int idx = ngroup*i+j;
          auto d = PolarizationPart(idx);
          av.add(sqrt(d.value), 0.5*d.error/sqrt(d.value));
          tt.add((*TIME)[idx].value);
          tmin = std::min(tmin, (*TIME)[idx].value - (*TIME)[idx].error);
          tmax = std::min(tmin, (*TIME)[idx].value + (*TIME)[idx].error);
        }
        g->SetPoint(i, tt.average(), av.average());
        g->SetPointError(i, 0,av.sigma_average());
      }
      g->SetMarkerStyle(21);
      g->Draw("ap");
      TF1 * f = new TF1("polpart",[&](double *x, double *p){ return sqrt(PolarizationPart(x[0])); },0, tmax,0);
      f->Draw("same");
			ECView->Update();
		}
	};
}
