/*
 * =====================================================================================
 *
 *       Filename:  RootFunctionFitter.h
 *
 *    Description:  Fit using Root function infrastructure
 *
 *        Version:  1.0
 *        Created:  2013-01-31 14:32 NSK
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Institute of Nuclear Physics, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#ifndef IBN_FIT_ROOTFUNCTION_FITTER_H
#define IBN_FIT_ROOTFUNCTION_FITTER_H
#include "Fitter.h"

#include <TF1.h>
#include <TVirtualFitter.h>

#include <TGraph.h>
#include <TGraphErrors.h>
#include <list>
#include <algorithm>

namespace fit
{

  /*
   * =====================================================================================
   *        Class:  RootFunctionFitter
   *  Description:  Подгоняет данные в графике при помощи ROOT-вой функции TF1.
   *                Данный класс нужен для унификации интерфейса подгонки с более 
   *                сложными случаями.
   * =====================================================================================
   */
  class RootFunctionFitter : public Fitter
  {
    TGraph * graph;
    std::unique_ptr<TF1> func;
    public:
    RootFunctionFitter(std::string fit_name, double(*funptr)(double*,double*), const ROOT::Minuit2::MnUserParameters & p,  bool s=true)
    {
			ROOT::Minuit2::MnPrint::SetLevel(0); //this code doest work
      type = ROOT_FUNCTION;
      defpar = p; //set default parameters which never changed during runtime
      inipar = defpar;
      func.reset(new TF1(fit_name.c_str(),funptr,0,1e4,p.Params().size()));
      for(unsigned i=0;i<p.Params().size(); i++)
      {
        func->SetParName(i,p.Name(i));
      }
      graph=0;
      name = func->GetName();
      func->SetNpx(1000);
			isscan = s;
    }

    RootFunctionFitter(TF1 * f, const ROOT::Minuit2::MnUserParameters & p,  bool s=true)
    {
			ROOT::Minuit2::MnPrint::SetLevel(0); //this code doest work
      type = ROOT_FUNCTION;
      defpar = p; //set default parameters which never changed during runtime
      inipar = defpar;
      func.reset(f);
      for(unsigned i=0;i<p.Params().size(); i++)
      {
        func->SetParName(i,p.Name(i));
      }
      graph=0;
      name = func->GetName();
      func->SetNpx(1000);
			isscan = s;
    }

    //void SetData(TGraph * g)
    //{
    //  graph=g;
    //}

    void Fit(void)
    {
      std::clog << LOG(5) << "RootFunctionFitter::Fit BEGIN" << std::endl;
      TVirtualFitter::SetMaxIterations 	(1000000);
      graph = fView->GetGraph(fDataName);
      if(graph == nullptr) 
      {
        std::cerr << LOG(NORMAL_LOG) << "Unable to fine graph " << fDataName << std::endl;
        return;
      }
      using namespace ROOT::Minuit2;
      MnUserParameters newpar=inipar;
      if(!newpar.Parameter(newpar.Index("T")).IsFixed() && isscan) newpar = FindJump();
      std::clog << LOG(6) << "RootFunctionFitter::Fit After find jump" << std::endl;
      if(tmin<tmax)
      {
        func->SetRange(tmin, tmax);
      }
      std::clog << LOG(6) << "RootFunctionFitter::Fit After func->SetRange" << std::endl;
      for(unsigned i=0;i<newpar.Params().size();i++)
      {
        const ROOT::Minuit2::MinuitParameter & mp = newpar.Parameter(i);
        const ROOT::Minuit2::MinuitParameter & ip = inipar.Parameter(i);
        func->SetParameter(i,mp.Value());
        func->SetParError(i,mp.Error());
        //if(mp.HasLimits()) 
        if(ip.HasLimits()) 
        {
          std::cout << "Setting par limit for varibale " << i << std::endl;
          func->SetParLimits(i, ip.LowerLimit(), ip.UpperLimit());
        }
        else func->SetParLimits(i,0,0);
        if(mp.IsFixed()) func->FixParameter(i,mp.Value());
        else func->ReleaseParameter(i);
      }
      std::clog << LOG(6) << "RootFunctionFitter::Fit After set parameters" << std::endl;
			const char * fitopt = "EX0RME";
      if(tmin < tmax)
      {
        graph->Fit(func.get(),fitopt);
        std::clog << LOG(6) << "RootFunctionFitter::Fit After graph->Fit(R)" << std::endl;
      }
      else 
      {
        graph->Fit(func.get(), fitopt);
        std::clog << LOG(6) << "RootFunctionFitter::Fit After graph->Fit" << std::endl;
      }
      minpar = newpar;
      for(unsigned i=0;i<newpar.Params().size();i++)
      {
        minpar.SetValue(i,func->GetParameter(i));
        minpar.SetError(i,func->GetParError(i));
      }
      std::clog << LOG(6) << "RootFunctionFitter::Fit After setting minpar" << std::endl;
      std::clog << LOG(5) << "RootFunctionFitter::Fit END" << std::endl;
    }

    /*  Ищет скачок, предполагая, что время обозначено параметром T */
    ROOT::Minuit2::MnUserParameters FindJump(void)
    {
      std::clog << LOG(5) << "RootFunctionFitter::FindJump BEGIN" << std::endl;
      using namespace ROOT::Minuit2;
      //Заполним список минимумов
      std::list<FunctionMinimum> fmlist;
      for(int i=0; i<graph->GetN();i++)
      {
        double t =  graph->GetX()[i]; //возьмем время в определённой точке
        if((tmin < tmax && tmin < t  && t< tmax) || tmin==tmax)
        {
          MnMigrad migrad(*this, inipar);
          migrad.SetValue("T",t);
          fmlist.push_back(migrad());
        }
      }
      //в этом списке найдем самый минимальный
      //ROOT::Minuit2::MnUserParameters par;
      return min_element (
          begin(fmlist), end(fmlist), 
          [](const FunctionMinimum & f1, const FunctionMinimum & f2) 
          { 
          return f1.Fval() < f2.Fval(); 
          }
          )->UserParameters();
      //clog << LOG(5) << "RootFunctionFitter::FindJump END" << endl;
    }

    double operator() (const std::vector<double> & par) const
    {
      double chi2=0;
      for(int i=0; i<graph->GetN();i++)
      {
        double t =  graph->GetX()[i]; //возьмем время в определённой точке
        if((tmin < tmax && tmin < t  && t< tmax) || tmin==tmax)
        {
          double y =  graph->GetY()[i];
          double dy = graph->GetErrorY(i);
          double yexp = func->EvalPar(&t,&par[0]);
          chi2+= pow((yexp - y)/dy,2.0);
        }
      }
      return chi2;
    }
    double Up() const { return 1.; }
  };
}
#endif
