// =====================================================================================
//
//       Filename:  SimpleAtanFitter.cpp
//
//    Description:  
//
//        Version:  1.0
//        Created:  20.04.2016 19:44:11
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================
#include "SimpleAtanFitter.h"
namespace fit
{
  SimpleAtanFitter::SimpleAtanFitter(void)
  {
    type = MINUIT2;
    name = "SIMPLE_ATAN";
    title = "Fit atan using new approximation";
    defpar.Add("T",        0, 100); //depolarization moment
    defpar.Add("DELTA",    0, 0.1); //depolarization effect
    defpar.Add("TAU",      50,   50); //depolarization time
    defpar.Add("C",        0,   1); //delta level before depolarization
    defpar.Add("K1",       0,   1); //slope before depolarization
    defpar.Add("K2",       0,   1); //slope after depolarization
    inipar=defpar;
    TIME=0;
    DELTA=0;
  }
  SimpleAtanFitter::~SimpleAtanFitter(void)
  {
  }

  void SimpleAtanFitter::Fit(void)
  {
    data::EnergyCalibrationStorage * EC = dynamic_cast<data::EnergyCalibrationStorage*>(fStorage);
    auto theData = fView->GetData(fDataName);
    if(theData == nullptr) 
    {
      std::clog << LOG(NORMAL_LOG) << "Unable to find data " << fDataName << std::endl;
      return;
    }
    TIME = &EC->t;
    TC = EC->cfg.count_time;
    DELTA  = theData->data;
    fBunch = theData->bunch;
    using namespace ROOT::Minuit2;
    MnUserParameters newpar=inipar;
    if(!inipar.Parameter(newpar.Index("T")).IsFixed()) newpar = FindJump();
    MnMigrad migrad(*this, newpar);
    FunctionMinimum minimum = migrad();
    minpar=minimum.UserParameters();
    SetMinPar();
  }

  void SimpleAtanFitter::SetParameters(const std::vector<double> & par) const
  {
    TD = par[0];
    D = par[1];
    TAU = par[2];
    C = par[3];
    K1 = par[4];
    K2 = par[5];
  }

  double SimpleAtanFitter::heaviside(double x) 
  {
    return atan(x)*M_1_PI + 0.5;
  };

  double SimpleAtanFitter::int1_heaviside(double x)
  {
    return x*heaviside(x) - 0.5*M_1_PI*log(x*x+1.);
  }

  double SimpleAtanFitter::int2_heaviside(double x)
  {
    return 0.5*(x*x-1.0)*heaviside(x) + 0.25 + 0.5*M_1_PI*x*(1.0 - log(x*x+1.0));
  }

  double SimpleAtanFitter::delta(double t) const 
  {
    double x = (t-TD)/TAU;
    double x0 = -TD/TAU;
    double k1 = K1*TAU;
    double k2 = K2*TAU;
    return C + D*heaviside(x) + k1*x + (k2-k1)*(int1_heaviside(x) - int1_heaviside(x0));
  }


  double SimpleAtanFitter::delta_average(double t, double T) const 
  {
    double x = (t-TD)/TAU;
    double x0 = -TD/TAU;
    double k1 = K1*TAU;
    double k2 = K2*TAU;
    double dx = T/TAU;
    return C - (k2-k1)*int1_heaviside(x0)  + 
      D*(int1_heaviside(x)-int1_heaviside(x-dx))/dx +
      0.5*k1*(x*x - (x-dx)*(x-dx))/dx +
      (k2-k1)*(int2_heaviside(x) - int2_heaviside(x-dx))/dx;
  }

  double SimpleAtanFitter::delta_average(double *X, double *P)
  {
    double t = X[0];
    double TD=P[0];
    double D=P[1];
    double TAU = P[2];
    double C = P[3];
    double K1 = P[4];
    double K2 = P[5];
    double T = P[6];
    double x = (t-TD)/TAU;
    double x0 = -TD/TAU;
    double k1 = K1*TAU;
    double k2 = K2*TAU;
    double dx = T/TAU;
    return C - (k2-k1)*int1_heaviside(x0)  + 
      D*(int1_heaviside(x)-int1_heaviside(x-dx))/dx +
      0.5*k1*(x*x - (x-dx)*(x-dx))/dx +
      (k2-k1)*(int2_heaviside(x) - int2_heaviside(x-dx))/dx;
  }

  double SimpleAtanFitter::operator() (const std::vector<double> & par) const
  {
    SetParameters(par);
    double chi2=0;
    for(unsigned i=0;i<TIME->size();i++)
    {
      double t = (*TIME)[i].value;
      double D = (*DELTA)[i].value;
      double dD = (*DELTA)[i].error;
      chi2+=pow((delta_average(t,TC) - D)/dD,2.0);
    }
    return chi2;
  } 

  double SimpleAtanFitter::Up() const { return 1.; }

  void SimpleAtanFitter::SetMinPar(void)
  {
    SetParameters(minpar.Params());
  }

  void SimpleAtanFitter::DrawFunction(void)
  {
    root_function.reset(new TF1("SimpleAtan", &SimpleAtanFitter::delta_average,0,1e5,7));
    std::vector<double> par = minpar.Params();
    for(int i=0;i<6;i++)
    {
      root_function->SetParameter(i, par[i]);
    }
    root_function->SetParameter(6,TC);
    for(int i=0;i<7;i++)
    {
      std::cout << "p"<<i << "=" << root_function->GetParameter(i) << std::endl;
    }
    root_function->Draw("same");
  }

  /*  Ищет скачок, предполагая, что время обозначено параметром T */
  ROOT::Minuit2::MnUserParameters SimpleAtanFitter::FindJump(void)
  {
    using namespace ROOT::Minuit2;
    //Заполним список минимумов
    std::list<FunctionMinimum> fmlist;
    for(unsigned i=0; i<TIME->size();i++)
    {
      double t =  (*TIME)[i].value; //возьмем время в определённой точке
      std::cout << i << " " << t << " " << std::endl;
      MnMigrad migrad(*this, inipar);
      migrad.SetValue("T",t);
      FunctionMinimum fm = migrad();
      fmlist.push_back(fm);
      std::cout << i << " " << t << " " << std::endl;
    }
    //в этом списке найдем самый минимальный
    return min_element 
      (
       begin(fmlist), end(fmlist), 
       [](const FunctionMinimum & f1, const FunctionMinimum & f2) 
       { 
       return f1.Fval() < f2.Fval(); 
       }
      )->UserParameters();
  }
}
