/*
 * =====================================================================================
 *
 *       Filename:  hev.h
 *
 *    Description:  Implementation of heaviside functions.
 *
 *        Version:  1.0
 *        Created:  12/04/2009 05:17:01 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */

#ifndef IBN_FIT_HEV_H
#define IBN_FIT_HEV_H

#include <cmath>


/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  heaviside
 *  Description:  Heaviside function
 * =====================================================================================
 */
inline double heaviside(double x)
{
  if(x<0) return 0;
  if(x==0) return 0.5;
  return 1;
}


/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  int_heaviside
 *  Description:  Integral of heaviside function
 * =====================================================================================
 */
inline double int_heaviside(double x)
{
  return x*heaviside(x);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  int2_heaviside
 *  Description:  Second integral of heaviside function
 * =====================================================================================
 */
inline double int2_heaviside(double x)
{
  return 0.5*x*x*heaviside(x);
}



/*-----------------------------------------------------------------------------
 *  Step function based on erf function.
 *-----------------------------------------------------------------------------*/

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  heaviside_erf
 *  Description:  Heaviside function
 * =====================================================================================
 */
inline double heaviside_erf(double x)
{
  return 0.5*(erf(x) + 1.0);
}


/* ===  FUNCTION  ======================================================================
 *         Name:  int_heaviside_erf
 *  Description:  Integral of heaviside function
 * =====================================================================================
 */
inline double int_heaviside_erf(double x)
{
  return x*heaviside_erf(x) - 0.25*M_2_SQRTPI*(1.0 - exp(-x*x));
}

/* ===  FUNCTION  ======================================================================
 *         Name:  int2_heaviside_erf
 *  Description:  Second Integral of heaviside function
 * =====================================================================================
 */
inline double int2_heaviside_erf(double x)
{
  return 0.25*(2*x*x+1.0)*heaviside_erf(x)-0.125 - x*0.125*M_2_SQRTPI*(2.0 - exp(-x*x));
}

/*-----------------------------------------------------------------------------
 *  Step function based on atan function.
 *-----------------------------------------------------------------------------*/

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  heaviside_atan
 *  Description:  Heaviside function
 * =====================================================================================
 */
inline double heaviside_atan(double x)
{
  return 0.5*(M_2_PI*atan(x) + 1.);
}


/* ===  FUNCTION  ======================================================================
 *         Name:  int_heaviside_atan
 *  Description:  Integral of heaviside function
 * =====================================================================================
 */
inline double int_heaviside_atan(double x)
{
  return x*heaviside_atan(x) - 0.5*M_1_PI*log(x*x+1.);
}

/* ===  FUNCTION  ======================================================================
 *         Name:  int2_heaviside_atan
 *  Description:  Second Integral of heaviside function
 * =====================================================================================
 */
inline double int2_heaviside_atan(double x)
{
  return 0.5*(x*x-1.0)*heaviside_atan(x)+x*(1.0 - log(x*x+1.0))*0.5*M_1_PI;
}

#endif //IBN_FIT_HEV_H
