// =====================================================================================
//
//       Filename:  BunchViewFrame.cpp
//
//    Description:  View load status
//
//        Version:  2.0
//        Created:  17.02.2016 12:01:45
//       Revision:  none
//       Compiler:  g++
//
//         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
//   Organization:  Budker Institute of Nuclear Physics
//
// =====================================================================================


#include <TGFrame.h>
#include <TGTextEntry.h>
#include <TGTableLayout.h>
#include <TGButton.h>
#include <TGLabel.h>

#include "EnergyCalibrationView.h"
#include "BunchViewFrame.h"
#include "data/PolarimeterStatus.h"

namespace gui
{
  BunchViewFrame::BunchViewFrame( const TGWindow *w, const std::vector<data::BunchInfo_t> & b, EnergyCalibrationView * ecv) 
    : TGGroupFrame(w,"Counter load information", kHorizontalFrame ) , BI(b),ECView(ecv)
  {
		int title_size = 1;
		auto layout = new TGTableLayout(this,18,BI.size()+title_size, kFALSE,0,kLHintsNormal);
    SetLayoutManager(layout);

		char default_str[128];
		sprintf(default_str, format_str.c_str(), 0.0);

		auto title_layout_opts=kLHintsExpandX;
		auto layout_opts=kLHintsExpandX | kLHintsFillX;


    std::map<int,std::string> MRow;
    BB.resize(BI.size());
    int item_index=0;
    for(unsigned b=0;b<BI.size();b++)
    {
      int row=0;
      //add the bunch numbers
      MRow[row]="bunch";
      auto bunch_number_lbl = new TGLabel(this, ((b<2 ? "e" : "p" )+std::to_string(b%2+1)).c_str());
      AddFrame(bunch_number_lbl, new TGTableLayoutHints(title_size+b, title_size+b+1, row, row+1, layout_opts) );
      row++;
      auto init = [&](Item_t & item,std::string title, std::string name, std::string units)
      {
        item.title=title;
        item.name=name+"-"+std::to_string(b+1);
        item.units=units;
        auto & tb = item.text_button;
        tb.reset(new TGTextButton(this, default_str, item_index));
        tb->SetTextJustify(kTextRight);
        AddFrame(tb.get(), new TGTableLayoutHints(title_size+b, title_size+b+1, row, row+1, layout_opts) );
        MRow[row]=title;
        row++;
        fBMap[item_index] = &item;
        item_index++;
      };
      //add bunch current
      init(BB[b].I,"I","current","mA");
      //init counter load
      BB[b].Cnt.resize(BI[0].counter.size());
      for(unsigned c=0;c<BB[b].Cnt.size();c++) init(BB[b].Cnt[c],"counter-"+std::to_string(c+1), "counter-"+std::to_string(c+1),"kHz");
      //init counter sum
      init(BB[b].CntSum,"counter-sum","counter_sum","kHz");
      //init delta in counter sum
      init(BB[b].DeltaCntSum,"delta","delta","%");
      //init coincidence load
      BB[b].Coinc.resize(BI[0].cncr.size());
      for(unsigned pair=0; pair< BB[b].Coinc.size();pair++) init(BB[b].Coinc[pair],"coinc-"+std::to_string(pair+1),
          "coincidence-"+std::to_string(pair+1),"kHz");
      //init coincidence sum
      init(BB[b].CoincSum,"coincidence-sum","coincidence_sum","kHz");
      //init delta in coincidence sum
      init(BB[b].DeltaCoincSum,"coincidence-delta","delta_coincidence_sum","%");
    }
    //now add titles
    for(int row=0;row<MRow.size();row++)
    {
      auto lbl = new TGLabel(this,MRow[row].c_str());
      AddFrame(lbl,  new TGTableLayoutHints(0, title_size, row, row+1, title_layout_opts) );
    }
    Layout();
    MapSubwindows();
    MapWindow();
  }

  BunchViewFrame::~BunchViewFrame(void) 
  {

  }


  void BunchViewFrame::Update(void)    
  {
    unsigned Nb = BB.size();
    unsigned Nc = BB[0].Cnt.size();
    unsigned Np = BB[0].Coinc.size();
    for(unsigned b=0; b<Nb; b++)
    {
      auto & B = BB[b];
      auto set_text = [&](Item_t & item, double data, double dmin=-1e10, double dmax=+1e10, double good_min=0, double good_max=0)
      {
        auto & tb = item.text_button;
        char str[128];
        sprintf(str,format_str.c_str(),data);
        tb->SetText(str);
        if(data > dmax || data < dmin) tb->SetTextColor(0xff0000); //set color red
        else 
        {
          if( data < good_max && data > good_min) tb->SetTextColor(0x228b22);
          else tb->SetTextColor(0x000000);
        }
      };
      set_text(B.I,BI[b].I*1e-3);
      double MaxLoad = 500; //kHz
      double MinLoad = 150; //kHz
      for(unsigned c=0;c<Nc; c++) 
      {
        if(BI[b].VviCounterStatus[c]  == 1) //if vvi is ON
        {
          set_text(B.Cnt[c],  BI[b].counter[c]*1e-3, 0, MaxLoad, MinLoad, MaxLoad);
        }
        else
        {
          B.Cnt[c].text_button->SetText("OFF");
        }
      }
      for(unsigned p=0;p<Np; p++) set_text(B.Coinc[p],   BI[b].cncr[p]*1e-3, 0, MaxLoad, MinLoad, MaxLoad);
      set_text(B.CntSum,BI[b].counter_sum*1e-3,0, MaxLoad*4, MinLoad*4,MaxLoad*4);
      set_text(B.CoincSum,BI[b].cncr_sum*1e-3, 0, MaxLoad*4, MinLoad*4,MinLoad*4);
      set_text(B.DeltaCntSum,BI[b].delta*100,-3, +3, 1, 3);
      set_text(B.DeltaCoincSum,BI[b].delta_cncr*100,-3,3,1,3);
    }
  }

  Bool_t BunchViewFrame::ProcessMessage(Long_t msg, Long_t parm1, Long_t parm2 )
  {
    switch( GET_MSG(msg) )
    {
      case kC_COMMAND:
        switch(GET_SUBMSG(msg))
        {
          case kCM_BUTTON:
            {
              auto & item = fBMap[parm1];
              if(ECView != nullptr) 
              {
                ECView->Add(item->name);
                ECView->Draw();
              }

            }
            break;
          case kCM_MENU:
            break;
        }
        break;
    }
    return true;
  }
}
