#include <cstdio>

#include <fstream>
#include <list>
#include <mutex>
#include <thread>

#include <boost/format.hpp>

#include "V4Par.h"
#include "vepp4/Stap.h"


#include "Canvas.h"
#include "Utils.h"

#include <TROOT.h>
#include <TSystem.h>
#include <TQObject.h>
#include <TGraphErrors.h>

#include <TStyle.h>
#include <TROOT.h>
#include <TCanvas.h>
#include <TGraphErrors.h>
#include <TH1F.h>
#include <TAxis.h>
#include <TObject.h>
#include <TRootCanvas.h>

#include "gui/EnergyCalibrationView.h"
extern std::recursive_mutex  RootProcessEventsMutex;

extern std::list <unsigned> BadPointList;
extern PolarimeterConfig PCFG;

using namespace std;

namespace gui
{
  const double XTIME = 0.5;
  const double YFIRST = 0.95;
  const double YSECOND = 0.91;
  const double XSECOND = 0.83;

  CanvasData::MeasurementParameters::MeasurementParameters(void)
  {
    x1=0.901;
    y1=0.7;
    x2=1.0;
    y2=0.9;

    pave = new TPaveText(x1,y1,x2,y2);
    pave->SetTextFont(102);
    pave->SetTextAlign(12);
    pave->SetBorderSize(1);

    pave->SetFillColor(0);
    Tcount = pave->AddText("TCOUNT");
    speed=pave->AddText("SPEED");
    step=pave->AddText("STEP");
    initial=pave->AddText("INITIAL");
    final=pave->AddText("FINAL");
    attenuation = pave->AddText("ATT");
		voltage = pave->AddText("2U");
    pave->SetLineColor(kWhite);

    Date=new TText(XTIME, YFIRST,"DATE");
    Run=new TText(XTIME+0.2, YSECOND,"RUN");
    Mode=new TText(XSECOND, YFIRST,"MODE");
    Es=new TText(XSECOND,YSECOND,"ESET");
		Shiftman = new TText(0.15, YSECOND, "SHIFTMAN");
		
		measurement_name = new TText(0.35, YSECOND, "");
		experiment_name = new TText(0.35, YFIRST,   "");

    ScanDir.reset(new TText(0.93,YSECOND,"TEST"));
    Reason.reset(new  TText(0.93,YFIRST,"NONE"));

    Date->SetNDC();
    Run->SetNDC();
    Mode->SetNDC();
    Es->SetNDC();
		Shiftman->SetNDC();
    ScanDir->SetNDC();
		measurement_name->SetNDC();
		experiment_name->SetNDC();
    Reason->SetNDC();

    SetTextSize(0.035);
  }

  CanvasData::MeasurementParameters::~MeasurementParameters(void)	
  {
    delete pave; //do not delete the TText which inside the pave
    delete Date;
    delete Run;
    delete Mode;
    delete Es;
		delete Shiftman;
		delete experiment_name;
		delete measurement_name;
  }

  void CanvasData::MeasurementParameters::Draw(TCanvas * c)	
  {
    c->cd();
    pave->Draw();
    Date->Draw();
    Run->Draw();
    Mode->Draw();
    Es->Draw();
		Shiftman->Draw();
    ScanDir->Draw();
		experiment_name->Draw();
		measurement_name->Draw();
    Reason->Draw();
    c->Update();
  }

  void set_text(TText *t, const char * buf)
  {
    t->SetText(t->GetX(), t->GetY(), buf);
  }

  void CanvasData::MeasurementParameters::Update(const data::EnergyCalibrationStorage * EC)
  {
    clog << LOG(5) << "MeasurementParameters::Update BEGIN" << endl;

    measurement::EnergyCalibrationConfig cfg=EC->cfg;
    measurement::DepolarizerConfig dcfg=cfg.depolarizer;

		set_text(measurement_name, EC->cfg.name.c_str());
		set_text(experiment_name, EC->cfg.experiment_name.c_str());

		clog << LOG(5) << "SHIFTMAN = " << EC->shiftman << std::endl;
    set_text(Shiftman, (boost::format("%-15s") % EC->shiftman.substr(0, 14)).str().c_str());
    
    //show vepp4 status
    set_text(Mode,EC->mode.c_str());
    set_text(Es, (boost::format("%-7.2f") % EC->setE).str().c_str());

    //show energy calibration run
    set_text(Run,(boost::format("Run %5d") % EC->cfg.run).str().c_str());

    //show begin time of energy calibration
    char time_string[1024];
    strftime(time_string,1024,"%F %T",localtime(&EC->begin_time.tv_sec));
    set_text(Date,time_string);

    //show scan parameters
    set_text(Tcount,      (boost::format("%-1s %9.1f %-3s") % "T"       % EC->cfg.count_time % "sec").str().c_str());
    set_text(speed,       (boost::format("%-5s %5.3f %-3s") % "SPEED"   % (dcfg.speed*1000)  % "#frac{keV}{sec}").str().c_str());
    set_text(step,        (boost::format("%-4s %6.3f %-3s") % "STEP"    % (dcfg.step*1000)   % "keV").str().c_str());
    set_text(initial,     (boost::format("%-2s %8.3f %-3s") % "E1"      % (dcfg.initial)     % "MeV").str().c_str());
    set_text(final,       (boost::format("%-2s %8.3f %-3s") % "E2"      % (dcfg.final)       % "MeV").str().c_str());
    set_text(attenuation, (boost::format("%-3s %7d %-3s")   % "ATT"     % (dcfg.attenuation)       % "dB").str().c_str());
    set_text(voltage,     (boost::format("%-3s %7.0f %-3s")   % "2U"      % (dcfg.voltage*1e3)       % "mV").str().c_str());

    //show direction information
    if(dcfg.initial < dcfg.final) set_text(ScanDir.get(),"UP");
    else set_text(ScanDir.get(),"DOWN");

    switch (cfg.reason)
    {
      case 0:
        set_text(Reason.get(),"");
        break;
      case 1:
        set_text(Reason.get(),"REGL");
        break;
      case 2:
        set_text(Reason.get(),"MODE");
        break;
      case 3:
        set_text(Reason.get(),"CYCL");
        break;
      case 100:
        set_text(Reason.get(),"OTHR");
        break;
    }

    clog << LOG(5) << "MeasurementParameters::Update END" << endl;
  }

  void CanvasData::MeasurementParameters::SetTextSize(double textsize) 
  {
    Mode->SetTextSize(textsize);
    Es->SetTextSize(textsize);
    Run->SetTextSize(textsize);
    Date->SetTextSize(textsize);
    Shiftman->SetTextSize(textsize);
    ScanDir->SetTextSize(textsize);
    measurement_name->SetTextSize(textsize);
    experiment_name->SetTextSize(textsize);
    Reason->SetTextSize(textsize);
  }


  void CanvasData::MeasurementParameters::RenewPaveLocation(void)
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);
  }

  void CanvasData::MeasurementParameters::RenewDirectionLocation(TCanvas * c)
  {
    //THis code draw new arrow or line every time.
    //double x=0.9;
    //double y=0.9;
    //double y2=1.0;
    //double R=0.05;
    //int linewidth=3;
    //int px  = c->UtoPixel(x);
    //int py  = c->UtoPixel(y);
    //int py2 = c->UtoPixel(y2);
    //int pR =  c->UtoPixel(R);
    //double ux,uy,uy2,uR;
    //ux = c->PixeltoX(px);
    //uy = c->PixeltoX(py);
    //uy2 = c->PixeltoX(py2);
    //uR = (y2-y1)/2.;
    ////arrow->DrawArrow(ux,uy,ux,uy2);
    //arrow->DrawLineNDC(x,y,x,y2);
  }




  CanvasData::MeasurementInfo::MeasurementInfo(void)
	{

    x1=0.901;
    y1=0.1;
    x2=1.0;
    y2=0.3;


    pave = new TPaveText(x1,y1,x2,y2,"");
    pave->SetFillColor(0);
    pave->SetLineColor(0);
    pave->SetTextFont(102);
    pave->SetTextAlign(12);
    pave->AddText("  LOAD, kHz");
    pave->SetBorderSize(1);

    Counter.resize(PCFG.counter_number/2);
    for(unsigned i=0 ; i<PCFG.counter_number/2 ; i++)
    {
      Counter[i] = pave->AddText("");
    }	
  }

  CanvasData::MeasurementInfo::~MeasurementInfo(void)
	{
    delete pave;
  }

  void CanvasData::MeasurementInfo::Draw(TCanvas * c)  
  {
    c->cd();
    pave->Draw();
  }

  void CanvasData::MeasurementInfo::Update(const data::EnergyCalibrationStorage * EC)
	{
    double count1;
    double count2;
		if(EC->t.size()>0)
		{
			for(unsigned i=0 ; i< Counter.size() ; i++)
			{
				count1 = fabs((EC->Count[0][2*i][0].value)*1e-3);
				count2 = fabs((EC->Count[0][2*i+1][0].value)*1e-3);
				Counter[i]->SetText(Counter[i]->GetX(), Counter[i]->GetY(),(boost::format("C_{%d,%d} %1.1f/%1.1f") % (2*i+1) % (2*i+2)% count1 % count2 ).str().c_str());
			}	
		}
  }

  void CanvasData::MeasurementInfo::RenewPaveLocation(void)
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);
  }


  CanvasData::NMR::NMR(void)   
  {
    x1=0.901;
    y1=0.6;
    x2=1.0;
    y2=0.7;
    pave = new TPaveText(x1,y1,x2,y2);
    pave->SetTextAlign(12);
    pave->SetFillColor(0);
    pave->SetLineColor(0);
    pave->SetBorderSize(1);
    pave->SetTextFont(102);
    pave->SetTextAlign(12);
    pave->AddText("   NMR, Gs");
    nmr["H"] = pave->AddText("H");
    nmr["NEM"] = pave->AddText("NEM");
    nmr["SEM"] = pave->AddText("SEM");
    //nmr["SIM"] = pave->AddText("SIM");
    //nmr["EM3"] = pave->AddText("EM3");
    //nmr["KEDR1"] = pave->AddText("KEDR1");
    //nmr["KEDR2"] = pave->AddText("KEDR2");
  }

  CanvasData::NMR::~NMR(void) 
  {
    delete pave;
  }

  void CanvasData::NMR::Draw(TCanvas * c)  
  {
    c->cd();
    pave->Draw();
  }

  void CanvasData::NMR::Update(const data::EnergyCalibrationStorage * EC)
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);

    map < std::string, unsigned > idx_map;
    idx_map["H"]=0;
    idx_map["NEM"]=1;
    idx_map["SEM"]=2;
    //idx_map["SIM"]=3;
    //idx_map["EM3"]=4;
    //idx_map["KEDR1"]=5;
    //idx_map["KEDR2"]=6;
    clog << LOG(5) << "NMR::Update: nmr.size()="<< nmr.size() << " EC.nmr.size() = " << EC->nmr.size() << "\n";
    for( auto & p : nmr)
    {
      TText * t = p.second;
      t->SetText(t->GetX(), t->GetY(),(boost::format("%-5s %8.3f") % p.first.c_str() % EC->nmr[idx_map[p.first]]).str().c_str());
    }
  }

  void CanvasData::NMR::RenewPaveLocation(void)
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);
  }

  CanvasData::VeppInfo::VeppInfo(void)
  {
    x1=0.901;
    y1=0.3;
    x2=1.0;
    y2=0.6;
    pave = new TPaveText(x1,y1,x2,y2);
    pave->SetTextAlign(12);
    pave->SetTextFont(102);
    pave->SetFillColor(0);
    pave->SetLineColor(0);
    pave->SetBorderSize(1);
    pave->AddText("  VEPP4 INFO");
  }
  CanvasData::VeppInfo::~VeppInfo(void)	
  {
    delete pave;
  }

  void CanvasData::VeppInfo::Draw(TCanvas *c) 
  {
    c->cd();
    pave->Draw();
  }

  void CanvasData::VeppInfo::Update(const data::EnergyCalibrationStorage *EC)	
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);

    V4Current v4cur;
		vepp4::Stap stap;

		auto fill_text  = [&](std::string name,  boost::format & fmt)
		{
			TText * ttext;
			auto it  = fText.find(name);
			if(it == fText.end())
			{
				ttext = pave->AddText(name.c_str());
			}
			else
			{
				ttext = it->second;
			}
			ttext->SetText(ttext->GetX(), ttext->GetY(),fmt.str().c_str());
		};

    std::cout << "Is fit = " << EC->isfit << std::endl;

		fill_text("RF", boost::format("%-2s %9.0f Hz") % "F_{RF}" % stap.RF);
		fill_text("U", boost::format("%-3s %8.0f kV ") % "U_{RF}" % v4cur.URF);
		fill_text("J", boost::format("%-3s %8.0f A ") % "J_{RF}" % v4cur.JRF);
		fill_text("SVED", boost::format("%-4s %s") % "SVED" % (stap.sved == 1? "ON" : "OFF"));
		fill_text("Tz",  boost::format("%-2s %9.4f   ") % "T_{z}" % v4cur.TZ());
		fill_text("Qx", boost::format("%-2s %9.4f") % "Qx"    % stap.QX);
		fill_text("Qz", boost::format("%-2s %9.4f") % "Qz"    % stap.QZ);
    if(!EC->isfit)
    {
      fill_text("I^{-}", boost::format("%-2s %4.2f/%4.2f mA") % "I^{-}" % (stap.Ie1*1e-3) % (stap.Ie2*1e-3));
      if(stap.Ip1<10)  stap.Ip1=0;
      if(stap.Ip2<10)  stap.Ip2=0;
      fill_text("I^{+}", boost::format("%-2s %4.2f/%4.2f mA") % "I^{+}" % (stap.Ip1*1e-3) % (stap.Ip2*1e-3));
    }
    else 
    {
      //find jump index
      double jump_time = EC->fit_param[0].value;
      //double jump_index =0;
      //for(int i=0;i<EC->t.size()-1;i++)
      //{
      //  if(t[i].value <= jump_time && jump_time < t[i+1].value) 
      //  {
      //    jump_index = i;
      //    break;
      //  }
      //}
      std::cout << "jump_time = " << jump_time<< std::endl;
      std::vector<double> I(EC->I.size());
      for(int i=0;i<I.size();i++)
      {
        try
        {
          I[i] = EC->GetValueByTime(jump_time, EC->I[i]);
        }
        catch(...)
        {
          I[i] = EC->I[i][0].value;
        }
      }
      fill_text("I^{-}", boost::format("%-2s %4.2f/%4.2f mA") % "I^{-}" % (I[0]*1e-3) % (I[1]*1e-3));
      fill_text("I^{+}", boost::format("%-2s %4.2f/%4.2f mA") % "I^{+}" % (I[2]*1e-3) % (I[3]*1e-3));
    }
		fill_text("v4Dout", boost::format("%-6s %5.2f C ") % "v4Dout" % V4TermoNew()["v4Dout"] );


		//if(v3p.ready())
		//{
		//	set_text(Q1_vepp3  , (boost::format("%-2s %9.4f ") % "Q_{1}" % v3p.Q1).str().c_str());
		//	set_text(Q2_vepp3  , (boost::format("%-2s %9.4f ") % "Q_{2}" % v3p.Q2).str().c_str());
		//	set_text(polarization_time, (boost::format("%-2s %9.4f ") % "#tau_{p}" % v3p.poltime).str().c_str());
		//	//Q2_vepp3->SetText((boost::format("%-2s %9.4f ") % "Q_{2}" % v3p.Q2).str().c_str());
		//	//polarization_time->SetText(.c_str());
		//}
  }

  void CanvasData::VeppInfo::RenewPaveLocation(void)
  {
    pave->SetX1NDC(x1);
    pave->SetX2NDC(x2);
    pave->SetY1NDC(y1);
    pave->SetY2NDC(y2);
  }

  Canvas::Canvas(Bool_t build )
    : TCanvas(build) 
  {
    Init();
  }

  Canvas::Canvas(const char* name, const char* title, Int_t form )
    :	TCanvas(name,title, form) 
  {
    Init();  
  }

  Canvas::Canvas(const char* name, const char* title, Int_t ww, Int_t wh)
    :	TCanvas(name,title,ww,wh) 
  { 
    Init(); 
  }

  Canvas::Canvas(const char* name, const char* title, Int_t wtopx, Int_t wtopy, Int_t ww, Int_t wh)
    :	TCanvas(name,title, wtopx, wtopy, ww, wh) 
  {
    Init(); 
  }


  Canvas::Canvas(const char* name, Int_t ww, Int_t wh, Int_t winid)
    :	TCanvas(name,  ww,  wh,  winid) 
  {
    Init();
  }


  Canvas::Canvas(data::EnergyCalibrationStorage *ec) 
  {
    clog << LOG(5) << "Canvas::Canvas(data::EnergyCalibrationStorage *ec) BEGIN" << endl;
    EC=ec;
    X=0;
    depFr.reset(new TLatex(0.6, 0.005, 0));
    MeEn.reset(new TLatex(0.1,0.01,0));
    depFr->SetTextSize(0.035);
    MeEn->SetTextSize(0.035);
    draw_notes=true;
    SetFillColor(0);
    clog << LOG(6) << "Canvas::UPdate before MeasPrm \n";
    MeasPrm.Update(EC);
    clog << LOG(6) << "Canvas::UPdate before MeasInfo\n";
    MeasInfo.Update(EC);
    clog << LOG(6) << "Canvas::UPdate before nmr.update\n";
    nmr.Update(EC);
    clog << LOG(6) << "Canvas::UPdate before vinfo update\n";
    vinfo.Update(EC);
    clog << LOG(5) << "Canvas::Canvas(data::EnergyCalibrationStorage *ec) END" << endl;
  }

  Canvas::Canvas(data::EnergyCalibrationStorage * ec ,  Data * x, string du_name): Canvas(ec)
  { 
    X=x;
    name = du_name;
    SetName(name.c_str());
    SetTitle(name.c_str());
  }

  Canvas::Canvas(data::EnergyCalibrationStorage * ec ,  data::Array * x, string nm) : Canvas(ec,new Data(*x,nm, nm),nm)
  { 
  }


  Canvas::Canvas(data::EnergyCalibrationStorage * ec ,  Data * x, string du_name, int ww, int wh): 
    Canvas(ec, x,du_name)
  { 
  }

  ////=====================================================================
  Canvas::Canvas(data::EnergyCalibrationStorage *ec , data::Array * x, string nm, int ww, int wh) :
    Canvas(ec, new Data(x,nm), nm, ww,wh)
  {
  }

  void monitorClose()
  {
    std::clog << LOG(DEBUG_LOG) << "################## CANVAS CLOSED #################" << endl;
  }

  void Canvas::Init(void)
  {
    SetGridx();
    SetGridy();
    SetBorderMode(0);
    EC=0;
    TRootCanvas *rc = (TRootCanvas*)GetCanvasImp();
    rc->Disconnect("CloseWindow()");
    rc->Disconnect("Close()");
  }





  Canvas::~Canvas(void)
  { 
    theEnergyCalibrationView->Erase(name);
  }

  void Canvas::ProcessedEvent(Int_t event,Int_t x,Int_t y ,TObject* selected)
  {
    signalProcessedEvent(event,x,y,selected); /*  send signal for everybody who interested */
    TCanvas::ProcessedEvent(event, x,y,selected);
  }


  void Canvas::DrawParameters(void)
  {
    MeasInfo.Draw(this);
    MeasPrm.Draw(this);
    nmr.Draw(this);
    vinfo.Draw(this);
  }

  void Canvas::RenewPaveLocation(void)
  {
    clog << LOG(5) << "Canvas::RenewPaveLocation" << endl;
    MeasPrm.RenewPaveLocation();
    MeasInfo.RenewPaveLocation();
    nmr.RenewPaveLocation();
    vinfo.RenewPaveLocation();
  }

  //=====================================================================
  Graph * Canvas::GetGraph(std::string name)	
  {
    Itr_t p = graphic.find(name);
    if(p!=graphic.end())
    {
      return graphic.find(name)->second.get(); 
    }
    else
    {
      return nullptr;
    }
  }
  //=====================================================================

  Graph * Canvas::AddGraph(std::shared_ptr<Graph> graph)
  {
    auto g = graphic.find(graph->GetName());
    if(g==graphic.end())
    {
      //Graph * graph = new Graph(X,y);
      graphic[graph->GetName()] = graph;
      Update();
      return graph.get();
    }
    return g->second.get();
  }

  Graph * Canvas::AddGraph(Data *y)
  {
    auto g = graphic.find(y->name);
    if(g==graphic.end())
    {
			std::shared_ptr<Graph> graph(new Graph(X,y));
      graphic[y->name] = graph;
      graph->SetMarkerStyle(23+graphic.size());
      Update();
      return graph.get();
    }
    return g->second.get();
  }

  Graph *  Canvas::AddGraph(std::string nm, data::Array *y)
  {
    return AddGraph(new Data(y, nm));
  }


  void Canvas::Update(void)	
  {
		std::clog << LOG(5) << "Canvas::Update thread=" << hex << std::this_thread::get_id() << dec << std::endl;
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    for(graphic_t::iterator p=graphic.begin(); p!=graphic.end(); ++p)
    {
      p->second->Update();
    }
    cd();
    unsigned old_n = note.GetSize();
    Graph * gr=graphic.begin()->second.get();
    //Поменяем масштаб графика, чтобы подписи частоты не перекрывали точки графика, для этого
    //от наименьшего значения графика отнимим полторы ошибки (спрятано в функции Graph->GetMinimum)
    //и отнимим часть занимаемую текстом (подпись энергии), приведённое на физический масштаб графика.
		double mmax = gr->GetMaximum(last_number_of_points);
		double mmin = gr->GetMinimum(last_number_of_points);
    double yscale=fabs(mmax-mmin);
    gr->SetMinimum(mmin-note.GetTotalTextSize()*yscale);
    for(unsigned i = old_n; i < (unsigned)gr->GetN(); i++)	
    {
      double tmp=EC->GetEnergy(EC->DepFr[i].value);
      if(EC->DepFr[i].value == 0 || X->show[i]==0) tmp=0;
      note.AddNote(tmp, (*X)[i].value,gr->GetY()[i]);
      note.SetY(gr->GetHistogram()->GetMinimum());
			if(fabs(fabs(EC->DepFr[i].value) - fabs(EC->MDepFr[i].value))>PCFG.measurement.max_allowed_frequency_difference)
			{
        //TEMPORARY SUPRESS RED COLOR
				//note.GetList().back().text.SetTextColor(kRed);
			}
    }
		if(gr->GetN()>last_number_of_points && last_number_of_points>2)
		{
			int size=gr->GetN();
			auto X = gr->GetX();
			double xmin = X[size-last_number_of_points];
			double xmax = X[size-1];
			double dx = (xmax-xmin);
			xmin -= dx*0.1;
			xmax += dx*0.1;
			gr->SetXRangeUser(xmin, xmax);  
			gr->SetYRangeUser(mmin-note.GetTotalTextSize()*yscale, mmax);  
		}
    TCanvas::Modified();
    TCanvas::Update();
    RenewPaveLocation();
    if(EC) MeasPrm.Update(EC);
    if(EC) MeasInfo.Update(EC);
  }

  void Canvas::Clear(void)
  {
    TCanvas::Clear();
    for(auto & p : graphic) p.second->Clear();
  }

  void Canvas::Draw(void)
  {
    clog << LOG(DEBUG_LOG) << "Canvas::Draw" << endl;
    Canvas::cd();
    for(GI g = graphic.begin(); g!=graphic.end(); ++g)
    {
      std::string graph_name = g->first;
      TGraphErrors * gr = g->second.get();
      if(!FindObject(gr->GetName()))
      {
        string draw_opt(g==graphic.begin() ? string("apZ") : string("pZ"));
        gr->Draw(draw_opt.c_str());
        if(gr->GetN()>0)
        {
          gr->GetXaxis()->SetTitle(X->axis_title.c_str());
        }
      }
    }
    if(draw_notes) note.Draw(); //narisuem podpisi dlya graphica
    DrawParameters();
    MeEn->Draw();
    depFr->Draw();
  }
  //=====================================================================



  //=====================================================================
  void Canvas::Draw(std::string name)	
  {
    Itr_t p = graphic.find(name);
    if(p==graphic.end()) return;
    cd();
    if(p==graphic.begin()) p->second->Draw("ap");	
    else p->second->Draw("p");
  }
  //=====================================================================




  ////=====================================================================
  //bool Canvas::IsCanvasExist(void)	
  //{
  //  return gROOT->FindObject(name.c_str());
  //}
  ////=====================================================================

  //=====================================================================
  void Canvas::SetMeasured(void)
  {
    clog << LOG(DEBUG_LOG) << "Canvas::SetMeasured\n";
    boost::format Efmt("E = %11.6f #pm %8.6f MeV");
    Efmt % EC->E.value % EC->E.error;
    MeEn->SetText(MeEn->GetX(),MeEn->GetY(),Efmt.str().c_str());
    MeEn->SetNDC();
    boost::format Ffmt("F = %+11.3f #pm %6.3f Hz");
    Ffmt % EC->Fr.value % EC->Fr.error;
    depFr->SetText(depFr->GetX(),depFr->GetY(),Ffmt.str().c_str());
    depFr->SetNDC();
    clog << LOG(DEBUG_LOG) << "Canvas::SetMeasured exit\n";
  }
  //=====================================================================


  PolarimeterEmbeddedCanvas::PolarimeterEmbeddedCanvas(const char *name, const TGWindow *p, UInt_t w,
      UInt_t h, UInt_t options,
      ULong_t back  ) :
    TRootEmbeddedCanvas(name,p,w,h,options, back){ Init();};

  void PolarimeterEmbeddedCanvas::Init(void)
  {
    GetCanvas()->SetGridx();
    GetCanvas()->SetGridy();
    GetCanvas()->cd();
  }

  MyCanvas::MyCanvas(void) : TGTransientFrame(gClient->GetRoot(),gClient->GetRoot(),600,300)
  {
    canvas = new TRootEmbeddedCanvas("test",this);
    AddFrame(canvas,new TGLayoutHints(kLHintsExpandX | kLHintsExpandY));
    Layout();
    MapSubwindows();
    MapWindow();
  }

}//end of namespace gui
