#ifndef IBN_GUI_POLCANVAS_H
#define IBN_GUI_POLCANVAS_H
#include <string>
#include <map>
#include <memory>

#include <time.h>
#include <iostream>

#include <boost/signals2.hpp>

#include <TGFrame.h>
#include <TCanvas.h>
#include <TRootCanvas.h>
#include <TRootEmbeddedCanvas.h>
#include <TPaveText.h>
#include <TText.h>
#include <TEllipse.h>
#include <TArrow.h>
#include <TLatex.h>

#include "Graph.h"
#include "Note.h"
#include "Utils.h"
#include "Data.h"
#include "Graph.h"
#include "V4Par.h"

namespace data
{
  class EnergyCalibrationStorage;
}
namespace gui   
{
  class EnergyCalibrationView;

	class CanvasData 
	{
		public:
			CanvasData(void){};
			virtual ~CanvasData(void){}
		private:
			class Info	
			{	 
				public:
					virtual ~Info(void){};
					virtual void Draw (TCanvas *) =0;
					virtual void Update(const data::EnergyCalibrationStorage *) =0 ;
			};

			class MeasurementParameters : public  Info 
		{
			/*
			 * =====================================================================================
			 *        Class:  MeasurementParameters
			 * 
			 *  Description:  Show measurement parameters like Tscan,Tcount, initial frequency of scaning,
			 *		final frequency of scaning, frequency step and depolarizer band width/2. 
			 *		display it in the pave. 
			 *		Moreover it shows the data, run number, mode, and set energy
			 *       Author:  Ivan Nikolaev
			 *      Created:  05.04.2004 16:17:53 NOVST
			 *     Revision:  none
			 * =====================================================================================
			 */
			TPaveText *pave;
			TText * Tcount;
			TText * step, *speed, *initial, *final;
			TText * attenuation;
			TText * voltage;
			TText * Date,*Run,*Mode,*Es, *Shiftman;
			TText * measurement_name;
			TText * experiment_name;
			float x1,x2,y1,y2;
			std::unique_ptr <TText> ScanDir; //label of scan direction
      std::unique_ptr <TText> Reason; //Energy calibration reason
			public:
			MeasurementParameters(void);
			~MeasurementParameters(void);
			void Draw(TCanvas * );
			void Update(const data::EnergyCalibrationStorage *);
			void SetTextSize(double textsize); 
			void RenewPaveLocation(void);
			void RenewDirectionLocation(TCanvas *);
			void SetDirection(int direction);
		};

			class NMR : public Info
		{
			TPaveText * pave;
			std::map<std::string, TText*> nmr;
			float x1,x2,y1,y2;
			public:
			NMR(void);
			~NMR(void);
			void Draw(TCanvas*);
			void Update(const data::EnergyCalibrationStorage *);
			void RenewPaveLocation(void);
		};

			class MeasurementInfo : public Info 
		{
			TPaveText * pave;
			std::vector<TText * > Counter;

			TText * Intens;
			float x1,x2,y1,y2;
			public:
			MeasurementInfo(void);
			~MeasurementInfo(void);
			void Draw(TCanvas * );
			void Update(const data::EnergyCalibrationStorage *);
			void RenewPaveLocation(void);
		};

			class VeppInfo : public Info    
		{
			TPaveText * pave;
			std::map<std::string, TText*> fText; 
			TText *Q1_vepp3;
			TText *Q2_vepp3;
			TText *polarization_time;
			//V3Polarization v3p;
			float x1,x2,y1,y2;

			public:
			VeppInfo (void);
			~VeppInfo (void);
			void Update(const data::EnergyCalibrationStorage * );
			void Draw(TCanvas * );
			void RenewPaveLocation(void);
		};

		protected: 

			MeasurementParameters MeasPrm;
			MeasurementInfo   MeasInfo;
			NMR nmr;
			VeppInfo vinfo;

		private:
			virtual void Init(void){};
			static const unsigned BUF = 128;
	};

  class MyCanvas : public TGTransientFrame
  {
    public:
      MyCanvas(void);
      ~MyCanvas(void){};
      void CloseWindow(void) {std::cout << "MyCanvas::CloseWindow" << std::endl; }
    private:
      TGTextButton * fit_button;
      TGTextButton * fit_panel_button;
      TRootEmbeddedCanvas * canvas;
  };

	class Canvas : public TCanvas, public CanvasData   
	{
		data::EnergyCalibrationStorage * EC;
		Data * X;
		std::string name;
		std::unique_ptr<TLatex> depFr; // частота деполяризации (вместе с ошибками)
		std::unique_ptr<TLatex> MeEn; // измеренная энергия(вместе с ошибками)
		bool  draw_notes;
		Note  note;
		typedef std::map<std::string, std::shared_ptr<Graph> > graphic_t;
		typedef  graphic_t::iterator Itr_t;
		typedef  graphic_t::iterator GI;
		public:
		graphic_t graphic;
		Canvas(Bool_t build = kTRUE);
		Canvas(const char* name, const char* title, Int_t form = 1);
		Canvas(const char* name, const char* title, Int_t ww, Int_t wh);
		Canvas(const char* name, const char* title, Int_t wtopx, Int_t wtopy, Int_t ww, Int_t wh);
		Canvas(const char* name, Int_t ww, Int_t wh, Int_t winid);
		Canvas(void);

		Canvas(data::EnergyCalibrationStorage * ec);
		Canvas(data::EnergyCalibrationStorage * ec,  Data * X, std::string nm);
		Canvas(data::EnergyCalibrationStorage * ec, Data * x,  std::string nm, int ww, int wh);
		Canvas(data::EnergyCalibrationStorage * ec, data::Array * x, std::string nm);
		Canvas(data::EnergyCalibrationStorage * ec, data::Array * x, std::string nm, int ww, int wh);
		virtual ~Canvas(void);
		virtual void ProcessedEvent(Int_t,Int_t,Int_t,TObject*);
		void RenewPaveLocation(void);
		void Update(void);
		void DrawParameters(void);
		boost::signals2::signal <void(void)> signal_erase;
		boost::signals2::signal <void(int,int,int,TObject *)> signalProcessedEvent;

		Graph * GetGraph(std::string name);


		Note * GetNote(void) { return & note; }
		data::EnergyCalibrationStorage * GetEnergyCalibration(void)	{ return EC; }	

		Graph * AddGraph(std::shared_ptr<Graph> graph);
		Graph * AddGraph(Data * Y);
		Graph * AddGraph(std::string nm, data::Array *y);

		void EraseGraph(std::string name) { graphic.erase(graphic.find(name)); }

		void Clear(void);

		void Draw(void);
		void Draw(std::string name);

		//bool IsCanvasExist(void);

		void SetMeasured(void);

		void SetDrawNotes(bool draw=true) { draw_notes = draw; } 


		void ShowLastPoints(int n=0) { last_number_of_points = n;}
    static gui::EnergyCalibrationView * theEnergyCalibrationView;
		private:
		int last_number_of_points=0;
		void Init(void);
	};// end of class Canvas

	class PolarimeterEmbeddedCanvas : public TRootEmbeddedCanvas, public CanvasData    
	{
		public:
			PolarimeterEmbeddedCanvas(const char *name, const TGWindow *p, UInt_t w,
					UInt_t h, UInt_t options = kSunkenFrame | kDoubleBorder,
					ULong_t back = GetDefaultFrameBackground()  );
			virtual ~PolarimeterEmbeddedCanvas(void){}
		private:
			void Init(void);
	};
} //end of namespace gui
#endif //IBN_GUI_POLCANVAS_H
