/*
 * =====================================================================================
 *
 *       Filename:  Data.h
 *
 *    Description:  Gui data.
 *
 *        Version:  1.0
 *        Created:  07/12/2010 03:39:48 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#ifndef IBN_POLARIMETER_GUI_DATA_H
#define IBN_POLARIMETER_GUI_DATA_H

#include <vector>
#include <string>

#include "data/Array.h"
#include "data/Calculator.h"
#include "data/EnergyCalibrationStorage.h"

#include <mutex>
#include <thread>

namespace gui
{
  /*
   * =====================================================================================
   *        Class:  Data
   *  Description:  Data list gui information. Name, axis name, etc...
   * =====================================================================================
   */
  class Data
  {
		private:
		std::mutex mutex;
    public:
    data::Array * data; /*  reference to the data */
    std::vector<int> show;
    std::string name;  /*  the name of the data  */
    std::string title; /* title of data */
    std::string axis_title; /*  name of axis */
		int bunch;
		int counter;
		int pair;
    public:
      Data(void);
      Data(std::string nm, std::string tl="", std::string at="");
      Data(data::Array * x, std::string nm="", std::string tl="", std::string at="");
      Data(data::Array & dt, std::string name, std::string axis_title,  int bunch=-1,  int counter=-1,  int pair=-1);
      //Data(const Data &d);
      virtual ~Data(void){}
      //Data & operator=(const Data & d);
      void SetAxisTitle(std::string s);
      data::Array & operator()(void);
      std::vector<double> & value(void);
      std::vector<double> & error(void);
      double *v(void);
      double *e(void);
      ibn::valer<data::Array::Value_t>::reference operator[](unsigned i);
      ibn::valer<data::Array::Value_t>::const_reference  operator[](unsigned i) const;
      virtual void Update(void) 
      { 
        std::clog << LOG(5) << "Data::Update  thread="<< std::this_thread::get_id() << std::endl;
				std::lock_guard<std::mutex> lock(mutex);
        data->Update(); 
        show.reserve(data->size());
        unsigned idx=show.size();
        for(;idx<data->size();idx++) show.push_back(1);
      }
      size_t size(void) const;
  };


  inline bool operator>(const Data & d1, const Data & d2)
  {
    return  d1.name > d2.name;
  }
  inline bool operator<(const Data & d1, const Data & d2)
  {
    return  d1.name < d2.name;
  }

  class NormData : public Data
  {
    data::Array * data1;
    data::Array * data2;
    public:
      NormData(data::Array & d1, data::Array & d2, std::string nm="", std::string tl="", std::string at="")
        : Data(nm,tl,at)
      {
        data1 = &d1;
        data2 = &d2;
        data = new data::Array;
        data->reserve(std::max(d1.size()*2, size_t(100)));
        //Update();
      }

      ~NormData(void)
      {
        delete data;
      }

      void Update(void)
      {
        std::clog << LOG(5) << "NormData::Update BEGIN" << std::endl;
        size_t new_size = std::min(data1->size(), data2->size());
        size_t old_size = data->size();
        for(size_t i=old_size;i<new_size;i++)
        {
          std::clog << LOG(6) << "NormData::Update: " << i << std::endl;
          data->push_back((*data1)[i]/(*data2)[i]);
        }
        show.reserve(data->size());
        for(unsigned idx=show.size(); idx<data->size();idx++) show.push_back(1);
        std::clog<< LOG(5) << "NormData::UPdate FINISH" << std::endl;
      }
  };

  class DeltaData : public Data
  {
    data::Array * data1;
    data::Array * data2;
    public:
      DeltaData(data::Array & d1, data::Array & d2, std::string nm="", std::string tl="", std::string at="")
        : Data(nm,tl,at)
      {
        data1 = &d1;
        data2 = &d2;
        data = new data::Array;
        data->reserve(std::max(d1.size()*2, size_t(100)));
        //Update();
      }

      ~DeltaData(void)
      {
        delete data;
      }

      void Update(void)
      {
        std::clog << LOG(5) << "DeltaData::Update BEGIN" << std::endl;
        size_t new_size = std::min(data1->size(), data2->size());
        size_t old_size = data->size();
        for(size_t i=old_size;i<new_size;i++)
        {
          data->push_back(data::GetDelta((*data1)[i], (*data2)[i]));
        }
        show.reserve(data->size());
        for(unsigned idx=show.size(); idx<data->size();idx++) show.push_back(1);
        std::clog<< LOG(5) << "DeltaData::UPdate FINISH" << std::endl;
      }
  };

  
  class EnergyData : public Data
  {
    data::Array * frequency;
    data::EnergyCalibrationStorage * EC;
    size_t proceed_size;
    public:
      EnergyData(data::EnergyCalibrationStorage * ec, data::Array & freq, std::string nm="", std::string tl="", std::string at="")
        : Data(nm,tl,at)
      {
        EC = ec;
        frequency = &freq;
        data = new data::Array;
        data->reserve(std::max(frequency->size()*2, size_t(100)));
        proceed_size = 0;
    //    Update();
      }
      ~EnergyData(void)
      {
        delete data;
      }
      void Update(void)
      {
        std::clog << LOG(5) << "EnergyData::Update " << std::endl;
        size_t new_size = frequency->size();
        for(size_t i=proceed_size;i<new_size;i++)
        {
          data::PhysValue F = (*frequency)[i];
          data::PhysValue E = EC->frequency2energy(F);
          data->push_back(E);
          show.push_back(F.value == 0 ? 0 : 1);
        }
        proceed_size = new_size;
      }
  };
}
#endif
