/*
 * =====================================================================================
 *
 *       Filename:  DataView.cpp
 *
 *    Description:  Data view implementation
 *
 *        Version:  1.0
 *        Created:  06/17/2010 04:55:12 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#include <thread>
#include <mutex>

#include <boost/format.hpp>
#include <boost/bind.hpp>

#include <TROOT.h>

#include "EnergyCalibrationView.h"

extern std::recursive_mutex  RootProcessEventsMutex;

#include <TClass.h>

#include "data/Calculator.h"

#include "PolarimeterConfig.h"

#include "data/VirtualArray.h"
#include "AntiSlopeData.h"
#include "VolumeData.h"

#include "Canvas.h"

#include "fit/FitConfig.h"

namespace gui
{


  EnergyCalibrationView  * Canvas::theEnergyCalibrationView = nullptr;

  EnergyCalibrationView::EnergyCalibrationView(data::EnergyCalibrationStorage * ec)	
  {
    EC = ec;
    storage=ec;
    counter_output.resize(EC->numbunch); 
    for(unsigned i=0; i< counter_output.size();++i) counter_output[i].resize(EC->numpair);
    for(std::vector < std::vector <int> >::iterator bunch=counter_output.begin(); bunch!=counter_output.end();bunch++)
      for(std::vector <int>::iterator pair=bunch->begin(); pair!=bunch->end();pair++)
        *pair=0;
    InitDataList();
    InitGraphList();
    edit_bad_points=false;
    updatable = false;
    draw_notes=true;
    auto_rare=PCFG.gui.autorare;
    startthreadid  = std::this_thread::get_id();
    Canvas::theEnergyCalibrationView = this;
  }

  EnergyCalibrationView::~EnergyCalibrationView(void)	
  {
		std::clog << LOG(DEBUG_LOG) << "~EnergyCalibrationView(void) BEGIN" << std::endl;
		Clear();
		Canvas::theEnergyCalibrationView = nullptr;
		std::clog << LOG(DEBUG_LOG) << "~EnergyCalibrationView(void) END" << std::endl;
	}

  /*  EnergyCalibrationView */
  Canvas * EnergyCalibrationView::Add(std::string du_name, Data *x, unsigned w, unsigned h)	
  {
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    Canvas * du;
    auto d = displays.find(du_name);
    if(d==displays.end())
    {
      du = new Canvas(EC,x,du_name, w,h);
			du->ShowLastPoints(last_number_of_points);
      displays.insert(make_pair(du_name, du));
    }
    else
    {
      du = d->second;
    }
    du->SetDrawNotes(draw_notes);
    return du;
  }


  Canvas * EnergyCalibrationView::Add(std::string canvas_name,  data::Array *x, unsigned w, unsigned h)	
  {
    return Add(canvas_name, new Data(*x,"time","secc"), w,h);
  }


  Canvas * EnergyCalibrationView::Add(
      std::string canvas_name,
      std::string graphic_name, 
      Data *x, 
      Data *y,
      unsigned w,
      unsigned h
      )
  {
    //boost::recursive_mutex::scoped_lock lock(RootProcessEventsMutex);
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    Canvas * d = Add(canvas_name,x,w,h);
    d->AddGraph(y);
    return d;
  }

  Canvas * EnergyCalibrationView::Add(
      std::string canvas_name,
      std::string graphic_name, 
      data::Array *x, 
      data::Array *y,
      unsigned w,
      unsigned h
      )
  {
    return Add(canvas_name, graphic_name, new Data(x,"time"), new Data(y, graphic_name), w,h);
  }

  Canvas * EnergyCalibrationView::Add(
      std::string canvas_name,
      std::shared_ptr<Graph> graph,
      unsigned w,
      unsigned h
      )
  {
    Canvas * d = Add(canvas_name,graph->X,w,h);
    d->AddGraph(graph);
    return d;
  }

  Canvas *  EnergyCalibrationView::Add ( std::string n )	
  {
    std::clog<< LOG(5) << "Canvas *  EnergyCalibrationView::Add ( std::string n ) BEGIN" << std::endl;
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    auto I=GraphList.find(n);
    Canvas * du =0;
    if(I==GraphList.end())
    {
      std::cerr << LOG(1) << "WARNING: Unable to find graphic " << n << '\n';
    }
    else
    {
      du =   Add ( n, I->second, 300,200); 
      du->signalProcessedEvent.connect(boost::bind(&EnergyCalibrationView::ProcessedEvent,this,du,_1,_2,_3,_4));
      //du->signalProcessedEvent.connect(void [=](Int_t event,Int_t x,Int_t y ,TObject* selected) { this->ProcessedEvent(du, event, x, y, selected); } );
    }
    std::clog<< LOG(5) << "Canvas *  EnergyCalibrationView::Add ( std::string n ) END" << std::endl;
    return du;
  }



	void EnergyCalibrationView::DeleteCanvas(std::string name)
	{
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
		auto d = displays.find(name);
		if(d != displays.end())
		{
			if(IsCanvasExist(d->first) && d->second)  delete d->second; //will be automaticaly erase canvas from list
		}
	}

    //void EnergyCalibrationView::Erase(std::string nm)	
    //{
		//	std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    //  auto d = displays.find(nm);
    //  if(d != displays.end())
    //  {
    //    if(IsCanvasExist(d->first) && d->second)  delete d->second;
    //    //displays.erase(d); 
    //  }
    //}


    //void EnergyCalibrationView::Erase(void)
    //{
		//	std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    //  for(auto d = displays.begin(); d!=displays.end();)
    //  {
    //    if(IsCanvasExist(d->first) && d->second)  delete d->second;
    //    //displays.erase(d++);
		//		d++;
    //  }
    //  displays.clear();
    //}

    //remove canvas from list and not delete it
    void EnergyCalibrationView::Erase(std::string name)
    {
			std::clog << LOG(DEBUG_LOG) << "EnergyCalibrationView::Erase BEGIN: canvas_name = "<< name << std::endl;
			std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      auto d = displays.find(name);
      if(d != displays.end())
      {
      	displays.erase(d);
			}
			std::clog << LOG(DEBUG_LOG) << "EnergyCalibrationView::Erase END" << std::endl;
    }

    Canvas * EnergyCalibrationView::GetCanvas(std::string name)	
    {
			std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      auto d = displays.find(name);
      if( d!= displays.end() && IsCanvasExist(d->first) && d->second) return d->second;
      return nullptr;
    }
    Graph * EnergyCalibrationView::GetGraph(std::string name)	
    {
			std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      auto c = GetCanvas(name);
      if(c==nullptr) return nullptr;
      auto g = c->GetGraph(name);
      if(g==nullptr) return nullptr;
      return g;
    }

    Data * EnergyCalibrationView::GetData(std::string name)
    {
      auto g = GetGraph(name);
      if(g==nullptr) return nullptr;
      return g->Y;
    }

    void EnergyCalibrationView::Update(void)
    {
      using namespace std;
		  std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
			std::clog << LOG(10) << "EnergyCalibrationView::Update BEGIN" << std::endl;
      for(auto d = std::begin(displays); d!=std::end(displays); d++)	
      {
        if(IsCanvasExist(d->first) && d->second) d->second->Update();
      }
      if(storage->t.size()!=0) // Print only nonempty data
      {
        bool isprint=false;
        for(unsigned bunch=0;bunch<storage->numbunch; bunch++)
          for(unsigned pair=0;pair<storage->numcounter/2;pair++)
            isprint=isprint || counter_output[bunch][pair];
        if(isprint)
        {
          std::clog  << setw(12) << "#bunch-pair"
            << setw(10) << "time,sec" << " "
            << setw(10) << "I,mA" << " "
            << setw(10) << "in,kHz" << " "
            << setw(10) << "out,kHz" << " "
            << setw(10) << "in*out,kHz"<< " "
            << setw(10) << "cor,kHz"<< " "
            << setw(10) << "sx,mm" << " "
            << setw(10) << "sz,mm" << " "
            << setw(10) << "sl,mm" << " "
            << std::endl;
        }
        for(unsigned bunch=0;bunch<storage->numbunch; bunch++)
        {
          for(unsigned pair=0;pair<storage->numcounter/2;pair++)
          {
            if(counter_output[bunch][pair])
            {
              std::clog << "bunch"<<bunch+1<<"-pair"<<pair+1
                << setw(10) << storage->t.back().value << " "
                << setw(10) << storage->I[bunch].back().value/1000.<< " "
                << setw(10) << storage->Count[bunch][pair*2].back().value/1000. << " "
                << setw(10) << storage->Count[bunch][pair*2+1].back().value/1000. << " "
                << setw(10) << storage->Coinc[bunch][pair].back().value/1000. << " "
                << setw(10) << storage->Cor[bunch][pair].back().value/1000. << " "
                << setw(10) << storage->SigmaX[bunch].back().value << " "
                << setw(10) << storage->SigmaZ[bunch].back().value << " "
                << setw(10) << storage->SigmaL[bunch].back().value << " "
                << "\n";
            }
          }
        }
      }
      updatable=false;
      if(auto_rare)
      {
        if(EC->N!=0 && EC->N%auto_rare == 0)
        {
          Rare(1+unsigned(EC->N/auto_rare));
        }
      }
    }


    void EnergyCalibrationView::Notify(void)
    {
      updatable = true;
    }


    void EnergyCalibrationView::Clear (void)
    {
      std::clog << LOG(5) << "EnergyCalibrationView::Clear BEGIN" << std::endl;
		  std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      std::list<std::string> du_names;
			for(auto & d: displays)
      {
        du_names.push_back(d.first);
        if(IsCanvasExist(d.first) && d.second) delete d.second;
        //displays.erase(d++);
        //if(d!=displays.end()) cout << "After erase d.first="<< d->first<< endl;
      }
      displays.clear();
      fit::BadPointList.clear();
      Draw();
      std::clog << LOG(5) << "EnergyCalibrationView::Clear END" << std::endl;
    }

    void EnergyCalibrationView::Draw(void)
    {
		  std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      for(auto d = displays.begin(); d!=displays.end(); d++)
      {
        if(IsCanvasExist(d->first) && d->second) d->second->Draw();
      }
    }

    bool EnergyCalibrationView::IsCanvasExist(const std::string & name)	
    {
      TObject * obj = gROOT->FindObject(name.c_str());
      if(obj)
      {
        return !strcmp(obj->ClassName(),"TCanvas");
      }
      return false;
    }


    void EnergyCalibrationView::Print(boost::filesystem::path print_dir)
    {
		  std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
			Update();
      if (!boost::filesystem::exists(print_dir) || !boost::filesystem::is_directory(print_dir))
      {
        std::cerr << LOG(NORMAL_LOG) << "WARNING: Print directory " << print_dir.string() <<  " doesnt exist. Unable to print" << std::endl;
        return;
      }
      char time_string[1024];
      strftime(time_string,1024,"-%F-%T-",localtime(&EC->begin_time.tv_sec));
      std::string file_prefix =std::to_string(EC->cfg.run)+std::string(time_string);
      for(auto d = displays.begin(); d!=displays.end(); d++)
      {
        if(IsCanvasExist(d->first) && d->second)
        {
          std::string pdf_name = file_prefix+d->first + ".pdf";
          std::string name =print_dir.string()+"/"+pdf_name;
          std::cout << LOG(NORMAL_LOG) << "Printing graphic " << d->first << " into file " << name << std::endl;
          d->second->GetCanvas()->Print(name.c_str());

          //send data to printer
          std::string command="lpr \'"+name+"\'";
          std::cout << LOG(NORMAL_LOG) << "Printing graphic " << d->first << " from file " << name << std::endl;
          system(command.c_str());
          std::cout << LOG(NORMAL_LOG) << "Printing DONE" << std::endl;

          //workaround energy calibration itself
          //and show result for KEDR
          if(d->first == "delta-1")
          {
            command = "ln -f -s " + pdf_name + " " + print_dir.string()+"/last.pdf";
            std::cout << LOG(NORMAL_LOG) << "Creating symbolic link to last.pdf" << std::endl;
            system(command.c_str());


            std::cout << LOG(NORMAL_LOG) << "Creating last.gif " << std::endl;
            std::string name =print_dir.string()+"/last.gif";

            command = "mv " + name + " " + print_dir.string()+"/previous.gif";
            system(command.c_str());

            d->second->GetCanvas()->Print(name.c_str());
          }
        }
      }
    }

    void EnergyCalibrationView::Rare(unsigned rare)
    {
      for(auto d = displays.begin(); d!=displays.end(); d++)
      {
        if(IsCanvasExist(d->first) && d->second) d->second->GetNote()->Rare(rare);
      }
    }

    void EnergyCalibrationView::ProcessedEvent(Canvas * c, Int_t event,Int_t x,Int_t y ,TObject* selected)
    {
		  std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
      if(!edit_bad_points) return;
      TClass * cl = selected->IsA();
      std::string classname(cl->GetName());
      if(event==1 && classname=="TGraphErrors")
      {
        TGraphErrors * g = (TGraphErrors * )(selected);
        //localize point to be deleted
        Int_t ipoint = -2;
        // start with a small window (in case the mouse is very close to one point)
        c->cd();
        for (Int_t i=0;i<g->GetN();i++)
        {
          Int_t dpx = x - gPad->XtoAbsPixel(gPad->XtoPad(g->GetX()[i]));
          Int_t dpy = y - gPad->YtoAbsPixel(gPad->YtoPad(g->GetY()[i]));
          if (dpx*dpx+dpy*dpy < 100) {ipoint = i; break;}
        }
        if(ipoint <0) return;
        double xdata=g->GetX()[ipoint];
        double ydata=g->GetY()[ipoint];
        std::string graph_name(g->GetName());
        Graph * du=0;
        for(display_t::iterator di = displays.begin(); di!=displays.end();di++)
        {
          Graph * p = di->second->GetGraph(g->GetName());
          if(p)
          {
            du=p;
            break;
          }
        }
        if(du==0)
        {
          std::clog << "Unable to find data unit for graph " << g->GetName() << std::endl;
          return;
        }
        int data_point=-1;
        for(unsigned i = 0; i!=du->xData.size();i++)
        {
          if(xdata==du->xData[i].value && ydata == du->yData[i].value)
          {
            data_point = i;
            break;
          }

        }
        if(data_point>-1) 
        {
          std::clog << "Removing graph point " << ipoint << std::endl;
          if(edit_bad_points) g->RemovePoint(ipoint);
          std::clog << "Select data point " << data_point << std::endl;
          if(find(fit::BadPointList.begin(), fit::BadPointList.end(),unsigned(data_point))==fit::BadPointList.end())
          {
            std::clog << LOG(1) << "Add new bad data point #" << data_point << " into BadPointList while graph point # " << ipoint << "\n";
            fit::BadPointList.push_back(data_point);
            fit::BadPointList.sort();
            fit::BadPointList.unique();
          }
          else
          {
            std::clog << LOG(1) << "Selected point #" << data_point <<  " already in BadPointList\n";
          }
        }
      }
    }

    void LoadBadPointList(std::istream & is)
    {
      unsigned run;
      std::clog << LOG(1) << "Reading bad point list: ";
      while(is>>run) 
      {
        fit::BadPointList.push_back(run);
        LOG << run << " ";
      }
      fit::BadPointList.sort();
      fit::BadPointList.unique();
      std::clog << LOG(1) << std::endl;
    }

    void SaveBadPointList(std::ostream & file)
    {
      unsigned i=0;
      for(std::list<unsigned>::iterator it=fit::BadPointList.begin();it!=fit::BadPointList.end();++it)
      {
        file << *it << " ";
        if(i%10==0 && i!=0) file << std::endl;
      }
      file << std::endl;
    }

  void EnergyCalibrationView::InitDataList(void)
  {
    //boost::recursive_mutex::scoped_lock lock(RootProcessEventsMutex);
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    NonBunchDataList.clear();
		CounterDataList.clear();
    BunchDataList.clear();
    CounterBunchDataList.clear();
    PairBunchDataList.clear();
    DataList.clear();

    typedef std::shared_ptr<Data> D_t;

    NonBunchDataList["time"]        = (D_t)new Data(storage->t,"time", "t, sec");
    NonBunchDataList["frequency"]   = (D_t)new Data(storage->DepFr,"frequency","f_{d}, Hz");
    NonBunchDataList["current-sum"] = (D_t)new Data(storage->Isum,"current-sum","I, mkA");
    NonBunchDataList["energy"]      = (D_t)new EnergyData(storage, storage->DepFr,"energy","E, MeV");
    NonBunchDataList["vacuum"]      = (D_t)new Data(storage->MediumVacuumPresure,"vacuum", "P, nTor");

		for(unsigned counter=0;counter<storage->numcounter; counter++)
		{
      std::string name = std::string("counter_position") + "-" + std::to_string(counter+1);
			CounterDataList[name][counter] = (D_t) new Data(storage->CounterPosition[counter], name,  "x,  mm");
		}


    for(unsigned bunch=0; bunch<storage->numbunch; bunch++)
    {
      auto add_bdl = [&] (data::Array & array, std::string name, std::string axis_name)
      {
        std::string Name =  name + "-" + std::to_string(bunch+1);
        BunchDataList[name][bunch] = (D_t)new Data(array, Name, axis_name, bunch);
      };
      std::string bs=std::to_string(bunch+1);
      add_bdl(storage->I[bunch], "current", "I, mkA");
      add_bdl(storage->SigmaX[bunch], "sigma_x", "#sigma_{x}, mkm");
      add_bdl(storage->SigmaZ[bunch], "sigma_z", "#sigma_{z}, mkm");
      add_bdl(storage->SigmaL[bunch], "sigma_l", "#sigma_{l}, cm");
      add_bdl(storage->CountSum[bunch],"counter_sum", "Hz");
      add_bdl(storage->CoincSum[bunch],"coincidence_sum", "Hz");
      add_bdl(storage->RndCoincSum[bunch],"random_coincidence_sum", "Hz");
      add_bdl(storage->CorSum[bunch],"correlated_sum", "Hz");
      add_bdl(storage->RndSum[bunch],"random_sum", "Hz");
      add_bdl(storage->DeltaSum[bunch],"delta", " ");
      add_bdl(storage->DeltaCoincSum[bunch],"delta_coincidence_sum", " ");
      add_bdl(storage->DeltaCorSum[bunch],"delta_correlated_sum", " ");
      add_bdl(storage->DeltaRndSum[bunch],"delta_random_sum", " ");
      ////***************************** WARING memory leak  
      //data::Array * count_norm = new data::FitSquareCurrentNormalizedArray(&storage->CountSum[bunch], &storage->t);
      ////*********************************************************
      //BunchDataList["counter_sum/square_current"][bunch]=(D_t)new NormData(, "counter_sum/square_current-"+bs, "Hz/mA^{2}");
      BunchDataList["counter_sum/random_sum"][bunch]=(D_t)new DeltaData(storage->CountSum[bunch], storage->RndSum[bunch],"counter_sum/random_sum-"+bs, "Hz");
      add_bdl(storage->CountSumNormI[bunch], "normalized_counter_sum", "Hz");
      add_bdl(storage->CountSumNormI2[bunch], "normalized2_counter_sum", "Hz");
      add_bdl(storage->CoincSumNormI[bunch], "normalized_coincidence_sum", "Hz");
      add_bdl(storage->CorSumNormI[bunch], "normalized_correlated_sum", "Hz");
      add_bdl(storage->RndSumNormI[bunch], "normalized_random_sum", "Hz");
      add_bdl(storage->RndCoincSumNormI[bunch], "normalized_random_coincidence_sum", "Hz");

      BunchDataList["beam_volume"][bunch] = (D_t) new VolumeData(&storage->SigmaX[bunch],&storage->SigmaZ[bunch],  &storage->t, "beam_volume-"+bs, " ");
      BunchDataList["antislope_counter_sum"][bunch] = (D_t) new AntiSlopeData(&storage->CountSum[bunch], &storage->t, "antislope_counter_sum-"+bs, " ");
      BunchDataList["antislope_coincidence_sum"][bunch] = (D_t) new AntiSlopeData(&storage->CoincSum[bunch], &storage->t, "antislope_coincidence_sum-"+bs, " ");
      BunchDataList["antislope_correlated_sum"][bunch] = (D_t) new AntiSlopeData(&storage->CorSum[bunch], &storage->t, "antislope_correlated_sum-"+bs, " ");

      for(unsigned counter = 0; counter<storage->numcounter; counter++)
      {
        auto add_bc = [&] (data::Array & array, std::string name, std::string axis_name)
        {
          std::string Name =  name + "-"+std::to_string(counter+1) + "-" + std::to_string(bunch+1);
          CounterBunchDataList[name][counter][bunch] = (D_t)new Data(array, Name, axis_name, bunch, counter);
        };
        add_bc(storage->CounterLoad[bunch][counter],"load", "Hz");
        add_bc(storage->Count[bunch][counter],"counter", "Hz");
        add_bc(storage->DeltaCount[bunch][counter],"delta_count", "Hz");
        add_bc(storage->Rnd[bunch][counter],"random", "Hz");
        add_bc(storage->DeltaRnd[bunch][counter],"delta_random", "Hz");
      }
      for(unsigned pair = 0; pair<storage->numpair; pair++)
      {
        auto add_pb = [&] (data::Array & array, std::string name, std::string axis_name)
        {
          std::string Name =  name + "-"+std::to_string(pair+1) + "-" + std::to_string(bunch+1);
          PairBunchDataList[name][pair][bunch] = (D_t)new Data(array, Name, axis_name, bunch, -1,  pair);
        };
        add_pb(storage->CoincLoad[bunch][pair],"coincidence_load", "Hz");
        add_pb(storage->Coinc[bunch][pair],"coincidence", "Hz");
        add_pb(storage->Cor[bunch][pair],"correlated", "Hz");
        add_pb(storage->RndCoinc[bunch][pair],"random_coincidence", "Hz");
        add_pb(storage->DeltaCoinc[bunch][pair],"delta_coincidence", " ");
        add_pb(storage->DeltaCor[bunch][pair],"delta_correlated", " ");
        add_pb(storage->DeltaRndCoinc[bunch][pair],"delta_random_coincidence", " ");
      }
    }

    for(auto & d : NonBunchDataList)
    {
      DataList.push_back(d.second);
    }

    for(auto & B : BunchDataList)
    {
      for(auto & d : B.second)
      {
        DataList.push_back(d.second);
      }
    }
    for(auto & D : CounterBunchDataList)
    {
      for(auto & counter : D.second)
      {
        for(auto & bunch : counter.second)
        {
          DataList.push_back(bunch.second);
        }
      }
    }
    for(auto & D : PairBunchDataList)
    {
      for(auto & pair : D.second)
      {
        for(auto & bunch : pair.second)
        {
          DataList.push_back(bunch.second);
        }
      }
    }
  }

  void EnergyCalibrationView::InitGraphList(void)
  {
    NonBunchGraphList.clear();
    BunchGraphList.clear();
    GraphList.clear();
    Data * T = NonBunchDataList["time"].get();
    typedef std::shared_ptr<Graph> G;

		//for(unsigned counter=0; counter<EC->numcounter; counter++)
		//{
		//	for(auto & cdl : CounterDataList)
		//	{
		//		CounterGraphList[cdl.first][counter] = G(new Graph(T, cdl.second[counter].get()));
		//	}
		//}
    for(unsigned bunch = 0; bunch < EC->numbunch; bunch++ )
    {
      for(auto & BDL : BunchDataList)
      {
        BunchGraphList[BDL.first][bunch] = G( new Graph(T,BDL.second[bunch].get()));
      }
      for(unsigned counter = 0; counter < EC->numcounter; counter++ )
      {
        for(auto & CBDL : CounterBunchDataList)
        {
          BunchGraphList[CBDL.first+"-"+std::to_string(counter+1)][bunch] = G( new Graph(T,CBDL.second[counter][bunch].get()));
        }
      }
      for(unsigned pair = 0; pair < EC->numpair; pair++ )
      {
        for(auto & CBDL : PairBunchDataList)
        {
          BunchGraphList[CBDL.first+"-"+std::to_string(pair+1)][bunch] = G( new Graph(T,CBDL.second[pair][bunch].get()));
        }
      }
    }

    for(unsigned bunch=0; bunch<EC->numbunch; bunch++) 
    {

      BunchGraphList["normalized-counter-sum-vs-I"][bunch] = G(new Graph(&EC->I[bunch],&EC->CountSumNormI[bunch]));
      BunchGraphList["normalized-coincidence-sum-vs-I"][bunch] = G(new Graph(&EC->I[bunch],&EC->CoincSumNormI[bunch]));
      BunchGraphList["normalized-correlated-sum-vs-I"][bunch] = G(new Graph(&EC->I[bunch],&EC->CorSumNormI[bunch]));
      BunchGraphList["normalized-random-sum-vs-I"][bunch] = G(new Graph(&EC->I[bunch],&EC->RndSumNormI[bunch]));
      BunchGraphList["normalized-random-coinc-sum-vs-I"][bunch] = G(new Graph(&EC->I[bunch],&EC->RndCoincSumNormI[bunch]));

      BunchGraphList["normalized2-counter-sum-vs-Ir"][bunch] = G(new Graph(&EC->IRatio[bunch],&EC->CountSumNormI2[bunch]));
      BunchGraphList["normalized2-coincidence-sum-vs-Ir"][bunch] = G(new Graph(&EC->IRatio[bunch],&EC->CoincSumNormI2[bunch]));
      BunchGraphList["normalized2-correlated-sum-vs-Ir"][bunch] = G(new Graph(&EC->IRatio[bunch],&EC->CorSumNormI2[bunch]));

      //BunchGraphList["normalized2-counter-sum"][bunch] = G(new Graph(T, &EC->CountSumNormI2[bunch]));
      BunchGraphList["normalized2-coincidence-sum"][bunch] = G(new Graph(T, &EC->CoincSumNormI2[bunch]));
      BunchGraphList["normalized2-correlated-sum"][bunch] = G(new Graph(T, &EC->CorSumNormI2[bunch]));
    }
    NonBunchGraphList["frequency"] = G(new Graph ( T, &EC->DepFr ));
    NonBunchGraphList["vacuum-medium"] = G(new Graph ( T, &EC->MediumVacuumPresure));
    NonBunchGraphList["Isum"] = G(new Graph ( T, &EC->Isum));

    for( std::map < std::string , data::Array>::iterator i=EC->Vacuum.begin(); i!=EC->Vacuum.end(); i++)
    {
      std::string name = "vacuum-"+i->first;
      NonBunchGraphList[name] = G(new Graph(T, &(i->second)));
    }
    for(size_t bunch =0 ; bunch<EC->numbunch ; bunch++)   
    {
      for(BunchGraphList_t::iterator I=BunchGraphList.begin(); I!=BunchGraphList.end(); I++) 
      {
        GraphList[I->first + '-' + setString(bunch+1,0,0)] = (I->second)[bunch];
      }
    }
    for(GraphList_t::iterator I=NonBunchGraphList.begin(); I!=NonBunchGraphList.end(); I++) 
    {
      GraphList[I->first]=I->second;
    }
  }

  void EnergyCalibrationView::OrderWindows(void)
  {
    double x = sqrt(displays.size());
    int ny = floor(x);
    int nx = ny+1;
    int WX = 1920-5*nx;
    int WY = 1000-30*ny;
    int wx = WX/nx;
    int wy = WY/(ny+1);
    int index = 0;
    for( auto & d : displays)
    {
      int ix = index % nx;
      int iy = index / nx;
      int x  = ix*(wx+5);
      int y  = iy*(wy+30);
      d.second->SetWindowPosition(x,y); 
      d.second->SetWindowSize(wx,wy);   
      index++;
    }
  }

}
