/*
 * =====================================================================================
 *
 *       Filename:  FitDialog.cpp
 *
 *    Description:  Fit Dialog implementation
 *
 *        Version:  1.0
 *        Created:  07/11/2010 01:34:59 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */


#include <string>
#include <vector>
#include <map>

#include "FitDialog.h"

#include <TGString.h>
#include <TGLabel.h>

#include "Utils.h"
#include "Log.h"

#include <iostream>
#include <map>
#include <regex>
#include <boost/regex.hpp>



#include <TGFrame.h>
#include <TGButton.h>
#include <TGTextEntry.h>
#include <TGNumberEntry.h>
#include <TGComboBox.h>

#include "fit/FitManager.h"
#include "gui/EnergyCalibrationView.h"
#include "data/VirtualArray.h"

#include "fit/RootFunctionFitter.h"


namespace gui
{

  std::map < unsigned , std::string > COMBO_FUNCTIONS_MAP;
  std::map < unsigned , std::string > COMBO_GRAPHICS_MAP;

  enum ENTRY
  {	
    TE_BEGIN_TIME,
    TE_END_TIME,
    TE_FORMULA
  };

  enum BUTTON
  {
    TB_RESET,
    TB_OK,
    TB_FIT,
    TB_CANCEL,
    TB_CLOSE
  };

  enum COMBO
  {
    GRAPH_COMBO,
    FUNCTIONS_COMBO,
    COMBO_DELTA,
    COMBO_TOUSCHEK1,
    COMBO_TOUSCHEK2,
    COMBO_TOUSCHEK_NORM1,
    COMBO_TOUSCHEK_NORM2,
    COMBO_DIRECT1,
    COMBO_DIRECT2,
    COMBO_RANDOM1,
    COMBO_RANDOM2,
    COMBO_DELTA_DIRECT,
    COMBO_DELTA_RANDOM,
    COMBO_DELTA_RANDOM_DIRECT,
    COMBO_IN,
    COMBO_OUT,
    COMBO_IN2,
    COMBO_OUT2,
    COMBO_NOSLOPE
  };

	int DRAW_CHECKBUTTON_ID;
	int FULL_DRAW_CHECKBUTTON_ID;

  /* varios case */	
  static const int ENUM_PARAMETER_SHIFT = 65535;

  FitDialog::FitDialog(const TGCompositeFrame *main, gui::EnergyCalibrationView * ecv, fit::FitManager *FM, bool iswait) : 
    TGTransientFrame(gClient->GetRoot(),main)
  {
    theFitManager = FM;
    EC = theFitManager->GetEnergyCalibration();
    ECView = ecv;

    LHGeneral = new TGLayoutHints(kLHintsTop | kLHintsExpandX,5,5,5,5);
    /* Field for inputing the fit range */
    GFTextEntry.reset(new TGGroupFrame(this,new TGHotString("Fit range"),kVerticalFrame));
    GFTextEntry->SetLayoutManager(new TGMatrixLayout(GFTextEntry.get(),2,2,10,10));

    TEBeginTime.reset(new TGNumberEntry(GFTextEntry.get(),0,8,TE_BEGIN_TIME,TGNumberFormat::kNESInteger));
    TEEndTime.reset(new TGNumberEntry(GFTextEntry.get(),0,8,TE_END_TIME,TGNumberFormat::kNESInteger));
    TEBeginTime->Associate(this);
    TEEndTime->Associate(this);
    GFTextEntry->AddFrame(new TGLabel(GFTextEntry.get(), new TGHotString("Begin time,s")));
    GFTextEntry->AddFrame(new TGLabel(GFTextEntry.get(), new TGHotString("End time,s")));
    GFTextEntry->AddFrame(TEBeginTime.get(), new TGLayoutHints(kLHintsExpandX | kLHintsLeft));
    GFTextEntry->AddFrame(TEEndTime.get(),   new TGLayoutHints(kLHintsExpandX | kLHintsRight));
    AddFrame(GFTextEntry.get(),LHGeneral);

    AddFrame(new TGLabel(this, new TGHotString("Graphics")), LHGeneral);
    FitCombo.reset(new TGComboBox(this, GRAPH_COMBO));
    AddFrame(FitCombo.get(),LHGeneral);
    FitCombo->Resize(200, 20);

    unsigned combo_id=0;	

    //Fill graphics combo
    int default_combo_id=0;
    for(auto & p : ECView->displays)
    {
      COMBO_GRAPHICS_MAP[combo_id] = p.first;
      FitCombo->AddEntry( p.first.c_str(), combo_id );
      if ( p.first == PCFG.gui.default_graph ) default_combo_id=combo_id;
      combo_id++;
    }
    FitCombo->Select(default_combo_id);


    AddFrame(new TGLabel(this, new TGHotString("Fit method")), LHGeneral);
    FunctionsCombo.reset(new TGComboBox(this, FUNCTIONS_COMBO));
    int i_fc=0;
    for (auto & f : theFitManager->GetFitList())
    {
      FunctionsCombo->AddEntry(f->name.c_str(),i_fc);
      COMBO_FUNCTIONS_MAP[i_fc] = f->name;
      if(theFitManager->GetFunctionName() == f->name ) FunctionsCombo->Select(i_fc);
      i_fc++;
    }
    FunctionsCombo->Resize(200, 20);
    AddFrame(FunctionsCombo.get(),LHGeneral);

    TEFormula.reset(new TGTextEntry(this,"[0]+[1]*(1 - exp(-x/[2]))", TE_FORMULA));
    AddFrame (TEFormula.get(), LHGeneral);

    ParameterFrame.reset(new TGGroupFrame (this,new TGHotString("Initial parameters value")));
    ParameterFrame->SetLayoutManager(new TGMatrixLayout(ParameterFrame.get(),0,6,5,2));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "fix"));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "initial"));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "error"));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "limit"));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "lower"));
    ParameterFrame->AddFrame(new TGLabel(ParameterFrame.get(), "upper"));

    pcfg = theFitManager->GetParameters();
    FillParametersFrame();
    AddFrame( ParameterFrame.get(), LHGeneral);
		DRAW_CHECKBUTTON_ID = 10*par_vgt.size();
		FULL_DRAW_CHECKBUTTON_ID = DRAW_CHECKBUTTON_ID +1;


    DrawHF.reset(new TGHorizontalFrame(this,0,0));

		DrawCheckButton.reset(new TGCheckButton ( DrawHF.get(), "draw result", DRAW_CHECKBUTTON_ID));
		DrawCheckButton->Resize(200,DrawCheckButton->GetDefaultHeight());
		DrawCheckButton->Associate(this);
		DrawCheckButton->SetState(kButtonDown);
		DrawHF->AddFrame(DrawCheckButton.get(),  LHGeneral);

		FullDrawCheckButton.reset(new TGCheckButton ( DrawHF.get(), "full draw", FULL_DRAW_CHECKBUTTON_ID));
		FullDrawCheckButton->Resize(200,FullDrawCheckButton->GetDefaultHeight());
		FullDrawCheckButton->Associate(this);
		FullDrawCheckButton->SetState(kButtonDown);
		DrawHF->AddFrame(FullDrawCheckButton.get(),  LHGeneral);

		AddFrame(DrawHF.get(),  LHGeneral);


    /* Buttons */
    HFTextButton.reset(new TGHorizontalFrame(this,0,0));
    TBReset.reset(new TGTextButton(	HFTextButton.get() ," Reset ",TB_RESET));
    TBReset->Associate(this);

    TBFit.reset(new TGTextButton(	HFTextButton.get() ," Fit ",TB_FIT));
    TBFit->Associate(this);

    TBClose.reset(new TGTextButton(	HFTextButton.get() ," Close ",TB_CLOSE));
    TBClose->Associate(this);

    LHButton = new TGLayoutHints(kLHintsTop| kLHintsExpandX | kLHintsExpandY,0,2,2,2);
    HFTextButton->AddFrame(TBReset.get(),LHButton);
    HFTextButton->AddFrame(TBFit.get(),LHButton);
    HFTextButton->AddFrame(TBClose.get(),LHButton);
    AddFrame(HFTextButton.get(),LHGeneral);

    // Layout and map all frames and vidget's in the window 
    FillCheckButtons();
    ShowParameters(pcfg);
    Layout();
    SetWindowName("Fit Panel");
    MapSubwindows();
    MapWindow();
    //Resize(GetDefaultSize());
    Move(main->GetX()+main->GetWidth(),main->GetY());
    Resize(211,510);
    if(iswait) fClient->WaitFor(this);
  };

  FitDialog::~FitDialog(void)
  {
    delete LHButton;
    delete LHGeneral;
  }


  void FitDialog::ShowParameters(const ROOT::Minuit2::MnUserParameters & upar)
  {
    std::clog << LOG(3) << "FitDialog::SetFitParameters" << std::endl;
    const unsigned NP  = upar.Params().size();
    for (unsigned i = 0; i<NP; i++)
    {
      const ROOT::Minuit2::MinuitParameter & mp = upar.Parameter(i);
      Parameters_t & p = par_vgt[i];
      p.fix->SetState(mp.IsFixed() ? kButtonDown : kButtonUp);
      p.value->SetText(mp.Name());
      p.name = mp.Name();
      p.fix->SetText(mp.Name());
      p.value->SetNumber(mp.Value());
      p.value->SetState(true);
      p.error->SetState(true);
      p.error->SetNumber(mp.Error());
      p.limit->SetState(mp.HasLimits() ? kButtonDown : kButtonUp);
      p.lower->SetState(mp.HasLowerLimit());
      p.upper->SetState(mp.HasUpperLimit());
      p.lower->SetNumber(mp.LowerLimit());
      p.upper->SetNumber(mp.UpperLimit());
      std::clog << LOG(4) << p.number << " " << p.name << " " << mp.Value() << std::endl;
    }
    //Disable remains widget
    for(unsigned i=NP;i<par_vgt.size();i++)
    {
      Parameters_t & p = par_vgt[i];
      p.fix->SetState(kButtonDisabled);
      p.limit->SetState(kButtonDisabled);
      p.value->SetState(false);
      p.lower->SetState(false);
      p.upper->SetState(false);
      p.error->SetState(false);
			p.value->SetText("");
			p.fix->SetText("");
			p.lower->SetNumber(0);
			p.upper->SetNumber(0);
			p.error->SetNumber(0);
			p.value->SetNumber(0);
    }
  }

  ROOT::Minuit2::MnUserParameters  FitDialog::GetParameters()
  {
    std::clog << LOG(5) << "FitDialog::GetParameters() BEGIN" << std::endl;
    const unsigned NP  = pcfg.Params().size();
    ROOT::Minuit2::MnUserParameters par;
    for ( unsigned i = 0; i<NP ; i++)
    {
      Parameters_t & pw = par_vgt[i];
      par.Add(pw.name, pw.value->GetNumber(), pw.error->GetNumber());
    }
    for ( unsigned i = 0; i<NP ; i++)
    {
      Parameters_t & pw = par_vgt[i];
      if(pw.fix->GetState() == kButtonDown) par.Fix(i);
      else 
      {
        if(par.Parameter(i).IsFixed()) par.Release(i);
      }
      if(pw.limit->GetState() == kButtonDown)
      {
        double lower_limit=pw.lower->GetNumber();
        double upper_limit=pw.upper->GetNumber();
        if(lower_limit!=upper_limit)
        {
          par.SetLimits(i,lower_limit,upper_limit);
        }
        else
        {
          par.RemoveLimits(i);
          pw.limit->SetState(kButtonUp);
          pw.upper->SetState(kButtonDisabled);
          pw.lower->SetState(kButtonDisabled);
        }
      }
      else 
      {
        par.RemoveLimits(i);
      }
    }
    return par;
  }

  void FitDialog::UpdateParIni(void)	
  {
    pcfg = GetParameters();
    theFitManager->SetParameters(pcfg);
  }

  void FitDialog::FillCheckButtons(void)	
  {
  }

  void FitDialog::UpdateCheckButtons(void)	
  {
  }


  void FitDialog::ClearParametersFrame(void)
  {
    std::clog<<LOG(DEBUG_LOG) << "FitDialog::ClearParametersFrame BEGIN" << std::endl;
    for(auto & p: par_vgt)
    {
      //std::clog<<LOG(NORMAL_LOG+1) << "FitDialog::ClearParametersFrame::removeFrom Frame" << std::endl;
      p.removeFrom(ParameterFrame.get());
    }
    par_vgt.clear();
    std::clog<< LOG(DEBUG_LOG) << "FitDialog::ClearParametersFrame END" << std::endl;
  }

  void FitDialog::FillParametersFrame(void)
  {
    const unsigned NP  = pcfg.Params().size();
    par_vgt.resize(NP);
    for(int i=0;i<par_vgt.size(); i++)
    {
      Parameters_t & p = par_vgt[i];
      p.number = i;
      p.fix.reset(new TGCheckButton ( ParameterFrame.get(), "", 2*i));
      p.fix->Resize(200,p.fix->GetDefaultHeight());
      p.fix->Associate(this);
      p.value.reset(new TGNumberEntry(ParameterFrame.get(),0,10,i+ENUM_PARAMETER_SHIFT,TGNumberFormat::kNESReal));
      p.value->SetText(p.name.c_str());
      p.value->Associate(this);
      p.error.reset(new TGNumberEntry(ParameterFrame.get(),0,10,i+ENUM_PARAMETER_SHIFT,TGNumberEntry::kNESReal));
      p.error->Associate(this);
      p.limit.reset(new TGCheckButton ( ParameterFrame.get(), "", 2*i+1));
      p.limit->Resize(200,p.limit->GetDefaultHeight());
      p.limit->Associate(this);
      p.lower.reset(new TGNumberEntry(ParameterFrame.get(),0,10,i+ENUM_PARAMETER_SHIFT,TGNumberEntry::kNESReal));
      p.lower->Associate(this);
      p.upper.reset(new TGNumberEntry(ParameterFrame.get(),0,10,i+ENUM_PARAMETER_SHIFT,TGNumberEntry::kNESReal));
      p.upper->Associate(this);
      p.addTo(ParameterFrame.get());
    }
    //ShowParameters(pcfg);
    Layout();
    MapSubwindows();
    MapWindow();
  }


  void FitDialog::UpdateFunctionCombo(void)
  {
    //Update fit functions combo;
    std::string selected_fit_method = COMBO_FUNCTIONS_MAP[FunctionsCombo->GetSelected()];
    std::cout << "Before theFitManager->SetFitFunction" << std::endl;
    theFitManager->SetFitFunction(selected_fit_method);
    UpdateGraphCombo();
    ClearParametersFrame();
    pcfg=theFitManager->GetParameters();
    FillParametersFrame();
    ShowParameters(pcfg);
    Layout();
    MapSubwindows();
    MapWindow();
    Resize(GetDefaultSize());
  }

  void FitDialog::UpdateGraphCombo(void)
  {
    std::string du_name = COMBO_GRAPHICS_MAP[FitCombo->GetSelected ()];
    canvas = ECView->GetCanvas(du_name);
    if(canvas == 0) 
    {
      std::cout << LOG(1) << "No data to fit :" << du_name << '\n';
      return;
    }
    TGraphErrors * graph = canvas->GetGraph(du_name);
    if(graph==0) 
    {
      std::cout << LOG(1) << "Unable to  find graphic\n";
      return;
    }
    theFitManager->SetName(du_name);
    theFitManager->SetGraph(graph);
    theFitManager->SetCanvas(canvas);
  }


  void FitDialog::SetTimeRange(void)
  {
    double tmin=TEBeginTime->GetNumber();
    double tmax=TEEndTime->GetNumber();
    if(tmin!=theFitManager->GetFitter()->tmin)
    {
      SetBeginTime();
    }
    if(tmax!=theFitManager->GetFitter()->tmax)
    {
      SetEndTime();
    }
  };

  void FitDialog::SetBeginTime(void)
  {
    double tmin=TEBeginTime->GetNumber();
    double tmax=TEEndTime->GetNumber();
    //modify end time if tmax < tmin
    if(tmax<tmin)
    {
      data::EnergyCalibrationStorage * EC = ECView->GetEnergyCalibrationStorage();
      tmax = EC->t[EC->t.size()-1].value;
      TEEndTime->SetNumber(tmax);
    }
    theFitManager->GetFitter()->tmin=TEBeginTime->GetNumber();
    theFitManager->GetFitter()->tmax=TEEndTime->GetNumber();
  };

  void FitDialog::SetEndTime(void)
  {
    double tmin=TEBeginTime->GetNumber();
    double tmax=TEEndTime->GetNumber();
    //modify begin time if tmax < tmin
    if(tmax<tmin)
    {
      data::EnergyCalibrationStorage * EC = ECView->GetEnergyCalibrationStorage();
      tmin = EC->t[0].value;
      TEBeginTime->SetNumber(tmin);
    }
    theFitManager->GetFitter()->tmin=TEBeginTime->GetNumber();
    theFitManager->GetFitter()->tmax=TEEndTime->GetNumber();
  }



  Bool_t FitDialog::ProcessMessage(Long_t msg, Long_t parm1, Long_t parm2)	
  {
    switch( GET_MSG(msg) )
    {
      case kC_COMMAND:
        switch (GET_SUBMSG(msg))
        {
          case kCM_COMBOBOX:
            switch ( parm1 )
            {
              case FUNCTIONS_COMBO:
                std::clog  << LOG(NORMAL_LOG) << "Select fit method: " << FunctionsCombo->GetSelected()<< std::endl;
                UpdateFunctionCombo();
                break;
              case GRAPH_COMBO:
                std::clog  << LOG(NORMAL_LOG) << "Graphic to fit: " << FunctionsCombo->GetSelected()<< std::endl;
                UpdateGraphCombo();
              default:	
                break;
            }
            break; //kCM_COMBOBOX
          case kCM_BUTTON:
            switch(parm1)
            {
              case TB_OK:
                break; // end of B_SCAN
              case TB_RESET:
                ShowParameters(theFitManager->GetFitter()->defpar);
                break; // end of BUTTON_RESET
              case TB_FIT:
                TBFit->SetState(kButtonEngaged);
                Fit();
                TBFit->SetState(kButtonUp);
                break;
              case TB_CLOSE:
                CloseWindow();
                break;
            } //end of switch parm1
            break; //end of kCM_BUTTON
          case kCM_CHECKBUTTON:
            {
              unsigned i = parm1/2;
							if(i<par_vgt.size())
							{
								Parameters_t & wgt = par_vgt[i];
								if(parm1%2)
								{
									switch(wgt.limit->GetState())
									{
										case kButtonUp:
											wgt.upper->SetState(false);
											wgt.lower->SetState(false);
											break;
										case kButtonDown:
											wgt.upper->SetState(true);
											wgt.lower->SetState(true);
											break;
										default:
											break;
									}
								}
							}
            }
            break;
        }
        break;

      case kC_TEXTENTRY:
        switch(GET_SUBMSG(msg))
        {
          case kTE_ENTER:
            switch(parm1)
            {
              case TE_BEGIN_TIME:
                SetBeginTime();
                break;
              case TE_END_TIME:
                SetEndTime();
                break;
              case TE_FORMULA:
                //UpdateFunctionCombo();
                UpdateParIni();
                theFitManager->Fit(TEFormula->GetText(), TEBeginTime->GetIntNumber(), TEEndTime->GetIntNumber());
                break;
            }
            break;
          case kTE_TAB:
            switch(parm1)
            {
              case TE_BEGIN_TIME:
                TEEndTime->GetNumberEntry()->SetFocus();
                break;
              case TE_END_TIME:
                break;
              case TE_FORMULA:
                //UpdateCombo();
                UpdateGraphCombo();
                UpdateParIni();
                theFitManager->Fit(TEFormula->GetText(), TEBeginTime->GetIntNumber(), TEEndTime->GetIntNumber());
                break;
            }
            break;
        }
        break;
    }//GET_MSG
    return kTRUE;
  };


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  FitDialog
   *      Method:  DrawBG
   * Description:  Draw background
   *--------------------------------------------------------------------------------------
   */
  void FitDialog::DrawBG(void)
  {
    std::string graph_name =  theFitManager->fcn_fit_graph_name;
    const unsigned BN=2;
    std::vector <TGraph*> graph(BN); /* Fitfunction graph will created here */
    std::vector <gui::Canvas * > display(BN); /*  display units */
    for(unsigned bunch=0;bunch<BN;bunch++)
    {
      std::string graph_bunch_name = graph_name + "-" + std::to_string(bunch+1);
      display[bunch]=ECView->Add(graph_bunch_name);
    }
    ECView->Draw();
    ECView->Update();
    //vector <data::PhysValue> K(4);
    //K[0]=theFitManager->kT;
    //K[1]=theFitManager->kG;
    //K[2]=theFitManager->kG2;
    //K[3]=theFitManager->kG3;
    double tmin = TEBeginTime->GetIntNumber();
    double tmax = TEEndTime->GetIntNumber();
    for(unsigned bunch=0; bunch<BN;bunch++)
    {
      graph[bunch] = new TGraph;
      unsigned point=0;
      for(unsigned i=0;i<EC->t.size();i++)
      {
        if(find(fit::BadPointList.begin(), fit::BadPointList.end(), i) != fit::BadPointList.end()) continue;
        if(EC->t[i].value < tmin || EC->t[i].value > tmax) continue;
        graph[bunch]->SetPoint(point++,EC->IRatio[bunch][i].value, theFitManager->NormalizedRate(i,bunch,theFitManager->fit_result).value);
      }
      graph[bunch]->SetLineColor(kBlue);
      graph[bunch]->SetLineWidth(3);
      display[bunch]->GetCanvas()->cd();
      graph[bunch]->Draw("l");
    }
  }

   /*--------------------------------------------------------------------------------------
   *       Class:  FitDialog
   *      Method:  DrawBG
   * Description:  Draw background
   *--------------------------------------------------------------------------------------
   */
  void FitDialog::DrawBG2(void)
  {
    std::string graph_name =  theFitManager->fcn_fit_graph_name;
    const unsigned BN=2;
    graph_tmp.resize(0);
    std::vector <gui::Canvas * > display(BN); /*  display units */
    graph_tmp.resize(BN);
    for(unsigned bunch=0;bunch<BN;bunch++)
    {
      std::string graph_bunch_name = graph_name + "-" + std::to_string(bunch+1);
      display[bunch]=ECView->Add(graph_bunch_name);
    }
    ECView->Draw();
    ECView->Update();
    double tmin = TEBeginTime->GetIntNumber();
    double tmax = TEEndTime->GetIntNumber();
    for(unsigned bunch=0; bunch<BN;bunch++)
    {
      graph_tmp[bunch].reset(new TGraph);
      unsigned point=0;
      for(unsigned i=0;i<EC->t.size();i++)
      {
        if(find(fit::BadPointList.begin(), fit::BadPointList.end(), i) != fit::BadPointList.end()) continue;
        if(EC->t[i].value < tmin || EC->t[i].value > tmax) continue;
        double I = EC->I[bunch][i].value;
        double rate= theFitManager->NormRate(i,bunch,theFitManager->fit_result).value;
        graph_tmp[bunch]->SetPoint(point++,I,rate);
      }
      graph_tmp[bunch]->SetLineColor(kBlue);
      graph_tmp[bunch]->SetLineWidth(3);
      display[bunch]->GetCanvas()->cd();
      graph_tmp[bunch]->Draw("l");
      display[bunch]->GetCanvas()->Modified();
      display[bunch]->GetCanvas()->Update();
    }
  }

  void FitDialog::CloseWindow(void)
  {	
    std::clog << LOG(4) << "FitDialog::CloseWindow" << std::endl;
    if(TBFit->GetState() == kButtonUp)
    {
      delete this;
    }
    else
    {
      std::cout << "Cant delete fit dialog, fit in process" << std::endl;
    }
  }

  void FitDialog::Fit(void)
  {
    std::string method_name = COMBO_FUNCTIONS_MAP[FunctionsCombo->GetSelected()];
    std::string display_unit_name = COMBO_GRAPHICS_MAP[FitCombo->GetSelected()];
    fit::Fitter * fitter = theFitManager->FindFitter(method_name);
    if(fitter == nullptr)
    {
      std::cerr << LOG(NORMAL_LOG) << "Unable to find fit method " << method_name << std::endl;
      return;
    }
    SetTimeRange();
    fitter->SetParameters(GetParameters());
    theFitManager->Fit(ECView, method_name, display_unit_name);
    ShowParameters(fitter->minpar);
    if(DrawCheckButton->GetState() == kButtonDown) fitter->Draw(FullDrawCheckButton->GetState() == kButtonDown);
  }


  void FitDialog::Parameters_t::addTo(TGCompositeFrame * f, TGLayoutHints *l)
  {
      f->AddFrame(fix.get(),l);
      f->AddFrame(value.get(),l);
      f->AddFrame(error.get(),l);
      f->AddFrame(limit.get(),l);
      f->AddFrame(lower.get(),l);
      f->AddFrame(upper.get(),l);
  }

  void FitDialog::Parameters_t::removeFrom(TGCompositeFrame *f)
  {
      f->RemoveFrame (fix.get());
      f->RemoveFrame (value.get());
      f->RemoveFrame (error.get());
      f->RemoveFrame (limit.get());
      f->RemoveFrame (lower.get());
      f->RemoveFrame (upper.get());
  }

} //namespace gui
