/*
 * =====================================================================================
 *
 *       Filename:  Graph.cpp
 *
 *    Description:  
 *
 *        Version:  1.0
 *        Created:  11.03.2013 17:54:08
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Institute of Nuclear Physics, Novosibirsk, Russia
 *
 * =====================================================================================
 */


#include "Graph.h"
#include "fit/FitConfig.h"
#include <iostream>
#include <mutex>
#include <thread>

#include <TAxis.h>
extern std::recursive_mutex  RootProcessEventsMutex;

namespace gui
{
  /* #####   Graph CLASS IMPLEMENTATIONS  -  EXPORTED CLASSES   ############################# */
  Graph::Graph(Data *x, Data *y) :
    X(x), Y(y), xData((*X)()), yData((*Y)())
  {
    index_max=-1;
    index_min=-1;
    last_size=0;
    SetName(Y->name.c_str());
    SetTitle(Y->title.c_str());
    SetMarkerStyle(24);
		SetMarkerSize(0.8);
    SetEditable(kFALSE);
    SetDraw(false);
    localX=false;
    localY=false;
//    Update();
  }

  Graph::Graph(data::Array *x, data::Array *y) :
    Graph(new Data(x,"x"), new Data(y,"y"))
  {
    localX=true;
    localY=true;
  }

  Graph::Graph(data::Array *x, Data *y) :
    Graph(new Data(x,"x"), y)
  {
    localX=true;
  }

  Graph::Graph(Data *x, data::Array *y) :
    Graph(x, new Data(y,"y"))
  {
    localY=true;
  }

  Graph::~Graph(void)
  { 
    if(localX) delete  X;
    if(localY) delete  Y;
  }

  void Graph::Update(void)	
  {
		std::clog << LOG(5) << "Graph::Update BEGIN" << std::endl;
		std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
    X->Update();
    Y->Update();
    //set new amount of point
    unsigned new_size = std::min(xData.size(), yData.size());
    if(new_size!=0 && index_max < 0 && index_min < 0)
    {
      index_max=0;
      index_min=0;
    }
    unsigned idx = GetN(); //индекс по записанным в TGraph данным
    std::clog << LOG(6) << "Graph::Update last_size = " << last_size << " new_size = " << new_size  << " idx = " << idx << std::endl;
    for(unsigned i = last_size; i!= new_size; i++)
    {
      if(find(fit::BadPointList.begin(), fit::BadPointList.end(), i) != fit::BadPointList.end()) continue;
      if(yData[i].value > yData[index_max].value) index_max=i;
      if(yData[i].value < yData[index_min].value) index_min=i;
      if(X->show[i] == 0 || Y->show[i] == 0) continue;
      SetPoint(idx, xData[i], yData[i]);
      idx++;
    }
    //TH1F * h = (TH1F*)GetHistogram();
    //SetHistogram(0);
    //delete h;
    last_size = new_size;
    std::clog << LOG(5) << "Graph::Update finish"<< std::endl;
  }

  void Graph::Clear(void)	
  {
    last_size =0;
    Set(0);
    SetMarkerStyle(24);
    SetEditable(kFALSE);
    index_min=-1;
    index_max=-1;
  }

	double Graph::GetMaximum(double n)
	{
		if(n<2) return GetMaximum();
		double m =  *std::max_element(yData.value(yData.size()-n),  yData.value(yData.size()-1));
		double error= *std::max_element(yData.error(yData.size()-n),  yData.error(yData.size()-1));
		return m+error*1.5;
	}
	double Graph::GetMinimum(double n)
	{
		if(n<2) return GetMinimum();
		double m= *std::min_element(yData.value(yData.size()-n),  yData.value(yData.size()-1));
		double error= *std::max_element(yData.error(yData.size()-n),  yData.error(yData.size()-1));
		return m-error*1.5;
	}

	void Graph::SetXRangeUser(double tmin,  double tmax)
	{
		if(!isdraw) return;
		GetXaxis()->SetRangeUser(tmin, tmax);
	}
	void Graph::SetYRangeUser(double tmin,  double tmax)
	{
		if(!isdraw) return;
		GetYaxis()->SetRangeUser(tmin, tmax);
	}
}
