#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/info_parser.hpp>
#include <boost/regex.hpp>
#include <boost/optional.hpp>

#include <TSystem.h>
#include <TGMsgBox.h>
#include <TGFileDialog.h>

#include "Log.h"
#include "PolarimeterConfig.h"
#include "data/EnergyCalibrationStorage.h"
#include "MeasurementManagerDialog.h"
#include "measurement/Manager.h"
#include "EnergyCalibrationView.h"

#include "MainWindow.h"
#include "MeasurementConfigDialog.h"
#include "MeasurementViewDialog.h"
#include "FitDialog.h"
#include "ECStatusDialog.h"
#include "ECReasonDialog.h"
#include "ExpectedEnergyDialog.h"

#include "vepp4/Stap.h"

#include "Event.h"
#include "fit/RootFunctionFitter.h"

extern std::recursive_mutex  RootProcessEventsMutex;

namespace gui
{
  MeasurementManagerDialog::MeasurementManagerDialog(MainWindow * md) 
  {
    std::clog<<LOG(3)<<"MeasurementManagerDialog::MeasurementManagerDialog START"<< std::endl;
    mode_dialog = md; 
    manager = new measurement::Manager(PCFG.bunch_number, PCFG.counter_number);
    EC = dynamic_cast<data::EnergyCalibrationStorage*>(manager->GetStorage("EnergyCalibration"));
		EC->cfg = LoadMeasurementConfig(PCFG.last_good_measurement_config_file.string());
    depolarisator = manager->GetDepolarisator();
    ECView = new EnergyCalibrationView(EC);

    new_point_connection = manager->EC->sig_new_point.connect([this]{ECView->Notify();});
    fitManager = new fit::FitManager(EC);
    isdepolon=false;

		std::clog << LOG(NORMAL_LOG) << "Load measurement config list" << std::endl;
		fMeasurementConfigList  = LoadMeasurementConfigList();
		for(auto & m : fMeasurementConfigList)
		{
			for(auto & p : m.second)
			{
				std::clog << LOG(NORMAL_LOG)<< m.first << " " << p.name << " speed=" << p.depolarizer.speed << " count="<< p.count_time << " attenuation = " << p.depolarizer.attenuation << " range = " << p.depolarizer.range << std::endl;
			}
		}
    calculator.SetTemperatureShiftParameter(PCFG.temperature_drift);
    calculator.SetN(PCFG.points_before_jump);
    std::clog<<LOG(3)<<"MeasurementManagerDialog::MeasurementManagerDialog FINISH" << std::endl; 
  }

  MeasurementManagerDialog::~MeasurementManagerDialog(void)
  {
    std::clog<<LOG(DEBUG_LOG)<<"MeasurementManagerDialog::~MeasurementManagerDialog START" << std::endl;
		new_point_connection.disconnect();
    delete ECView;
    delete fitManager;
    delete manager;
    std::clog<<LOG(DEBUG_LOG)<<"MeasurementManagerDialog::~MeasurementManagerDialog FINISH" << std::endl;
  }

	void MeasurementManagerDialog::Stop(void)
	{
		manager->Stop();
		depolarisator->Stop();
	}

	measurement::EnergyCalibrationConfig  MeasurementManagerDialog::LoadMeasurementConfig(std::string filename)
	{
		measurement::EnergyCalibrationConfig cfg;
		try
		{
			boost::property_tree::ptree pt;
			boost::property_tree::info_parser::read_info(filename, pt);
			pt >> cfg;
		}
		catch (std::runtime_error e)
		{
			std::cerr << LOG(NORMAL_LOG) << "ERROR in reading old energy calibration: " << e.what() << std::endl;
		}
		return cfg;
	}

	void MeasurementManagerDialog::SaveMeasurementConfig(const measurement::EnergyCalibrationConfig & cfg,  std::string filename)
	{
		boost::property_tree::ptree pt;
		pt << cfg; //fill  property tree
		boost::property_tree::info_parser::write_info(filename, pt);
	}


	//read the measurement config from config file
	MeasurementManagerDialog::Mlist_t MeasurementManagerDialog::LoadMeasurementConfigList(void)
	{
		Mlist_t mp;
		boost::property_tree::ptree tree;
		boost::property_tree::info_parser::read_info(PCFG.measurement_config_file.string(), tree);
		auto pt = tree.get_child("EnergyCalibration");
		for(auto & m : pt)
		{
			std::list<measurement::EnergyCalibrationConfig> lst;
			measurement::EnergyCalibrationConfig cfg;
			m.second>>cfg;
			if(cfg.name=="") cfg.name = m.first;
			//int pass=0;
			for(auto & p : m.second)
			{
				if(!p.second.empty())
				{
          measurement::EnergyCalibrationConfig c=cfg;
          c.name = "";
          p.second >> c;
          if(c.name=="") c.name = m.first+"."+p.first;
          lst.push_back(c);
				}
			}
			if(lst.empty()) lst.push_back(cfg);
			mp.push_back(make_pair(m.first, lst));
		}
		return mp;
	}


	int MeasurementManagerDialog::Measure(int i) //start measurement number i
	{
		if(i>=fMeasurementConfigList.size()) throw std::runtime_error("MeasurementManagerDialog::Measure bad measurement number " + to_string(i));
		std::string name  = fMeasurementConfigList[i].first;
		auto lst   = fMeasurementConfigList[i].second; //pickupt some config from list
		bool count_only= name == "TEST" ? true : false;
		manager->EC->ECStorage->isdepolon=true;
		int pass=0;
    int reason=0;
    bool higher_scan = false;
    bool lower_scan = false;
    std::pair<double,double> base_range(std::numeric_limits<double>::max(), std::numeric_limits<double>::min());
    if(name != "TEST")
    {
      new ECReasonDialog(mode_dialog,&reason);
      if(reason<0) return BACK;
    }
		for(const auto cfg: lst) //loop over different passes for calibration
		{
			if(name=="TEST")
			{
				isdepolon=false;
				manager->EC->ECStorage->isdepolon=false;
			}
			else
			{
				isdepolon=true;
				manager->EC->ECStorage->isdepolon=true;
			}
			measurement::EnergyCalibrationConfig last_cfg = LoadMeasurementConfig(PCFG.last_good_measurement_config_file.string());
			measurement::EnergyCalibrationConfig last_bad_cfg = LoadMeasurementConfig(PCFG.last_bad_measurement_config_file.string());
			measurement::EnergyCalibrationConfig current_cfg = cfg;
			current_cfg.run = std::max(last_cfg.run, last_bad_cfg.run);  //takin into account both bad and good runs
			data::GetNMR(&current_cfg);
			data::GetRF(&current_cfg);
			vepp4::Stap v4_stap;
			EC->mode = v4_stap.mode;
			EC->setE = v4_stap.Eset;	
			VeppDuty vduty;
			current_cfg.shiftman=vduty.duty_translit;
			int retval=NONE;
			if(PCFG.restore) goto RESTO1;
			calculator.CalculateNewEnergy(&last_cfg, &current_cfg);
			fitManager->SetFitFunction(current_cfg.fit_method);
E_DIALOG:
			if(name!="TEST") //skip configuration of expected energy
			{
				new ExpectedEnergyDialog(mode_dialog,&last_cfg,&current_cfg,&retval);
				if (retval== BACK) return retval;
			}
			calculator.CalculateEnergyRange(&current_cfg);
      //here save the base scan range
M_DIALOG:
			current_cfg.run++;  //increase the run number
			retval = NONE;
			mode_dialog->Update(); 
			ECView->ShowLastPoints(0);
			new MeasurementConfigDialog(mode_dialog,  &current_cfg,  &retval,  count_only);

      base_range.first  = std::min(base_range.first, std::min(current_cfg.depolarizer.initial, current_cfg.depolarizer.final));
      base_range.second  = std::max(base_range.second, std::max(current_cfg.depolarizer.initial, current_cfg.depolarizer.final));

			if(name == "TEST" && retval == BACK ) return retval;
			if(retval == BACK) goto E_DIALOG;
      current_cfg.reason = reason;
			EC->cfg=current_cfg;  //store current_cfg into EnergyCalibrationStorage
      EC->cfg.type=name;
			EC->shiftman = current_cfg.shiftman;
      pass++;
			EC->cfg.pass=pass;
RESTO1:
			depolarisator->SetConfig(EC->cfg.depolarizer);
			if(PCFG.restore) depolarisator->SetInitialFrequency(EC->DepFr[EC->N-1].value);
			isdepolon=EC->isdepolon;
			manager->Start("EnergyCalibration");
			retval=NONE;
			Loop(mode_dialog->mview);
			retval=mode_dialog->mview->retval;
			if(name == "TEST" && (retval == BACK || retval == NEXT) ) return retval;
			if (retval == BACK) goto M_DIALOG;
FIT_LABEL:
      new FitDialog(mode_dialog, ECView, fitManager,true);
			ECStatus status;
			new ECStatusDialog(mode_dialog,  &status);
      if(status.back_to_fit) goto FIT_LABEL;
			EC->success = status.success;
			EC->cfg.depolarizer.voltage = status.U;
			//EC->cfg.pass=pass;
			EC->cfg.energy=EC->E;
			if(EC->success)
			{
        if(PCFG.debug.skip_data_storing)
        {
          std::clog << LOG(1) << "Skip data storing as defined in command line" << std::endl;
        }
        else
        {
          Save2FS();
          Save2DB();
          if(status.print) Print();
        }
				SaveMeasurementConfig(EC->cfg, PCFG.last_good_measurement_config_file.string());
				current_cfg.date=EC->cfg.date;
				current_cfg.energy=EC->E;
			}
			else
			{
        if(PCFG.debug.skip_data_storing)
        {
          std::clog << LOG(1) << "Skip data storing as defined in command line" << std::endl;
        }
        else
        {
				  SaveBad2FS();
          if(status.print) PrintBad();
        }
				SaveMeasurementConfig(EC->cfg, PCFG.last_bad_measurement_config_file.string());
        int r;
        if(!higher_scan) 
        {
          new TGMsgBox(gClient->GetRoot(),mode_dialog,"Change range", "Try higher energy range?",kMBIconQuestion,kMBYes | kMBNo, &r);
          if( r == kMBYes) 
          {
            current_cfg.depolarizer.initial = base_range.second-0.03;
            current_cfg.depolarizer.final = base_range.second+0.5;
            higher_scan = true;
            goto M_DIALOG;
          }
        }
        if(!lower_scan)
        {
          new TGMsgBox(gClient->GetRoot(),mode_dialog,"Change range", "Try lower energy range?",kMBIconQuestion,kMBYes | kMBNo, &r);
          if( r == kMBYes) 
          {
            current_cfg.depolarizer.initial = base_range.first+0.03;
            current_cfg.depolarizer.final = base_range.first-0.5;
            lower_scan = true;
            goto M_DIALOG;
          }
        }
			}
		}
		return NONE;
	}



  void MeasurementManagerDialog::Loop(MeasurementViewDialog *mview) 
  {
		clog << LOG(5) << "MeasurementManagerDialog::Loop BEGIN " << endl;
    ibn::timer timer;
    struct graph_list_info_t
    {
      std::string name;
      int x,y;
      unsigned wx,wy;
    };
    //std::list<std::string> previouse_graph_list;
    std::list<graph_list_info_t> previouse_graph_list;
    //remember old graph list
    for(auto & d: ECView->displays)
    {
      graph_list_info_t g;
      g.name = d.first;
      //d.second->GetCanvasPar(g.x,g.y,g.wx,g.wy);
      g.wx = d.second->GetWindowWidth();
      g.wy = d.second->GetWindowHeight();
      g.x = d.second->GetWindowTopX();
      g.y = d.second->GetWindowTopY();
      previouse_graph_list.push_back(g);
    }
    ECView->Clear();
    ECView->InitDataList();
    ECView->InitGraphList();
    if( PCFG.gui.remember_graph_list)
    {
      for( auto & g : previouse_graph_list)
      {
        ECView->Add(g.name);
        ECView->GetCanvas(g.name)->SetWindowPosition(g.x,g.y);
        ECView->GetCanvas(g.name)->SetWindowSize(g.wx,g.wy);
      }
    }
    ECView->Add(PCFG.gui.default_graph);
    ECView->Draw();

    int sleep_time = 10000; //mks
    ibn::timer status_timer;
    while(manager->EC->isMeasure())
    {
      usleep(sleep_time);
      {
				std::lock_guard<std::recursive_mutex> lock(RootProcessEventsMutex);
        gSystem->ProcessEvents();
      }
      if(status_timer.elapsed()>1)
      {
        mode_dialog->Update();
        status_timer.restart();
      }

      if(timer.elapsed() >= EC->cfg.count_time*3.93 && isdepolon )
      {		
        isdepolon=false;
        depolarisator->Start();
      }
    }
    clog << LOG(5) << "MeasurementManagerDialog::Loop: FINISH" << endl;
  }


  void MeasurementManagerDialog::Update(void)
  {
    if(ECView->IsUpdatable()) 
    {
      ECView->Update();
      if(PCFG.gui.autofit)
      {
        //fit::RootFunctionFitter * fitter = (fit::RootFunctionFitter*) fitManager->FindFitter("LINEAR_JUMP");
        //fitManager->SetCountTime();
        //string canvas_name="delta-1";
        //Canvas * canvas = ECView->GetCanvas(canvas_name);
        //if(!canvas) return;
        //string graph_name = canvas_name;
        //TGraphErrors * graph = canvas->GetGraph(graph_name);
        //if(graph->GetN()>4)
        //{
        //  fitter->SetData(graph);
        //  fitter->Fit();
        //  fitManager->FillEnergyCalibrations(fitter->minpar.Params(),fitter->minpar.Errors());
        //  canvas->SetMeasured();
        //}
      }
    }
  }


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  MeasurementManagerDialog
   *      Method:  Save2FS
   * Description:  Save data into dir sdir with suffix
   *--------------------------------------------------------------------------------------
   */
  void MeasurementManagerDialog::Save2FS(boost::filesystem::path sdir, const char * suffix="") 
  {
    using namespace boost::filesystem;
    auto choose_file  = [&](path & file) -> boost::optional<path>
    {
      TGFileInfo fi;
      fi.fIniDir    = StrDup(file.parent_path().string().c_str());
      fi.fFilename  = StrDup(file.filename().c_str());
      new TGFileDialog(gClient->GetRoot(), mode_dialog, kFDSave, &fi);
      if(!fi.fFilename) 
      {
        clog << LOG(NORMAL_LOG) << "Cancel saving" << endl;
        return boost::none;
      }
      file=path(fi.fFilename);
      return file;
    };

    auto make_path = [&](path dir, std::string prefix, std::string suffix, std::string extention) -> boost::optional<path>
    {
      int retval;
      if (!exists(dir))
      {
        clog << LOG(NORMAL_LOG) << "No such directory " << dir.string() <<  ". Creating new one... " << endl;
        create_directory(dir);
      }
      path file = dir/(prefix+std::to_string(EC->cfg.run)+suffix+extention);
      if(!is_directory(dir))
      {
        cerr << LOG(NORMAL_LOG) << dir.string() << " Is not directory. Choose file" << endl;
        new TGMsgBox(gClient->GetRoot(), mode_dialog, "Polarimeter",(dir.string() + " is not directory. Manualy choose file?").c_str(), kMBIconQuestion, kMBOk, &retval );
        if(retval==kMBNo) return boost::none;
        return choose_file(file);
      }
      else 
      {
        if(exists(file)) 
        {
          new TGMsgBox(gClient->GetRoot(), mode_dialog, "File exists",("File " + file.string() + " exists. Overwrite?").c_str(), kMBIconQuestion, kMBYes | kMBNo, &retval );
          if(retval==kMBNo) return choose_file(file);
        }
      }
      return file;
    };
    auto pol_file = make_path(sdir,"","",".pol");
    if(pol_file)
    {
      clog << LOG(NORMAL_LOG) << "Saving data into file " << (*pol_file).string() << endl;
      EC->Save2FSSerial(*pol_file);
    }


    auto xml_file = make_path(sdir/"xml","","",".xml");

    if(xml_file)
    {
      clog << LOG(NORMAL_LOG) << "Saving data into XML file " << (*xml_file).string() << endl;
      EC->Save2XML(xml_file.get().string());
    }
    std::cout << "pol_file = " << (*pol_file).string() << "   xml_file = " << (*xml_file).string() << std::endl;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  MeasurementManagerDialog
   *      Method:  Save2FS
   * Description:  Save GOOD data
   *--------------------------------------------------------------------------------------
   */
  void MeasurementManagerDialog::Save2FS()
  {
    clog << LOG(NORMAL_LOG) << "Saving GOOD data" << endl;
    Save2FS(PCFG.save_dir);
    clog << LOG(NORMAL_LOG) << "Saving data for UTUG"<< endl;
    EC->Save4Utug();
    //clog << LOG(NORMAL_LOG) << "Save data for Anton Bogomyagkov to /home/vepp4/kadrs/v4-energy-calibration.pol"<< endl;
    //EC->Save2kadrs();
    clog << LOG(NORMAL_LOG) << "Saving DONE" << endl;
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  MeasurementManagerDialog
   *      Method:  Save2FS
   * Description:  Save BAD data
   *--------------------------------------------------------------------------------------
   */
  void MeasurementManagerDialog::SaveBad2FS(void) 
  {
    clog << LOG(NORMAL_LOG) << "Saving BAD data" << endl;
    char time_string[1024];
    std::string prefix = "-bad-";
    strftime(time_string,1024,"%F-%T",localtime(&EC->begin_time.tv_sec));
    prefix+=string(time_string);
    boost::filesystem::path sdir(PCFG.save_dir/"bad");
    Save2FS(PCFG.save_dir/"bad", prefix.c_str());
  }

  void MeasurementManagerDialog::Save2DB(void)  
  {
    std::clog<<LOG(DEBUG_LOG)<<"MeasurementManagerDialog::Save2DB BEGIN" << std::endl;
    EC->Save2DB("vepp4pol");
  }


  /*
   *--------------------------------------------------------------------------------------
   *       Class:  MeasurementManagerDialog
   *      Method:  Print
   * Description:  Print good data into file and to printer
   *--------------------------------------------------------------------------------------
   */
  void MeasurementManagerDialog::Print(void)
  {
    using namespace boost::filesystem;
    if (!exists(PCFG.print_dir))
    {
      clog << LOG(1) << "No such directory " << PCFG.print_dir.string() <<  ". Creating new one " << endl;
      create_directory(PCFG.print_dir);
    }
    ECView->Print(PCFG.print_dir);
  }

  /*
   *--------------------------------------------------------------------------------------
   *       Class:  MeasurementManagerDialog
   *      Method:  PrintBad
   * Description:  Print bad data into file and to printer
   *--------------------------------------------------------------------------------------
   */
  void MeasurementManagerDialog::PrintBad(void)    
  {
    using namespace boost::filesystem;
    path pbad(PCFG.print_dir/boost::filesystem::path("bad"));
    //check existing of good print directory
    if (!exists(PCFG.print_dir))
    {
      clog << LOG(1) << "No such directory " << PCFG.print_dir.string() <<  ". Creating new one " << endl;
      create_directory(PCFG.print_dir);
    }
    if(!is_directory(PCFG.print_dir))
    {
      cerr << LOG(1) << PCFG.print_dir.string() << " Is not directory. Print nothing." << endl;
      return;
    }
    //check existing of bad print directory
    if (!exists(pbad))
    {
      clog << LOG(1) << "No such directory " << pbad.string() <<  ". Creating new one " << endl;
      create_directory(pbad);
    }
    ECView->Print(pbad);
  };
}

