/*
 * =====================================================================================
 *
 *       Filename:  ScanConfigFrame.cpp
 *
 *    Description:  Implementation fo scan config frame dialog
 *
 *        Version:  1.0
 *        Created:  07/09/2010 05:02:13 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */

#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <bitset>

#include <TGButton.h>
#include <TGNumberEntry.h>
#include <TGLabel.h>
#include <TGMenu.h>
#include <TSystem.h>

#include "Log.h"
#include "data/Calculator.h"
#include "ScanConfigFrame.h"
#include "measurement/DepolarizerConfig.h"


using namespace std;

namespace gui
{
  class ScanConfigFrame::Item
  {
    public:
      string name; //the name of the item
      bool  isdiff; // is it speed or step
      bool  isrange; //is this a range data;
      double energy;//MeV?
      double frequency; //kHz?
      measurement::DepolarizerConfig * cfg;
      mutable auto_ptr<TGHorizontalFrame> frame; //frame where item is located
      mutable auto_ptr<TGLabel> label; //label
      mutable auto_ptr<TGNumberEntry> enrg_ne; //energy text entry
      mutable auto_ptr<TGNumberEntry> freq_ne; //energy text entry
      Item(void);
      Item(const std::string & s );
      Item( const Item & item);
      void set_source(double * s);//set source data
      void update_source(void); //update source data
      void read_source(void); //read data from source and fill the number entries
      void read_energy(void); //read enregy from number entry and fill frequency data
      void read_frequency(void); //read frequency from numbe entry and fill energy data
			void set_energy_scale(double scale) { ESCALE=scale; }
			void set_frequency_scale(double scale) { FSCALE=scale; }
    private:
      double e2f(double energy); // energy to frequency convetion
      double f2e(double frequency); // frequency to energy conversion
      double de2f(double energy); // energy step or speed to frequency step or speed convertion
      double df2e(double frequency); //frequency step or speed to energy step or speed conversion
      double *source; //usualy this is the pointer to depolarizer config member.
      double ESCALE;
      double FSCALE;
      void Init(void);
  };

  /*  Widget IDs */
  enum
  {
    FREQLOCK_CB,
    SCANDIRECTION_CB,
    ATTENUATION_TE,
    HARMNUM_TE, 
		MEV_RB,  //RadioButton to check the MeV
		KEV_RB, 
		KHZ_CB, 
		UP_RB, 
		DOWN_RB
  };

  enum TextEntryType
  {
    SIGNLE_TE_TYPE,
    ENERGY_TE_TYPE,
    FREQ_TE_TYPE
  };

  const int ENERGY_ID=0;
  const int FREQ_ID=16;
  const int SNGL_ID=32;

  TextEntryType TEType(int parm)
  {
    if(ENERGY_ID <= parm && parm <FREQ_ID) return ENERGY_TE_TYPE;
    if(FREQ_ID   <= parm && parm < SNGL_ID) return FREQ_TE_TYPE;
    if(SNGL_ID   <= parm) return SIGNLE_TE_TYPE;
    return SIGNLE_TE_TYPE;
  }

  ScanConfigFrame::ScanConfigFrame(const TGWindow *w, measurement::DepolarizerConfig & depcfg,  int opts) :
    TGVerticalFrame(w,312,425),
    cfg(depcfg)
  {
    std::clog<<LOG(DEBUG_LOG)<<"ScanConfigFrame::ScanConfigFrame START\n";
		fOptions = opts;
		fEnergyUnits = MeV; //default energy units
    initial_config=cfg;
    HWSetAttenuation=nullptr;
    HWGetAttenuation=nullptr;
    cfg.range = fabs(cfg.final - cfg.initial);
    feItems.push_back(Item("Speed"));
    feItems.back().isdiff=true;
    feItems.back().set_source( & cfg.speed);
    feItems.push_back(Item("Step"));
    feItems.back().set_source( & cfg.step);
    feItems.back().isdiff=true;
		if((fOptions & SKIP_WIDTH) == 0)
		{
			feItems.push_back(Item("Width"));
			feItems.back().set_source( & cfg.width);
		}
    feItems.back().isdiff=true;
    feItems.push_back(Item("Expected"));
    feItems.back().set_source( &cfg.expected);
    feItems.back().isrange=true;
    feItems.push_back(Item("Range"));
    feItems.back().set_source( &cfg.range);
    feItems.back().isdiff=true;
    feItems.back().isrange=true;
    feItems.push_back(Item("Initial"));
    feItems.back().set_source( & cfg.initial);
    feItems.push_back(Item("Final"));
    feItems.back().set_source( & cfg.final);
		if((fOptions & SKIP_CURRENT_FREQUENCY) ==0 )
		{
			feItems.push_back(Item("Current"));
			feItems.back().set_source( & cfg.current);
		}
		if((fOptions & SKIP_MEAUSURED_FREQUENCY) ==0 )
		{
			feItems.push_back(Item("Measured"));
			feItems.back().set_source( & cfg.measured);
		}


    for(unsigned i = 0 ; i < feItems.size(); ++i) Imap[feItems[i].name]  = &feItems[i];

    lf.reset(new TGLayoutHints(kLHintsTop |kLHintsExpandX,2,2,2,2));
    ll.reset(new TGLayoutHints(kLHintsLeft,2,2,2,2));
    lr.reset(new TGLayoutHints(kLHintsRight,2,2,2,2));


    att_hf.reset(new TGHorizontalFrame(this, 100,20));
    AddFrame(att_hf.get(), lf.get());
    att_lbl.reset(new TGLabel(att_hf.get(),"Attenuation, dB"));
    att_ne.reset(new TGNumberEntry(att_hf.get(),cfg.attenuation,12,ATTENUATION_TE+SNGL_ID,TGNumberFormat::kNESInteger));
    att_hf->AddFrame(att_lbl.get(), ll.get());
    att_hf->AddFrame(att_ne.get(),lr.get());
    att_ne->Associate(this);

    harm_hf.reset(new TGHorizontalFrame(this, 100,20));
    AddFrame(harm_hf.get(), lf.get());
    harm_lbl.reset(new TGLabel(harm_hf.get(),"Harmonic number"));
    harm_ne.reset(new TGNumberEntry(harm_hf.get(),cfg.harmonic_number,12,HARMNUM_TE+SNGL_ID,TGNumberFormat::kNESInteger));
    harm_hf->AddFrame(harm_lbl.get(), ll.get());
    harm_hf->AddFrame(harm_ne.get(),lr.get());
    harm_ne->Associate(this);


    tf.reset(new TGHorizontalFrame(this, 100,20));
    AddFrame(tf.get(),lf.get());

		MeV_rb.reset(new TGRadioButton(tf.get(), "MeV", MEV_RB));
		keV_rb.reset(new TGRadioButton(tf.get(), "keV", KEV_RB));
		kHz_cb.reset(new TGCheckButton(tf.get(), "kHz  ", KHZ_CB));
    down_rb.reset(new TGRadioButton(tf.get(), "Down", DOWN_RB ));
    up_rb.reset(new TGRadioButton(tf.get(), "Up", UP_RB ));
		MeV_rb->Associate(this);
		keV_rb->Associate(this);
		kHz_cb->Associate(this);
		down_rb->Associate(this);
		up_rb->Associate(this);
		tf->AddFrame(MeV_rb.get(),  lr.get());
		tf->AddFrame(keV_rb.get(),  lr.get());
		tf->AddFrame(kHz_cb.get(),  lr.get());

		tf->AddFrame(up_rb.get(),  ll.get());
		tf->AddFrame(down_rb.get(),  ll.get());
		MeV_rb->SetState(kButtonDown);
		keV_rb->SetState(kButtonUp);
		kHz_cb->SetState(kButtonUp);
		UpdateDirectionCB();

		//auto pm = new TGPopupMenu(fClient->GetRoot());
		//auto sb = new TGSplitButton(tf.get(), new TGHotString("split"), pm, true);
		//pm->AddEntry("1", 1);
		//pm->AddEntry("2", 1);
		//pm->AddEntry("3", 1);
		//sb->Associate(this);
		//tf->AddFrame(sb, lr.get());



    for(unsigned i = 0; i< feItems.size(); ++i)
    {
      Item & fe =feItems[i];
      fe.cfg = &cfg;
			fe.set_energy_scale(fEnergyUnits);
      fe.frame.reset(new TGHorizontalFrame(this,100,200));
      fe.label.reset(new TGLabel(fe.frame.get(), fe.name.c_str()));
      fe.enrg_ne.reset(new TGNumberEntry(fe.frame.get(),0,12,i+ENERGY_ID,TGNumberFormat::kNESReal));
      fe.freq_ne.reset(new TGNumberEntry(fe.frame.get(),0,12,i+FREQ_ID,TGNumberFormat::kNESReal));
      fe.read_source();
      fe.enrg_ne->Associate(this);
      fe.freq_ne->Associate(this);
      fe.frame->AddFrame(fe.label.get(), ll.get());
      fe.frame->AddFrame(fe.enrg_ne.get(), lr.get());
      fe.frame->AddFrame(fe.freq_ne.get(), lr.get());
      fe.freq_ne->SetState(kFALSE);
      AddFrame(fe.frame.get(),lf.get());
			if(fe.name=="Width")
			{
				fe.enrg_ne->SetState(kFALSE);
			}
    }
		//Resize(GetDefaultSize());
    std::clog<<LOG(3)<<"ScanConfigFrame::ScanConfigFrame FINISH\n";
  }

  ScanConfigFrame::~ScanConfigFrame(void)
  {
  }

  Bool_t ScanConfigFrame::ProcessMessage(Long_t msg, Long_t parm1, Long_t parm2)
	{
		int idx=0;
		bool chfoc=true;
		switch( GET_MSG(msg) )
		{
			case kC_TEXTENTRY:
				switch(GET_SUBMSG(msg))
				{
					case kTE_TEXTCHANGED:
						chfoc=false;
						switch(parm1)
						{
							case ATTENUATION_TE+SNGL_ID:
								cfg.attenuation = att_ne->GetIntNumber();
								//att_ne->GetNumberEntry()->SetBackgroundColor(0xe0e0e0);
								//att_ne->GetNumberEntry()->SetBackgroundColor(0xffcccc);
								break;
							case HARMNUM_TE+SNGL_ID:
								cfg.harmonic_number = harm_ne->GetIntNumber();
								Update();
								break;
						}
						break;
					case kTE_ENTER:
						switch(parm1)
						{
							case ATTENUATION_TE+SNGL_ID:
								//att_ne->GetNumberEntry()->SetBackgroundColor(0xffffff);
								cfg.attenuation = att_ne->GetIntNumber();
								clog << LOG(1) << "Set attenuation: " << cfg.attenuation << endl;
								if(HWSetAttenuation && HWGetAttenuation)
								{
									att_ne->GetNumberEntry()->SetBackgroundColor(0xccffcc);
									HWSetAttenuation(cfg.attenuation);
									att_ne->GetNumberEntry()->SetBackgroundColor(0xffffff);
									clog << LOG(1) << "Set hw attenuation: " << HWGetAttenuation() << endl;
								}
								if(chfoc)harm_ne->GetNumberEntry()->SetFocus();
								break;
							case HARMNUM_TE+SNGL_ID:
								cfg.harmonic_number = harm_ne->GetIntNumber();
								Update();
								clog << LOG(1) << "Set harmonic number: " << cfg.harmonic_number << endl;
								if(chfoc)feItems[0].enrg_ne->GetNumberEntry()->SetFocus();
								break;
							default:
								switch(TEType(parm1))
								{
									case ENERGY_TE_TYPE:
										idx = parm1-ENERGY_ID;
										feItems[idx].read_energy();
										if(chfoc)feItems[((parm1+1)-ENERGY_ID)%feItems.size()].enrg_ne->GetNumberEntry()->SetFocus();
										break;
									case FREQ_TE_TYPE:
										idx = parm1 - FREQ_ID;
										feItems[idx].read_frequency();
										if(chfoc)feItems[((parm1+1)-FREQ_ID)%feItems.size()].freq_ne->GetNumberEntry()->SetFocus();
										break;
									default:
										break;
								}
								feItems[idx].update_source();
								if(feItems[idx].isrange)
								{
									double sign = cfg.initial < cfg.final ? 1 : -1;
									cfg.initial =cfg.expected-sign*fabs(cfg.range)/2.;
									cfg.final = cfg.expected+sign*fabs(cfg.range)/2.;
									for(unsigned i = 0 ;i<feItems.size();++i)
										if(feItems[i].name =="Initial" || feItems[i].name == "Final")
											feItems[i].read_source();
								}
								else 
								{
									cfg.expected = (cfg.initial + cfg.final)/2.;
									cfg.range = fabs(cfg.initial - cfg.final);
									for(unsigned i = 0 ;i<feItems.size();++i)
										if(feItems[i].name =="Expected" || feItems[i].name == "Range")
											feItems[i].read_source();
								}
								UpdateDirectionCB();
								break;
						}
						break;
					case kTE_TAB:
						{ 
							switch(parm1)
							{
								case ATTENUATION_TE+SNGL_ID:
									harm_ne->GetNumberEntry()->SetFocus();
									break;
								case HARMNUM_TE+SNGL_ID:
									feItems[0].enrg_ne->GetNumberEntry()->SetFocus();
									break;
								default:
									switch(TEType(parm1))
									{
										case ENERGY_TE_TYPE:
											feItems[((parm1+1)-ENERGY_ID)%feItems.size()].enrg_ne->GetNumberEntry()->SetFocus();
											break;
										case FREQ_TE_TYPE:
											feItems[((parm1+1)-FREQ_ID)%feItems.size()].freq_ne->GetNumberEntry()->SetFocus();
											break;
										default:
											break;
									}
									break;
							}
							break; //end of kTE_TAB
						}
				} // end of swith get sub messg
				break;
			case kC_COMMAND:
				switch(GET_SUBMSG(msg))
				{
					case kCM_CHECKBUTTON:
						switch(parm1)
						{
							case KHZ_CB:
								if(kHz_cb->IsOn())
									UnlockFrequency();
								else
									LockFrequency();
								break;
						}
						break;
					case kCM_RADIOBUTTON:
						switch (parm1)
						{
							case MEV_RB:
								keV_rb->SetState(kButtonUp);
								fEnergyUnits=MeV;
								SetEnergyScale(MeV);
								break;
							case KEV_RB:
								MeV_rb->SetState(kButtonUp);
								fEnergyUnits=keV;
								SetEnergyScale(keV);
								break;
							case UP_RB:
								if(cfg.initial > cfg.final) swap(cfg.initial, cfg.final);
								Imap["Initial"]->read_source();
								Imap["Final"]->read_source();
								UpdateDirectionCB();
								break;
							case DOWN_RB:
								if(cfg.initial < cfg.final) swap(cfg.initial, cfg.final);
								Imap["Initial"]->read_source();
								Imap["Final"]->read_source();
								UpdateDirectionCB();
								break;
						}
						break;

				}

				break;
		}
		return kTRUE;
	}

  void ScanConfigFrame::Update(void)
  {
    for(unsigned i =0; i< feItems.size(); ++i) feItems[i].read_source();
    att_ne->SetIntNumber(unsigned(cfg.attenuation));
    harm_ne->SetIntNumber(cfg.harmonic_number);
		//Erange=fabs(cfg.final-cfg.initial);
  }

  void ScanConfigFrame::Reset(void)
  {
    cfg=initial_config;
    Update();
  }

  void ScanConfigFrame::SetAttCallBack(void(*hw_set_fun)(double), double(hw_get_fun)(void))
  {
    HWSetAttenuation = hw_set_fun;
    HWGetAttenuation = hw_get_fun;
  }

  void ScanConfigFrame::UpdateDirectionCB(void)
  {
		if(cfg.final > cfg.initial) 
		{
			up_rb->SetState(kButtonDown);
			down_rb->SetState(kButtonUp);
		}
		else
		{
			up_rb->SetState(kButtonUp);
			down_rb->SetState(kButtonDown);
		}
		if(fOptions & DIRECTION_SWAP_DISABLE)
		{
			up_rb->SetState(kButtonDisabled);
			down_rb->SetState(kButtonDisabled);
		}
  }


	void ScanConfigFrame::SetEnergyScale(double scale)
	{
		for(auto & item : feItems)
		{
			item.set_energy_scale(scale);
		}
		Update();
	}


  ScanConfigFrame::Item::Item(const std::string & s ) : name(s)
  {
    Init();
  }

  ScanConfigFrame::Item::Item(void)
  {
    Init();
  } 

  void ScanConfigFrame::Item::Init(void)
  {
    isdiff = false;
    isrange = false;
    source =0;
    ESCALE=1; //MeV
    FSCALE=1e3; //kHz
  }

  ScanConfigFrame::Item::Item( const Item & item) : 
    name(item.name), frame(item.frame), label(item.label), enrg_ne (item.enrg_ne), freq_ne(item.freq_ne)
  {
    source = item.source;
    energy= item.energy;
    frequency = item.frequency;
    ESCALE = item.ESCALE;
    FSCALE = item.FSCALE;
    isdiff = item.isdiff;
    isrange = item.isrange;
  }

  void ScanConfigFrame::Item::update_source(void)
  {
    *source = energy;
    string unit=" MeV";
    if(name == "Speed") unit = " MeV/sec";
    clog << LOG(1) << "Set energy" << name << " to " <<  energy << unit << endl;
  }

  void ScanConfigFrame::Item::read_source(void)
  {
    energy = *source;
    frequency = isdiff ? de2f(energy) : e2f(energy);
		//if(!isdiff)
		//{
		//	if (fabs(frequency) < 1e-10) frequency=0;
		//	if (fabs(energy) <1e-10) energy =0;
		//}
    enrg_ne->SetNumber(energy/ESCALE);
    freq_ne->SetNumber(frequency/FSCALE);
  }
  void ScanConfigFrame::Item::read_energy(void)
  {
    energy =  enrg_ne->GetNumber()*ESCALE;
    frequency = isdiff ? de2f(energy) : e2f(energy);
    if(isrange) frequency=fabs(frequency);
    freq_ne->SetNumber(frequency/FSCALE);
  }
  void ScanConfigFrame::Item::read_frequency(void)
  {
    frequency = freq_ne->GetNumber()*FSCALE;
    energy = isdiff ? df2e(frequency) : f2e(frequency);
    if(isrange) energy = fabs(energy);
    enrg_ne->SetNumber(energy/ESCALE);
  }
  void ScanConfigFrame::Item::set_source(double * s) 
  {
    source = s;
  }

  double ScanConfigFrame::Item::e2f(double energy)
  {
    return data::energy2frequency(cfg, energy);
  }

  double ScanConfigFrame::Item::f2e(double frequency)
  {
    return data::frequency2energy(cfg, frequency);
  }

  double ScanConfigFrame::Item::de2f(double energy_step)
  {
    return data::denergy2frequency(cfg, energy_step);
  }

  double ScanConfigFrame::Item::df2e(double frequency_step)
  {
    return data::dfrequency2energy(cfg, frequency_step);
  }

	void ScanConfigFrame::LockFrequency(void)
	{
		for(auto & item : feItems) item.freq_ne->SetState(kFALSE);
	}

	void ScanConfigFrame::LockEnergy(void)
	{
		for(auto & item : feItems) item.enrg_ne->SetState(kFALSE);
	}

	void ScanConfigFrame::UnlockFrequency(void)
	{
		for(auto & item : feItems) item.freq_ne->SetState(kTRUE);
	}

	void ScanConfigFrame::UnlockEnergy(void)
	{
		for(auto & item : feItems) item.enrg_ne->SetState(kTRUE);
	}

	void ScanConfigFrame::Lock(void)
	{
		LockFrequency();
		LockEnergy();
		//freqlock_cb->SetState(kButtonDisabled);
		kHz_cb->SetState(kButtonDisabled);
		//scandir_cb->SetState(kButtonDisabled);
		harm_ne->SetState(kFALSE);
		Unlock("Final");
	}

	void ScanConfigFrame::Unlock(void)
	{
		UnlockFrequency();
		UnlockEnergy();
	}

	void ScanConfigFrame::Lock(std::string name)
	{
		for(auto & item: feItems)
		{
			if(item.name == name)
			{
				item.enrg_ne->SetState(kFALSE);
			}
		}
	}

	void ScanConfigFrame::Unlock(std::string name)
	{
		for(auto & item: feItems)
		{
			if(item.name == name)
			{
				item.enrg_ne->SetState(kTRUE);
			}
		}
	}

}
