/*
 * =====================================================================================
 *
 *       Filename:  CurrentMeasurement.cpp
 *
 *    Description:  CurrentMeasurement
 *
 *        Version:  1.0
 *        Created:  06/01/2010 04:20:32 AM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */
#include <fstream>
#include <sstream>
#include <algorithm>
#include <map>

#include "ibn/timer.h"
#include "CurrentMeasurement.h"
#include "Log.h"
#include "PolarimeterConfig.h"

#include <thread>
#include <mutex>

using namespace std;
namespace measurement
{
  CurrentMeasurement::CurrentMeasurement(void) 
  {
    Ia.resize(4); /*  average current */
    Ic.resize(4);
    A.resize(Ia.size()); /*  four bunch */
    ismeasure=false;
  };

  CurrentMeasurement::~CurrentMeasurement(void)
  {
    Stop();
  }

  void CurrentMeasurement::Start(void)
  {
    if(!fThread.joinable())
    {
      fThread  = std::thread([this]{ old_measure();} );
    }
  }

  void CurrentMeasurement::Stop(void)
  {
    clog << LOG(DEBUG_LOG) << "CurrentMeasurement::Stop BEGIN" <<endl;
    ismeasure=false;
    if(fThread.joinable())
    {
      fThread.join();
    }
    clog << LOG(DEBUG_LOG) << "CurrentMeasurement::Stop FINISH" <<endl;
  }

  bool CurrentMeasurement::extract_current(string name, const map<string, vector<double> > &dm)
  {
    vector <double> v(4);
    map<string, vector<double> >::const_iterator pv = dm.find(name);
    if(pv==dm.end()) return false;
    v[0]=pv->second[8];
    v[1]=pv->second[10];
    v[2]=pv->second[9];
    v[3]=pv->second[11];
    fill(v);
    return true;
  }

  void CurrentMeasurement::measure(void)
  {
    ismeasure=true;
    while(ismeasure)
    {
      if(PCFG.name=="vepp4") 
      {
        std::ifstream file(PCFG.vepp4.orbit_file.string().c_str());
        if(!file)
        {
          cerr << LOG(1) << "ERROR: Unable to open file: " <<  PCFG.vepp4.orbit_file.string() << endl;
        }
        file.ignore(65535,'\n'); /*  ignore date */
        std::map < std::string, std::vector <double> > dm; //data map
        string line;
        while(getline(file,line))
        { 
          string name,data_str;
					std::istringstream is(line);
          is >> name;
          vector<double> v; 
          while(is>>data_str)
          {
            if(data_str.find('#')==data_str.npos) v.push_back(std::stod(data_str));
            else v.push_back(-99999);
          }
          dm[name] = v;
        }
        file.close(); /* no need file anymore */
        vector <double> v(4);
        bool status=false;
        status = status || extract_current("SEP0",dm);
        //status = status || extract_current("NEP0",dm);
        status = status || extract_current("NEP5",dm);
        if(!status)
        {
          cerr << LOG(1) << "WARNING: Unable to read current from file: " << PCFG.vepp4.orbit_file.string() << " try read old." << endl;
          read_old();
        }
      }
      if(PCFG.name=="vepp3")
      {
          read_vepp3_current();
      }
      ibn::sleep(MEASUREMENT_TIME,&ismeasure,REACTION_TIME);
    }
  }

  void CurrentMeasurement::old_measure(void)
  {
    ismeasure=true;
    while(ismeasure)
    {
      read_old();
      ibn::sleep(MEASUREMENT_TIME,&ismeasure,REACTION_TIME);/* reaction time is 10msec */
    }
  }

  void CurrentMeasurement::read_old(void)
  {
    //const char * fname ="/home/vepp4/kadrs/stap.dat";
		std::cerr << LOG(DEBUG_LOG) << "CurrentMeasurement::read_old BEGIN" << std::endl; 
    ifstream file(PCFG.vepp4.stap_file.string().c_str());
    //ifstream file("home/vepp4/kadrs/stap.dat");
    if(!file)
    {
      cerr << LOG(3) << "ERROR: Unable to open file: " << PCFG.vepp4.stap_file.string()  << endl;
    }
    for(unsigned line=0;line<45;line++) file.ignore(1024,'\n');
    vector <double> I(Ia.size());
    for(unsigned i=0;i<I.size();i++)
    {
      file >> I[i]; file.ignore(1024,'\n');
      I[i]=I[i]/1000.;
    }
    file.close(); /* no need file anymore */
    fill(I);
		std::cerr << LOG(DEBUG_LOG) << "CurrentMeasurement::read_old END" << std::endl; 
  }


  void CurrentMeasurement::read_vepp3_current(void)
  {
    //const char * fname ="/home/vepp4/kadrs/stap.dat";
    ifstream file(PCFG.vepp4.stap_file.string().c_str());
    if(!file)
    {
      cerr << "ERROR: Unable to open file: " << PCFG.vepp4.stap_file.string()  << endl;
    }
    for(unsigned line=0;line<156;line++) file.ignore(1024,'\n');
    vector <double> I(Ia.size());
    for(unsigned i=0;i<I.size();i++)
    {
      file >> I[i];
      file.ignore(1024,'\n');
      I[i]/=1000.;
    }
    file.close(); /* no need file anymore */
    fill(I);
  }

  void CurrentMeasurement::fill(const vector<double> & I)
  {
		  std::lock_guard<std::mutex> lock(mutex);
      size_t size = std::min(I.size(), Ic.size());
      for(unsigned bunch=0;bunch<size;bunch++) 
      {
        Ic[bunch]=fabs(I[bunch])*1000;
        A[bunch].add(Ic[bunch]);
      }
  }

  data::PhysValue  CurrentMeasurement::GetCurrent(unsigned bunch)
  {
    return Ia[bunch];
  }

  double CurrentMeasurement::GetInstantCurrent(unsigned bunch)
  {
    return Ic[bunch];
  }
  
  void CurrentMeasurement::MakeNewPoint(void)
  {
		std::lock_guard<std::mutex> lock(mutex);
    for(unsigned bunch=0;bunch<Ia.size();bunch++) 
    {
      Ia[bunch]=data::PhysValue(A[bunch].average(), max(A[bunch].sigma_average(),0.5/sqrt(1.0*A[bunch].size())));
    }
    /* reset averagers */
    for(unsigned i=0;i<A.size();i++) A[i].reset();
  }
};

