/*
 * =====================================================================================
 *
 *       Filename:  DepolarizerConfig.h
 *
 *    Description:  Deporizer Configuration
 *
 *        Version:  1.0
 *        Created:  03/28/2008 02:41:18 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physics
 *
 * =====================================================================================
 */

#ifndef  IBN_DEPOLARIZER_CONFIG_INC
#define  IBN_DEPOLARIZER_CONFIG_INC

#include <iostream>
#include <sstream>
#include <iomanip>
#include "Config.h"
#include "data/Data.h"
#include "Log.h"


#include "PhysConst.h"
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/access.hpp>
#include "data/Storage.h"

using namespace boost::serialization;
namespace measurement
{
	class DepolarizerConfig  : public Config
	{
		enum
		{
			EGEN  =  1, 
			PGEN  =  2, 
			EPGEN =  3, 
			SEL   =  4, 
			ESEL  =  8,  //time selective for electrons only
			PSEL  = 16,  //time selective for positrons positrons
		};
		public:
			~DepolarizerConfig(void){}
			std::string generator_name; //generator name
			int generator_type; //generator type
			double scan_time;
			double attenuation;
			double speed; //scan speed in Hz/sec for frequency or Mev/sec for energy units
			double step; //step for scanning
			double initial; //initial scan energy
			double final; //final scan energy
			double width; //depolarizer width (currently not used)
			double expected; //expected energy(frequency) 
			double range; //energy scan range abs(final - initial)
			double current=0; //current energy (frequency) (setup)
			double measured=0;//measured by frequencymeter or the result of energy measurement
			int    direction=-1; //scan direction
			int harmonic_number;
			double RF; //rf frequency
			data::PhysValue revolution_frequency;
			int algorithm;
			int isenergy;
			double voltage;  //2U voltage of the depolarizer
			std::vector<double>  shift;

			/* *****************************************************************************************
			 * calculate frequency (Hz) from energy (MeV) using revolution frequency and harmonic number
			 ******************************************************************************************/
			double frequency(double E) const;

			/* *****************************************************************************************
			 * calculate energy (MeV) from  frequency (Hz) using revolution frequency and harmonic number
			 ******************************************************************************************/
			double energy(double f) const;

			/* *****************************************************************************************
			 * calculate step frequency (Hz) from step energy (MeV) using revolution frequency and harmonic number
			 ******************************************************************************************/
			double delta_frequency(double dE) const;

			/* *****************************************************************************************
			 * calculate energy step (MeV) from  frequency step(Hz) using revolution frequency and harmonic number
			 ******************************************************************************************/
			double delta_energy(double df) const;



		private:
			friend class boost::serialization::access;
			template<class Archive>
				void serialize(Archive & ar,  const unsigned int version)
				{
					std::clog << LOG(DEBUG_LOG) << "DepolarizerConfig::serialize  version=" << version << std::endl;
					ar & BOOST_SERIALIZATION_NVP(initial);
					std::clog << LOG(DEBUG_LOG+1) << "  initial = " << initial << std::endl;
					ar & BOOST_SERIALIZATION_NVP(final);
					std::clog << LOG(DEBUG_LOG+1) << "  final = " << final << std::endl;
					ar & BOOST_SERIALIZATION_NVP(speed);
					std::clog << LOG(DEBUG_LOG+1) << "  speed = " << speed << std::endl;
					ar & BOOST_SERIALIZATION_NVP(scan_time);
					std::clog << LOG(DEBUG_LOG+1) << "  scan_time = " << scan_time << std::endl;
					ar & BOOST_SERIALIZATION_NVP(step);
					std::clog << LOG(DEBUG_LOG+1) << "  step = " << step << std::endl;
					ar & BOOST_SERIALIZATION_NVP(attenuation);
					std::clog << LOG(DEBUG_LOG+1) << "  attenuation = " << attenuation << std::endl;
					ar & BOOST_SERIALIZATION_NVP(RF);
					std::clog << LOG(DEBUG_LOG+1) << "  RF = " << RF << std::endl;
					ar & BOOST_SERIALIZATION_NVP(revolution_frequency);
					std::clog << LOG(DEBUG_LOG+1) << "  revolution_frequency = " << revolution_frequency << std::endl;
					ar & BOOST_SERIALIZATION_NVP(harmonic_number);
					std::clog << LOG(DEBUG_LOG+1) << "  harmonic_number = " << harmonic_number << std::endl;
					ar & BOOST_SERIALIZATION_NVP(isenergy);
					std::clog << LOG(DEBUG_LOG+1) << "  isenergy = " << isenergy << std::endl;
					if(version<11) return;
					ar & BOOST_SERIALIZATION_NVP(algorithm);
					std::clog << LOG(DEBUG_LOG+1) << "  algorithm = " << algorithm << std::endl;
					ar & BOOST_SERIALIZATION_NVP(voltage);
					std::clog << LOG(DEBUG_LOG+1) << "  voltage = " << voltage << std::endl;
					ar & BOOST_SERIALIZATION_NVP(expected);
					std::clog << LOG(DEBUG_LOG+1) << "  expected = " << expected << std::endl;
					ar & BOOST_SERIALIZATION_NVP(measured);
					std::clog << LOG(DEBUG_LOG+1) << "  measured = " << measured << std::endl;
					ar & BOOST_SERIALIZATION_NVP(direction);
					std::clog << LOG(DEBUG_LOG+1) << "  direction = " << direction << std::endl;
					ar & BOOST_SERIALIZATION_NVP(width);
					std::clog << LOG(DEBUG_LOG+1) << "  width = " << width << std::endl;
					ar & BOOST_SERIALIZATION_NVP(generator_name);
					std::clog << LOG(DEBUG_LOG+1) << "  generator_name = " << generator_name << std::endl;
					ar & BOOST_SERIALIZATION_NVP(generator_type);
					std::clog << LOG(DEBUG_LOG+1) << "  generator_type = " << generator_type << std::endl;
				}
	}; 

	inline std::ostream & operator<<(std::ostream & os, const measurement::DepolarizerConfig& dcfg)
	{
		int w=15;
		os << std::left; //выравнивание по левому краю, так удобнее
		os << std::setw(w) <<  dcfg.attenuation << " # dB, attenuation\n";
		os << std::setw(w) <<  dcfg.speed << " # Hz, frequency speed \n";
		os << std::setw(w) <<  dcfg.step << " # Hz, frequency step\n";
		os << std::setw(w) <<  dcfg.width << " # Hz, frequency width\n";
		os << std::setw(w) <<  dcfg.scan_time << " # sec, scan time\n";
		os << std::setw(w) <<  dcfg.initial << " # Hz, initial frequency\n";
		os << std::setw(w) <<  dcfg.final << " # Hz, final frequency\n";
		os << std::setw(w) <<  dcfg.speed << " # Hz, frequency speed\n";
		os << std::setw(w) <<  dcfg.harmonic_number << " # harmonic number\n";
		os << std::setw(w) <<  dcfg.algorithm << " # scan algorithm \n";
		os << std::setw(w) <<  dcfg.voltage << " # 2U,  V \n";
		os << std::setw(w) <<  dcfg.expected<< " # Expected energy,  MeV \n";
		os << std::setw(w) <<  dcfg.measured<< " # measured energy,  MeV \n";
		os << std::setw(w) <<  dcfg.range<< " # range energy,  MeV \n";
		return os;
	}

	inline std::istream & operator>>(std::istream & is, measurement::DepolarizerConfig & dcfg)
	{
		is >> dcfg.attenuation ; is.ignore(1024,'\n');
		is >> dcfg.step ; is.ignore(1024,'\n');
		is >> dcfg.scan_time ; is.ignore(1024,'\n');
		is >> dcfg.initial ; is.ignore(1024,'\n');
		is >> dcfg.final ; is.ignore(1024,'\n');
		is >> dcfg.speed ; is.ignore(1024,'\n');
		is >> dcfg.harmonic_number ; is.ignore(1024,'\n');
		is >> dcfg.algorithm;        is.ignore(1024,'\n');
		is >> dcfg.voltage;        is.ignore(1024,'\n');
		is >> dcfg.expected;        is.ignore(1024,'\n');
		is >> dcfg.measured;        is.ignore(1024,'\n');
		is >> dcfg.range;        is.ignore(1024,'\n');
		return is;
	}
}

BOOST_CLASS_VERSION(measurement::DepolarizerConfig ,data::Storage::VERSION);

extern measurement::DepolarizerConfig to_frequency(const measurement::DepolarizerConfig &);
extern measurement::DepolarizerConfig to_energy(const measurement::DepolarizerConfig &);

#endif   /* ----- #ifndef IBN_DEPOLARIZER_CONFIG_INC  ----- */
