#include <unistd.h>
#include <signal.h>

#include <iostream>
#include <string>

#include <boost/asio.hpp>
#include <boost/format.hpp>
#include <TRandom.h>

#include "Log.h"
#include "V4Par.h"
#include "data/EnergyCalibrationStorage.h"
#include "data/Data.h"

#include "camac/ModuleInventory.h"

#include "Manager.h"
#include "EnergyCalibrationCamac.h"
#include "EnergyCalibration.h"
#include "BeamSizeMeasurement.h"
#include "CurrentMeasurement.h"
#include "FrequencyMeasurement.h"
#include "CountMeasurement.h"

using namespace std;

namespace measurement
{

  EnergyCalibration::EnergyCalibration(unsigned nbunch, unsigned ncounter) : 
    numbunch(nbunch), 
    numcounter(ncounter)
  {
    LOG(DEBUG_LOG) << clog << "EnergyCalibration::EnergyCalibration: BEGIN" << endl;
    is_delay_line_measure=false;
    ismeasure = false;
    ispause = false;
    is_new_point = false;

    /* Init data storage */
    ECStorage = new data::EnergyCalibrationStorage(numbunch, numcounter);
    LOG(DEBUG_LOG) << clog << "EnergyCalibration::EnergyCalibration: END" << endl;
  }


  EnergyCalibration::~EnergyCalibration(void)
  {
    clog << LOG(DEBUG_LOG) << "~EnergyCalibration BEGIN" << endl;
		Stop();
    delete ECStorage;
    clog << LOG(DEBUG_LOG) << "~EnergyCalibration FINISH" << endl;
  }

  void EnergyCalibration::Measure(void)
  {
    std::clog << LOG(2) << "EnergyCalibration::CountMeasurement: BEGIN " << hex << std::this_thread::get_id() << dec << std::endl;
    if(!PCFG.restore && boost::filesystem::exists(PCFG.tmp_save_file)) 
    {
      boost::filesystem::remove(PCFG.last_tmp_save_file);
      boost::filesystem::rename(PCFG.tmp_save_file, PCFG.last_tmp_save_file);
    }

    ECStorage->nmr=V4nmr().nmr_data; //save NMR value at the begining of the measurement

    //calculate maximum number of points
    unsigned max_point_number = (unsigned)(EnergyCalibration::MAX_MEASUREMENT_TIME/ECStorage->cfg.count_time);
    unsigned j=0; 
    int n_cur=0; //current point number
    //std::cout.precision(10);
    is_new_point=false;
    for(j=0 ; ismeasure && j<=max_point_number ; ++j) 
    {
      point_timer.restart();
      double delay=j*5;
      if(is_delay_line_measure) ECCamac->SetDelay(delay); //5ns delay for every new point
      fCountMeasurement->Wait();
      if(!ismeasure) break;
      fCurrentMeasurement->MakeNewPoint();
      fBeamSizeMeasurement->MakeNewPoint();
			fFrequencyMeasurement->MakeNewPoint();
      if(!ispause) 
      {
        MakeNewPoint(); 
        ECStorage->SaveCurrent(PCFG.tmp_save_file.string());
        n_cur = ECStorage->N-1;
        double E= ECStorage->DepFr[n_cur].value == 0 ? 0 : ECStorage->frequency2energy(ECStorage->DepFr[n_cur]).value ;
        if(!is_delay_line_measure)
        {
          std::clog << LOG (NORMAL_LOG) << 
					//boost::format("#%5d: %8.2f sec, %11.6f MeV, %+11.3f Hz,  %+11.3f") %  n_cur  % ECStorage->t[n_cur].value % E % ECStorage->DepFr[n_cur].value % ECStorage->MDepFr[n_cur].value;
						boost::format("#%5d: %8.2f sec, %11.6f MeV, %+11.3f Hz,  %+11.3f") %  n_cur  % ECStorage->t[n_cur].value % E % ECCamac->depolarizer->GetFrequency() % ECStorage->MDepFr[n_cur].value
						<< std::endl;
					//unsigned tmp = (unsigned)fabs(ECCamac->depolarizer->GetFrequency());
					//std::cout << std::bitset<32>(tmp) << std::endl;
        }
				else
        {
          std::cout << LOG (NORMAL_LOG) << "# " << n_cur << " : " << 
            setprecision(4) << std::setw(6) << ECStorage->t[n_cur].value<<" sec, " <<
            setprecision(10) << std::setw(10) << delay << " ns\n";
        }
        std::cout << flush;
      } 
      else 
      {
        //ECCamac->GetData();//read counter load from camac
        //ECCamac->GetCoinc(); //Read coincidence from camac
      }
    }
    std::clog << LOG(2) << "EnergyCalibration::CountMeasurement::operator(): END\n";
  }

  //========================================
  inline void EnergyCalibration::Stop(void)
  {
    std::clog << LOG(DEBUG_LOG) << "EnergyCalibration::Stop" << std::endl;
    if(ismeasure) 
    {
      UnPause();
      ismeasure = false;
			//raise(SIGUSR1);
			//ECCamac->StopTimer();
      fCountMeasurement->Notify();
      if(fThread.joinable()) 
      {
        fThread.join();
      }
      is_new_point=false;
    }
    std::clog << LOG(DEBUG_LOG) << "EnergyCalibration::Stop: FINISH" << std::endl;
  }


  //========================================
  inline void EnergyCalibration::Prepare()
  {
    if(!PCFG.restore)
    {
      unsigned NMAX = unsigned(MAX_MEASUREMENT_TIME/ECStorage->cfg.count_time);
      ECStorage->reserve(NMAX);
    }
  }
  //========================================

  //========================================
  inline void EnergyCalibration::Continue(void)
  {
    ismeasure = true;
    ispause = false;
    ECStorage->issave = false;

    if(ECStorage->issave) ECStorage->cfg.run++;
  }
  //========================================



  //========================================
  bool EnergyCalibration::isMeasure(void)
  {
    return ismeasure;
  }	
  //========================================




  //========================================
  inline bool EnergyCalibration::isPause(void)
  {
    return ispause;
  }	
  //========================================





  //========================================
  inline void EnergyCalibration::Pause(void)
  {
    ispause = true;
  }
  //========================================


  //========================================
  inline void EnergyCalibration::UnPause(void)
  {
    ispause = false;
  }
  //========================================

  //Point measurement are done. Now get the data and proceed it
  //========================================
  void EnergyCalibration::MakeNewPoint(void)
  {
		//read CountStorage
    fCountStorage->lock();
		//read the time
    ECStorage->Begin_time = fCountStorage->measurement_begin_time;
		double bt  =  fCountStorage->to_double(fCountStorage->point_begin_time - fCountStorage->measurement_begin_time);
		double et  =  fCountStorage->to_double(fCountStorage->point_end_time - fCountStorage->measurement_begin_time);
		//read the frequency
		double begin_frequency = fCountStorage->point_begin_frequency;
		double end_frequency = fCountStorage->point_end_frequency;
		//read the data
    std::vector< std::vector<Value_t> > Data =  fCountStorage->Count; //Read counte load from camac
    std::vector< std::vector<Value_t> > Coinc = fCountStorage->Coinc; //Read coincidence from camac
		//now unlock storage
    fCountStorage->unlock();
		//calculate point_time and real count time
		double point_time = et*0.5 + bt*0.5;
		double count_time = et-bt; //this count time could be wrong 

		//fill the time
    //ECStorage->t.push_back(point_time,count_time*0.5);
    ECStorage->t.push_back(bt+ECStorage->cfg.count_time*0.5,ECStorage->cfg.count_time*0.5);

		//calculate frequency
		double speed = ECCamac->depolarizer->GetScanSpeed();
		if(begin_frequency == 0 && end_frequency   !=0 ) begin_frequency = end_frequency - speed*ECStorage->cfg.count_time;
		if(  end_frequency == 0 && begin_frequency !=0 ) end_frequency = begin_frequency + speed*ECStorage->cfg.count_time;
		double point_frequency = 0.5*begin_frequency + 0.5*end_frequency;
		
		//fill the setup depolarizer frequency
    ECStorage->DepFr.push_back(point_frequency,0);
    //ECStorage->DepFr.push_back(ECCamac->depolarizer->GetFrequency(),0); //point end frequency

		//fill measured depolarizer frequency
    ECStorage->MDepFr.push_back(fFrequencyMeasurement->GetCurrentFrequency());

		//fill resonator  (181'801'160 Hz)  frequency
    ECStorage->RF.push_back(GetRadioFrequency(),0.);

		//fill counter position
		for(unsigned counter=0;counter<4;counter++)
		{
			ECStorage->CounterPosition[counter].push_back(0, 0);
		};
		std::ifstream motor_file(PCFG.motor_file.string());
		if(!motor_file)
		{
			std::clog << LOG(NORMAL_LOG) <<  "WARNING: Unable to open file: " << PCFG.motor_file.string() << endl;
			for(unsigned counter=4;counter<numcounter;counter++)
			{
				ECStorage->CounterPosition[counter].push_back(0, 0);
			};
		}
		else
		{
			for(unsigned counter=4;counter<numcounter;counter++)
			{
				double x; //mm
				motor_file >> x;
				motor_file.ignore(1024, '\n');
				ECStorage->CounterPosition[counter].push_back(x, 0.01);
			};
		}


    if(measurement::Manager::isnocamac)
    {
      double T=ECStorage->cfg.count_time;
      const double TAU = 1000;
      const double LOAD=5e5*T/(1+ECStorage->t.back().value/TAU); //load of counters
      const double CLOAD=1e4*T/(1+ECStorage->t.back().value/TAU); //load of coincedence scheme
      const double DELTA=0.01; //depolarisation effect
      const unsigned jump_point=10;
      double N0=818924*T;
			bool isjump=false;
			double delta=0;
      gRandom->SetSeed(ECStorage->t.back().value);
      for (unsigned bunch=0 ; bunch<ECStorage->numbunch ; ++bunch) 
      {
        for (unsigned counter=0 ; counter<ECStorage->numcounter ; ++counter)
        {     
          delta = DELTA;
          if(ECStorage->N>jump_point && bunch%2==0) 
					{
						//if(!isjump) 
						//{
						//	delta = gRandom->Uniform(0, DELTA);
						//	isjump=true;
						//}
						//else
						//{
						//	delta = DELTA;
						//}
						//N*=(1.+delta); //make jump
            delta = 0;
					}
          double load = LOAD*(1-delta);
          double N=gRandom->Gaus(load, sqrt(load*(1.-load/N0)));
          Data[bunch][counter] = N;
        }
				isjump =false;
        for (unsigned pair=0 ; pair<ECStorage->numpair ; ++pair)
        {     
          double N=gRandom->Gaus(CLOAD, sqrt(CLOAD*(1.-CLOAD/N0)));
          if(ECStorage->N>jump_point && bunch%2==0) 
					{
						if(!isjump) 
						{
							delta = gRandom->Uniform(0, DELTA);
							isjump=true;
						}
						else
						{
							delta = DELTA;
						}
						N*=(1.+delta); //make jump
					}
          Coinc[bunch][pair] = N;
        }
      }
    }
    for (unsigned bunch=0 ; bunch<ECStorage->numbunch ; bunch++) 
    {
      /* loop over counters */
      for (unsigned counter=0 ; counter<ECStorage->numcounter ; counter++)
      {     
        double T=ECStorage->cfg.count_time;
        double N = Data[bunch][counter];
        double N0 = 818924.*T;
        ECStorage->CounterLoad[bunch][counter].push_back(data::get_load(N0,N)/T);
        ECStorage->Count[bunch][counter].push_back(data::get_real_count(N0, N)/T);
      }
      /* loop over pair of counter */
      for (unsigned pair=0 ; pair<ECStorage->numpair ;pair++)
      {     
        double T=ECStorage->cfg.count_time;
        double N = Coinc[bunch][pair];
        //std::cout << "pair=" << pair << " bunch = " << bunch << "  N = " << (long)N << std::endl;
        double N0 = 818924.*T;
        ECStorage->CoincLoad[bunch][pair].push_back(data::get_load(N0,N)/T);
        ECStorage->Coinc[bunch][pair].push_back(data::get_real_count(N0,N)/T);
      }
    }
    V4Vacuum v4vacuum;
    ECStorage->MediumVacuumPresure.push_back(v4vacuum.Pa.value, 0);
    for(std::map<std::string, double>::iterator vcm=v4vacuum.P.begin(); vcm!=v4vacuum.P.end();++vcm)
    {
      ECStorage->Vacuum[vcm->first].push_back(vcm->second,0);
    }

    ECStorage->N=ECStorage->t.size();//save number of points

    //Calculate delta's
    unsigned n=ECStorage->t.size()-1; //valid index for following using
    for (unsigned bunch=0 ; bunch < ECStorage->numbunch ; bunch++) 
    {
      data::PhysValue sum_count, sum_count_norm_bunch;
      data::PhysValue sum_cncr, sum_cncr_norm_bunch;
      unsigned norm_bunch=ECStorage->cfg.bunch_norm[bunch]-1;
      ECStorage->I[bunch].push_back(fCurrentMeasurement->GetCurrent(bunch));
      for (unsigned counter=0 ; counter<ECStorage->numcounter ; counter++) 
      {
        ECStorage->DeltaCount[bunch][counter].push_back
          (  
           data::GetDelta(  ECStorage->Count[bunch][counter][n], ECStorage->Count[norm_bunch][counter][n])
          );
        sum_count = sum_count + ECStorage->Count[bunch][counter][n];
        sum_count_norm_bunch = sum_count_norm_bunch + ECStorage->Count[norm_bunch][counter][n];	
      }

      for (unsigned p=0 ; p<ECStorage->numcounter/2 ; ++p) 
      {
        ECStorage->DeltaCoinc[bunch][p].push_back(  data::GetDelta(  ECStorage->Coinc[bunch][p][n], ECStorage->Coinc[norm_bunch][p][n]));
        sum_cncr = sum_cncr + ECStorage->Coinc[bunch][p][n];
        sum_cncr_norm_bunch = sum_cncr_norm_bunch + ECStorage->Coinc[norm_bunch][p][n];	
      }
      ECStorage->CountSum[bunch].push_back(sum_count);	    
      ECStorage->DeltaSum[bunch].push_back(data::GetDelta( sum_count, sum_count_norm_bunch));
      ECStorage->CoincSum[bunch].push_back(sum_cncr);	    
      ECStorage->DeltaCoincSum[bunch].push_back(data::GetDelta( sum_cncr, sum_cncr_norm_bunch));
    }

    /* Fill NMR data */
    FillNMR();

    /*  Fill beem size */
    for (unsigned bunch=0 ; bunch<ECStorage->numbunch ; bunch++)
    {
      ECStorage->SigmaX[bunch].push_back(fBeamSizeMeasurement->GetX(bunch));
      ECStorage->SigmaL[bunch].push_back(fBeamSizeMeasurement->GetY(bunch));
      ECStorage->SigmaZ[bunch].push_back(fBeamSizeMeasurement->GetZ(bunch));
    }


    ECStorage->FillCalculableData(ECStorage->t.size()-1);

    ECCamac->depolarizer->WriteCurrentFrequencyToFile("/home/vepp4/kadrs/depfreq.dat");

    if(ismeasure) sig_new_point();

    //testing code for sending information about depolarizer frequency map
    /*  
    try 
    {
      boost::asio::io_service service;
      boost::asio::ip::tcp::endpoint ep(boost::asio::ip::address::from_string(PCFG.lsrp.ip_address), PCFG.lsrp.port);
      boost::asio::ip::tcp::socket sock(service);
      sock.connect(ep);
      //boost::asio::write(sock,boost::asio::buffer("Hello Laser Polarimeter!"));

      auto & dfreq = ECCamac->depolarizer->HighResolutionDepFreq;
      if(!dfreq.empty())
      {
        auto low = dfreq.lower_bound(last_time_sent_to_lsrp);
        auto up =  (dfreq.end()--);
        struct item_t
        {
          long int t;
          double E;
          double f;
        };
        if(low!=up)
        {
          auto last_prepared_time=last_time_sent_to_lsrp;
          std::vector<item_t> v;
          v.reserve(10000); 
          for(auto p = low; p!=up; p++)
          {
            auto t   = p->first.time_since_epoch().count(); //this is the time since epoch (1970) in nanoseconds
            double f = p->second; //this is the depolarizer frequency in Hz
            double E = ECStorage->cfg.depolarizer.energy(f); //the corresponding energy
            v.push_back({t,E,f});
            last_prepared_time= p->first;
          }
          boost::asio::write(sock,boost::asio::buffer(v));
          last_time_sent_to_lsrp = last_prepared_time;
        }
      }
    }
    catch(boost::system::system_error &e)
    {
      std::cout << LOG(1) <<  e.what() << std::endl;
    }
    */
  }
  //========================================



  //========================================
  //Из файла вычитывается частота резонатораaa
  double EnergyCalibration::GetRadioFrequency(void)
  {
    double rf, lost;
    std::ifstream FRF("/home/vepp4/kadrs/RF.dat");
    if(!FRF)
    {
      std::cerr << LOG(1) << "Unable to open file with radiofrequency /home/vepp4/kadrs/RF.dat, return dummy 181801160 Hz\n";
      return 181801160.;
    }
    FRF >> rf ;
    FRF.close();
    /* this is some special code for the resonator frequency 
     * it result from frequencymeter acually has a finit ralevant digit 
     * and i should define some lost part of frequency */
    if( rf < 1e8) lost = 100e6;
    if( rf < 1e7) lost = 180e6;
    if( rf < 1e6) lost = 181e6;
    if( rf < 1e5) lost = 181.8e6;
    /* add this lost part */
    rf += lost;
    return rf;
  }
  //=======================================


  //=======================================
  //=======================================



  inline void EnergyCalibration::Start(void)
  {
    if(ismeasure) 
    {
      std::cout << LOG(1) << "Already measure counting rate\n";
      return;
    }
    //StopBackgroundMeasurement();
    ECStorage->isfit=false;
    if(!PCFG.restore) 
    {
      global_timer.restart();
      Global_timer.restart();
      ECStorage->begin_time = global_timer.get_begin_time();
      ECStorage->Begin_time = Global_timer.get_begin_time();
      ECStorage->HRDepFreq = &(ECCamac->depolarizer->HighResolutionDepFreq);
    } 
    else 
    {
      timeval tmp_time;
      gettimeofday(&tmp_time, 0);
      ECStorage->begin_time.tv_sec = tmp_time.tv_sec - time_t(ECStorage->t[ECStorage->t.size()-1].value);
      ECStorage->begin_time.tv_usec = 0;
      global_timer.set_begin_time(ECStorage->begin_time);
      Global_timer.restart(ECStorage->Begin_time);
    }
    ismeasure = true;
    ispause = false;
    ECStorage->issave = false;
    if(!PCFG.restore) ECStorage->Clear();
    Prepare();
    if(ECStorage->issave) ECStorage->cfg.run++;
    fCountMeasurement->Restart(ECStorage->cfg.count_time);
		fThread = std::thread([this]{Measure();});
		return;
  }

  void EnergyCalibration::SetNewPoint(void)
  {
    is_new_point = true;
    std::clog << LOG(DEBUG_LOG) << "EnergyCalibration::SetNewPoint: is_new_point = " << is_new_point << std::endl;
  };

  bool EnergyCalibration::isNewPoint(void)	
  {
    if(is_new_point)
    {
      is_new_point=false;
      return true;
    }
    return false;
  }

  //=======================================
  inline data::EnergyCalibrationStorage* EnergyCalibration::GetStorage()
  {
    return ECStorage;
  }
  //=======================================


  //Заполнить вектор с НМРами
  //=======================================
  void EnergyCalibration::FillNMR(void)
  {
    V4nmr n;
    for(size_t i=0; i<ECStorage->NMR.size();i++)
    {
      ECStorage->NMR[i].push_back(n.nmr_data[i]);
    }   
  }       
  //=======================================


  //void EnergyCalibration::SendCount2Odrenok(const vector <double> & N, const vector<double> & delta)
  //{
  //  int fd = open("/home/shiftman/KEDR/POLA",O_WRONLY);
  //  if(fd<0)
  //  {
  //    static int first_error_open_POLA=0;
  //    if(!first_error_open_POLA) 
  //    {
  //      cerr << "ERROR: Unable to open file  /home/shiftman/KEDR/POLA for writing bunch counting" << endl;
  //      first_error_open_POLA++;
  //    }
  //    return;
  //  }
  //  static unsigned myworkingflag=0;
  //  if(myworkingflag==16777216) myworkingflag=0;
  //  u24_t workingflag= intu24(myworkingflag++);
  //  u24_t Nodr[4];
  //  u24_t Dodr[4];
  //  u24_t udar1,udar2;
  //  for(unsigned i=0; i<N.size();i++)
  //  {
  //    Nodr[i]=intu24(static_cast<unsigned>(N[i]/1000));
  //  }
  //  for(unsigned i=0; i<delta.size();i++) Dodr[i]=intu24(unsigned(delta[i]*100*100));
  //  //Karnaev's numbering: electrons my 1,2 = karnaevs electron 1,3. positron my 3,4 karnaevs positron = 2,4
  //  if(delta[0]<0.01)udar1=intu24(3);
  //  else udar1=intu24(1);	
  //  if(delta[2]<0.01)udar2=intu24(4);
  //  else udar2=intu24(2);
  //  int rc=0;
  //  rc=write(fd,&workingflag,3);
  //  if(rc!=3)
  //  {
  //    cerr << "ERROR: Writing flag to POLA" << endl;
  //  }
  //  if(write(fd,&Nodr[0],3)!=3)
  //  {
  //    cerr << "ERROR: Writing N0 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&Nodr[1],3)!=3)
  //  {
  //    cerr << "ERROR: Writing N1 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&Dodr[0],3)!=3)
  //  {
  //    cerr << "ERROR: Writing  delta12 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&udar1,3)!=3)
  //  {
  //    cerr << "ERROR: Writing udar1 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&Nodr[2],3)!=3)
  //  {
  //    cerr << "ERROR: Writing N3 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&Nodr[3],3)!=3)
  //  {
  //    cerr << "ERROR: Writing N4 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&Dodr[2],3)!=3)
  //  {
  //    cerr << "ERROR: Writing delta34 to POLA: " << strerror(errno) << endl;
  //  }
  //  if(write(fd,&udar2,3)!=3)
  //  {
  //    cerr << "ERROR: Writing udar2 to POLA: " << strerror(errno) << endl;
  //  }
  //  close(fd);
  //}


  ///* 
  // * ===  FUNCTION  ======================================================================
  // *         Name:  IsUDAR
  // *  Description:  Check whether or not Karnaev did UDAR to removes particles from a bunch
  // * =====================================================================================
  // */
  //bool EnergyCalibration::IsUDAR(void)	
  //{
  //  static unsigned myflag=0;
  //  int fd = open("/home/shiftman/KEDR/UDAR",O_RDONLY);
  //  if(fd<0)
  //  {
  //    cerr << "ERROR: Unable to open file  /home/shiftman/KEDR/UDAR for reading" << endl;
  //    close(fd);
  //    return false;
  //  }
  //  u24_t flag;
  //  u24_t udar;
  //  if(read(fd,&flag,3)!=3)cerr << "ERROR: Unable to read 3 bytes from UDAR" << endl;
  //  if(read(fd,&udar,3)!=3)cerr << "ERROR: Unable to read 3 bytes from UDAR" << endl;
  //  unsigned tmp=u24int(flag);
  //  if(myflag==tmp)
  //  {
  //    close(fd);
  //    return false; //file was not changed, no udar return false
  //  }
  //  myflag=tmp;
  //  ///unsigned myudar myudar=u24int(udar);
  //  close(fd);
  //  return true;
  //}

	//void EnergyCalibration::StartFrequencyMeasurement(void)
	//{
	//	fFrequencyMeasurement.SetFrequencymeter(ECCamac->frequencymeter);
	//	fFrequencyMeasurement.Start();
	//}

	//void EnergyCalibration::StopFrequencyMeasurement(void)
	//{
	//	fFrequencyMeasurement.Stop();
	//}
}//end of namespace measurement

