/*
 * =====================================================================================
 *
 *       Filename:  EnergyCalibrationCamac.cpp
 *
 *    Description:  Driver for camac modules.
 *
 *        Version:  1.0
 *        Created:  06/07/2010 04:13:37 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physycs, Novosibirsk, Russia
 *
 * =====================================================================================
 */


#ifndef SMV_MEASUREMENT_ENERGYCALIBRATIONCAMAC_CPP
#define SMV_MEASUREMENT_ENERGYCALIBRATIONCAMAC_CPP

#include <iostream>
#include <string>
#include <vector>
#include <string>
#include <sstream>

#include <TRandom.h>

#include "ibn/timer.h"

#include "Utils.h"
#include "Log.h"

#include "camac/InitModules.h"
//#include "camac/ModuleInventory.h"
//#include "camac/Counter.h"
//#include "camac/Vvi.h"
//#include "camac/Kvch.h"

#include "EnergyCalibrationCamac.h"
#include "Manager.h"
#include "Depolarisator.h"


#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/property_tree/info_parser.hpp>




extern bool measurement::Manager::isnocamac;

extern double FREQUENCY_SHIFT;
using namespace std;
TRandom RANDOM;
const double LOAD=100000;

namespace camac
{



	std::map <std::string,  std::shared_ptr<camac::ModuleInventory> > InitInfoModules(std::string filename)
	{
		boost::property_tree::ptree tree;
		boost::property_tree::info_parser::read_info(filename, tree);
		return camac::InitModules(tree, !measurement::Manager::isnocamac);
	}


	std::map <std::string,  std::shared_ptr<camac::ModuleInventory> > InitXmlModules(std::string filename)
	{
		boost::property_tree::ptree tree;
		boost::property_tree::xml_parser::read_xml(filename, tree);
		return camac::InitModules(tree, !measurement::Manager::isnocamac);
	}

  typedef double Value_t; 
  //=========================================================//
	EnergyCalibrationCamac::EnergyCalibrationCamac(unsigned nbunch, unsigned ncounter)  : numbunch(nbunch), numcounter(ncounter)
	{
		std::clog << LOG(1) << "Configurate camac hardware for energy calibration: nbunch=" << nbunch << ", ncounter=" << ncounter << std::endl;


		fModules = InitInfoModules(PCFG.camac_config_file.string());
		for(auto & m : fModules)
		{
			std::clog << LOG(DEBUG_LOG)<< m.first << " " << m.second.get() << " " << m.second.use_count() << std::endl;
		}

		frequencymeter = getModule<camac::Freq>("Frequencymeter.DepolarizerFrequencymeter");


		std::clog << LOG(2) << "Configurate depolarizer." << std::endl;
		depolarizer=new measurement::Depolarisator;
		depolarizer->SetMinimumAttenuation(PCFG.min_attenuation);
		std::clog << LOG(2) << "Configurate generator." << std::endl;
		depolarizer->generator=getModule<camac::Generator>("Generator."+PCFG.measurement.generator);
		clog << LOG(1) << "Generator type is: "  << PCFG.measurement.generator << endl;

		generatorDDS300 = getModule<camac::Generator>("Generator.Generator0");
		generatorDDS500 = getModule<camac::Generator>("Generator.Generator1");
		generatorDDS500_2 = getModule<camac::Generator>("Generator.Generator2");

		double_generator = new camac::GeneratorArray("SelectiveGeneratorDouble");
		double_generator->Add(generatorDDS500);
		double_generator->Add(generatorDDS500_2, FREQUENCY_SHIFT);

		std::clog << LOG(2) << "Configurate attenuator." << std::endl;

		for(int i=0;i<3;i++)
		{
			depolarizer->attenuator[i].AddAttenuator(getModule<camac::Attenuator>("Attenuator.Attenuator"+to_string(i)+"1"));
			depolarizer->attenuator[i].AddAttenuator(getModule<camac::Attenuator>("Attenuator.Attenuator"+to_string(i)+"3"));
		}
		depolarizer->attenuator[0].SetAttenuation(50);

		std::clog << LOG(2) << "Configurate delay line." << std::endl;

		DepolarizerDelayLine=getModule<camac::DelayLine>("DelayLine.DepolarizerDelayLine");
		DepolarizerDelayLine2=getModule<camac::DelayLine>("DelayLine.DepolarizerDelayLine2");

		std::clog << LOG(2) << "Configurate timers" << std::endl;
		Timer_.resize(2);
		Timer_[0]=getModule<camac::Timer>("Timer.timer0");
		Timer_[1]=getModule<camac::Timer>("Timer.timer1");

		Concur.resize(numbunch, std::vector<ModuleInventory*>(numcounter)); 
		ConcurCoinc.resize(numbunch, std::vector<ModuleInventory*>(numcounter/2)); 
		Count.resize(numbunch, std::vector<CounterC0303*>(numcounter));
		CounterCoinc.resize(numbunch, std::vector<CounterC0303*>(numcounter/2));
		Vvi.resize(numcounter);	
		Diskr.resize(numcounter);

		std::string name;

		std::clog << LOG(2) << "Configurate bunch delay line" << std::endl;
		DelayLineBunch.resize(nbunch);
		for(unsigned b=0;b <nbunch; b++) DelayLineBunch[b].resize(unsigned(numcounter/2));
		for(unsigned i=1; i <= numcounter ; i+=2)
		{
			for(unsigned j=1; j <= nbunch ; j++)
			{
				std::string name = "DelayLine.counter"+to_string(i)+"and"+to_string(i+1)+"bunch"+to_string(j); // std::string("counter")+setString(i,0,0)+std::string("and")+setString(i+1,0,0)+std::string("bunch")+setString(j,0,0));
				DelayLineBunch[j-1][unsigned((i-1)/2.)]=getModule<camac::DelayLine_L0606>(name);
			}
		}


		//std::clog << LOG(2) << "Configurate counter delay line" << std::endl;
		//DelayLineCounter.push_back( new DelayLineL0101("counter5or6"));
		//DelayLineCounter.push_back( new DelayLineL0101("counter7or8"));
		//ModuleList.push_back(DelayLineCounter[0]);
		//ModuleList.push_back(DelayLineCounter[1]);

		std::clog << LOG(2) << "Configurate counters vvi diskr" << std::endl;
		for (unsigned j=0 ; j < numcounter ; ++j)
		{
			name="counter" + to_string(j+1);
			Vvi[j] = getModule<camac::Vvi>("Vvi."+name);
			Diskr[j] = getModule<camac::Diskriminator>("Diskriminator."+name);
			for (unsigned i=0 ; i < numbunch ; ++i)
			{
				name = "counter"+to_string(j+1)+"bunch"+to_string(i+1);//counterNbunchM
				Concur[i][j] = getModule<camac::Concur>("Concurence."+name);
				Count[i][j]	= getModule<camac::CounterC0303>("Counter."+name);
			}
		}

		std::clog << LOG(2) << "Configurate coincidence scheme" << std::endl;
		for(unsigned b=0;b<numbunch; b++)
			for(unsigned pr=0;pr<numcounter/2; pr++) 
			{
				name = "counter"+to_string(2*pr+1)+"and"+to_string(2*pr+2)+"bunch"+to_string(b+1);
				ConcurCoinc[b][pr] = getModule<camac::Concur>("Concurence."+name);
				CounterCoinc[b][pr] = getModule<camac::CounterC0303>("Counter."+name);
			}
		std::clog << LOG(DEBUG_LOG) << "EnergyCalibrationCamac: END" << std::endl;
	}
  //=========================================================//

std::map<std::string,  std::shared_ptr<camac::ModuleInventory> > CAMAC;


  //=========================================================//
  EnergyCalibrationCamac::~EnergyCalibrationCamac()
  {
		std::clog << LOG(5) << "EnergyCalibrationCamac::~EnergyCalibrationCamac() BEGIN" << std::endl;
		//CAMAC=fModules;
		std::clog << LOG(5) << "EnergyCalibrationCamac::~EnergyCalibrationCamac() END" << std::endl;
  }
  //=========================================================//


  //=========================================================//
  std::vector< std::vector<Value_t> > EnergyCalibrationCamac::GetData(void)
  {

    std::cout<<LOG(3)<<"EnergyCalibrationCamac::GetData: Begin" << std::endl;
    std::vector<  std::vector<Value_t> > Data;
    Data.resize(numbunch, std::vector<Value_t>(numcounter));

    if(!measurement::Manager::isnocamac)
    {
			std::lock_guard<std::mutex> lock(counter_mutex);
      for(unsigned bunch=0 ; bunch<numbunch ;bunch++)
      {
        for(unsigned counter=0 ; counter<numcounter ; counter++)
        {
          Data[bunch][counter] = Count[bunch][counter]->GetCount();
          Count[bunch][counter]->ResetCount();
        }	
      }
    }
    else
    {
      for(unsigned bunch=0 ; bunch<numbunch ;bunch++)
      {
        for(unsigned counter=0 ; counter<numcounter ; counter++)
        {
          Data[bunch][counter] = LOAD*RANDOM.Gaus(1.,1./sqrt(LOAD));
        }	
      }
    }
    std::cout<<LOG(3)<<"EnergyCalibrationCamac::GetData: END" << std::endl;
    return Data;
  }
  //=========================================================//

  //=========================================================//
  std::vector< std::vector<Value_t> > EnergyCalibrationCamac::GetCoinc(void)
  {

    std::cout<<LOG(3)<<"EnergyCalibrationCamac::GetCoinc(void)" << std::endl;
    std::vector<  std::vector<Value_t> > Data;
    Data.resize(numbunch, std::vector<Value_t>(numcounter/2));


    if(!measurement::Manager::isnocamac) //без оператор хочет, то запускаем без камака
    {
      for(unsigned bunch=0 ; bunch<numbunch ; bunch++)
      {
        for(unsigned pair=0 ; pair<numcounter/2 ; pair++)
        {
					std::lock_guard<std::mutex> lock(coincidence_counter_mutex);
          Data[bunch][pair] = CounterCoinc[bunch][pair]->GetCount();
          CounterCoinc[bunch][pair]->ResetCount();
        }	
      }
      return Data;
    }
    else
    {
      for(unsigned bunch=0 ; bunch<numbunch ; bunch++)
      {
        for(unsigned pair=0 ; pair<numcounter/2 ; pair++)
        {
          Data[bunch][pair] = LOAD*RANDOM.Gaus(1.,1./sqrt(LOAD));
        }	
      }
      return Data;
    }
  }
  //=========================================================//

  //=========================================================//
  std::vector<std::vector<Value_t> > EnergyCalibrationCamac::GetDataWithout_reset(void)
  {

    std::vector<  std::vector<Value_t> > Data;
    Data.resize(numbunch, std::vector<Value_t>(numcounter));


    if(!measurement::Manager::isnocamac)
    {
      for(unsigned i=0 ; i<numbunch ; i++)
      {
        for(unsigned j=0 ; j<numcounter ; j++)
        {
          Data[i][j] = Count[i][j]->GetLoad();
        }	
      }
      return Data;
    }
    else //без оператор хочет, то запускаем без камака
    {
      for(unsigned i=0 ; i<numbunch ; i++)
      {
        for(unsigned j=0 ; j<numcounter ; j++)
        {
          Data[i][j] = LOAD*RANDOM.Gaus(1.,1./sqrt(LOAD));
        }	
      }
      return Data;
    }
  }

  //=========================================================//
  void EnergyCalibrationCamac::ResetCounters(void)
  {
    for(unsigned b=0 ; b<Count.size() ; b++)
      for(unsigned c=0 ; c<Count[b].size() ; c++)
        Count[b][c]->ResetCount();

    for(unsigned b=0 ; b<CounterCoinc.size() ; b++)
      for(unsigned c=0 ; c<CounterCoinc[b].size() ; c++)
        CounterCoinc[b][c]->ResetCount();
  }

  //=========================================================//
  std::vector<std::vector<Value_t> > EnergyCalibrationCamac::GetCoincWithout_reset(void)
  {
    std::vector<  std::vector<Value_t> > Data;
    Data.resize(numbunch, std::vector<Value_t>(numcounter/2));


    if(!measurement::Manager::isnocamac)
    {
      for(unsigned i=0 ; i<numbunch ; i++)
      {
        for(unsigned j=0 ; j<numcounter/2 ; j++)
        {
          Data[i][j] = CounterCoinc[i][j]->GetLoad();
        }	
      }
      return Data;
    }
    else 
    {
      for(unsigned i=0 ; i<numbunch ; i++)
      {
        for(unsigned j=0 ; j<numcounter/2 ; j++)
        {
          Data[i][j] = LOAD*RANDOM.Gaus(1.,1./sqrt(LOAD));
        }	
      }
      return Data;
    }
  }

  //=========================================================//



  void EnergyCalibrationCamac::StartTimer(void)
  {
    if(!(measurement::Manager::isnocamac))
    {
      Timer * timer;
      for(unsigned i=0;i<Timer_.size();i++)
      {
        timer = dynamic_cast<Timer*>(Timer_[i]);
        timer->Start();
      }
    }
  }

	void EnergyCalibrationCamac::StopTimer(void)
	{
		//if(timer_is_started) raise(SIGUSR1);
	}

  void EnergyCalibrationCamac::WaitTimer(const volatile bool & iswait )
  {
		Timer_.back()->Wait(iswait);
  }

  void EnergyCalibrationCamac::StartTimer(double time, volatile bool & iswait)
  {
		//signal(SIGUSR1, [](int)
		//		{
		//			std::cout << "Catch SIGUSR1" << std::endl;
		//		});
		//timer_is_started=true;
    if(!(measurement::Manager::isnocamac))
    {
			for(auto timer : Timer_)
			{
				timer->Start(time);
			}
      Timer_[0]->Wait(iswait);
			//for(auto timer : Timer_)
			//{
			//	timer->Wait(iswait);
			//	//timer_is_started=false;
			//}
    } 
    else 
    {
      ibn::timer timer;
      while(iswait && timer.elapsed() < time ) usleep(100000); // Спать 100000 мкc (0.1 s) 
    }
  }



  bool EnergyCalibrationCamac::GetStatusTimer(void)
  {
    if(!(measurement::Manager::isnocamac))
    {
      return Timer_.back()->GetStatus();
    }
    return true;
  }



  void EnergyCalibrationCamac::OnVvi(unsigned vvi_number, unsigned volt) //Switch on vvi. The first parameter is vvi number, the second - volt
  {
    if(!(measurement::Manager::isnocamac))
    {
      camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[vvi_number]); 
      //vvi->ResetOverload();
      vvi->SetVolts(volt);
      if(!vvi->IsOn()) vvi->On();	
    }   
  }


  void EnergyCalibrationCamac::SetDiskr(unsigned num_counter, unsigned treshold)
  {
    if(!(measurement::Manager::isnocamac))
    {
      camac::DiskrD0302* diskr = dynamic_cast<camac::DiskrD0302*>(Diskr[num_counter-1]);
      diskr->SetThreshold(treshold);
    }
  }


  void EnergyCalibrationCamac::SetBunchDelay(unsigned counter_pair, unsigned bunch, unsigned delay)
  {
    DelayLineBunch[bunch][counter_pair]->SetDelay(delay);
  }

  void EnergyCalibrationCamac::SetDelay(unsigned delay)
  {
    for(unsigned bunch = 0; bunch<numbunch; bunch++)
      for(unsigned pair = 0; pair < numcounter/2; ++pair)
        DelayLineBunch[bunch][pair]->SetDelay(delay);
  }

  unsigned EnergyCalibrationCamac::GetBunchDelay(unsigned counter_pair, unsigned bunch)
  {
    return DelayLineBunch[bunch][counter_pair]->GetDelay();
  }


	camac::Address EnergyCalibrationCamac::GetVviAddress(unsigned counter) const
	{
		const camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		return vvi->GetAddress();
	}

	double EnergyCalibrationCamac::GetVoltage(unsigned counter) const 
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		return vvi->GetVolts();
	}

	void EnergyCalibrationCamac::SetVoltage(unsigned counter,  double voltage)
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		vvi->SetVolts(voltage);
	}

	void EnergyCalibrationCamac::On(unsigned counter)
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		vvi->On();
	}
	void EnergyCalibrationCamac::Off(unsigned counter)
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		vvi->Off();
	}

	void EnergyCalibrationCamac::Reset(unsigned counter)
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		vvi->ResetOverload();
	}
	bool EnergyCalibrationCamac::IsOn(unsigned counter) const 
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		return vvi->IsOn();
	}
	bool EnergyCalibrationCamac::IsOverload(unsigned counter) const
	{
		camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
		return vvi->GetStatus();
	}

  std::vector<int> EnergyCalibrationCamac::GetVviStatus(void)
  {
    std::vector<int>  vvi_status(Vvi.size());
    for(int counter=0;counter<Vvi.size();counter++)
    {
      camac::Vvi_B0308* vvi = dynamic_cast<camac::Vvi_B0308*>(Vvi[counter]);
      vvi_status[counter] = vvi->IsOn(); 
    }
    return vvi_status;
  }


	void EnergyCalibrationCamac::ShowDepolarizerAmplitude(int channel)
	{
		getModule<camac::Kvch>("Kvch.KvchControlDepolarizerAmplitude")->Commutate(channel, 0);
	}

	void EnergyCalibrationCamac::ShowElectronDepolarizerAmplitude(void)
	{
		ShowDepolarizerAmplitude(0);
	}

	void EnergyCalibrationCamac::ShowPositronDepolarizerAmplitude(void)
	{
		ShowDepolarizerAmplitude(1);
	}

	int EnergyCalibrationCamac::WhichAmplitudeToShow(void)
	{
		return getModule<camac::Kvch>("Kvch.KvchControlDepolarizerAmplitude")->GetState(0);
	}
	
	template<class CamacModule>
		CamacModule * EnergyCalibrationCamac::getModule(std::string name)
		{
			if(fModules[name])
			{
				clog << LOG(DEBUG_LOG) << "Found module " << name << " " << fModules[name].use_count() << std::endl;
				auto m =  dynamic_cast<CamacModule*>(fModules[name].get());
				if(m==nullptr)
				{
					std::cerr << LOG(1) << "Unable to cast " <<  name << " : " << fModules[name]->GetName() << std::endl;
					return nullptr;
				}
				return m;
			}
			else
			{
				throw std::runtime_error("Unable to find camac module name " + name);
			}
			return  nullptr;
		}
}

#endif //SMV_MEASUREMENT_ENERGYCALIBRATIONCAMAC_CPP


