#ifndef SMV_MEASUREMENT_MANAGER_CPP
#define SMV_MEASUREMENT_MANAGER_CPP


#include <csignal>
#include <iostream>
#include <string>
#include <map>

#include "Manager.h"
#include "EnergyCalibrationCamac.h"
#include "EnergyCalibration.h"
#include "CountMeasurement.h"
#include "CurrentMeasurement.h"
#include "BeamSizeMeasurement.h"
#include "FrequencyMeasurement.h"
#include "StatusMeasurement.h"

measurement::Manager * MEASUREMENT_MANAGER = nullptr;

void sigint_handler(int )
{
  std::clog << LOG(0) << "Catch SIGINT signal (CTRL-C). Turn OFF the depolarizer device." << std::endl;
  MEASUREMENT_MANAGER->GetDepolarisator()->Stop();
  measurement::StatusMeasurement* mstatus = (measurement::StatusMeasurement*) MEASUREMENT_MANAGER->GetMeasurement("StatusMeasurement");
  mstatus->SendDepolarizerStatusToLsrp(1);
  exit(0);
}

namespace measurement
{
  const double PRECISION=10e-12; 

  Manager::Manager(unsigned nbunch, unsigned ncounter) :
    numbunch(nbunch),  //Number of bunches in the storage ring
    numcounter(ncounter) //Number of counters
  {
    std::clog<<LOG(DEBUG_LOG)<<"Manager::Manager: START" << std::endl;

    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating camac."<< std::endl;
    fCamac.reset(new camac::EnergyCalibrationCamac(nbunch, ncounter));


    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating CountMeasurement"<< std::endl;
    auto countMeasurement = new CountMeasurement;
		fMeasurementList["Count"].reset(countMeasurement);
		countMeasurement->SetCamac(fCamac.get());

    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating EnergyCalibration"<< std::endl;
		EC = new EnergyCalibration(numbunch, numcounter);
    EC->SetCamac(fCamac.get());
		fMeasurementList["EnergyCalibration"].reset(EC);

    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating CurrentMeasurement"<< std::endl;
    auto currentMeasurement = new CurrentMeasurement;
    fMeasurementList["CurrentMeasurement"].reset(currentMeasurement);

    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating BeamSizeMeasurement"<< std::endl;
    auto beamSizeMeasurement = new BeamSizeMeasurement;
    fMeasurementList["BeamSizeMeasurement"].reset(beamSizeMeasurement);

    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Creating FrequencyMeasurement"<< std::endl;
    auto frequencyMeasurement = new FrequencyMeasurement;
    frequencyMeasurement->SetFrequencymeter(fCamac->frequencymeter);
    fMeasurementList["FrequencyMeasurement"].reset(frequencyMeasurement);


    EC->SetCurrentMeasurement(currentMeasurement);
    EC->SetBeamSizeMeasurement(beamSizeMeasurement);
    EC->SetFrequencyMeasurement(frequencyMeasurement);
    EC->SetCountStorage(countMeasurement->GetStorage());
    EC->SetCountMeasurement(countMeasurement);


    std::clog << LOG(DEBUG_LOG+1) << "Manager::Manager Create StatusMeasurement" << std::endl;
    auto statusMeasurement = new StatusMeasurement;
    statusMeasurement->SetCamac(fCamac.get());
    statusMeasurement->SetCurrentMeasurement(currentMeasurement);
    statusMeasurement->SetEnergyCalibrationStorage(EC->GetStorage());
		statusMeasurement->SetCountMeasurement(countMeasurement);
    fMeasurementList["StatusMeasurement"].reset(statusMeasurement);

    MEASUREMENT_MANAGER = this;
    signal ( SIGINT, sigint_handler);
    std::clog<<LOG(DEBUG_LOG)<<"Manager::Manager: FINISH" << std::endl;
  }
  //================================================================================================//

  Manager::~Manager(void)
  {
		Stop();
  }

	void Manager::Stop(void)
	{
		std::clog << LOG(DEBUG_LOG) << "measurement::Manager::Stop BEGIN" << std::endl;
		for(auto & item : fMeasurementList)
		{
			auto & measurement = item.second;
			measurement->Stop();
		}
		std::clog << LOG(DEBUG_LOG) << "measurement::Manager::Stop END" << std::endl;
	}

  void Manager::Start(std::string namemeasure)
  {
    fMeasurementList[namemeasure]->Start();
  }

  void Manager::Stop(std::string namemeasure)
  {
		fMeasurementList[namemeasure]->Stop();
  }


  //Паузем измерение. Пауза организована с помощью блокировок
  void Manager::Pause(std::string namemeasure)
  {
		fMeasurementList[namemeasure]->Pause();
  }

  inline void Manager::UnPause(std::string namemeasure)
  {
		fMeasurementList[namemeasure]->UnPause();
  }


  //Если после нажатия кнопки остановка измерения захоти продолжить его, то жмем сюды
  inline void Manager::Continue(std::string namemeasure)
  {
		fMeasurementList[namemeasure]->Continue();
  }


  //Получи указатель на хранилище измеренных (и уже обработанных) данных
  data::Storage* Manager::GetStorage(std::string namemeasure)
  {
    return fMeasurementList[namemeasure]->GetStorage(); 
  }


  bool Manager::IsMeasure(std::string measure_name)
  {
		return fMeasurementList["EnergyCalibration"]->IsMeasure();
  }


  Depolarisator* Manager::GetDepolarisator(void)
  {
    return EC->ECCamac->depolarizer;
  }

  data::Status & Manager::GetStatus(void)
  {
    return *dynamic_cast<StatusMeasurement*>(fMeasurementList["StatusMeasurement"].get())->GetStorage();
  }


}//end of namespace measurement


#endif //SMV_MESUREMENT_MANAGER_CPP
