/*
 * =====================================================================================
 *
 *       Filename:  polarimeter.cpp
 *
 *    Description:  Touschek polarimeter DAQ
 *
 *        Version:  0.12.1
 *        Created:  06/27/2010 06:06:36 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physics, Novosibirsk, Russia
 *
 * =====================================================================================
 */


#include <iostream>
#include <fstream>
#include <thread>
#include <cstdlib>
#include <mutex>


void monitorClose()
{
  using namespace std;
  //printf("canvas closed\n");
  std::cout << "################## CANVSA CLOSED #################" << endl;
}

#include <boost/program_options.hpp>
#include <boost/program_options/positional_options.hpp>
#include <boost/program_options/parsers.hpp>
#include <boost/program_options/errors.hpp>

#include <boost/filesystem.hpp>



#include <TROOT.h>
#include <TApplication.h>
#include <TSystem.h>
#include <TStyle.h>
#include <TRootCanvas.h>
#include <TError.h>

#include "version.h"
#include "ibn/valer.h"
#include "ibn/opt.h"
#include "Log.h"

#include "PolarimeterConfig.h"
#include "gui/MainWindow.h"
#include "data/EnergyCalibrationStorage.h"
#include "measurement/Manager.h"
//#include "gui/MeasurementDialog.h"
#include "gui/MeasurementManagerDialog.h"
#include "gui/EnergyCalibrationView.h"

#include "gui/Canvas.h"
#include "fit/FitManager.h"

double  FREQUENCY_SHIFT=0;

std::recursive_mutex  RootProcessEventsMutex;


extern void InitGui();
VoidFuncPtr_t initfuncs[] = { InitGui, 0 };
TROOT root("polarimeter","polarimeter", initfuncs);

namespace gui
{
extern void LoadBadPointList(std::istream &);
}

PolarimeterConfig PCFG;

//bool IS_RESTORE=false; 


void signal_int_handler(int a)
{
  std::cout << "CNTRL-C pressed turn off the depolarizer" << std::endl;

  exit(1);
}


bool camac::Module::XQ_ERROR_EXCEPTION=true;

int main(int argc, char **argv)
{
	gErrorIgnoreLevel = 1001;
	std::string pol_file_name; //file name to load data
  std::string bad_point_file_name;
  std::string run_string; //run number
  std::string fit_method;
  std::string source_file, target_file="tmp.pol";
  std::string config_file_name;
	std::ios_base::sync_with_stdio(false);
  namespace po=boost::program_options;
	namespace fs=boost::filesystem;
  po::options_description opt_desc("Allowed options:");
  po::options_description fopt_desc("Config file options:");
  po::positional_options_description pos_desc;
  po::variables_map opt; //options
  po::variables_map fopt;//file options.
  opt_desc.add_options()
    ("help","Print this help")
		("version", "Print polarimeter version")
    ("nocamac,n",  "Turn off CAMAC operations")
    ("restore", "Restore process after crash")
    ("name",po::value<std::string>(&PCFG.name)->default_value("vepp4"),"The name of polarimeter")
    ("config",po::value<fs::path>(&PCFG.config_file),"Main configuration file")
    ("load,l",po::value<std::string>(&pol_file_name),"Load data from file")
    ("run,r",po::value<std::string>(&run_string),"Load run")
		("fit", "Automaticaly fit after load")
    ("fit-method",po::value<std::string>(&fit_method)->default_value("LINEAR_JUMP"), "Method use to fit")
    ("bad-point-list,b",po::value<std::string>(&bad_point_file_name), "Load bad point list from file")
    ("frequency-shift,F",po::value<double>(&FREQUENCY_SHIFT)->default_value(0),"Frequency shift in Hz between two selective depolarizer")
    ("convert","convert data from old format to new format")
    ("source",po::value<std::string>(&source_file)->default_value(""),"source file")
    ("target",po::value<std::string>(&target_file)->default_value("tmp.pol"),"target file")
    ("graph",po::value<std::string>()->default_value("delta-1"),"Graphic to draw")
    ("debug",  po::value<int>(&PCFG.debug.level)->default_value(1),"Output log level")
    ("test-canvas","test for mycanvas")
    ("vvi", po::value<std::string> ( &PCFG.vvi ) -> default_value("on"), " Vvi command off, on (default)")
    ("skip-data-storing",   "Skip data storing")
		;
    

  fopt_desc.add_options()
    ("nocamac,n",  "Turn off CAMAC operations")
    ("work_dir",                po::value<fs::path> (&PCFG.work_dir),"Working directory")
    ("save_dir",                po::value<fs::path> (&PCFG.save_dir),     "Relative Saving directory")
    ("print_dir",               po::value<fs::path> (&PCFG.print_dir),     "Relative Printing directory")
    ("camac_config_file",       po::value<fs::path> (&PCFG.camac_config_file), "Camac configuration file")
    ("measurement_config_file", po::value<fs::path> (&PCFG.measurement_config_file), "Measurement configuration file")
    ("last_good_measurement_config_file", po::value<fs::path> (&PCFG.last_good_measurement_config_file), "Last good Measurement configuration file")
    ("last_bad_measurement_config_file", po::value<fs::path> (&PCFG.last_bad_measurement_config_file), "Last good Measurement configuration file")
    ("log_file",                po::value<fs::path> (&PCFG.log_file),     "Logging file")
    ("tmp_save_file",           po::value<fs::path> (&PCFG.tmp_save_file),"Temporory file")
    ("last_tmp_save_file",      po::value<fs::path> (&PCFG.last_tmp_save_file),"Temporory file")
    ("motor_file",              po::value<fs::path> (&PCFG.motor_file),"Counter position file")
    ("depfreq_file",            po::value<fs::path> (&PCFG.high_resolution_depfreq_file),"High resolution depolarization frequency file")
    ("vepp4.stap_file",         po::value<fs::path> (&PCFG.vepp4.stap_file)->default_value("/home/vepp4/kadrs/stap.dat"),"VEPP4 stap file")
    ("vepp4.eCCD_file",         po::value<fs::path> (&PCFG.vepp4.eCCD_file)->default_value("/home/vepp4/kadrs/CCD-VEPP4e.dat"), "VEPP4 CCD electron file")
    ("vepp4.vepp4.pCCD_file",   po::value<fs::path> (&PCFG.vepp4.pCCD_file)->default_value("/home/vepp4/kadrs/CCD-VEPP4p.dat"), "VEPP4 CCD positron file")
    ("vepp4.orbit_file",        po::value<fs::path> (&PCFG.vepp4.orbit_file)->default_value("/home/vepp4/kadrs/orbi-v4-new"), "VEPP4 orbit file")
    ("vepp4.RF_file",           po::value<fs::path> (&PCFG.vepp4.RF_file)->default_value("/home/vepp4/kadrs/RF.dat"), "VEPP4 RF file")
    ("lsrp.ip_address",         po::value<std::string> (&PCFG.lsrp.ip_address)->default_value("192.168.163.79"), "rokk1m-lsrp ip adress")
    ("lsrp.port",               po::value<int> (&PCFG.lsrp.port)->default_value(9090), "rokk1m-lsrp ip adress")

    ("debug.level",             po::value<int>(&PCFG.debug.level),"Output log level")
    ("bunch_number",            po::value<unsigned> (&PCFG.bunch_number)->default_value(4),    "Number of bunches")
    ("counter_number",          po::value<unsigned> (&PCFG.counter_number)->default_value(8),  "Counter number")
    ("pair_number",             po::value<unsigned> (&PCFG.pair_number)->default_value(4),     "Counter pair number")
    ("calculator.temperature_drift",  po::value<double> (&PCFG.temperature_drift)->default_value(0),     "Temperature drift")
    ("calculator.points_before_jump", po::value<unsigned> (&PCFG.points_before_jump)->default_value(10),     "Points before jump")
    ("gui.fit.max_par_number",  po::value<unsigned>(&PCFG.gui.fit_max_par_number)->default_value(18),"Number of parameters in FitDialog")
    ("gui.autorare",            po::value<unsigned>(&PCFG.gui.autorare)->default_value(50),"Rare number")
    ("gui.autofit",             po::value<bool>(&PCFG.gui.autofit)->default_value(false),"Autofit by linear jump")
    ("gui.default_graph",       po::value<std::string>(&PCFG.gui.default_graph)->default_value("delta-1"),"Default graphic")
    ("gui.show_last_point",     po::value<int>(&PCFG.gui.show_last_point)->default_value(0),"Number of last point to show for TEST mode")
    ("gui.note_format",         po::value<std::string>(&PCFG.gui.note_format),"Format for labels of depolarizer energy (printf notation)")
    ("min_attenuation",         po::value<double>(&PCFG.min_attenuation)->default_value(10), "Minimum attenuation")
		("measurement.generator",   po::value<std::string>(&PCFG.measurement.generator)->default_value("Generator1"), "Default used generator")
		("measurement.max_allowed_frequency_difference", po::value<double>(&PCFG.measurement.max_allowed_frequency_difference)->default_value(25), "Maximul allowed frequency difference between measured and set frequency") 
    ;
  
  pos_desc.add("load", -1);

  try
  {
    po::store(po::command_line_parser(argc, argv).options(opt_desc).positional(pos_desc).run(), opt);
    if (opt.count("help"))
    {
      std::clog << "Polarimeter is the program for VEPP-4M energy measurement by resonance depolarization method\n"
        << "AUTHORS:  Ivan Nikolaev and Maxim Struchalin\n"
				<< "PHONE: 48-37\n"
        << "CELL PHONE:  +7(913)987-6819,\n"
        << "EMAIL: I.B.Nikolaev@inp.nsk.su.\n";
			std::clog << std::endl;
      std::clog << opt_desc;
      std::clog << fopt_desc;
      exit(0);
    }
		if(opt.count("version"))
		{
			std::clog << "Polarimeter version: " << POLARIMETER_VERSION <<  ". compiled at " << __TIMESTAMP__ << "." << std::endl;
			exit(0);
		}
		if(!fs::exists(PCFG.config_file) && !PCFG.config_file.has_root_name())  //no such configuration file try to add $HOME before
		{
			PCFG.config_file = fs::path(getenv("HOME"))/PCFG.config_file;
		}
		if(fs::exists(PCFG.config_file))
		{
			std::ifstream config_file(PCFG.config_file.string());
			if(!config_file)
			{
				std::cerr << "ERORR: Unable to open configuration file: " << PCFG.config_file << std::endl;
			}
			po::store(po::parse_config_file(config_file,fopt_desc,true), fopt);
			config_file.clear();
			config_file.seekg(std::ios_base::beg);
			po::store(po::parse_config_file(config_file,opt_desc,true), opt);
		}
    po::notify(opt);
    po::notify(fopt);

		std::clog << "Starting polarimeter (version "<< POLARIMETER_VERSION << ") compiled at " << __TIMESTAMP__  << "." << std::endl;
    PCFG.prepare();
		std::ofstream log_stream;
		if(fs::exists(PCFG.log_file))
		{
			log_stream.open(PCFG.log_file.string(),std::ios_base::app);
		}
		else
		{
			log_stream.open(PCFG.log_file.string(),std::ios_base::out);
		}
		if(!log_stream)
		{
			std::cerr << "ERROR: Unable to open log file: " << PCFG.log_file << std::endl;
			exit(1);
		}
		LOG.set_log(&log_stream);
    LOG.set_level(PCFG.debug.level);

		PCFG.restore = opt.count("restore");
    PCFG.debug.skip_data_storing = opt.count("skip-data-storing");

    if(opt.count("convert"))
    {
      data::EnergyCalibrationStorage ec(4,8);
      ec.Load(source_file);
      ec.Save2FSSerial(target_file);
      exit(0);
    }
    gStyle->SetTitleAlign(13);
    gStyle->SetTitleX(0);
    TApplication theApp("polarimeter", &argc, argv);
    PCFG.nocamac = opt.count("nocamac");
    measurement::Manager::isnocamac = PCFG.nocamac;
    gui::MainWindow main_window(gClient->GetRoot());

    //TCanvas * c = new TCanvas;
    //TRootCanvas *rc = (TRootCanvas*)c->GetCanvasImp();
    //rc->Disconnect("CloseWindow()");
    //c->Disconnect("CloseWindow()");
    //rc->Connect("CloseWindow()", 0, 0, "monitorClose()");
    if(opt.count("bad-point-list"))
    {
      clog << LOG(1) << " Load bad point list from file: " << bad_point_file_name << endl;
			boost::filesystem::path fpath(bad_point_file_name);
      std::ifstream file(fpath.string().c_str());
      if(!file)
      {
        cerr << LOG(1) << "Unable to open file: " << fpath.string() << endl;
      }
      gui::LoadBadPointList(file);
    }
    if(opt.count("load") || opt.count("run")) 
    {
			boost::filesystem::path runfile(pol_file_name);
      gui::MeasurementManagerDialog * manager = main_window.gmanager.get();
      data::EnergyCalibrationStorage * storage = manager->EC;
			boost::filesystem::path rundir(PCFG.save_dir);

      if(opt.count("run")) 
      {
				fs::recursive_directory_iterator it(rundir);
				fs::recursive_directory_iterator endit;
				for(;it!=endit; ++it)
				{
					if(!is_directory(*it))
					{
						auto file = it->path();
						auto p = file.filename();
						for(;!p.extension().empty();p=p.stem())
						{
							if(p.stem() == run_string)
							{
								//if(p.extension() == ".pol" || p.extension() == ".xml")
								if(p.extension() == ".pol")
								{
									std::clog << LOG(1) << "Try to load file: " << file << " " << p <<  " " << p.extension() << std::endl;
									runfile = file;
									break;
								}
							}
						}
						if(!runfile.empty()) break;
					}
				}
				if(runfile.empty()) 
				{
					cerr<< LOG(1) << "Unable to find file for run " << run_string << endl;
					exit(1);
				}
      }
      else 
      {
        LOG(DEBUG_LOG) << clog << "Option load" << endl;
        runfile=pol_file_name;
        if(!boost::filesystem::exists(runfile))
        {
          cerr << "There is no such file: " << runfile.string() << endl;
          exit(1);
        }
      }
      clog << LOG(1) << "Loading file: " << runfile.string() << endl;
      storage->Load(runfile.string());
      manager->ECView->InitDataList();
      manager->ECView->InitGraphList();
      manager->ECView->Add("delta-1");
      manager->ECView->Add(opt["graph"].as<string>());
      manager->ECView->Draw();
			if(opt.count("fit")) 
      {
        manager->fitManager->Fit(manager->ECView,opt["fit-method"].as<std::string>(), "delta-1");
        manager->fitManager->GetFitter()->Draw(1);
      }
    }
    bool always=true;
    if(opt.count("test-canvas"))
    {
      new gui::MyCanvas;
    }
    main_window.Run(always);
    //theApp.Run(kTRUE);
    clog << LOG(1) << "POLARIMETER END THE JOB. GOOD BYE!" << endl;
  }
  catch (boost::program_options::error & po_error)
  {
    cerr << "ERROR: options: "<< po_error.what() << endl;
  }
  catch(boost::filesystem::filesystem_error & e)
  {
    cerr << e.what() << endl;
  }
  catch(camac::Crate::Error & e)
  {
    std::cerr << LOG(0) << "polarimeter: camac crate error: " <<e.what() << '\n';
    std::cerr << LOG(0) << "check crate power or check ppi-crate cable" << '\n';
    return (1);
  }
  catch(camac::Module::Error & e)
  {
    std::cerr << LOG(0) <<"polarimeter: camac module error: " <<e.what() << '\n';
    return (1);
  }
  catch(std::runtime_error &e)
  {
    std::cerr <<LOG(0) << "polarimeter: runtime error: " << e.what() << '\n';
    return (2);
  }
  catch(std::out_of_range &e)
  {
    std::cerr <<LOG(0) << "polarimeter: data is out of range: " << e.what() << '\n';
    return (2);
  }
  catch(std::logic_error &e)
  {
    std::cerr << LOG(0) <<"polarimeter: logic error:  " <<e.what() << '\n';
    return (2);
  }
  catch(std::exception &e)
  {
    cerr << LOG(0) << e.what() << "\n";
    exit(0);
  }
  catch(...)
  {
    std::cerr << LOG(0) << "polarimeter: some unrecognize exeption " << '\n';
    return (3);
  }
  return 0;
}
