/*
 * =====================================================================================
 *
 *       Filename:  count.cpp
 *
 *    Description:     ޣ   
 *		           .
 *
 *        Version:  1.0
 *        Created:  10/05/2008 01:23:08 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *        Company:  Budker Institute of Nuclear Physics
 *
 * =====================================================================================
 */

#include <iostream>
#include <fstream>
#include <iomanip>

#include "../camac/Counter.h"
#include "../camac/Timer.h"

#include <unistd.h>
#include <sys/time.h>

#include <boost/program_options.hpp>
#include <boost/program_options/errors.hpp>


#include "/usr/local/include/odplib.h"
#include "/usr/local/include/icl.h"

using namespace std;

enum {
  ERROR, ERROR_BOOST, ERROR_CAMAC, ERROR_FILE
};

void PrintUsage(boost::program_options::options_description & opt_desc) 
{
  cout << "Usage: program --timer <address> --counter <address> [--time time>] [--file <file>]" << endl;
  cout << opt_desc;
  cout << "Example: "<< endl;
  cout << "program --timer 0:1:21 --counter 0:1:17:0 --time 0.1" << endl;
  exit(ERROR);
}

inline double get_real_count(double N0, double N) 
{
  return - N0 * log ( 1. - N/N0);
}

inline double get_real_count_error(double N0, double N) 
{
  return sqrt( N / (1. - N/N0) );
}

namespace ibn {
  class timer {
    public:
      timer(void)  {restart();};

      void restart(void) 
      {
        gettimeofday(&begin_time,0);
      }


      double elapsed(void)
      {
        timeval current_time;
        gettimeofday(&current_time,0);
        return double(current_time.tv_sec - begin_time.tv_sec)+ double(current_time.tv_usec-begin_time.tv_usec)*1e-6;
      }

      timeval get_begin_time(void) const { return begin_time; }

      void set_begin_time(timeval t) 
      {
        begin_time=t;
      }

    private:
      timeval begin_time;	
  };
};

int main(int argc, char ** argv)    {
  namespace po=boost::program_options;
  po::options_description opt_desc("Allowed options:");
  double count_time; //count time
  ostream * os = &cout;
  opt_desc.add_options()
    ("help","Print this help")
    ("synch","Synchronization with BANK")
    ("timer",po::value<string>(),"Address of timer in format interface:crate:station")
    ("counter",po::value<string>(),"Address of counter in format interface:crate:station:channel")
    ("time",po::value<double>(&count_time)->default_value(1.0),"Measurement time [sec]")
    ("file",po::value<string>(),"Output file");

  po::variables_map opt;
  try {
    po::store(po::parse_command_line(argc, argv, opt_desc), opt);
    po::notify(opt);

    if (opt.count("file"))
    {
      string file_name = opt["file"].as<string>();
      ofstream * file = new ofstream(file_name.c_str());
      if(!file)
      {
        cerr << "Unable to upen file " << file_name << endl;
        exit(ERROR_FILE);
      }
      os=file;
    }

    if (opt.count("help")) PrintUsage(opt_desc);
    if (!opt.count("timer") || !opt.count("counter" )) PrintUsage(opt_desc);
    string timer_addr_str = opt["timer"].as<string>();
    string counter_addr_str = opt["counter"].as<string>();

    camac::Timer timer("TIMER",camac::Address(timer_addr_str.c_str()),count_time);
    clog << "Init timer with address: " << timer_addr_str;
    clog << ", count_time=" << count_time << endl;
    camac::CounterC0303 counter("COUNTER",camac::Address(counter_addr_str.c_str()));
    clog << "Init counter with address: " << counter_addr_str << endl;
    if(opt.count("synch")) {
      int socket = odp_socket(ODP_DGRAM, ODP_MS);
      if(socket<0) {
        cerr << strerror(errno) << "\n";
        exit(-1);
      }
      int rc = odp_binda(socket, "POLCOUNT");
      if(rc<0) {
        cerr << strerror(errno) << "\n";
        exit(-1);
      }


      char addr[6] = {0,0,0,0,0,0};
      rc = odp_getaddra("MSVEPP-4",addr);
      if(rc<0) {
        cerr << strerror(errno) << "\n";
        exit(-1);
      }
      rc= odp_connecta(socket,addr,"0000BANK");
      odp_listen( socket, 100);
      clog << "Waiting message from BANK..." << endl;
      //rc = waitms_sgn(socket,0);
      rc = odp_wait(socket,0);
      if(rc<0) {
        cerr << strerror(errno) << "\n";
        exit(-1);
      }
      //odp_close_socket(socket);
      //clog << "Wait for sycnhronization"<<endl;
      //do {
      //    flag=0;
      //    ifstream synch("/home/vepp4/kadrs/polcan_synch.dat");
      //    if(!synch) continue;
      //    synch>>flag;
      //    synch.close();
      //} while(flag!=1);
      //ofstream synch("/home/vepp4/kadrs/polcan_synch.dat");
      //synch<<'0'<<endl;
      clog << "Start"<< endl;
    }
    ibn::timer tm;
    os->precision(15);
    unsigned w=25;
    (*os) << '#' << setw(w) << "elapsed time, s" << setw(w) << "load, Hz" << setw(w) << "error, Hz" << endl;
    while(true) {
      counter.ResetCount();
      timer.Start(count_time);
      timer.Wait();
      unsigned N = counter.GetCount();
      (*os) << setw(w) << tm.elapsed()-count_time/2.;
      //(*os) << setw(w) << double(N)/count_time;
      //(*os) << setw(w) << sqrt(double(N))/count_time;
      double N0=count_time*818924;
      (*os) << setw(w) << get_real_count(N0,N)/count_time;
      (*os) << setw(w) << get_real_count_error(N0,N)/count_time;
      (*os) << endl;
    }
  }
  catch (boost::program_options::error po_error) {
    cerr << "ERROR: options: "<< po_error.what() << endl;
    exit(ERROR_BOOST);
  }
  catch(camac::Crate::Error e)    {
    cerr  << "ERROR: CAMAC crate: " << e.what() << endl;
    cerr  << "check crate power or check ppi-crate cable" << endl;
    exit(ERROR_CAMAC);
  }
  catch(camac::Module::Error e)    {
    cerr << "ERROR: CAMAC Module: " << e.what() << endl;
    exit(ERROR_CAMAC);
  }

  return 0;
}

