/*
 * =====================================================================================
 * 
 *        Filename:  utug.cpp
 * 
 *     Description:  Server for depolarizing beam during injection
 * 
 *         Version:  1.0
 *         Created:  05.04.2005 11:02:14 NOVST
 *        Revision:  none
 *        Compiler:  gcc
 * 
 *          Author:  Ivan Nikolaev (), I.B.Nikolaev@inp.nsk.su
 *         Company:  BINP
 * 
 * =====================================================================================
 */

#include <iostream>
#include <string>
#include <vector>


#include <boost/program_options.hpp>



#include "../measurement/Depolarisator.h"

#include "timer.h"

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <wait.h>


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>


bool VERBOSE=false;


camac::Generator * generator=0;

measurement::Depolarisator * depolarizer=0;

measurement::Depolarisator * DEPOLARIZER=0;
/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  term_handler
 *  Description:  We should tern off Depolarizer since receiving SIGTERM
 * =====================================================================================
 */
void sigint_handler(int i)
{
  if(DEPOLARIZER)
  {
    cout << "Stop scan by user" << endl;
    DEPOLARIZER->StopScan();
    DEPOLARIZER->Scan.Join();
  }
  exit(0);
}

using namespace std;

class Communication 
{
    vector <char> buf;
    string file_name;
    int fd;
    void Close(void);
    void Open(int flag);
    void Read(void);
    void Write(void);
    public:
        Communication(const char *fname);
        ~Communication(void);
        bool IsStart(void);
        void SetProceed(void);
        void SetComplete(void);
};

Communication::Communication(const char *fname) 
{
    file_name = fname;
    buf.resize(6);
}

Communication::~Communication(void) {};

void Communication::Open(int flag)   
{
    fd = open(file_name.c_str(),flag);
    if(fd<0)
    {
        cerr << "ERROR: Unable to open file " << file_name << endl;
    }
}

void Communication::Read(void)   
{
    Open(O_RDONLY);
    size_t n=read(fd,&buf[0],buf.size());
    Close();
    if(VERBOSE) cout << "Read " << n << "bytes\n";
    if(n!=buf.size())
    {
        cerr << "ERROR: unable to read byte\n";
    }
}

void Communication::Write(void)  
{
    Open(O_WRONLY);
    size_t n=write(fd,&buf[0],buf.size());
    Close();
    if(n!=buf.size())
    {
        std::cerr << "ERROR: unable to write byte\n";
    }
}

void Communication::Close(void)  
{
    close(fd);
    fd=0;
}

bool Communication::IsStart(void)    
{
    Read();
    if(buf[5]==1) return true;
    return false;
}

void Communication::SetProceed(void)    
{
    buf[5]=2;
    Write();
}

void Communication::SetComplete(void)   
{
    buf[2]=0;
    buf[5]=0;
    Write();
}


unsigned Log::debug_level=1;
bool Log::is_time=false;
std::string Log::program_name="";

static bool NOCAMAC=false;
static bool MASTER=false;


void Off(void) 
{
  try
  {
    generator->Off();
  } 
  catch(runtime_error e)
  {
    cerr << e.what() << endl;
    exit(1);
  }
}


void Scan(data::DepolarizerConfig & cfg)    
{
  //camac::Address dep_adr("0:2:6:-1");
  //camac::Address a1_adr("0:2:7:-1");
  //camac::Address a3_adr("0:2:8:-1");
  try 
  {
    //measurement::Depolarisator depol(dep_adr,a1_adr,a3_adr);
    //DEPOLARIZER=&depol;
    DEPOLARIZER=depolarizer;
    depolarizer->SetStrategy(measurement::Depolarisator::STRATEGY(cfg.algorithm));
    depolarizer->StartScan(&cfg);

    ibn::timer timer;
    double f;
    while(depolarizer->IsScan()) 
    {
      usleep(int(cfg.scan_time*1e6));
      if(VERBOSE) 
      {
        f = depolarizer->GetCurrentFrequency();
        cout << setprecision(4) << setw(7) << right << timer.elapsed() << " s";
        cout << setprecision(9) << setw(12) <<  right << f << " Hz";
        cout << setprecision(10)<< setw(12) << right << data::frequency2energy(f,cfg.revolution_frequency,cfg.harmonic_number).value << " MeV";
        cout << setw(3) << right<<  depolarizer->attenuator.GetAttenuation() << " dB";
        cout << endl;
      }
    }
  } 
  catch(camac::Crate::Error e)
  {
    std::cerr << "polarimeter: camac crate error: " <<e.what() << '\n';
    std::cerr << "check crate power or check ppi-crate cable" << '\n';
  }
  catch(camac::Module::Error e)
  {
    std::cerr  <<"polarimeter: camac module error: " <<e.what() << '\n';
  }
  catch(std::runtime_error e)
  {
    std::cerr << "polarimeter: runtime error: " << e.what() << '\n';
  }
  std::cout << "End scan\n";
}

void open_file_error(std::string filename)
{
  std::cerr  << "ERROR: Unable to open file " << filename << std::endl;
  exit(1);
};


data::DepolarizerConfig read_config(void)
{
  //       .
  std::string last_file_name="/home/vepp4/kadrs/vepp4energy.dat";
  ifstream ifs(last_file_name.c_str());
  if(!ifs) open_file_error(last_file_name);
  time_t t; // ,   
  int run;  // 
  data::PhysValue E; // , MeV
  //      ,       
  data::PhysValue F; //Hz
  // :
  // unixtime # 2008-12-01 23:23:23
  // run      # run number of RDM energy calibrations
  // E+-dE    # measured energy and error
  // F+-dF    # measured frequency and error
  static const unsigned BUFSIZE=1024;
  ifs >> t;	ifs.ignore(BUFSIZE,'\n');
  ifs >> run; ifs.ignore(BUFSIZE,'\n');
  ifs >> E;	ifs.ignore(BUFSIZE,'\n');
  ifs >> F;	ifs.ignore(BUFSIZE,'\n');
  ifs.close();

  if(VERBOSE) {
    cout << "Read config: " << endl;
    cout << t << endl;
    cout << run << endl;
    cout << E << endl;
    cout << F << endl;
  }

  data::DepolarizerConfig cfg;
  cfg.harmonic_number=cfg.harmonic_number;
  cfg.revolution_frequency=data::PhysValue(818924,0);
  //  
  std::string config_file_name="/home/vepp4/dds/utug.cfg";
  ifs.open(config_file_name.c_str());
  if(!ifs) open_file_error(config_file_name);
  //  
  // UTUG        #    .
  // scan_speed  #  , /c
  // scan_step   #  , 
  // range       # ,     , .
  // attenuation #  , dB

  // 
  double range=0;
  string strategy_name;
  ifs.ignore(BUFSIZE,'\n');
  ifs >> cfg.speed;		ifs.ignore(BUFSIZE,'\n');
  ifs >> cfg.step;            ifs.ignore(BUFSIZE,'\n');
  ifs >> range;		ifs.ignore(BUFSIZE,'\n');
  ifs >> strategy_name;	ifs.ignore(BUFSIZE,'\n');
  ifs >> cfg.attenuation;	ifs.ignore(BUFSIZE,'\n');


  map < string , int > algo_map;
  algo_map["PILA"]=measurement::Depolarisator::LINEAR;
  algo_map["GAUS_RANDOM"]=measurement::Depolarisator::GRND;
  algo_map["UNIFORM_RANDOM"]=measurement::Depolarisator::URND;
  algo_map["MONO"]=measurement::Depolarisator::MONO;

  cfg.algorithm=algo_map.find(strategy_name)->second;



  //  
  if(cfg.step==0) {
    cerr << "ERROR: ZERO frequency step" << endl;
    exit(1);
  }
  cfg.scan_time = cfg.step/cfg.speed;



  cfg.initial=F.value-range/2.;
  cfg.final=F.value+range/2.;
  cfg.harmonic_number=4;

  if(VERBOSE) cout << cfg << endl;
  return cfg;
}

int main(int argc, char ** argv)    
{
  namespace po=boost::program_options;
  po::variables_map opt;
  try 
  {
    /* initalize hardware */
    camac::Address dep_adr("0:2:6:-1");
    generator = new camac::Generator_DDS300("UTUG_GENERATOR",dep_adr);
    depolarizer = new measurement::Depolarisator;
    depolarizer->SetGenerator(generator);


    po::options_description opt_desc("Allowed options");
    opt_desc.add_options()
      ("help","Print this help")
      ("master","Start depolarization without waiting SVED")
      ("off","Turn off depolarizer")
      ("verbose","verbose mode")
      ;
    po::store(po::parse_command_line(argc, argv, opt_desc), opt);
    po::notify(opt);
    if (opt.count("help")) 
    {
      cout << opt_desc << "\n";
      return 1;
    }
    if (opt.count("off")) 
    {
      cout << "Turn depolarizer OFF" << endl;
      generator->Off();
      return 0;
    }
    VERBOSE=opt.count("verbose");
    NOCAMAC=opt.count("nocamac");
    MASTER=opt.count("master");
  } 
  catch(po::error e)
  {
    cerr <<  e.what() << endl;
    exit(1);
  }

  //  ,     CONTROL-C
  struct sigaction sa;
  sa.sa_handler = sigint_handler;
  sigaction(SIGINT, &sa, 0);

  if(opt.count("master")) 
  {
    cout << "Force scan..." << endl;
    data::DepolarizerConfig cfg;
    cfg=read_config();
    Scan(cfg);
    exit(1);
  } 
  else 
  {
    string file_name="/home/shiftman/KEDR/UTUG";
    Communication com(file_name.c_str());
    while(true)
    {
      read_config();
      if(com.IsStart())   {
        com.SetProceed();
        data::DepolarizerConfig cfg;
        cfg=read_config();
        Scan(cfg);
        com.SetComplete();
      }
      sleep(1);
    }
  }
}
